/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/color-gray.C,v 1.2 1995/08/16 00:44:54 olson Exp (LBL)";
#endif

#include <stdio.h>
#include <osfcn.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "color.h"
#include "color-models.h"
#include "endian.h"

GrayColorModel::GrayColorModel()
{
	/*XXX*/
	if (depth_ < 8)
		return;
	if (depth_ > 8) {
		XVisualInfo vi;
		int screen = DefaultScreen(dpy_);
		XMatchVisualInfo(dpy_, screen, depth_, PseudoColor, &vi);
		visual_ = vi.visual;
		if (visual_ == 0)
			return;
	}
	valid_ = 1;
}

GrayColorModel::~GrayColorModel()
{
	if (!valid_)
		return;
	/*XXX*/
}

Colorizer* GrayColorModel::allocate_422(int inw, int inh, int scale) const
{
	return (gray(inw, inh, scale));
}

Colorizer* GrayColorModel::allocate_411(int inw, int inh, int scale) const
{
	return (gray(inw, inh, scale));
}

class GrayMapper : public Colorizer {
protected:
	GrayMapper(const ColorModel& cm, int w, int h, int scale)
		: Colorizer(w, h, scale), cm_(cm) { } 
	const ColorModel& cm_;
};

class FullGrayMapper : public GrayMapper {
    public:
	FullGrayMapper(const ColorModel& cm, int w, int h, int scale)
		: GrayMapper(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class DecGrayMapper : public GrayMapper {
    public:
	DecGrayMapper(const ColorModel& cm, int w, int h, int scale)
		: GrayMapper(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

/* Interpolating GrayMapper. */
class IntGrayMapper : public GrayMapper {
    public:
	IntGrayMapper(const ColorModel& cm, int w, int h, int scale)
		: GrayMapper(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

Colorizer* ColorModel::gray(int inw, int inh, int scale) const
{
	GrayMapper* p;
	if (scale == 0)
		p = new FullGrayMapper(*this, inw, inh, 0);
	else if (scale > 0)
		p = new DecGrayMapper(*this, inw, inh, scale);
	else
		p = new IntGrayMapper(*this, inw, inh, scale);

	return (p);
}

void DecGrayMapper::colorize(u_char* data, const u_char *yp,
			     const u_char* up, const u_char *vp,
			     int x, int y, int width, int height) const
{
	height >>= scale_;
	width >>= scale_;
	width = (width + 3) & ~3;

	u_int* xip = (u_int*)data;
	register int scaler = 1 << scale_;
	
	yp += y * inw_ + x;
	xip += (((y * inw_ >> scale_) + x) >> scale_) / sizeof(*xip);
	u_int* xstart = xip;
	const int stride = inw_ >> scale_;
	const u_short* const map = cm_.graylut();
	int e = 0;
	while (--height >= 0) {
		for (int w = width; w > 0; w -= 4) {
			u_int out = 0;
			e += *yp;
			yp += scaler;
			e = map[e];
			SPLICE(out, e & 0xff, 24);
			e >>= 8;
			e += *yp;
			yp += scaler;
			e = map[e];
			SPLICE(out, e & 0xff, 16);
			e >>= 8;
			e += *yp;
			yp += scaler;
			e = map[e];
			SPLICE(out, e & 0xff, 8);
			e >>= 8;
			e += *yp;
			yp += scaler;
			e = map[e];
			SPLICE(out, e & 0xff, 0);
			e >>= 8;

			*xip++ = out;
		}
		yp += (inw_ - width) << scale_;
		xip = xstart + stride / sizeof(*xip);
		xstart = xip;
	}
}

void FullGrayMapper::colorize(u_char* data, const u_char *yp,
			      const u_char* up, const u_char *vp,
			      int x, int y, int width, int height) const
{
	u_int* xip = (u_int*)data;
	yp += y * inw_ + x;
	xip += (y * inw_ + x) / 4;
	const u_short* const map = cm_.graylut();
	int len = width * height;
	int w = width;
	int e = 0;
	for (; len > 0; len -= 4) {
		u_int out = 0;
		e += yp[0];
		e = map[e];
		SPLICE(out, e & 0xff, 24);
		e >>= 8;
		e += yp[1];
		e = map[e];
		SPLICE(out, e & 0xff, 16);
		e >>= 8;
		e += yp[2];
		e = map[e];
		SPLICE(out, e & 0xff, 8);
		e >>= 8;
		e += yp[3];
		e = map[e];
		SPLICE(out, e & 0xff, 0);
		e >>= 8;

		*xip++ = out;
		
		yp += 4;
		w -= 4;
		if (w <= 0) {
			yp += inw_ - width;
			xip += (inw_ - width) / 4;
			w = width;
		}
	}
}

void IntGrayMapper::colorize(u_char* data, const u_char *yp,
			     const u_char* up, const u_char *vp,
			     int x, int y, int width, int height) const
{
	u_int* xip = (u_int*)data;
	yp += y * inw_ + x;
	xip += (4 * y * inw_ + 2 * x) / 4;
	const int stride = (inw_ * 2) / 4;
	const u_short* const map = cm_.graylut();
	u_int* xstart = xip;
	int w = width;
	int e = 0;
	for (int len = width * height; len > 0; len -= 2) {
		int v = yp[0];
		e += v;
		e = map[e];
		u_int out0 = 0;
		SPLICE(out0, e & 0xff, 24);
		e >>= 8;
		e += v;
		e = map[e];
		u_int out1 = 0;
		SPLICE(out1, e & 0xff, 24);
		e >>= 8;
		e += v;
		e = map[e];
		SPLICE(out1, e & 0xff, 16);
		e >>= 8;
		e += v;
		e = map[e];
		SPLICE(out0, e & 0xff, 16);
		e >>= 8;
		
		v = yp[1];
		e += v;
		e = map[e];
		SPLICE(out0, e & 0xff, 8);
		e >>= 8;
		e += v;
		e = map[e];
		SPLICE(out1, e & 0xff, 8);
		e >>= 8;
		e += v;
		e = map[e];
		SPLICE(out1, e & 0xff, 0);
		e >>= 8;
		e += v;
		e = map[e];
		SPLICE(out0, e & 0xff, 0);
		e >>= 8;

		xip[0] = out0;
		xip[stride] = out1;
		
		yp += 2;
		xip += 1;

		w -= 2;
		if (w <= 0) {
			yp += inw_ - width;
			xip = xstart + 2 * stride;
			xstart = xip;
			w = width;
		}
	}
}
