/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/color-lut.C,v 1.3 1995/08/26 21:15:39 olson Exp (LBL)";

#include <osfcn.h>
#include <sys/file.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "color-models.h"
#include "Tcl.h"

LUTColorModel::LUTColorModel() : lut_(0)
{
	if (depth_ < 8) {
		visual_ = 0;
		return;
	}
	if (depth_ > 8) {
		XVisualInfo vi;
		int screen = DefaultScreen(dpy_);
		XMatchVisualInfo(dpy_, screen, depth_, PseudoColor, &vi);
		visual_ = vi.visual;
		if (visual_ == 0)
			return;
	}
	file_ = Tcl::instance().attr("colorFile");
}

LUTColorModel::~LUTColorModel()
{
	delete lut_;
}

int LUTColorModel::allocate(int ncolor, const u_char* cmap, u_char* lut)
{
	int LUT[256];
	int i;
	for (i  = 0; i < ncolor; ++i) {
		int r = cmap[0];
		int g = cmap[1];
		int b = cmap[2];
		/*XXX boost luminance to make up for CCIR 601 YCrCb*/
#ifdef notdef
		if (r < 240 && g < 240 && b < 240) {
			r += 16;
			g += 16;
			b += 16;
		}
#endif
		int pixel = alloc_color(r, g, b);
		if (pixel < 0)
			return (-1);
		cmap += 3;
		LUT[i] = pixel;
	}
	/*
	 * Map from colormap indices to allocated pixels.
	 */
	for (i = 0; i < 65536; ++i)
		lut[i] = LUT[lut[i]];

	return (0);
}

int LUTColorModel::installLUT(int def_ncolor, const u_char* def_cmap,
			      const u_char* def_lut)
{
	delete lut_;
	lut_ = new u_char[65536];
	u_char cmap[3*255];
	int ncolor = -1;
	if (file_ != 0) {
		ncolor = readLUT(file_, cmap);
		if (ncolor < 0) {
			fprintf(stderr, "vic: couldn't parse %s\n", file_);
			fprintf(stderr, "vic: using builtin color table\n");
		}
	}
	if (ncolor < 0) {
		ncolor = def_ncolor;
		memcpy(cmap, def_cmap, 3 * ncolor);
		memcpy(lut_, def_lut, 65536);
	}
	if (allocate(ncolor, cmap, lut_) < 0) {
		delete lut_;
		lut_ = 0;
		return (-1);
	}
	return (0);
}

int LUTColorModel::readLUT(const char* file, u_char* cmap)
{
	int fd = open(file, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "vic: %s: ", file);
		perror("");
		return (-1);
	}
	u_char c;
	if (read(fd, (char*)&c, 1) != 1)
		return (-1);

	int ncolor = c;
	int cnt = 3 * ncolor;
	if (read(fd, (char*)cmap, cnt) != cnt ||
	    read(fd, (char*)lut_, 65536) != 65536)
		return (-1);

	return (ncolor);
}

int LUTColorModel::newLUT(const char* file)
{
	file_ = file;
	free_colors(); 
	if (alloc_colors() < 0) {
		fprintf(stderr, "vic: could not reallocate color map.\n");
		fprintf(stderr, "vic: kill the interfering application or\n");
		fprintf(stderr, "vic: restart vic with a private colormap.\n");
		exit(1);
	}
	alloc_grays();

	return (0);
}
