/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/color-models.h,v 1.2 1995/08/16 00:44:56 olson Exp (LBL)
 */

#ifndef vic_color_models_h
#define vic_color_models_h

#include "color.h"

class TrueColorModel : public ColorModel {
public:
	TrueColorModel();
	~TrueColorModel();
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const;
	virtual Colorizer* gray(int inw, int inh, int scale) const;
	virtual int alloc_colors();
	virtual int alloc_grays();
	inline u_int omask() const { return (omask_); }
	inline u_int pmask() const { return (pmask_); }
	inline const u_int* uvtab() const { return (&uvtab_[0]); }
protected:
	Colorizer* allocate16_411(int inw, int inh, int scale) const;
	u_int omask_;
	u_int pmask_;
	u_int uvtab_[65536];
};

class TrueGrayColorModel : public TrueColorModel {
public:
	TrueGrayColorModel();
	virtual int alloc_grays();
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const;
};

class LUTColorModel : public ColorModel {
public:
	LUTColorModel();
	~LUTColorModel();
	inline const u_char* lut() const { return (lut_); }
	virtual int newLUT(const char* file);
protected:
	int installLUT(int def_ncolor, const u_char* def_cmap,
		       const u_char* def_lut);
	int allocate(int ncolor, const u_char* cmap, u_char* lut);
	int readLUT(const char* file, u_char* cmap);
	const char* file_;
	u_char* lut_;
};

class DitherColorModel : public LUTColorModel {
public:
	DitherColorModel(int mapno = 0);
	virtual int alloc_colors();
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const;
};

class QuantColorModel : public LUTColorModel {
public:
	QuantColorModel();
	virtual int alloc_colors();
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const;
};

class EDColorModel : public LUTColorModel {
public:
	EDColorModel();
	virtual int alloc_colors();
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const;
	inline const u_int* lut() const { return (lut_); }
private:
	u_int* lut_;
};

class GrayColorModel : public ColorModel {
public:
	GrayColorModel();
	~GrayColorModel();
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const;
};

class MonoColorModel : public ColorModel {
public:
	MonoColorModel();
	~MonoColorModel();
	virtual int alloc_grays();
	Colorizer* mono(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const;
};

#endif
