/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/color-mono.C,v 1.2 1995/08/16 00:44:57 olson Exp (LBL)";
#endif

#include <stdio.h>
#include <osfcn.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "color.h"
#include "color-models.h"
#include "endian.h"

MonoColorModel::MonoColorModel()
{
	if (depth_ == 1)
		valid_ = 1;
}

MonoColorModel::~MonoColorModel()
{
	if (!valid_)
		return;
	/*XXX*/
}

int MonoColorModel::alloc_grays()
{
	return (0);
}

Colorizer* MonoColorModel::allocate_422(int inw, int inh, int scale) const
{
	return (mono(inw, inh, scale));
}

Colorizer* MonoColorModel::allocate_411(int inw, int inh, int scale) const
{
	return (mono(inw, inh, scale));
}

class MonoMapper : public Colorizer {
protected:
	MonoMapper(const ColorModel& cm, int w, int h, int scale)
		: Colorizer(w, h, scale), cm_(cm) { } 
	const ColorModel& cm_;
};

class FullMonoMapper : public MonoMapper {
    public:
	FullMonoMapper(const ColorModel& cm, int w, int h, int scale)
		: MonoMapper(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class DecMonoMapper : public MonoMapper {
    public:
	DecMonoMapper(const ColorModel& cm, int w, int h, int scale)
		: MonoMapper(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

/* Interpolating MonoMapper. */
class IntMonoMapper : public MonoMapper {
    public:
	IntMonoMapper(const ColorModel& cm, int w, int h, int scale)
		: MonoMapper(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

Colorizer* MonoColorModel::mono(int inw, int inh, int scale) const
{
	MonoMapper* p;
	if (scale == 0)
		p = new FullMonoMapper(*this, inw, inh, 0);
	else if (scale > 0)
		p = new DecMonoMapper(*this, inw, inh, scale);
	else
		p = new IntMonoMapper(*this, inw, inh, scale);

	return (p);
}

#define ONEPIX(out, in, n)  \
	e = (e & 0xff) + (in); \
	(out) |= (e & 0x100) >> (n);

void DecMonoMapper::colorize(u_char* data, const u_char *yp,
			     const u_char* up, const u_char *vp,
			     int x, int y, int width, int height) const
{
}

void FullMonoMapper::colorize(u_char* data, const u_char *yp,
			      const u_char* up, const u_char *vp,
			      int x, int y, int width, int height) const
{
}

void IntMonoMapper::colorize(u_char* data, const u_char *yp,
			     const u_char* up, const u_char *vp,
			     int x, int y, int width, int height) const
{
}
