/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/color-quant.C,v 1.2 1995/08/16 00:44:59 olson Exp (LBL)";
#endif

#include <stdio.h>
#include <osfcn.h>
#include <sys/file.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "color.h"
#include "color-models.h"
#include "endian.h"
#include "Tcl.h"

extern "C" int quant_ncolor;
extern "C" u_char quant_cmap[];
extern "C" u_char quant_lut[];

QuantColorModel::QuantColorModel()
{
	/*
	 * parent constructor zeros this if there was a problem
	 */
	if (visual_ == 0)
		return;

	valid_ = 1;
}

int QuantColorModel::alloc_colors()
{
	return (installLUT(quant_ncolor, quant_cmap, quant_lut));
}

class Quantizer : public Colorizer {
protected:
	Quantizer(const QuantColorModel& cm, int w, int h, int scale)
		: Colorizer(w, h, scale), cm_(cm) { } 
	const QuantColorModel& cm_;
};

class Quantizer_422 : public Quantizer {
    public:
	Quantizer_422(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class Dec2Quantizer_422 : public Quantizer {
    public:
	Dec2Quantizer_422(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class Dec4Quantizer_422 : public Quantizer {
    public:
	Dec4Quantizer_422(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class Int2Quantizer_422 : public Quantizer {
    public:
	Int2Quantizer_422(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class Quantizer_411 : public Quantizer {
    public:
	Quantizer_411(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class Dec2Quantizer_411 : public Quantizer {
    public:
	Dec2Quantizer_411(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class Dec4Quantizer_411 : public Quantizer {
    public:
	Dec4Quantizer_411(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

class Int2Quantizer_411 : public Quantizer {
    public:
	Int2Quantizer_411(const QuantColorModel& cm, int w, int h, int scale)
		: Quantizer(cm, w, h, scale) { }
	virtual void colorize(u_char* out,
			    const u_char* yplane,
			    const u_char* uplane,
			    const u_char* vplane,
			    int x, int y, int width, int height) const;
};

Colorizer* QuantColorModel::allocate_422(int inw, int inh, int scale) const
{
	Quantizer* p;
	switch (scale) {

	case 0:
		p = new Quantizer_422(*this, inw, inh, 0);
		break;

	case 1:
		p = new Dec2Quantizer_422(*this, inw, inh, 1);
		break;

	case 2:
		p = new Dec4Quantizer_422(*this, inw, inh, 2);
		break;

	case -1:
		p = new Int2Quantizer_422(*this, inw, inh, -1);
		break;

	default:
		if (scale > 0)
			p = new Dec2Quantizer_422(*this, inw, inh, scale);
		else
			p = new Int2Quantizer_422(*this, inw, inh, scale);
		break;
	}
	return (p);
}

Colorizer* QuantColorModel::allocate_411(int inw, int inh, int scale) const
{
	Quantizer* p;
	switch (scale) {

	case 0:
		p = new Quantizer_411(*this, inw, inh, 0);
		break;

	case 1:
		p = new Dec2Quantizer_411(*this, inw, inh, 1);
		break;

	case 2:
		p = new Dec4Quantizer_411(*this, inw, inh, 2);
		break;

	case -1:
		p = new Int2Quantizer_411(*this, inw, inh, -1);
		break;

	default:
		if (scale > 0)
			p = new Dec2Quantizer_411(*this, inw, inh, scale);
		else
			p = new Int2Quantizer_411(*this, inw, inh, scale);
		break;
	}
	return (p);
}

#if BYTE_ORDER == LITTLE_ENDIAN
#define PIX0(i) ((i) >> 2)
#define PIX1(i) ((i) >> 10)
#define PIX2(i) ((i) >> 18)
#define PIX3(i) ((i) >> 26)
#define UV0 ((v << 3) & 0xffc0)
#define UV1 (u & 0xffc0)
#define UV2 ((v >> 13) & 0xffc0)
#define UV3 ((u >> 16) & 0xffc0)
#else
#define PIX0(i) ((i) >> 26)
#define PIX1(i) ((i) >> 18)
#define PIX2(i) ((i) >> 10)
#define PIX3(i) ((i) >> 2)
#define UV0 ((u >> 16) & 0xffc0)
#define UV1 ((v >> 13) & 0xffc0)
#define UV2 (u & 0xffc0)
#define UV3 ((v << 3) & 0xffc0)
#endif

void Quantizer_422::colorize(register u_char* data,
			     register const u_char *yp,
			     register const u_char* up,
			     register const u_char *vp,
			     int x, int y, int width, int height) const
{
#ifdef DEBUG_QUANT
	register const u_char* lut = cm_.lut();
	for (int y = inh_; --y >= 0; ) {
		for (int x = inw_; x > 0; x -= 2) {
			int u = *up++;
			int v = *vp++;
			int uv = (u + 4 >> 3) << 11;
			uv |= (v + 4 >> 3) << 6;

			int y = *yp++;
			*data++ = lut[uv | (y + 2 >> 2)];
			y = *yp++;
			*data++ = lut[uv | (y + 2 >> 2)];
		}
	}
#else
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = pstart >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register u_int* xip = (u_int*)data + (pstart >> 2);
	register int w = width;
	register const u_char* lut = cm_.lut();
	for (register int len = w * height; len > 0; len -= 8) {
		register u_int u = *(const u_int*)up;
		register u_int v = *(const u_int*)vp;
		register u_int y1;
		register u_int out = 0;
		register const u_char* y2r;

		/*
		 * we do 8 pixels (4 u & v values) per iteration.
		 * the yuv to rgb conversion table is 16 bits (6
		 * y, 5 u, 5 v).  grab 4 u & v values and
		 * construct the appropriate 16 bit representation
		 * (minus the y values).
		 */
		y1 = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
		v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
		u = y1;

		/*
		 * grab first 4 y values & map them to rgb.
		 */
		y1 = *(const u_int*)yp;
		y2r = lut + UV0;
		SPLICE(out, y2r[PIX0(y1) & 0x3f], 24);
		SPLICE(out, y2r[PIX1(y1) & 0x3f], 16);

		y2r = lut + UV1;
		SPLICE(out, y2r[PIX2(y1) & 0x3f], 8);
		SPLICE(out, y2r[PIX3(y1) & 0x3f], 0);
		xip[0] =  out;

		/*
		 * grab last 4 y values & map them to rgb.
		 */
		y1 = *(const u_int*)(yp + 4);
		y2r = lut + UV2;
		out = 0;
		SPLICE(out, y2r[PIX0(y1) & 0x3f], 24);
		SPLICE(out, y2r[PIX1(y1) & 0x3f], 16);

		y2r = lut + UV3;
		SPLICE(out, y2r[PIX2(y1) & 0x3f], 8);
		SPLICE(out, y2r[PIX3(y1) & 0x3f], 0);
		xip[1] =  out;

		xip += 2;
		yp += 8;
		up += 4;
		vp += 4;

		w -= 8;
		if (w <= 0) {
			w = width;
			register int pstride = iw - w;
			register int cstride = pstride >> 1;
			yp += pstride;
			up += cstride;
			vp += cstride;
			xip += (pstride >> 2);
		}
	}
#endif
}

void Dec2Quantizer_422::colorize(u_char* data, const u_char *yp,
				   const u_char* up, const u_char *vp,
				   int x, int y, int width, int height) const
{
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = pstart >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register u_int* xip = (u_int*)(data + (((y * iw >> 1) + x) >> 1));
	register int w = width;
	register const u_char* lut = cm_.lut();
	for (register int len = w * height >> 1; len > 0; len -= 8) {
		register u_int u = *(const u_int*)up;
		register u_int v = *(const u_int*)vp;
		register u_int y1, out = 0;
		register const u_char* y2r;

		y1 = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
		v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
		u = y1;

		y1 = *(const u_int*)yp;
		y2r = lut + UV0;
		SPLICE(out, y2r[PIX0(y1) & 0x3f], 24);

		y2r = lut + UV1;
		SPLICE(out, y2r[PIX2(y1) & 0x3f], 16);

		y1 = *(const u_int*)(yp + 4);
		y2r = lut + UV2;
		SPLICE(out, y2r[PIX0(y1) & 0x3f], 8);

		y2r = lut + UV3;
		SPLICE(out, y2r[PIX2(y1) & 0x3f], 0);

		*xip++ = out;
		yp += 8;
		up += 4;
		vp += 4;

		w -= 8;
		if (w <= 0) {
			w = width;
			register int pstride = 2 * iw - w;
			register int cstride = pstride >> 1;
			yp += pstride;
			up += cstride;
			vp += cstride;
			xip += (iw - w) >> 3;
		}
	}
}

void Dec4Quantizer_422::colorize(u_char* data, const u_char *yp,
				const u_char* up, const u_char *vp,
				int x, int y, int width, int height) const
{
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = pstart >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register int w = width;
	register const u_char* lut = cm_.lut();
	register u_int doff = ((y * iw >> 2) + x) >> 2;
	register int len = w * height >> 2;
	if ((w & 0xf) != 0 || (len & 0xf) != 0 || (doff & 3) != 0) {
		register u_short* xip = (u_short*)(data + doff);
		for (; len > 0; len -= 8) {
			register u_int u = *(const u_int*)up;
			register u_int v = *(const u_int*)vp;
			register u_int y1;
			register u_short out = 0;
#if BYTE_ORDER == LITTLE_ENDIAN
			v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
#else
			u = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
#endif
			y1 = (u_int)(yp[0] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV0], 8);
			y1 = (u_int)(yp[4] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV2], 0);
			*xip++ = out;
			yp += 8;
			up += 4;
			vp += 4;
			w -= 8;
			if (w <= 0) {
				w = width;
				register int pstride = 4 * iw - w;
				register int cstride = pstride >> 1;
				yp += pstride;
				up += cstride;
				vp += cstride;
				xip += (iw - w) >> 3;
			}
		}
	} else {
		register u_int* xip = (u_int*)(data + doff);
		for (; len > 0; len -= 16) {
			register u_int u = *(const u_int*)up;
			register u_int v = *(const u_int*)vp;
			register u_int y1, out = 0;
#if BYTE_ORDER == LITTLE_ENDIAN
			v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
#else
			u = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
#endif
			y1 = (u_int)(yp[0] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV0], 24);
			y1 = (u_int)(yp[4] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV2], 16);
			u = *(const u_int*)(up + 4);
			v = *(const u_int*)(vp + 4);
#if BYTE_ORDER == LITTLE_ENDIAN
			v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
#else
			u = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
#endif
			y1 = (u_int)(yp[8] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV0], 8);
			y1 = (u_int)(yp[12] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV2], 0);
			*xip++ = out;
			yp += 16;
			up += 8;
			vp += 8;
			w -= 16;
			if (w <= 0) {
				w = width;
				register int pstride = 4 * iw - w;
				register int cstride = pstride >> 1;
				yp += pstride;
				up += cstride;
				vp += cstride;
				xip += (iw - w) >> 4;
			}
		}
	}
}

void Int2Quantizer_422::colorize(register u_char* data,
				register const u_char *yp,
				register const u_char* up,
				register const u_char *vp,
				int x, int y, int width, int height) const
{
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = pstart >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register u_int* xip = (u_int*)(data + (((y * iw << 1) + x) << 1));
	register int w = width;
	register u_int e1 = yp[0] >> 2;
	register const u_char* lut = cm_.lut();

	for (register int len = width * height; len > 0; len -= 8) {
		register u_int u = *(const u_int*)up;
		register u_int v = *(const u_int*)vp;
		register u_int y1, e2, out, p;
		register const u_char* y2r;
		register u_int* xip2 = xip + (iw >> 1);

		y1 = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
		v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
		u = y1;

		y1 = *(const u_int*)yp;
		out = 0;
		y2r = lut + UV0;
		e2 = PIX0(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX1(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip++ = out;
		*xip2++ = out;

		out = 0;
		y2r = lut + UV1;
		e2 = PIX2(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX3(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip++ = out;
		*xip2++ = out;

		yp += 4;
		y1 = *(const u_int*)yp;
		out = 0;
		y2r = lut + UV2;
		e2 = PIX0(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX1(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip++ = out;
		*xip2++ = out;

		out = 0;
		y2r = lut + UV3;
		e2 = PIX2(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX3(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip++ = out;
		*xip2++ = out;

		yp += 4;
		up += 4;
		vp += 4;

		w -= 8;
		if (w <= 0) {
			w = width;
			register u_int pstride = iw - w;
			register u_int cstride = pstride >> 1;
			yp += pstride;
			e1 = yp[0] >> 2;
			up += cstride;
			vp += cstride;
			xip += (iw + pstride) >> 1;
		}
	}
}


void Quantizer_411::colorize(register u_char* data,
			     register const u_char *yp,
			     register const u_char* up,
			     register const u_char *vp,
			     int x, int y, int width, int height) const
{
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = ((y * iw >> 1) + x) >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register u_int* xip = (u_int*)(data + pstart);
	register int w = width;
	register const u_char* lut = cm_.lut();
	for (register int len = width * height; len > 0; len -= 16) {
		register u_int u = *(const u_int*)up;
		register u_int v = *(const u_int*)vp;
		register u_int y1, y2, out1, out2;
		register const u_char* y2r;
		register u_int* xip2 = xip + (iw >> 2);

		y1 = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
		v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
		u = y1;

		y1 = *(const u_int*)yp;
		y2 = *(const u_int*)(yp + iw);
		out1 = 0;
		out2 = 0;
		y2r = lut + UV0;
		SPLICE(out1, y2r[PIX0(y1) & 0x3f], 24);
		SPLICE(out1, y2r[PIX1(y1) & 0x3f], 16);
		SPLICE(out2, y2r[PIX0(y2) & 0x3f], 24);
		SPLICE(out2, y2r[PIX1(y2) & 0x3f], 16);

		y2r = lut + UV1;
		SPLICE(out1, y2r[PIX2(y1) & 0x3f], 8);
		SPLICE(out1, y2r[PIX3(y1) & 0x3f], 0);
		SPLICE(out2, y2r[PIX2(y2) & 0x3f], 8);
		SPLICE(out2, y2r[PIX3(y2) & 0x3f], 0);

		*xip++ = out1;
		*xip2++ = out2;

		yp += 4;
		y1 = *(const u_int*)yp;
		y2 = *(const u_int*)(yp + iw);
		out1 = 0;
		out2 = 0;
		y2r = lut + UV2;
		SPLICE(out1, y2r[PIX0(y1) & 0x3f], 24);
		SPLICE(out1, y2r[PIX1(y1) & 0x3f], 16);
		SPLICE(out2, y2r[PIX0(y2) & 0x3f], 24);
		SPLICE(out2, y2r[PIX1(y2) & 0x3f], 16);

		y2r = lut + UV3;
		SPLICE(out1, y2r[PIX2(y1) & 0x3f], 8);
		SPLICE(out1, y2r[PIX3(y1) & 0x3f], 0);
		SPLICE(out2, y2r[PIX2(y2) & 0x3f], 8);
		SPLICE(out2, y2r[PIX3(y2) & 0x3f], 0);

		*xip++ = out1;
		*xip2++ = out2;

		yp += 4;
		up += 4;
		vp += 4;

		w -= 8;
		if (w <= 0) {
			w = width;
			register int pstride = 2 * iw - w;
			register int cstride = (iw - w) >> 1;
			yp += pstride;
			up += cstride;
			vp += cstride;
			xip += (pstride >> 2);
		}
	}
}

void Dec2Quantizer_411::colorize(u_char* data, const u_char *yp,
				const u_char* up, const u_char *vp,
				int x, int y, int width, int height) const
{
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = ((y * iw >> 1) + x) >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register u_int* xip = (u_int*)(data + (((y * iw >> 1) + x) >> 1));
	register int w = width;
	register const u_char* lut = cm_.lut();
	for (register int len = w * height >> 1; len > 0; len -= 8) {
		register u_int u = *(const u_int*)up;
		register u_int v = *(const u_int*)vp;
		register u_int y1, out = 0;
		register const u_char* y2r;

		y1 = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
		v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
		u = y1;

		y1 = *(const u_int*)yp;
		y2r = lut + UV0;
		SPLICE(out, y2r[PIX0(y1) & 0x3f], 24);

		y2r = y2r = lut + UV1;
		SPLICE(out, y2r[PIX2(y1) & 0x3f], 16);

		y1 = *(const u_int*)(yp + 4);
		y2r = lut + UV2;
		SPLICE(out, y2r[PIX0(y1) & 0x3f], 8);

		y2r = lut + UV3;
		SPLICE(out, y2r[PIX2(y1) & 0x3f], 0);

		*xip++ = out;
		yp += 8;
		up += 4;
		vp += 4;

		w -= 8;
		if (w <= 0) {
			w = width;
			register int pstride = 2 * iw - w;
			register int cstride = (iw - w) >> 1;
			yp += pstride;
			up += cstride;
			vp += cstride;
			xip += cstride >> 2;
		}
	}
}

void Dec4Quantizer_411::colorize(u_char* data, const u_char *yp,
				const u_char* up, const u_char *vp,
				int x, int y, int width, int height) const
{
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = ((y * iw >> 1) + x) >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register int w = width;
	register const u_char* lut = cm_.lut();
	register u_int doff = ((y * iw >> 2) + x) >> 2;
	register int len = w * height >> 2;
	if ((w & 0xf) != 0 || (len & 0xf) != 0 || (doff & 3) != 0) {
		register u_short* xip = (u_short*)(data + doff);
		for (; len > 0; len -= 8) {
			register u_int u = *(const u_int*)up;
			register u_int v = *(const u_int*)vp;
			register u_int y1;
			register u_short out = 0;
#if BYTE_ORDER == LITTLE_ENDIAN
			v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
#else
			u = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
#endif
			y1 = (u_int)(yp[0] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV0], 8);
			y1 = (u_int)(yp[4] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV2], 0);
			*xip++ = out;
			yp += 8;
			up += 4;
			vp += 4;
			w -= 8;
			if (w <= 0) {
				w = width;
				register int pstride = 4 * iw - w;
				register int cstride = iw - (w >> 1);
				yp += pstride;
				up += cstride;
				vp += cstride;
				xip += (iw - w) >> 3;
			}
		}
	} else {
		register u_int* xip = (u_int*)(data + doff);
		for (; len > 0; len -= 16) {
			register u_int u = *(const u_int*)up;
			register u_int v = *(const u_int*)vp;
			register u_int y1, out = 0;
#if BYTE_ORDER == LITTLE_ENDIAN
			v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
#else
			u = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
#endif
			y1 = (u_int)(yp[0] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV0], 24);
			y1 = (u_int)(yp[4] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV2], 16);
			u = *(const u_int*)(up + 4);
			v = *(const u_int*)(vp + 4);
#if BYTE_ORDER == LITTLE_ENDIAN
			v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
#else
			u = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
#endif
			y1 = (u_int)(yp[8] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV0], 8);
			y1 = (u_int)(yp[12] & 0xfc) >> 2;
			SPLICE(out, lut[y1 + UV2], 0);
			*xip++ = out;
			yp += 16;
			up += 8;
			vp += 8;
			w -= 16;
			if (w <= 0) {
				w = width;
				register int pstride = 4 * iw - w;
				register int cstride = iw - (w >> 1);
				yp += pstride;
				up += cstride;
				vp += cstride;
				xip += (iw - w) >> 4;
			}
		}
	}
}

void Int2Quantizer_411::colorize(register u_char* data,
				   register const u_char *yp,
				   register const u_char* up,
				   register const u_char *vp,
				   int x, int y, int width, int height) const
{
	register u_int iw = inw_;
	u_int pstart = y * iw + x;
	u_int cstart = ((y * iw >> 1) + x) >> 1;
	yp += pstart;
	up += cstart;
	vp += cstart;
	register u_int* xip = (u_int*)(data + 4 * y * iw + 2 * x);
	register int w = width;
	register const u_char* lut = cm_.lut();
	register u_int e1 = (yp[0] & 0xfc) >> 2, o1 = (yp[iw] & 0xfc) >> 2;
	for (register int len = width * height; len > 0; len -= 16) {
		register u_int u = *(const u_int*)up;
		register u_int v = *(const u_int*)vp;
		register u_int y1, y2, e2, o2, out, p;
		register const u_char* y2r;
		register u_int* xip2 = xip + (iw >> 1);

		y1 = (u & 0xf800f800) | ((v & 0xf800f800) >> 5);
		v = ((u & 0x00f800f8) << 5) | (v & 0x00f800f8);
		u = y1;

		/* double pixels 0 & 1 of the `even' row */
		y1 = *(const u_int*)yp;
		out = 0;
		y2r = lut + UV0;
		e2 = PIX0(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX1(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip = out;
		*xip2 = out;

		/* double pixels 0 & 1 of the `odd' row */
		y2 = *(const u_int*)(yp + iw);
		out = 0;
		o2 = PIX0(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[o2], 16);
		o1 = PIX1(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[o1], 0);
		xip[iw] = out;
		xip2[iw] = out;
		++xip;
		++xip2;

		/* double pixels 2 & 3 of the `even' row */
		out = 0;
		y2r = lut + UV1;
		e2 = PIX2(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX3(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip = out;
		*xip2 = out;

		/* double pixels 2 & 3 of the `odd' row */
		out = 0;
		o2 = PIX2(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[o2], 16);
		o1 = PIX3(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[o1], 0);
		xip[iw] = out;
		xip2[iw] = out;
		++xip;
		++xip2;

		/* double pixels 4 & 5 of the `even' row */
		yp += 4;
		y1 = *(const u_int*)yp;
		out = 0;
		y2r = lut + UV2;
		e2 = PIX0(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX1(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip = out;
		*xip2 = out;

		/* double pixels 4 & 5 of the `odd' row */
		y2 = *(const u_int*)(yp + iw);
		out = 0;
		o2 = PIX0(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[o2], 16);
		o1 = PIX1(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[o1], 0);
		xip[iw] = out;
		xip2[iw] = out;
		++xip;
		++xip2;

		/* double pixels 6 & 7 of the `even' row */
		out = 0;
		y2r = lut + UV3;
		e2 = PIX2(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[e2], 16);
		e1 = PIX3(y1) & 0x3f;
		p = ((e1 + e2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[e1], 0);
		*xip = out;
		*xip2 = out;

		/* double pixels 6 & 7 of the `odd' row */
		out = 0;
		o2 = PIX2(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 24);
		SPLICE(out, y2r[o2], 16);
		o1 = PIX3(y1) & 0x3f;
		p = ((o1 + o2) >> 1) & 0x3f;
		SPLICE(out, y2r[p], 8);
		SPLICE(out, y2r[o1], 0);
		xip[iw] = out;
		xip2[iw] = out;
		++xip;
		++xip2;

		up += 4;
		vp += 4;
		yp += 4;

		w -= 8;
		if (w <= 0) {
			w = width;
			register u_int pstride = 2 * iw - w;
			register u_int cstride = (iw - w) >> 1;
			yp += pstride;
			e1 = (yp[0] & 0xfc) >> 2;
			o1 = (yp[iw] & 0xfc) >> 2;
			up += cstride;
			vp += cstride;
			xip += (8 * iw - 2 * w) >> 2;
		}
	}
}
