/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/color.h,v 1.2 1995/08/16 00:45:05 olson Exp (LBL)
 */

#ifndef vic_color_h
#define vic_color_h

#include "iohandler.h"

extern "C" {
#include <tk.h>
}

#define GREY_LEVELS 32
#define YUVLIM(x) (((x) < 0) ? 0 : ((x) > 127) ? 127 : (x))


class Colorizer;

class MedianCutHandler : public IOHandler {
    public:
	MedianCutHandler();
	~MedianCutHandler();
	void dispatch(int mask);
	inline int child() const { return (child_); }
	int dump(const u_int* histogram);
	int fork(int eflag);
    private:
	int child_;
	int channel_;
	char infile_[80];
	char outfile_[80];
};

/*
 * Classes for mapping YUV image data into a representation that
 * can be rendered on the display.  The ColorModel class represents
 * a finite-element color cube, which might be uniformaly sampled,
 * or biased; or, a map from YUV space to 24-bit true color.
 * The Colorizer class represents the agents that do the work of
 * mapping from YUV space to the ColorModel space.  For example,
 * they might do dithering or simple (non-uniform) quantization.
 */

class ColorModel {
 public:
	struct color {
		short r;
		short g;
		short b;
		short y;
		short u;
		short v;
	};
	virtual ~ColorModel();
	static inline ColorModel& instance() { return (*instance_); }
	inline Colormap colormap() const { return (colormap_); }
	inline Visual* visual() const { return (visual_); }
	inline int depth() const { return (depth_); }
	virtual Colorizer* allocate_422(int inw, int inh, int scale) const = 0;
	virtual Colorizer* allocate_411(int inw, int inh, int scale) const = 0;
	virtual Colorizer* gray(int inw, int inh, int scale) const;

	static void init(int model);
	static void release();
	static void use_private_map(int v) { use_private_map_ = v; }

#define COLORMODEL_TRUE		0
#define COLORMODEL_DITHER	1
#define COLORMODEL_QUANTIZE	2
#define COLORMODEL_MONO		3
#define COLORMODEL_GRAY		4
#define COLORMODEL_TRUEGRAY	5
#define COLORMODEL_JV		6
#define COLORMODEL_BEST		7
#define COLORMODEL_ED		8
	static const char* cmtoa(int cm);
	static int atocm(const char*);
	inline const u_short* graylut() const { return (grey_lut_); }
	inline int valid() const { return (valid_); }
	virtual int alloc_colors();
	virtual int alloc_grays();
	const color* lookup(int pixel) const { return (&colors_[pixel]); }
	virtual int newLUT(const char* file);
	inline int procBusy() const { return (proc_.child() != 0); }
	void fork(const u_int* histogram);
 protected:
	ColorModel();
	static ColorModel* instance_;
	static int use_private_map_;

	void private_map();
	void free_colors();
	void rgb_to_yuv(color& c) const;
	int yuv_to_rgb(color& c) const;

	/*
	 * Return the n'th allocated color.
	 */
	inline const color& colors(int n) const { return (colors_[pixel_[n]]); }

	int alloc_color(color&);
	int alloc_color(int r, int g, int b);
	
	int valid_;
	Display* dpy_;
	Colormap colormap_;
	Visual* visual_;
	int depth_;
	double gamma_;

	int ncolor_;
	color colors_[256];
	u_long pixel_[256];
	u_short grey_lut_[512];

	/*
	 * Hooks for median cut.
	 */
	int eflag_;
 private:
	MedianCutHandler proc_;
};

inline int ColorModel::alloc_color(int r, int g, int b)
{
	color c;
	c.r = r;
	c.g = g;
	c.b = b;
	return (alloc_color(c));
}

class Colorizer {
 public:
	virtual void colorize(u_char* out,
			      const u_char* yplane,
			      const u_char* uplane,
			      const u_char* vplane,
			      int x, int y, int width, int height) const = 0;
	inline int scale() const { return (scale_); }
 protected:
	Colorizer(int inw, int inh, int scalebits);
	int inw_;
	int inh_;
	int scale_;
};

#endif
