/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Network Research
 *	Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/dct.C,v 1.3 1995/08/26 21:15:43 olson Exp (LBL)";
#endif

/*
 * Macros for fix-point (integer) arithmetic.  FP_NBITS gives the number
 * of binary digits past the decimal point.  FP_MUL computes the product
 * of two fixed point numbers.  A fixed point number and an integer
 * can be directly multiplied to give a fixed point number.  FP_SCALE
 * converts a floating point number to fixed point (and is used only
 * at startup, not by the dct engine).  FP_NORM converts a fixed
 * point number to scalar by rounding to the closest integer.
 * FP_NORM converts a fixed point number to scalar by  truncating
 * toward zero.  FP_MUL2 multiplies to fixed point numbers, but
 * retains more precision on the second argument.
 */
#define FP_NBITS 12
#define FP_NORM(v)	((v) + (1 << (FP_NBITS-1)) >> FP_NBITS)
#define FP_NORMT(t)	(((t) >> FP_NBITS) + (((u_int)(t)) >> 31))
#define FP_MUL(a, b)	(((a) >> FP_NBITS / 2) * \
			 ((b) >> (FP_NBITS - FP_NBITS / 2)))
#define FP_SCALE(v)	(int)((double)(v) * (1 << FP_NBITS) + 0.5)

#define FP_MUL2(a, b)	(((a) >> FP_NBITS) * (b))

#include "endian.h"

#define M(n) ((m0 >> (n)) & 1)

/*
 * This macro stolen from nv.
 */
/* Sick little macro which will limit x to [0..255] with logical ops */
#define LIMIT8(x, t) ((t = (x)), (t &= ~(t>>31)), (t | ~((t-256) >> 31)))
#define LIMIT(x, t) (LIMIT8((x), t) & 0xff)

#include <sys/types.h>
#include "dct.h"

/* row order */
const int ROWZAG[] = {
	0,  1,  8, 16,  9,  2,  3, 10,
	17, 24, 32, 25, 18, 11,  4,  5,
	12, 19, 26, 33, 40, 48, 41, 34,
	27, 20, 13,  6,  7, 14, 21, 28,
	35, 42, 49, 56, 57, 50, 43, 36,
	29, 22, 15, 23, 30, 37, 44, 51,
	58, 59, 52, 45, 38, 31, 39, 46,
	53, 60, 61, 54, 47, 55, 62, 63,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0
};
/* column order */
const int COLZAG[] = {
	0, 8, 1, 2, 9, 16, 24, 17,
	10, 3, 4, 11, 18, 25, 32, 40,
	33, 26, 19, 12, 5, 6, 13, 20,
	27, 34, 41, 48, 56, 49, 42, 35,
	28, 21, 14, 7, 15, 22, 29, 36,
	43, 50, 57, 58, 51, 44, 37, 30,
	23, 31, 38, 45, 52, 59, 60, 53,
	46, 39, 47, 54, 61, 62, 55, 63,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0
};

#define A1 FP_SCALE(0.7071068)
#define A2 FP_SCALE(0.5411961)
#define A3 A1
#define A4 FP_SCALE(1.3065630)
#define A5 FP_SCALE(0.3826834)
#define B0 0.353553
#define B1 0.254898
#define B2 0.270598
#define B3 0.300672
#define B4 0.353553
#define B5 0.449988
#define B6 0.653281
#define B7 1.28146

/*
 * Output multipliers for AA&N DCT
 * (i.e., first stage multipliers for inverse DCT).
 */
static double first_stage[8] = { B0, B1, B2, B3, B4, B5, B6, B7, };

/*
 * The first_stage array crossed with itself.  This allows us
 * to embed the first stage multipliers of the row pass by
 * computing scaled versions of the columns.
 */
static int cross_stage[64] = {
	FP_SCALE(B0 * B0),
	FP_SCALE(B0 * B1),
	FP_SCALE(B0 * B2),
	FP_SCALE(B0 * B3),
	FP_SCALE(B0 * B4),
	FP_SCALE(B0 * B5),
	FP_SCALE(B0 * B6),
	FP_SCALE(B0 * B7),

	FP_SCALE(B1 * B0),
	FP_SCALE(B1 * B1),
	FP_SCALE(B1 * B2),
	FP_SCALE(B1 * B3),
	FP_SCALE(B1 * B4),
	FP_SCALE(B1 * B5),
	FP_SCALE(B1 * B6),
	FP_SCALE(B1 * B7),

	FP_SCALE(B2 * B0),
	FP_SCALE(B2 * B1),
	FP_SCALE(B2 * B2),
	FP_SCALE(B2 * B3),
	FP_SCALE(B2 * B4),
	FP_SCALE(B2 * B5),
	FP_SCALE(B2 * B6),
	FP_SCALE(B2 * B7),

	FP_SCALE(B3 * B0),
	FP_SCALE(B3 * B1),
	FP_SCALE(B3 * B2),
	FP_SCALE(B3 * B3),
	FP_SCALE(B3 * B4),
	FP_SCALE(B3 * B5),
	FP_SCALE(B3 * B6),
	FP_SCALE(B3 * B7),

	FP_SCALE(B4 * B0),
	FP_SCALE(B4 * B1),
	FP_SCALE(B4 * B2),
	FP_SCALE(B4 * B3),
	FP_SCALE(B4 * B4),
	FP_SCALE(B4 * B5),
	FP_SCALE(B4 * B6),
	FP_SCALE(B4 * B7),

	FP_SCALE(B5 * B0),
	FP_SCALE(B5 * B1),
	FP_SCALE(B5 * B2),
	FP_SCALE(B5 * B3),
	FP_SCALE(B5 * B4),
	FP_SCALE(B5 * B5),
	FP_SCALE(B5 * B6),
	FP_SCALE(B5 * B7),

	FP_SCALE(B6 * B0),
	FP_SCALE(B6 * B1),
	FP_SCALE(B6 * B2),
	FP_SCALE(B6 * B3),
	FP_SCALE(B6 * B4),
	FP_SCALE(B6 * B5),
	FP_SCALE(B6 * B6),
	FP_SCALE(B6 * B7),

	FP_SCALE(B7 * B0),
	FP_SCALE(B7 * B1),
	FP_SCALE(B7 * B2),
	FP_SCALE(B7 * B3),
	FP_SCALE(B7 * B4),
	FP_SCALE(B7 * B5),
	FP_SCALE(B7 * B6),
	FP_SCALE(B7 * B7),
};

/*
 * Map a quantization table in natural, row-order,
 * into the qt input expected by rdct().
 */
void
rdct_fold_q(const int* in, int* out)
{
	for (int i = 0; i < 64; ++i) {
		/*
		 * Fold column and row passes of the dct.
		 * By scaling each column DCT independently,
		 * we pre-bias all the row DCT's so the
		 * first multiplier is already embedded
		 * in the temporary result.  Thanks to
		 * Martin Vetterli for explaining how
		 * to do this.
		 */
		int k = COLZAG[i];
		double v = double(in[k]);
		v *= first_stage[k & 7];
		v *= first_stage[k >> 3];
		out[k] = FP_SCALE(v);
	}
}

/*
 * Just like rdct_fold_q() but we divide by the quantizer.
 */
void
fdct_fold_q(const int* in, int* out, int* outinv)
{
	for (int i = 0; i < 64; ++i) {
		int k = COLZAG[i];
		double v = first_stage[k >> 3];
		v *= first_stage[k & 7];
		v /= double(in[k]);
		out[k] = FP_SCALE(v);
		outinv[k] = FP_SCALE(1./v);
	}
}

void dcsum(int dc, u_char* in, u_char* out, int stride)
{
	for (int k = 8; --k >= 0; ) {
		int t;
#ifdef INT_64
		/*XXX assume little-endian */
		INT_64 i = *(INT_64*)in;
		INT_64 o = (INT_64)LIMIT(dc + (i >> 56), t) << 56;
		o |=  (INT_64)LIMIT(dc + (i >> 48 & 0xff), t) << 48;
		o |=  (INT_64)LIMIT(dc + (i >> 40 & 0xff), t) << 40;
		o |=  (INT_64)LIMIT(dc + (i >> 32 & 0xff), t) << 32;
		o |=  (INT_64)LIMIT(dc + (i >> 24 & 0xff), t) << 24;
		o |=  (INT_64)LIMIT(dc + (i >> 16 & 0xff), t) << 16;
		o |=  (INT_64)LIMIT(dc + (i >> 8 & 0xff), t) << 8;
		o |=  (INT_64)LIMIT(dc + (i & 0xff), t);
		*(INT_64*)out = o;
#else
		u_int o = 0;
		u_int i = *(u_int*)in;
		SPLICE(o, LIMIT(dc + EXTRACT(i, 24), t), 24);
		SPLICE(o, LIMIT(dc + EXTRACT(i, 16), t), 16);
		SPLICE(o, LIMIT(dc + EXTRACT(i, 8), t), 8);
		SPLICE(o, LIMIT(dc + EXTRACT(i, 0), t), 0);
		*(u_int*)out = o;

		o = 0;
		i = *(u_int*)(in + 4);
		SPLICE(o, LIMIT(dc + EXTRACT(i, 24),  t), 24);
		SPLICE(o, LIMIT(dc + EXTRACT(i, 16), t), 16);
		SPLICE(o, LIMIT(dc + EXTRACT(i, 8), t), 8);
		SPLICE(o, LIMIT(dc + EXTRACT(i, 0), t), 0);
		*(u_int*)(out + 4) = o;
#endif
		in += stride;
		out += stride;
	}
}

void dcsum2(int dc, u_char* in, u_char* out, int stride)
{
	for (int k = 8; --k >= 0; ) {
		int t;
		u_int o = 0;
		SPLICE(o, LIMIT(dc + in[0], t), 24);
		SPLICE(o, LIMIT(dc + in[1], t), 16);
		SPLICE(o, LIMIT(dc + in[2], t), 8);
		SPLICE(o, LIMIT(dc + in[3], t), 0);
		*(u_int*)out = o;

		o = 0;
		SPLICE(o, LIMIT(dc + in[4], t), 24);
		SPLICE(o, LIMIT(dc + in[5], t), 16);
		SPLICE(o, LIMIT(dc + in[6], t), 8);
		SPLICE(o, LIMIT(dc + in[7], t), 0);
		*(u_int*)(out + 4) = o;

		in += stride;
		out += stride;
	}
}

void dcfill(int DC, u_char* out, int stride)
{
	int t;
	u_int dc = DC;
	dc = LIMIT(dc, t);
	dc |= dc << 8;
	dc |= dc << 16;
#ifdef INT_64
	INT_64 xdc = dc;
	xdc |= xdc << 32;
	*(INT_64 *)out = xdc;
	out += stride;
	*(INT_64 *)out = xdc;
	out += stride;
	*(INT_64 *)out = xdc;
	out += stride;
	*(INT_64 *)out = xdc;
	out += stride;
	*(INT_64 *)out = xdc;
	out += stride;
	*(INT_64 *)out = xdc;
	out += stride;
	*(INT_64 *)out = xdc;
	out += stride;
	*(INT_64 *)out = xdc;
#else
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
	out += stride;
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
	out += stride;
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
	out += stride;
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
	out += stride;
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
	out += stride;
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
	out += stride;
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
	out += stride;
	*(u_int*)out = dc;
	*(u_int*)(out + 4) = dc;
#endif
}

/*
 * This routine mixes the DC & AC components of an 8x8 block of
 * pixels.  This routine is called for every block decoded so it
 * needs to be efficient.  It tries to do as many pixels in parallel
 * as will fit in a word.  The one complication is that it has to
 * deal with overflow (sum > 255) and underflow (sum < 0).  Underflow
 * & overflow are only possible if both terms have the same sign and
 * are indicated by the result having a different sign than the terms.
 * Note that underflow is more worrisome than overflow since it results
 * in bright white dots in a black field.
 * The DC term and sum are biased by 128 so a negative number has the
 * 2^7 bit = 0.  The AC term is not biased so a negative number has
 * the 2^7 bit = 1.  So underflow is indicated by (DC & AC & sum) != 0;
 */
#define MIX_LOGIC(sum, a, b, omask, uflo) \
{ \
	sum = a + b; \
	uflo = (a ^ b) & (a ^ sum) & omask; \
	if (uflo) { \
		if ((b = uflo & a) != 0) { \
			/* integer overflows */ \
			b |= b >> 1; \
			b |= b >> 2; \
			b |= b >> 4; \
			sum |= b; \
		} \
		if ((uflo &=~ b) != 0) { \
			/* integer underflow(s) */ \
			uflo |= uflo >> 1; \
			uflo |= uflo >> 2; \
			uflo |= uflo >> 4; \
			sum &= ~uflo; \
		} \
	} \
}
/*
 * Table of products of 8-bit scaled coefficients
 * and idct coefficients (there are only 33 unique
 * coefficients so we index via a compact ID).
 */
extern "C" u_char multab[];
/*
 * Array of coefficient ID's used to index multab.
 */
extern "C" u_int dct_basis[64][64 / sizeof(u_int)];

/*XXX*/
#define LIMIT_512(s) ((s) > 511 ? 511 : (s) < -512 ? -512 : (s))

void
bv_rdct1(int dc, short* bp, int acx, u_char* out, int stride)
{
	u_int omask = 0x80808080;
	u_int uflo;
	u_int* vp = dct_basis[acx];
	int s = LIMIT_512(bp[acx]);
	s = (s >> 2) & 0xff;
	/* 66 unique coefficients require 7 bits */
	char* mt = (char*)&multab[s << 7];

	dc |= dc << 8;
	dc |= dc << 16;
	for (int k = 8; --k >= 0; ) {
		u_int v = *vp++;
		u_int m = mt[v >> 24] << SHIFT(24) |
			mt[v >> 16 & 0xff] << SHIFT(16) |
			mt[v >> 8 & 0xff] << SHIFT(8) |
			mt[v & 0xff] << SHIFT(0);
		MIX_LOGIC(v, dc, m, omask, uflo);
		*(u_int*)out = v;
		v = *vp++;
		m = mt[v >> 24] << SHIFT(24) |
			mt[v >> 16 & 0xff] << SHIFT(16) |
			mt[v >> 8 & 0xff] << SHIFT(8) |
			mt[v & 0xff] << SHIFT(0);
		MIX_LOGIC(v, dc, m, omask, uflo);
		*(u_int*)(out + 4) = v;
		out += stride;
	}
}

/* XXX this version has to be exact */
void
bv_rdct2(int dc, short* bp, int ac0, u_char* in, u_char* out, int stride)
{
	int s0 = LIMIT_512(bp[ac0]);
	s0 = (s0 >> 2) & 0xff;
	/* 66 unique coefficients require 7 bits */
	const char* mt = (const char*)&multab[s0 << 7];
	const u_int* vp0 = dct_basis[ac0];

	dc |= dc << 8;
	dc |= dc << 16;
	u_int omask = 0x80808080;
	u_int uflo;
	for (int k = 8; --k >= 0; ) {
		u_int v, m, i;
		v = *vp0++;
		m = mt[v >> 24] << SHIFT(24) | mt[v >> 16 & 0xff] << SHIFT(16) |
		    mt[v >> 8 & 0xff] << SHIFT(8) | mt[v & 0xff] << SHIFT(0);
		MIX_LOGIC(v, dc, m, omask, uflo);
		i = in[0] << SHIFT(24) | in[1] << SHIFT(16) |
		    in[2] << SHIFT(8) | in[3] << SHIFT(0);
		MIX_LOGIC(m, i, v, omask, uflo);
		*(u_int*)out = m;

		v = *vp0++;
		m = mt[v >> 24] << SHIFT(24) | mt[v >> 16 & 0xff] << SHIFT(16) |
		    mt[v >> 8 & 0xff] << SHIFT(8) | mt[v & 0xff] << SHIFT(0);
		MIX_LOGIC(v, dc, m, omask, uflo);
		i = in[4] << SHIFT(24) | in[5] << SHIFT(16) |
		    in[6] << SHIFT(8) | in[7] << SHIFT(0);
		MIX_LOGIC(m, i, v, omask, uflo);
		*(u_int*)(out + 4) = m;
		out += stride;
		in += stride;
	}
}

/* XXX this version has to be exact */
void
bv_rdct3(int dc, short* bp, int ac0, int ac1, u_char* in, u_char* out, int stride)
{
	int s0 = LIMIT_512(bp[ac0]);
	s0 = (s0 >> 2) & 0xff;
	/* 66 unique coefficients require 7 bits */
	char* mt0 = (char*)&multab[s0 << 7];

	int s1 = LIMIT_512(bp[ac1]);
	s1 = (s1 >> 2) & 0xff;
	char* mt1 = (char*)&multab[s1 << 7];

	u_int* vp0 = dct_basis[ac0];
	u_int* vp1 = dct_basis[ac1];

	for (int k = 8; --k >= 0; ) {
		int t;
		u_int v0 = *vp0++;
		u_int v1 = *vp1++;
		s0 = mt0[v0 >> 24] + mt1[v1 >> 24] + in[0] + dc;
		u_int m = LIMIT(s0, t) << SHIFT(24);
		s0 = mt0[v0 >> 16 & 0xff] + mt1[v1 >> 16 & 0xff] + in[1] + dc;
		m |= LIMIT(s0, t) << SHIFT(16);
		s0 = mt0[v0 >> 8 & 0xff] + mt1[v1 >> 8 & 0xff] + in[2] + dc;
		m |= LIMIT(s0, t) << SHIFT(8);
		s0 = mt0[v0 & 0xff] + mt1[v1 & 0xff] + in[3] + dc;
		m |= LIMIT(s0, t) << SHIFT(0);
		*(u_int*)out = m;

		v0 = *vp0++;
		v1 = *vp1++;
		s0 = mt0[v0 >> 24] + mt1[v1 >> 24] + in[4] + dc;
		m = 0;
		m |= LIMIT(s0, t) << SHIFT(24);
		s0 = mt0[v0 >> 16 & 0xff] + mt1[v1 >> 16 & 0xff] + in[5] + dc;
		m |= LIMIT(s0, t) << SHIFT(16);
		s0 = mt0[v0 >> 8 & 0xff] + mt1[v1 >> 8 & 0xff] + in[6] + dc;
		m |= LIMIT(s0, t) << SHIFT(8);
		s0 = mt0[v0 & 0xff] + mt1[v1 & 0xff] + in[7] + dc;
		m |= LIMIT(s0, t) << SHIFT(0);
		*(u_int*)(out + 4) = m;

		out += stride;
		in += stride;
	}
}

/*
 * A 2D Inverse DCT based on a column-row decomposition using
 * Arai, Agui, and Nakajmia's 8pt 1D Inverse DCT, from Fig. 4-8
 * Pennebaker & Mitchell (i.e., the pink JPEG book).  This figure
 * is the forward transform; reverse the flowgraph for the inverse
 * (you need to draw a picture).  The outputs are DFT coefficients
 * and need to be scaled according to Eq [4-3].
 *
 * The input coefficients are, counter to tradition, in column-order.
 * The bit mask indicates which coefficients are non-zero.  If the
 * corresponding bit is zero, then the coefficient is assumed zero
 * and the input coefficient is not referenced and need not be defined.
 * The 8-bit outputs are computed in row order and placed in the
 * output array pointed to by p, with each of the eight 8-byte lines
 * offset by "stride" bytes.
 *
 * qt is the inverse quantization table in column order.  These
 * coefficients are the product of the inverse quantization factor,
 * specified by the jpeg quantization table, and the first multiplier
 * in the inverse DCT flow graph.  This first multiplier is actually
 * biased by the row term so that the columns are pre-scaled to
 * eliminate the first row multiplication stage.
 *
 * The output is biased by 128, i.e., [-128,127] is mapped to [0,255],
 * which is relevant to jpeg.
 */
void
#ifdef INT_64
rdct(register short *bp, INT_64 m0, u_char* p, int stride, const int* qt)
#else
rdct(register short *bp, u_int m0, u_int m1, u_char* p,
     int stride, const int* qt)
#endif
{
	/*
	 * First pass is 1D transform over the columns.  Note that
	 * the coefficients are stored in column order, so even
	 * though we're accessing the columns, we access them
	 * in a row-like fashion.  We use a column-row decomposition
	 * instead of a row-column decomposition so we can splice
	 * pixels in an efficient, row-wise order in the second
	 * stage.
	 *
	 * The inverse quantization is folded together with the
	 * first stage multiplier.  This reduces the number of
	 * multiplies (per 8-pt transform) from 11 to 5 (ignoring
	 * the inverse quantization which must be done anyway).
	 *
	 * Because we compute appropriately scaled column DCTs,
	 * the row DCTs require only 5 multiplies per row total.
	 * So the total number of multiplies for the 8x8 DCT is 80
	 * (ignoring the checks for zero coefficients).
	 */
	int tmp[64];
	int* tp = tmp;
	int i;
	for (i = 8; --i >= 0; ) {
		if ((m0 & 0xfe) == 0) {
			/* AC terms all zero */
			int v = M(0) ? qt[0] * bp[0] : 0;
			tp[0] = v;
			tp[1] = v;
			tp[2] = v;
			tp[3] = v;
			tp[4] = v;
			tp[5] = v;
			tp[6] = v;
			tp[7] = v;
		} else {
			int t0, t1, t2, t3, t4, t5, t6, t7;

			if ((m0 & 0xaa) == 0)
				t4 = t5 = t6 = t7 = 0;
			else {
				t0 = M(5) ? qt[5] * bp[5] : 0;
				t2 = M(1) ? qt[1] * bp[1] : 0;
				t6 = M(7) ? qt[7] * bp[7] : 0;
				t7 = M(3) ? qt[3] * bp[3] : 0;

				t4 = t0 - t7;
				t1 = t2 + t6;
				t6 = t2 - t6;
				t7 += t0;

				t5 = t1 - t7;
				t7 += t1;

				t2 = FP_MUL(-A5, t4 + t6);
				t4 = FP_MUL(-A2, t4);
				t0 = t4 + t2;
				t1 = FP_MUL(A3, t5);
				t2 += FP_MUL(A4, t6);
			
				t4 = -t0;
				t5 = t1 - t0;
				t6 = t1 + t2;
				t7 += t2;
			}

#ifdef notdef
			if ((m0 & 0x55) == 0)
				t0 = t1 = t2 = t3 = 0;
			else {
#endif
				t0 = M(0) ? qt[0] * bp[0] : 0;
				t1 = M(4) ? qt[4] * bp[4] : 0;
				int x0 = t0 + t1;
				int x1 = t0 - t1;

				t0 = M(2) ? qt[2] * bp[2] : 0;
				t3 = M(6) ? qt[6] * bp[6] : 0;
				t2 = t0 - t3;
				t3 += t0;

				t2 = FP_MUL(A1, t2);
				t3 += t2;

				t0 = x0 + t3;
				t1 = x1 + t2;
				t2 = x1 - t2;
				t3 = x0 - t3;
#ifdef notdef
			}
#endif

			tp[0] = t0 + t7;
			tp[7] = t0 - t7;
			tp[1] = t1 + t6;
			tp[6] = t1 - t6;
			tp[2] = t2 + t5;
			tp[5] = t2 - t5;
			tp[3] = t3 + t4;
			tp[4] = t3 - t4;
		}
		tp += 8;
		bp += 8;
		qt += 8;

		m0 >>= 8;
#ifndef INT_64
		m0 |= m1 << 24;
		m1 >>= 8;
#endif
	}
	tp -= 64;
	/*
	 * Second pass is 1D transform over the rows.  Note that
	 * the coefficients are stored in column order, so even
	 * though we're accessing the rows, we access them
	 * in a column-like fashion.
	 *
	 * The pass above already computed the first multiplier
	 * in the flow graph.
	 */
	for (i = 8; --i >= 0; ) {
		int t0, t1, t2, t3, t4, t5, t6, t7;

		t0 = tp[8*5];
		t2 = tp[8*1];
		t6 = tp[8*7];
		t7 = tp[8*3];

#ifdef notdef
		if ((t0 | t2 | t6 | t7) == 0) {
			t4 = t5 = 0;
		} else 
#endif
{
			t4 = t0 - t7;
			t1 = t2 + t6;
			t6 = t2 - t6;
			t7 += t0;

			t5 = t1 - t7;
			t7 += t1;

			t2 = FP_MUL(-A5, t4 + t6);
			t4 = FP_MUL(-A2, t4);
			t0 = t4 + t2;
			t1 = FP_MUL(A3, t5);
			t2 += FP_MUL(A4, t6);
			
			t4 = -t0;
			t5 = t1 - t0;
			t6 = t1 + t2;
			t7 += t2;
		}

		t0 = tp[8*0];
		t1 = tp[8*4];
		int x0 = t0 + t1;
		int x1 = t0 - t1;

		t0 = tp[8*2];
		t3 = tp[8*6];
		t2 = t0 - t3;
		t3 += t0;

		t2 = FP_MUL(A1, t2);
		t3 += t2;

		t0 = x0 + t3;
		t1 = x1 + t2;
		t2 = x1 - t2;
		t3 = x0 - t3;

#ifdef INT_64
		/*XXX assume little-endian */
		u_int v = FP_NORM(t0 + t7) + 128;
		int oflo = v;
		INT_64 pix = v;
		v = FP_NORM(t1 + t6) + 128;
		oflo |= v;
		pix |= (INT_64)v << 8;
		v = FP_NORM(t2 + t5) + 128;
		oflo |= v;
		pix |= (INT_64)v << 16;
		v = FP_NORM(t3 + t4) + 128;
		oflo |= v;
		pix |= (INT_64)v << 24;
		v = FP_NORM(t3 - t4) + 128;
		oflo |= v;
		pix |= (INT_64)v << 32;
		v = FP_NORM(t2 - t5) + 128;
		oflo |= v;
		pix |= (INT_64)v << 40;
		v = FP_NORM(t1 - t6) + 128;
		oflo |= v;
		pix |= (INT_64)v << 48;
		v = FP_NORM(t0 - t7) + 128;
		oflo |= v;
		pix |= (INT_64)v << 56;
		if (oflo & ~0xff) {
			/*
			 * Spend more cycles to handle overflow.
			 * This rarely happens, and would never happen
			 * if quantization tables and DCT implementations
			 * were exactly matched, but this is not utopia.
			 */
			v = FP_NORM(t0 + t7) + 128;
			pix = LIMIT(v, x0);
			v = FP_NORM(t1 + t6) + 128;
			pix |= (INT_64)LIMIT(v, x0) << 8;
			v = FP_NORM(t2 + t5) + 128;
			pix |= (INT_64)LIMIT(v, x0) << 16;
			v = FP_NORM(t3 + t4) + 128;
			pix |= (INT_64)LIMIT(v, x0) << 24;
			v = FP_NORM(t3 - t4) + 128;
			pix |= (INT_64)LIMIT(v, x0) << 32;
			v = FP_NORM(t2 - t5) + 128;
			pix |= (INT_64)LIMIT(v, x0) << 40;
			v = FP_NORM(t1 - t6) + 128;
			pix |= (INT_64)LIMIT(v, x0) << 48;
			v = FP_NORM(t0 - t7) + 128;
			pix |= (INT_64)LIMIT(v, x0) << 56;
		} 
		*(INT_64*)p = pix;
#else
		u_int v = FP_NORM(t0 + t7) + 128;
		int oflo = v;
		u_int pix = 0;
		SPLICE(pix, v, 24);
		v = FP_NORM(t1 + t6) + 128;
		oflo |= v;
		SPLICE(pix, v, 16);
		v = FP_NORM(t2 + t5) + 128;
		oflo |= v;
		SPLICE(pix, v, 8);
		v = FP_NORM(t3 + t4) + 128;
		oflo |= v;
		SPLICE(pix, v, 0);
		if (oflo & ~0xff) {
			/*
			 * Spend more cycles to handle overflow.
			 * This rarely happens, and would never happen
			 * if quantization tables and DCT implementations
			 * were exactly matched, but this is not utopia.
			 */
			v = FP_NORM(t0 + t7) + 128;
			v = LIMIT(v, x0);
			pix = 0;
			SPLICE(pix, v, 24);
			v = FP_NORM(t1 + t6) + 128;
			v = LIMIT(v, x0);
			SPLICE(pix, v, 16);
			v = FP_NORM(t2 + t5) + 128;
			v = LIMIT(v, x0);
			SPLICE(pix, v, 8);
			v = FP_NORM(t3 + t4) + 128;
			v = LIMIT(v, x0);
			SPLICE(pix, v, 0);
		}
		*(u_int*)p = pix;

		v = FP_NORM(t3 - t4) + 128;
		oflo = v;
		pix = 0;
		SPLICE(pix, v, 24);
		v = FP_NORM(t2 - t5) + 128;
		oflo |= v;
		SPLICE(pix, v, 16);
		v = FP_NORM(t1 - t6) + 128;
		oflo |= v;
		SPLICE(pix, v, 8);
		v = FP_NORM(t0 - t7) + 128;
		oflo |= v;
		SPLICE(pix, v, 0);
		if (oflo & ~0xff) {
			v = FP_NORM(t3 - t4) + 128;
			v = LIMIT(v, x0);
			pix = 0;
			SPLICE(pix, v, 24);
			v = FP_NORM(t2 - t5) + 128;
			v = LIMIT(v, x0);
			SPLICE(pix, v, 16);
			v = FP_NORM(t1 - t6) + 128;
			v = LIMIT(v, x0);
			SPLICE(pix, v, 8);
			v = FP_NORM(t0 - t7) + 128;
			v = LIMIT(v, x0);
			SPLICE(pix, v, 0);
		}
		*(u_int*)(p + 4) = pix;
#endif
		++tp;
		p += stride;
	}
}

/*
 * Inverse 2-D transform, similar to routine above (see comment above),
 * but more appropriate for H.261 instead of JPEG.  This routine does
 * not bias the output by 128, and has an additional argument which is
 * an input array which gets summed together with the inverse-transform.
 * For example, this allows motion-compensation to be folded in here,
 * saving an extra traversal of the block.  The input pointer can be
 * null, if motion-compensation is not needed.
 *
 * This routine does not take a quantization table, since the H.261
 * inverse quantizer cannot be carried out by a simple multiply
 * (and easily implemented via table lookup in the decoder anyway).
 */
void
#ifdef INT_64
rdct(register short *bp, INT_64 m0, u_char* p, int stride, const u_char* in)
#else
rdct(register short *bp, u_int m0, u_int m1, u_char* p, int stride, const u_char *in)
#endif
{
	int tmp[64];
	int* tp = tmp;
	int* qt = cross_stage;
	int i;
	for (i = 8; --i >= 0; ) {
		if ((m0 & 0xfe) == 0) {
			/*
			 * All ac terms are zero.
			 */
			int v = M(0) ? qt[0] * bp[0] : 0;
			tp[0] = v;
			tp[1] = v;
			tp[2] = v;
			tp[3] = v;
			tp[4] = v;
			tp[5] = v;
			tp[6] = v;
			tp[7] = v;
		} else {
			int t0, t1, t2, t3, t4, t5, t6, t7;

			if ((m0 & 0xaa) == 0)
				t4 = t5 = t6 = t7 = 0;
			else {
				t0 = M(5) ? qt[5] * bp[5] : 0;
				t2 = M(1) ? qt[1] * bp[1] : 0;
				t6 = M(7) ? qt[7] * bp[7] : 0;
				t7 = M(3) ? qt[3] * bp[3] : 0;

				t4 = t0 - t7;
				t1 = t2 + t6;
				t6 = t2 - t6;
				t7 += t0;

				t5 = t1 - t7;
				t7 += t1;

				t2 = FP_MUL(-A5, t4 + t6);
				t4 = FP_MUL(-A2, t4);
				t0 = t4 + t2;
				t1 = FP_MUL(A3, t5);
				t2 += FP_MUL(A4, t6);
			
				t4 = -t0;
				t5 = t1 - t0;
				t6 = t1 + t2;
				t7 += t2;
			}

			if ((m0 & 0x55) == 0)
				t0 = t1 = t2 = t3 = 0;
			else {
				t0 = M(0) ? qt[0] * bp[0] : 0;
				t1 = M(4) ? qt[4] * bp[4] : 0;
				int x0 = t0 + t1;
				int x1 = t0 - t1;

				t0 = M(2) ? qt[2] * bp[2] : 0;
				t3 = M(6) ? qt[6] * bp[6] : 0;
				t2 = t0 - t3;
				t3 += t0;

				t2 = FP_MUL(A1, t2);
				t3 += t2;

				t0 = x0 + t3;
				t1 = x1 + t2;
				t2 = x1 - t2;
				t3 = x0 - t3;
			}

			tp[0] = t0 + t7;
			tp[7] = t0 - t7;
			tp[1] = t1 + t6;
			tp[6] = t1 - t6;
			tp[2] = t2 + t5;
			tp[5] = t2 - t5;
			tp[3] = t3 + t4;
			tp[4] = t3 - t4;
		}
		qt += 8;
		tp += 8;
		bp += 8;
		m0 >>= 8;
#ifndef INT_64
		m0 |= m1 << 24;
		m1 >>= 8;
#endif
	}
	tp -= 64;
	/*
	 * Second pass is 1D transform over the rows.  Note that
	 * the coefficients are stored in column order, so even
	 * though we're accessing the rows, we access them
	 * in a column-like fashion.
	 */
	for (i = 8; --i >= 0; ) {
		int t0, t1, t2, t3, t4, t5, t6, t7;

		t0 = tp[8*5];
		t2 = tp[8*1];
		t6 = tp[8*7];
		t7 = tp[8*3];

		if ((t0 | t2 | t6 | t7) == 0) {
			t4 = t5 = 0;
		} else 
		{
			t4 = t0 - t7;
			t1 = t2 + t6;
			t6 = t2 - t6;
			t7 += t0;

			t5 = t1 - t7;
			t7 += t1;

			t2 = FP_MUL(-A5, t4 + t6);
			t4 = FP_MUL(-A2, t4);
			t0 = t4 + t2;
			t1 = FP_MUL(A3, t5);
			t2 += FP_MUL(A4, t6);
			
			t4 = -t0;
			t5 = t1 - t0;
			t6 = t1 + t2;
			t7 += t2;
		}

		t0 = tp[8*0];
		t1 = tp[8*4];
		int x0 = t0 + t1;
		int x1 = t0 - t1;

		t0 = tp[8*2];
		t3 = tp[8*6];
		t2 = t0 - t3;
		t3 += t0;

		t2 = FP_MUL(A1, t2);
		t3 += t2;

		t0 = x0 + t3;
		t1 = x1 + t2;
		t2 = x1 - t2;
		t3 = x0 - t3;

		if (in != 0) {
#ifdef INT_64
			/*XXX assume little-endian */
			u_int v = FP_NORM(t0 + t7) + in[0];
			int oflo = v;
			INT_64 pix = v;
			v = FP_NORM(t1 + t6) + in[1];
			oflo |= v;
			pix |= (INT_64)v << 8;
			v = FP_NORM(t2 + t5) + in[2];
			oflo |= v;
			pix |= (INT_64)v << 16;
			v = FP_NORM(t3 + t4) + in[3];
			oflo |= v;
			pix |= (INT_64)v << 24;
			v = FP_NORM(t3 - t4) + in[4];
			oflo |= v;
			pix |= (INT_64)v << 32;
			v = FP_NORM(t2 - t5) + in[5];
			oflo |= v;
			pix |= (INT_64)v << 40;
			v = FP_NORM(t1 - t6) + in[6];
			oflo |= v;
			pix |= (INT_64)v << 48;
			v = FP_NORM(t0 - t7) + in[7];
			oflo |= v;
			pix |= (INT_64)v << 56;
			if (oflo & ~0xff) {
				v = FP_NORM(t0 + t7);
				pix = LIMIT(v + in[0], t3);
				v = FP_NORM(t1 + t6);
				pix |= (INT_64)LIMIT(v + in[1], t3) << 8;
				v = FP_NORM(t2 + t5);
				pix |= (INT_64)LIMIT(v + in[2], t3) << 16;
				v = FP_NORM(t3 + t4);
				pix |= (INT_64)LIMIT(v + in[3], t3) << 24;
				v = FP_NORM(t3 - t4);
				pix |= (INT_64)LIMIT(v + in[4], t3) << 32;
				v = FP_NORM(t2 - t5);
				pix |= (INT_64)LIMIT(v + in[5], t3) << 40;
				v = FP_NORM(t1 - t6);
				pix |= (INT_64)LIMIT(v + in[6], t3) << 48;
				v = FP_NORM(t0 - t7);
				pix |= (INT_64)LIMIT(v + in[7], t3) << 56;
			}
			*(INT_64*)p = pix;
#else
			u_int v = FP_NORM(t0 + t7) + in[0];
			int oflo = v;
			u_int pix = 0;
			SPLICE(pix, v, 24);
			v = FP_NORM(t1 + t6) + in[1];
			oflo |= v;
			SPLICE(pix, v, 16);
			v = FP_NORM(t2 + t5) + in[2];
			oflo |= v;
			SPLICE(pix, v, 8);
			v = FP_NORM(t3 + t4) + in[3];
			oflo |= v;
			SPLICE(pix, v, 0);
			if (oflo & ~0xff) {
				v = FP_NORM(t0 + t7) + in[0];
				v = LIMIT(v, x0);
				pix = 0;
				SPLICE(pix, v, 24);
				v = FP_NORM(t1 + t6) + in[1];
				v = LIMIT(v, x0);
				SPLICE(pix, v, 16);
				v = FP_NORM(t2 + t5) + in[2];
				v = LIMIT(v, x0);
				SPLICE(pix, v, 8);
				v = FP_NORM(t3 + t4) + in[3];
				v = LIMIT(v, x0);
				SPLICE(pix, v, 0);
			}
			*(u_int*)p = pix;

			v = FP_NORM(t3 - t4) + in[4];
			oflo = v;
			pix = 0;
			SPLICE(pix, v, 24);
			v = FP_NORM(t2 - t5) + in[5];
			oflo |= v;
			SPLICE(pix, v, 16);
			v = FP_NORM(t1 - t6) + in[6];
			oflo |= v;
			SPLICE(pix, v, 8);
			v = FP_NORM(t0 - t7) + in[7];
			oflo |= v;
			SPLICE(pix, v, 0);
			if (oflo & ~0xff) {
				v = FP_NORM(t3 - t4) + in[4];
				v = LIMIT(v, x0);
				pix = 0;
				SPLICE(pix, v, 24);
				v = FP_NORM(t2 - t5) + in[5];
				v = LIMIT(v, x0);
				SPLICE(pix, v, 16);
				v = FP_NORM(t1 - t6) + in[6];
				v = LIMIT(v, x0);
				SPLICE(pix, v, 8);
				v = FP_NORM(t0 - t7) + in[7];
				v = LIMIT(v, x0);
				SPLICE(pix, v, 0);
			}
			*(u_int*)(p + 4) = pix;
#endif
			in += stride;
		} else {
#ifdef INT_64
			/*XXX assume little-endian */
			u_int v = FP_NORM(t0 + t7);
			u_int oflo = v;
			INT_64 pix = v;
			v = FP_NORM(t1 + t6);
			oflo |= v;
			pix |= (INT_64)v << 8;
			v = FP_NORM(t2 + t5);
			oflo |= v;
			pix |= (INT_64)v << 16;
			v = FP_NORM(t3 + t4);
			oflo |= v;
			pix |= (INT_64)v << 24;
			v = FP_NORM(t3 - t4);
			oflo |= v;
			pix |= (INT_64)v << 32;
			v = FP_NORM(t2 - t5);
			oflo |= v;
			pix |= (INT_64)v << 40;
			v = FP_NORM(t1 - t6);
			oflo |= v;
			pix |= (INT_64)v << 48;
			v = FP_NORM(t0 - t7);
			oflo |= v;
			pix |= (INT_64)v << 56;
			if (oflo & ~0xff) {
				v = FP_NORM(t0 + t7);
				v = LIMIT(v, x0);
				pix = v;
				v = FP_NORM(t1 + t6);
				v = LIMIT(v, x0);
				pix |= (INT_64)v << 8;
				v = FP_NORM(t2 + t5);
				v = LIMIT(v, x0);
				pix |= (INT_64)v << 16;
				v = FP_NORM(t3 + t4);
				v = LIMIT(v, x0);
				pix |= (INT_64)v << 24;
				v = FP_NORM(t3 - t4);
				v = LIMIT(v, x0);
				pix |= (INT_64)v << 32;
				v = FP_NORM(t2 - t5);
				v = LIMIT(v, x0);
				pix |= (INT_64)v << 40;
				v = FP_NORM(t1 - t6);
				v = LIMIT(v, x0);
				pix |= (INT_64)v << 48;
				v = FP_NORM(t0 - t7);
				v = LIMIT(v, x0);
				pix |= (INT_64)v << 56;
			}
			*(INT_64*)p = pix;
#else
			u_int v = FP_NORM(t0 + t7);
			int oflo = v;
			u_int pix = 0;
			SPLICE(pix, v, 24);
			v = FP_NORM(t1 + t6);
			oflo |= v;
			SPLICE(pix, v, 16);
			v = FP_NORM(t2 + t5);
			oflo |= v;
			SPLICE(pix, v, 8);
			v = FP_NORM(t3 + t4);
			oflo |= v;
			SPLICE(pix, v, 0);
			if (oflo & ~0xff) {
				v = FP_NORM(t0 + t7);
				v = LIMIT(v, x0);
				pix = 0;
				SPLICE(pix, v, 24);
				v = FP_NORM(t1 + t6);
				v = LIMIT(v, x0);
				SPLICE(pix, v, 16);
				v = FP_NORM(t2 + t5);
				v = LIMIT(v, x0);
				SPLICE(pix, v, 8);
				v = FP_NORM(t3 + t4);
				v = LIMIT(v, x0);
				SPLICE(pix, v, 0);
			}
			*(u_int*)p = pix;

			v = FP_NORM(t3 - t4);
			oflo = v;
			pix = 0;
			SPLICE(pix, v, 24);
			v = FP_NORM(t2 - t5);
			oflo |= v;
			SPLICE(pix, v, 16);
			v = FP_NORM(t1 - t6);
			oflo |= v;
			SPLICE(pix, v, 8);
			v = FP_NORM(t0 - t7);
			oflo |= v;
			SPLICE(pix, v, 0);
			if (oflo & ~0xff) {
				v = FP_NORM(t3 - t4);
				v = LIMIT(v, x0);
				pix = 0;
				SPLICE(pix, v, 24);
				v = FP_NORM(t2 - t5);
				v = LIMIT(v, x0);
				SPLICE(pix, v, 16);
				v = FP_NORM(t1 - t6);
				v = LIMIT(v, x0);
				SPLICE(pix, v, 8);
				v = FP_NORM(t0 - t7);
				v = LIMIT(v, x0);
				SPLICE(pix, v, 0);
			}
			*(u_int*)(p + 4) = pix;
#endif
		}
		++tp;
		p += stride;
	}
}

void fdct(const u_char* in, int stride, short* out, int* qt, int* qinv)
{
	int tmp[64];
	int* tp = tmp;
	int i;
	for (i = 8; --i >= 0; ) {
		int x0 = in[0];
		int t7 = in[7];
		int t0 = x0 + t7;
		t7 = x0 - t7;
		x0 = in[1];
		int t6 = in[6];
		int t1 = x0 + t6;
		t6 = x0 - t6;
		x0 = in[2];
		int t5 = in[5];
		int t2 = x0 + t5;
		t5 = x0 - t5;
		x0 = in[3];
		int t4 = in[4];
		int t3 = x0 + t4;
		t4 = x0 - t4;

		t4 += t5;
		t5 += t6;
		t6 += t7;

		x0 = t6 - t4;

		x0 *= -A5;
		t4 *= A2;
		t4 += x0;
		t6 *= A4;
		t6 += x0;
		x0 = A3 * t5;
		t7 <<= FP_NBITS;

		t5 = x0 + t7;
		t7 -= x0;

		x0 = t0 + t3;
		t3 = t0 - t3;
		t0 = t1 + t2;
		t2 = t1 - t2;

		t2 += t3;
		t2 *= A1;
		t3 <<= FP_NBITS;

		tp[0] = (x0 + t0) << FP_NBITS;
		tp[1] = t5 + t6;
		tp[2] = t3 + t2;
		tp[3] = t7 - t4;
		tp[4] = (x0 - t0) << FP_NBITS;
		tp[5] = t7 + t4;
		tp[6] = t3 - t2;
		tp[7] = t5 - t6;
    
		in += stride;
		tp += 8;
	}
	tp -= 64;

	for (i = 8; --i >= 0; ) {
		int x0 = tp[8*0];
		int t7 = tp[8*7];
		int t0 = x0 + t7;
		t7 = x0 - t7;
		x0 = tp[8*1];
		int t6 = tp[8*6];
		int t1 = x0 + t6;
		t6 = x0 - t6;
		x0 = tp[8*2];
		int t5 = tp[8*5];
		int t2 = x0 + t5;
		t5 = x0 - t5;
		x0 = tp[8*3];
		int t4 = tp[8*4];
		int t3 = x0 + t4;
		t4 = x0 - t4;

		t4 += t5;
		t5 += t6;
		t6 += t7;

		x0 = t6 - t4;

		x0 = FP_MUL(-A5, x0);
		t4 = FP_MUL(A2, t4);
		t4 += x0;
		t6 = FP_MUL(A4, t6);
		t6 += x0;
		x0 = FP_MUL(A3, t5);

		t5 = x0 + t7;
		t7 -= x0;

		x0 = t0 + t3;
		t3 = t0 - t3;
		t0 = t1 + t2;
		t2 = t1 - t2;

		t2 += t3;
		t2 = FP_MUL(A1, t2);

		/*
		 * This macro does the combined descale-and-quantize
		 * multiply.  It special cases coefs. smaller than
		 * the quantization level since they will result in 0.
		 * Use NORMT macro to do truncation toward zero instead
		 * of rounding.  This makes the H.261 dead-band quantizer
		 * work right.
		 */
#define FWD_DandQ(v, tq, iq, t, q) \
	  (t=(v), q = (iq), (t < q && -t < q)? 0 : FP_NORMT(FP_MUL2(t, tq)))

		register int t, q;
		out[0] = FWD_DandQ(x0 + t0, qt[0], qinv[0], t, q);
		out[1] = FWD_DandQ(t5 + t6, qt[1], qinv[1], t, q);
		out[2] = FWD_DandQ(t3 + t2, qt[2], qinv[2], t, q);
		out[3] = FWD_DandQ(t7 - t4, qt[3], qinv[3], t, q);
		out[4] = FWD_DandQ(x0 - t0, qt[4], qinv[4], t, q);
		out[5] = FWD_DandQ(t7 + t4, qt[5], qinv[5], t, q);
		out[6] = FWD_DandQ(t3 - t2, qt[6], qinv[6], t, q);
		out[7] = FWD_DandQ(t5 - t6, qt[7], qinv[7], t, q);

		out += 8;
		++tp;
		qt += 8;
	}
}
