/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/decoder-h261.C,v 1.4 1995/08/26 21:15:46 olson Exp (LBL)";

#include <osfcn.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <netinet/in.h>
#include "inet.h"
#include "packet.h"
#include "video.h"
#include "decoder-h261.h"
#include "color.h"
#include "renderer.h"

#ifdef VIC_ANALOG
extern int analog_yuv_init(long width, long height);
extern void analog_yuv_config(long width, long height);
extern void analog_yuv_out(const u_char *yp, const u_char *up,
			   const u_char *vp,
			   long width, long height);
extern void analog_yuv_close();
#endif

u_char* H261Decoder::wrkbuf_;
int H261Decoder::wrkbuflen_;

H261Decoder::H261Decoder(const u_char* pkthdr)
	: SoftwareDecoder(2), codec_(0)
{
	int flags = htons(*(const u_short*)pkthdr);
	fmt_ = flags & 7;
	if (fmt_ != IT_QCIF && fmt_ != IT_CIF) {
		valid_ = 0;
		return;
	}
	duplicates_ = 0;
	fragsizechange_ = 0;
	badbitoffset_ = 0;
	badsize_ = 0;
	fmtchange_ = 0;
	seqno_ = 0;

	codec_ = new P64Decoder(fmt_);
	codec_->renderer(this);
	inw_ = codec_->width();
	inh_ = codec_->height();
	packetsize_ = 0;
	for (int i = 0; i < H261_SLOTS; ++i) {
		slot_[i].len = 0;
		slot_[i].flags = 0;
		slot_[i].seqno = 0;
		slot_[i].bp = 0;
	}
}

H261Decoder::~H261Decoder()
{
	delete codec_;
	delete slot_[0].bp;
}

void H261Decoder::counter_names(char* wrk) const
{
	strcpy(wrk, "Bad-PSC Bad-GOB Bad-syntax Bad-fmt Bad-size Bad-bitoff \
Dups Size-change Fmt-change");
}

int H261Decoder::counter_query(const char* name, u_int& v) const
{
	if (strcasecmp(name, "bad-psc") == 0)
		v = codec_->bad_psc();
	else if (strcasecmp(name, "bad-GOB") == 0)
		v = codec_->bad_GOBno();
	else if (strcasecmp(name, "bad-syntax") == 0)
		v = codec_->bad_bits();
	else if (strcasecmp(name, "bad-fmt") == 0)
		v = codec_->bad_fmt();
	else if (strcasecmp(name, "bad-size") == 0)
		v = badsize_;
	else if (strcasecmp(name, "bad-bitoff") == 0)
		v = badbitoffset_;
	else if (strcasecmp(name, "dups") == 0)
		v = duplicates_;
	else if (strcasecmp(name, "size-change") == 0)
		v = fragsizechange_;
	else if (strcasecmp(name, "fmt-change") == 0)
		v = fmtchange_;
	else
		return (-1);
	return (0);
}

int H261Decoder::colorhist(u_int* hist) const
{
	colorhist_411_556(hist, codec_->yplane(), codec_->uplane(),
			  codec_->vplane(), inw_, inh_);
	return (1);
}

/*
 * Packets waiting for reassembly are kept in 'slots'
 * indexed by the low bits of the sequence number.
 * To try to avoid copying packets twice, we assume
 * that the packet source is breaking packets at some
 * fixed, max size (which we can determine from the size
 * of a packet with the SOG marker) and we allocate the buffer 
 * for slots to be this size.  As long as a packet doesn't
 * wrap, this should make the final packets contiguous.
 */
inline u_char* H261Decoder::reassemble(u_char* bp, int& cc, u_int seqno,
				       u_int& rflags)
{
	register int flags = rflags;
	if (packetsize_ != cc) {
		if (packetsize_ == 0) {
			if ((flags & 0x80) == 0)
				return (0);
			packetsize_ = cc;
			u_int len = cc * H261_SLOTS;
			u_char* bp = new u_char[len];
			for (int i = 0; i < H261_SLOTS; ++i) {
				slot_[i].bp = bp;
				bp += cc;
			}
			if (len > wrkbuflen_) {
				delete wrkbuf_;
				wrkbuf_ = new u_char[len];
				wrkbuflen_ = len;
			}
		} else if ((flags & 0x08) == 0 || cc > packetsize_) {
			++fragsizechange_;
			delete slot_[0].bp;
			packetsize_ = 0;
			for (int i = 0; i < H261_SLOTS; ++i) {
				slot_[i].len = 0;
				slot_[i].flags = 0;
				slot_[i].seqno = 0;
				slot_[i].bp = 0;
			}
			return (0);
		}
	}
	register u_int l = seqno & H261_SLOTMASK;
	if (slot_[l].seqno == seqno) {
		/* got a duplicate */
		++duplicates_;
		return (0);
	}
	slot_[l].len = cc;
	slot_[l].seqno = seqno;
	slot_[l].flags = flags;
	memcpy((char*)slot_[l].bp, (char*)bp, cc);

	/*
	 * Don't bother trying to reassemble packet unless it
	 * has an EOG marker or we have the next packet.
	 */
	if ((flags & 0x08) == 0 &&
	    ((seqno + 1) & 0xffff) !=
	    slot_[(l + 1) & H261_SLOTMASK].seqno)
		return (0);

	/* scan backward for BOG */
	register u_int pseq = seqno;
	register u_int f = l;
	while ((slot_[f].flags & 0x80) == 0) {
		pseq = (pseq - 1) & 0xffff;
		f = pseq & H261_SLOTMASK;
		if (slot_[f].seqno != pseq)
			/* packet missing */
			return (0);
	}

	/* scan forward for EOG */
	while ((slot_[l].flags & 0x08) == 0) {
		seqno = (seqno + 1) & 0xffff;
		l = seqno & H261_SLOTMASK;
		if (slot_[l].seqno != seqno)
			/* packet missing */
			return (0);
	}

	if (l < f) {
		/* packet wraps around buffer, have to copy it */
		u_char* wp = wrkbuf_;
		register u_int i = f - 1;
		do {
			i = (i + 1) & H261_SLOTMASK;
			register int slen = slot_[i].len;
			memcpy((char*)wp, (char*)slot_[i].bp, slen);
			wp += slen;
		} while (i != l);
		bp = wrkbuf_;
		cc = wp - bp;
	} else {
		bp = slot_[f].bp;
		cc = slot_[l].bp - bp + slot_[l].len;
	}

	/*
	 * At this point, all we care about is ebit and sbit.
	 */
	flags = slot_[f].flags & 0x70;
	flags |= slot_[l].flags & 0x07;
	rflags = flags;

	return (bp);
}


void H261Decoder::render(int x, int y, int w, int h,
			 u_char* yp, u_char* up, u_char* vp)
{
if (x < 0 || y < 0 || x > (352 - 16) || y > (288 - 16))
 abort();
	x >>= 3;
	y >>= 3;
	rvts_[y << 7 | x] = now_;
	rvts_[y << 7 | x + 1] = now_;
	rvts_[y + 1 << 7 | x] = now_;
	rvts_[y + 1 << 7 | x + 1] = now_;
}

int H261Decoder::recv(rtphdr* rh, u_char* vh, u_char* bp, int cc, int jitter)
{	
	register u_int flags = ntohs(*(u_short*)vh);
	register u_int fmt = flags & 7;
	if (fmt != fmt_) {
		++fmtchange_;
		if (fmt > 1) {
			if (badsize_ == 0) {
				fprintf(stderr,
					"vic: unsupported H.261 format (%d)\n",
					fmt);
			}
			++badsize_;
			return (0);
		}
		/*
		 * Should probably try to avoid flip-flopping here
		 * when format change accompanies packet reordering.
		 */
		fmt_ = fmt;
		delete codec_;
		codec_ = new P64Decoder(fmt_);
		codec_->renderer(this);
		inw_ = codec_->width();
		inh_ = codec_->height();
		/*
		 * resize all the renderers
		 */
		for (SoftwareRenderer* p = engines_; p != 0; p = p->next_)
			p->resize(inw_, inh_, color_);
		render_frame(codec_->yplane(), codec_->uplane(),
			     codec_->vplane());
	}
	register u_int seqno = ntohs(rh->rh_seqno);
	/*
	 * if the packet contains both start & end gobs we can
	 * just render it.  Otherwise we wait until we get the
	 * start and end markers and all the packets in between.
	 */
	flags >>= 8;
	if ((flags & 0x88) != 0x88) {
		/*
		 * if the packet has a non-zero sbit but no SOG or
		 * a non-zero ebit but no EOG, toss it.
		 */
		if (((flags & 0x70) != 0 && (flags & 0x80) == 0) ||
		    ((flags & 0x07) != 0 && (flags & 0x08) == 0)) {
			++badbitoffset_;
			return (0);
		}
		u_int rflags = flags;
		bp = reassemble(bp, cc, seqno, rflags);
		if (bp == 0)
			return (0);
		flags = rflags;
	}
	int fwd = (((seqno - seqno_) << 16 >> 16) > 0) ? 1 : 0;
	seqno_ = seqno;
	if (codec_->decode(bp, cc, (flags >> 4) & 7, flags & 7, fwd)) {
#ifdef VIC_ANALOG
	    if (analog_)
	    {
		analog_yuv_out(codec_->yplane(), codec_->uplane(),
			       codec_->vplane(),
			       (long) inw_, (long) inh_);
	    }
#endif
		render_frame(codec_->yplane(), codec_->uplane(),
			     codec_->vplane());
		return (1);
	}
	return (0);
}

SoftwareRenderer* H261Decoder::create(VideoWindow* w)
{
	return (new SoftwareRenderer(w, inw_, inh_, color_, 411));
}

void H261Decoder::redraw(SoftwareRenderer *p)
{
	p->dither(0, 0, inw_, inh_, codec_->yplane(), codec_->uplane(),
		  codec_->vplane());
	p->render();
	p->purge();
}

void H261Decoder::toggle_analog()
{
#ifdef VIC_ANALOG
    analog_ = 1 - analog_;

    printf("%x::toggle_analog(): nv analog_=%d\n", this, analog_);

    if (analog_)
    {
	if (!analog_yuv_init((long) inw_, (long) inh_))
	{
	    analog_ = 0;
	}
    }
    else
    {
	analog_yuv_close();
    }
#endif
}

