/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef vic_decoder_h261_h
#define vic_decoder_h261_h

#include "decoder.h"
#include "p64/p64.h"

class SoftwareRenderer;

class H261Decoder : public SoftwareDecoder, public P64Renderer {
public:
	H261Decoder(const u_char* pkthdr);
	virtual ~H261Decoder();
	int colorhist(u_int* hist) const;

	virtual void counter_names(char*) const;
	virtual int counter_query(const char* name, u_int& value) const;
	virtual void toggle_analog();
protected:
	virtual void render(int x, int y, int w, int h,
			    u_char* yp, u_char* up, u_char* vp);
	virtual void redraw(SoftwareRenderer *p);
	virtual SoftwareRenderer* create(VideoWindow*);

	int recv(rtphdr*, u_char* vh, u_char* pkt, int len, int jitter);
	u_char* reassemble(u_char* bp, int& cc, u_int seqno, u_int& flags);

	u_int fmt_;		/* image type (CIF/QCIF) */

	P64Decoder* codec_;

	u_int duplicates_;	/* duplicate packets dropped */
	u_int badbitoffset_;	/* unexpected non-zero sbits or ebits */
	u_int badsize_;		/* packet dropped for unsupported fmt */
	u_int fragsizechange_;	/* packet dropped because frag size changed */
	u_int fmtchange_;	/* # times fmt changed */
	u_short seqno_;		/* last rtp seqno */
	/*
	 * Reassembly buffer 'slots' for resequencing & reassembling
	 * gobs split across packet boundaries.
	 */
#define H261_SLOTS 64
#define H261_SLOTMASK (H261_SLOTS - 1)
	static u_char* wrkbuf_;
	static int wrkbuflen_;
	int packetsize_;	/* max size of gob fragments */
	struct slot {
		int len;
		u_short flags;
		u_short seqno;
		u_char* bp;
		int pad;
	} slot_[H261_SLOTS];
};
#endif
