/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/decoder-jpeg.C,v 1.4 1995/08/25 17:38:55 olson Exp (LBL)";

#include <osfcn.h>
#include <stdio.h>
#include <string.h>
#include "inet.h"
#include "packet.h"
#include "video.h"
#include "decoder-jpeg.h"
#include "color.h"
#include "renderer.h"
#include "endian.h"
#include "Tcl.h"

#ifdef VIC_ANALOG
extern int analog_init(long width, long height, long q);
extern void analog_config(long width, long height, long q);
extern void analog_out(u_char *frame, int len);
extern void analog_close();
#endif

MotionJpegDecoder::MotionJpegDecoder(const u_char* vh)
	: SoftwareDecoder(sizeof(jpeghdr)), codec_(0)
{
#ifdef DEBUG
    printf("new mjdecoder %x\n", this);
#endif
	const jpeghdr* p = (const jpeghdr*)vh;
	inq_ = p->q;
	inw_ = p->width << 3;
	inh_ = p->height << 3;
	color_ = 1;
	type_ = p->type;
	/*XXX use p->type here; config should have JFIF sample factors */
	JpegDecoder::defaults(config_);
	config_.width = inw_;
	config_.height = inh_;
	JpegDecoder::quantizer(config_, inq_);
	/*XXX*/
	if (p->type == 1) {
		config_.comp[0].hsf = 2;
		config_.comp[0].vsf = 2;
		config_.comp[1].hsf = 1;
		config_.comp[1].vsf = 1;
		config_.comp[2].hsf = 1;
		config_.comp[2].vsf = 1;
	}
	Tcl& tcl = Tcl::instance();
	codec_ = JpegDecoder::create(config_);
	int q = atoi(tcl.attr("softJPEGthresh"));
	if (q < 0)
		q = JpegDecoder::q_to_thresh(inq_);
	codec_->thresh(q);
	codec_->cthresh(atoi(tcl.attr("softJPEGcthresh")));
	codec_->renderer(this);

}

void MotionJpegDecoder::toggle_analog()
{
#ifdef VIC_ANALOG
    analog_ = 1 - analog_;

#ifdef DEBUG
    printf("%x::toggle_analog(): mj analog_=%d\n", this, analog_);
#endif

    if (analog_)
    {
	if (!analog_init(inw_, inh_, inq_))
	    analog_ = 0;
    }
    else
    {
	analog_close();
    }
#endif
}

MotionJpegDecoder::~MotionJpegDecoder()
{
	delete codec_;
}

void MotionJpegDecoder::info(char* wrk) const
{
	sprintf(wrk, "q=%d, type=%d", inq_, type_);
}

void MotionJpegDecoder::counter_names(char* wrk) const
{
	strcpy(wrk, "Bad-Offset Huge-Frame");
}

int MotionJpegDecoder::counter_query(const char* name, u_int& v) const
{
	if (strcasecmp(name, "bad-offset") == 0)
		v = jpeg_badoff_;
	else if (strcasecmp(name, "huge-frame") == 0)
		v = jpeg_hugefrm_;
	else
		return (-1);

	return (0);
}

int MotionJpegDecoder::colorhist(u_int* hist) const
{
	if (codec_->decimation() == 411)
		colorhist_411_556(hist, codec_->yplane(), codec_->uplane(),
				  codec_->vplane(), inw_, inh_);
	else
		colorhist_422_556(hist, codec_->yplane(), codec_->uplane(),
				  codec_->vplane(), inw_, inh_);
	return (1);
}

void MotionJpegDecoder::render(int x, int y, int w, int h,
			       u_char* yp, u_char* up, u_char* vp)
{
	x >>= 3;
	y >>= 3;
	rvts_[y << 7 | x] = now_;
	rvts_[y << 7 | x + 1] = now_;
	if (h > 8) {
		rvts_[y + 1 << 7 | x] = now_;
		rvts_[y + 1 << 7 | x + 1] = now_;
	}
}

void MotionJpegDecoder::scale_rate(int level)
{
	decimate_ = level;
}

int MotionJpegDecoder::recv(rtphdr* rh, u_char* vh, u_char* bp, int cc,
			    int late)
{	
	/*XXX and drop whole frame*/
	if (late)
		return (0);

	jpeghdr* p = (jpeghdr*)vh;
	int reconfig = 0;
	if (inq_ != p->q) {
		reconfig = 1;
		inq_ = p->q;
#ifdef DEBUG
		printf("new q=%d reconfig=%d\n", inq_, reconfig);
#endif
		JpegDecoder::quantizer(config_, inq_);
	}
	if (inw_ != p->width << 3 || inh_ != p->height << 3) {
		reconfig = 1;
		inw_ = p->width << 3;
		inh_ = p->height << 3;
#ifdef DEBUG
		printf("new size %dx%d reconfig=%d\n", inw_, inh_, reconfig);
#endif
		config_.width = inw_;
		config_.height = inh_;
		/*
		 * resize all the renderers
		 */
		for (SoftwareRenderer* p = engines_; p != 0; p = p->next_)
			p->resize(inw_, inh_, color_);
		render_frame(codec_->yplane(),
			     codec_->uplane(),
			     codec_->vplane());
	}

#ifdef VIC_ANALOG
	if (reconfig && analog_)
	{
#ifdef DEBUG
	    printf("Reconfigure to %dx%d q=%d\n", inw_, inh_, inq_);
#endif
	    analog_config(inw_, inh_, inq_);
	}
#endif

	if (reconfig) {
		delete codec_;
		codec_ = JpegDecoder::create(config_);
		Tcl& tcl = Tcl::instance();
		int q = atoi(tcl.attr("softJPEGthresh"));
		if (q < 0)
			q = JpegDecoder::q_to_thresh(inq_);
		codec_->thresh(q);
		codec_->cthresh(atoi(tcl.attr("softJPEGcthresh")));
		codec_->renderer(this);
	}

	/*
	 * This is a hack.  If we don't have any active windows,
	 * (only the thumbnail)  and we have updated it recently, bail.
	 */
	if (engines_ == stamp_ && stamp_->next_ == 0 && !update_stamp_)
		/*XXX approximation */
		return (ntohs(rh->rh_flags) & RTP_M);		

	bp = reassemble(rh, vh, bp, cc);
	if (bp == 0)
		return (0);

#ifdef VIC_ANALOG
	if (analog_)
	{
	    analog_out(bp, cc);
	}
	else
#endif
	{
	    codec_->decode(bp, cc);
	    render_frame(codec_->yplane(), codec_->uplane(), codec_->vplane());
	}
	
	return (1);
}

SoftwareRenderer* MotionJpegDecoder::create(VideoWindow* w)
{
	return (new SoftwareRenderer(w, inw_, inh_, color_,
				     codec_->decimation()));
}

void MotionJpegDecoder::redraw(SoftwareRenderer *p)
{
	p->dither(0, 0, inw_, inh_, codec_->yplane(), codec_->uplane(),
		  codec_->vplane());
	p->render();
}
