/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef vic_decoder_jpeg_h
#define vic_decoder_jpeg_h

#include "decoder.h"
#include "reasm-jpeg.h"
#include "jpeg/jpeg.h"

class SoftwareRenderer;

class MotionJpegDecoder : public SoftwareDecoder,
			  public JpegReassembler,
			  public JpegRenderer {
public:
	MotionJpegDecoder(const u_char* pkthdr);
	virtual ~MotionJpegDecoder();
	virtual void scale_rate(int level);
	void resize(int width, int height);
	int colorhist(u_int* hist) const;

	virtual void info(char*) const;
	virtual void counter_names(char*) const;
	virtual int counter_query(const char* name, u_int& value) const;
	virtual void toggle_analog();
protected:
	virtual void render(int x, int y, int w, int h,
			    u_char* yp, u_char* up, u_char* vp);
	int recv(rtphdr*, u_char* vh, u_char* data, int len, int jitter);
	virtual void redraw(SoftwareRenderer *p);
	virtual SoftwareRenderer* create(VideoWindow*);

	int inq_;		/* input quantization */
	int type_;		/* JPEG/RTP parameters type code */

	JpegDecoder* codec_;
	JpegDecoder::config config_;

	int needstamp_;		/* HACK */
};
#endif
