/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/decoder-jv.C,v 1.2 1995/08/16 00:45:19 olson Exp (LBL)";

#include <osfcn.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include "inet.h"
#include "packet.h"
#include "video.h"
#include "decoder-jv.h"
#include "color.h"
#include "renderer.h"
#include "Tcl.h"

extern "C" {
#include <sys/ipc.h>
#include <sys/shm.h>
#ifdef mips
#include <mips/cachectl.h>
int cacheflush(char* addr, int len, int cache);
#endif
}

class JvRenderer : public Renderer, public IOHandler {
public:
	JvRenderer(VideoWindow* w, int inw, int inh, int inq, int ncolor);
	~JvRenderer();
	JvRenderer* next_;
        /*
	 * This structures should be protected,
	 * but DEC's C++ compiler cannot hack it.
	 */
	struct pbuf {
		dmabuf db;
		int len;
		pbuf* next;
	};
	void recv(u_char* frame, int len);
	int match(VideoWindow*, int inq, int inw, int inh);
	void reset(int inq, int inw, int inh);
	void sleep();
protected:
	virtual void dispatch(int);

	void render();
	void decompress();
	int waitd();

	int setcolors(int ncolor);

	int decompressing_;	/* true when decomp request is pending */
	int rw_;		/* width of rendered image */
	int rh_;		/* height of rendered image */
	int q_;			/* input quantization */
	int server_;		/* J-Video server */
	int pad_;		/* horizontal padding for output image */

	SharedVideoImage* vbd_;
	SharedVideoImage* vbx_;
	dmabuf vb0_;
	dmabuf vb1_;

	void allocdma(dmabuf&, int type, int w, int h, int rmid = 1);
	void freedma(dmabuf&);
	void allocimage();

	pbuf pbase_[3];
	pbuf* pf_;
	pbuf* pi_;
	pbuf* pbd_;

	void enque(pbuf* pb);
	pbuf* deque();
	pbuf* qh_;
	pbuf* qt_;
};

JvRenderer::JvRenderer(VideoWindow* window, int inw, int inh, int inq, int nc)
	: Renderer(window, inw, inh),
	  q_(inq), decompressing_(0), next_(0)
{
	/*
	 * Open server connection.  Note that we
	 * create one connection per video window.
	 * We need to do this because the J-Video
	 * server caches the decoding state on
	 * a per socket basis.  For instance, the
	 * server automatically reprograms the
	 * hardware scaling unit each time we issue
	 * a decompression request.
	 */
	server_ = JvsOpen("", JVS_SOCKET);
	if (server_ < 0) {
		/*XXX*/
		return;
	}

	init(window);

	qh_ = qt_ = 0;
	vbd_ = 0;
	vbx_ = 0;

	/*XXX width/height bogus*/
	allocdma(pbase_[0].db, JVS_JPEG, 640, 480);
	allocdma(pbase_[1].db, JVS_JPEG, 640, 480);
	allocdma(pbase_[2].db, JVS_JPEG, 640, 480);

	pbase_[0].next = 0;
	pbase_[1].next = &pbase_[0];
	pf_ = &pbase_[1];
	pi_ = &pbase_[2];

	/*
	 * If jvdriver already has an installed colormap, use those
	 * colors.  Otherwise, set our default.
	 */
	if (setcolors(0) == 0)
		setcolors(nc);

	link(server_, TK_READABLE);

	rw_ = -1;
	reset(q_, inw_, inh_);
}

JvRenderer::~JvRenderer()
{
	if (server_ >= 0) {
		if (decompressing_)
			(void)waitd();
		freedma(pbase_[0].db);
		freedma(pbase_[1].db);
		freedma(pbase_[2].db);
		if (vbx_ != 0) {
			/*XXX*/
			int s0 = vb0_.shmid;
			int s1 = vb1_.shmid;
			delete vbx_;
			delete vbd_;
			(void)JvsDeallocBuf(server_, s0);
			(void)JvsDeallocBuf(server_, s1);
		}
		/* ~IOHandler() has already unlinked us */
		(void)close(server_);
	}
}

void JvRenderer::sleep()
{
	if (decompressing_)
		(void)waitd();
	windows_ = 0;
}

inline int JvRenderer::match(VideoWindow* w, int inq, int inw, int inh)
{
	return (inq == q_ && inw == inw_ && inh == inh_ &&
		Renderer::match(w));
}

int JvRenderer::setcolors(int ncolors)
{
	int mono = 0;
	if (JvsColormap(server_, ":0", Tk_Colormap(windows_->tkwin()),
			&ncolors, mono) < 0) {
		fprintf(stderr, "vic: JvsColormap failed\n");
		/*XXX*/
		exit(1);
	}
	return (ncolors);
}

void JvRenderer::allocdma(dmabuf& d, int type, int w, int h, int rmid)
{
	if (JvsAllocateBuf(server_, JVS_OUTPUT, type, w, h, &d.shmid) < 0) {
		fprintf(stderr, "vic: JvsAllocateBuf failed\n");
		exit(1);
	}
	d.bp = (u_char *)shmat(d.shmid, 0, 0);
	if (d.bp == (u_char*)-1) {
		perror("vic: shmat");
		exit(1);
	}
	/*
	 * remove the shared memory segment now -- right after
	 * it has been allocated.  the idea is that we'd like
	 * the shared memory segments to go away if we exit
	 * abonormally, and since they will still be attached to
	 * our address space the kernel won't deallocate it
	 * until we exit.  the ultrix man page doesn't seems
	 * to indicate that this is necessarily the case, but
	 * a quick glance at the kernel code for the shmctl()
	 * system call verifies it.
	 *
	 * we don't always do this -- if we need to share them
	 * with X we must do the XShmAttach first.
	 * This might file since SharedVideoImage does it too.
	 */
	if (rmid)
		(void)shmctl(d.shmid, IPC_RMID, 0);
}

void JvRenderer::freedma(dmabuf& d)
{
	(void)JvsDeallocBuf(server_, d.shmid);
	if (shmdt((char*)d.bp) < 0)
		perror("shmdt");
}

void JvRenderer::reset(int inq, int inw, int inh)
{
	if (decompressing_)
		waitd();

	/*
	 * Make sure all outstanding putimage's finish.
	 * Syncing with the first window, syncs the server,
	 * and hence all of them.
	 */
	if (windows_)
		windows_->sync();
	
	q_ = inq;
	inw_ = inw;
	inh_ = inh;

	int w = width_;
	int h = height_;

	if (JvsSetDecomp(server_, q_, inw_, inh_, &w, &h, 8, 0, 0, &pad_) < 0)
		fprintf(stderr, "vic: JvsSetDecomp failed\n");

	if (w != rw_ || h != rh_) {
		rw_ = w;
		rh_ = h;
		allocimage();
	}
	if (qh_ != 0)
		decompress();
}

void JvRenderer::allocimage()
{
	if (vbx_ != 0) {
		/*XXX*/
		int s0 = vb0_.shmid;
		int s1 = vb1_.shmid;
		if (JvsDeallocBuf(server_, s0) < 0 || 
		    JvsDeallocBuf(server_, s1) < 0) {
			fprintf(stderr, "vic: JvsDeallocBuf failed\n");
			/*XXX*/
			exit(1);
		}
		delete vbx_;
		delete vbd_;
	}
	Tk_Window tk = windows_->tkmain();

	imagesize_ = (rw_ + pad_) * rh_;
	allocdma(vb0_, JVS_DITHERED, rw_ + pad_, rh_, 0);
	allocdma(vb1_, JVS_DITHERED, rw_ + pad_, rh_, 0);
	vbd_ = new SharedVideoImage(tk, rw_ + pad_, rh_, vb0_.bp, vb0_.shmid);
	vbx_ = new SharedVideoImage(tk, rw_ + pad_, rh_, vb1_.bp, vb1_.shmid);

	/*
	 * Once the X server has attached the shm segments,
	 * we rmid them so they will go away when we exit.
	 * The sync is to make the X server do the attach
	 * before we do the rmid.
	 */
	XSync(Tk_Display(tk), 0);
	(void)shmctl(vbx_->shmid(), IPC_RMID, 0);
	(void)shmctl(vbd_->shmid(), IPC_RMID, 0);
}

void JvRenderer::dispatch(int)
{
	int s = waitd();
	if (windows_ == 0)
		return;
	if (s >= 0 && !windows_->outboard())
		render();

	/*
	 * If another frame is ready, continue decompression.
	 */
	if (qh_ != 0)
		decompress();
}

void JvRenderer::decompress()
{
	decompressing_ = 1;
	pbd_ = deque();
	if (windows_->outboard()) {
		if (JvsStartAnalogDecomp(server_, pbd_->db.shmid,
					 pbd_->len) < 0)
			fprintf(stderr, "vic: JvsStartDecomp failed\n");
	} else {
		if (JvsStartDecomp(server_, pbd_->db.shmid, vbd_->shmid(),
				   pbd_->len) < 0)
			fprintf(stderr, "vic: JvsStartDecomp failed\n");
#ifdef mips
		if (cacheflush((char*)vbd_->buffer(),
			       imagesize_, DCACHE) < 0) {
			perror("cacheflush");
			exit(1);
		}

#endif
	}
}

/*
 * Wait for the most recently issued J-Video decompression
 * request to finish.  On exit, vbd is null, vbx contains
 * the image being copied to the framebuffer.  Returns 0
 * if vbx is a new image, or -1 if vbx is an old image (e.g.,
 * the last ShmPutImage has not finished yet).
 */
int JvRenderer::waitd()
{
	decompressing_ = 0;
	/*
	 * Decompression is done.  Put the packet
	 * buffer back on the free list.
	 */
	pbd_->next = pf_;
	pf_ = pbd_;
	pbd_ = 0;

	int cshmid;
	int dshmid;
	if (JvsWaitDecomp(server_, &cshmid, &dshmid) < 0) {
		/*
		 * The decompression timed out.
		 * The hardware can be flaky.
		 * Oh well, don't do anything drastic.
		 */
		fprintf(stderr, "vic: JvsWaitDecomp failed\n");/*XXX*/
		
		return (-1);
	}
	/*XXX sanity check shmid's? */
	return (0);
}

void JvRenderer::render()
{
	/*
	 * Swap the decompression and displayed
	 * images, and fire off the XShmPutImage.
	 */
	SharedVideoImage* t = vbx_;
	vbx_ = vbd_;
	vbd_ = t;

	for (VideoWindow* p = windows_; p != 0; p = p->next_)
		p->render(vbx_);

	purge();
}

void JvRenderer::enque(pbuf* pb)
{
	if (qh_ == 0)
		qh_ = qt_ = pb;
	else {
		qt_->next = pb;
		qt_ = pb;
	}
	pb->next = 0;
	if (!decompressing_)
		decompress();
}

JvRenderer::pbuf* JvRenderer::deque()
{
	pbuf* p = qh_;
	qh_ = p->next;
	return (p);
}

void JvRenderer::recv(u_char* frame, int cc)
{
	/* Copy packet data into a dma buffer */
	memcpy((char*)pi_->db.bp, (char*)frame, cc);
	pi_->len = cc;
	/*XXX this is a gross test condition */
	if (pf_ != 0) {
		enque(pi_);
		pi_ = pf_;
		pf_ = pf_->next;
	}
}

JvRenderer* JvDecoder::free_;
int JvDecoder::nfree_;

extern int quality_to_qfactor(int);

JvDecoder::JvDecoder(const u_char* pkthdr)
	: Decoder(sizeof(jpeghdr)), engines_(0)
{
	const jpeghdr* p = (jpeghdr*)pkthdr;
	if (p->type != 0) {
		valid_ = 0;
		return;
	}
	/*XXX probe jvdriver */
	int server = JvsOpen("", JVS_SOCKET);
	if (server < 0) {
		valid_ = 0;
		return;
	}
	(void)close(server);

	inq_ = p->q;
	inw_ = p->width << 3;
	inh_ = p->height << 3;
	update_stamp_ = 1;
}

JvDecoder::~JvDecoder()
{
	/* XXX spin through renderers and delete them */
	JvRenderer* p = engines_;
	while (p != 0) {
		JvRenderer* n = p->next_;
		release(p);
		p = n;
	}
}

void JvDecoder::release(JvRenderer* p)
{
	p->sleep();
	p->next_ = free_;
	free_ = p;
	++nfree_;
	/* XXX throw away if nfree to big */
}

/*
 * Receive a packet.  Reassemble it and hand it off to the
 * hardware for decoding.
 */
int JvDecoder::recv(rtphdr* rh, u_char* vh, u_char* bp, int cc, int)
{
	jpeghdr* p = (jpeghdr*)vh;
	if (inq_ != p->q || inw_ != p->width << 3 || inh_ != p->height << 3) {
		/*
		 * The JPEG parameters have changed.
		 * It's easy enough to reprogram the renderers,
		 * so just do that.
		 */
		inq_ = p->q;
		inw_ = p->width << 3;
		inh_ = p->height << 3;
		for (JvRenderer* r = engines_; r != 0; r = r->next_)
			r->reset(inq_, inw_, inh_);
	}
	bp = reassemble(rh, vh, bp, cc);
	if (bp != 0) {
		for (JvRenderer* r = engines_; r != 0; r = r->next_)
			if (r != stamp_)
				r->recv(bp, cc);
		if (update_stamp_) {
			update_stamp_ = 0;
			stamp_->recv(bp, cc);
		}
		return (1);
	}
	return (0);
}

JvRenderer* JvDecoder::allocate(VideoWindow* w)
{
	JvRenderer* p;
	for (JvRenderer** pp = &free_; (p = *pp) != 0; pp = &p->next_) {
		if (p->match(w, inq_, inw_, inh_)) {
			*pp = (*pp)->next_;
			p->init(w);
			return (p);
		}
	}
	for (pp = &free_; (p = *pp) != 0; pp = &p->next_) {
		if (p->Renderer::match(w)) {
			*pp = (*pp)->next_;
			p->init(w);
			p->reset(inq_, inw_, inh_);
			return (p);
		}
	}
	p = free_;
	if (p != 0) {
		free_ = p->next_;
		p->init(w);
		p->reset(inq_, inw_, inh_);
		return (p);
	}
	int ncolors = atoi(Tcl::instance().attr("jvColors"));
	return (new JvRenderer(w, inw_, inh_, inq_, ncolors));
}

void JvDecoder::attach(VideoWindow* w, int stamp)
{
	for (JvRenderer* p = engines_; p != 0; p = p->next_) {
		if (p->Renderer::match(w)) {
			p->attach(w);
			return;
		}
	}
	p = allocate(w);
	p->next_ = engines_;
	engines_ = p;
	if (stamp)
		stamp_ = p;
}

void JvDecoder::detach(VideoWindow* w)
{
	JvRenderer** np = &engines_;
	JvRenderer* p;
	for (;;) {
		p = *np;
		/*XXX*/
		if (p == 0)
			return;
		if (p->contains(w))
			break;
		np = &p->next_;
	}
	p->detach(w);
	/*
	 * Clear out the image so that the Video class does
	 * not try to use it before another window gets attached
	 * and replaces the image.
	 */
	w->setimage(0);
	if (p->empty()) {
		*np = p->next_;
		release(p);
	}
}

/*
 * Open a dummy socket to jvdriver and allocate the number of
 * colors we want to use.  We need jvdriver to allocate the
 * colors first (because it wants free cells), then when the
 * dithering code goes to allocate the same colors, it will
 * succeed.  If we do this in the other order, jvdriver will
 * just give up because it won't see enough free cells.
 *
 * We leave the socket open indefinitely.
 */
void
reserve_jv_colors()
{
	int fd = JvsOpen("", JVS_SOCKET);
	if (fd < 0)
		return;

	int mono = 0;
	/*XXX display name*/
	Tcl& tcl = Tcl::instance();
	int ncolors = atoi(tcl.attr("jvColors"));
	if (JvsColormap(fd, ":0", Tk_Colormap(tcl.tkmain()),
			&ncolors, mono) < 0) {
		fprintf(stderr, "vic: JvsColormap failed\n");
		/*XXX*/
		exit(1);
	}
}
