/*
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/decoder-scr.C,v 1.2 1995/08/16 00:45:24 olson Exp (LBL)";

#include <osfcn.h>
#include <string.h>
#include <sys/param.h>
#include "inet.h"
#include "packet.h"
#include "video.h"
#include "decoder-scr.h"
#include "color.h"
#include "renderer.h"
#include "endian.h"

ScrDecoder::ScrDecoder(const u_char* pkthdr) : PlaneDecoder(sizeof(nvhdr))
{
	const nvhdr* ph = (nvhdr*)pkthdr;
	int w = ntohs(ph->width) & 0x7fff;
	int h = ntohs(ph->height);
	if ((unsigned)w > 1000 || (unsigned)h > 1000) {
		/*XXX*/
		w = 320;
		h = 240;
	}
	resize(w, h);
}

int ScrDecoder::recv(rtphdr* rh, u_char* vh, u_char* bp, int cc, int jitter)
{	
	nvhdr* ph = (nvhdr*)vh;
	decode(bp, cc, ph);
	if (ntohs(rh->rh_flags) & RTP_M) {
		render_frame(yplane_, uplane_, vplane_);
		return (1);
	}
	return (0);
}

u_char* ScrDecoder::decode_block(u_char* bp, u_char* end, int width)
{
	/*XXX check for whole block here*/
	if (end - bp < 2)
		return (end);

	int x0 = bp[0];
	int y0 = bp[1];
	bp += 4;

	if (x0 > inw_ / 8 || y0 >= inh_ / 8) {
		/*XXX*/
		return 0;
	}

	int offset = y0 * 8 * width + x0 * 8;
	n_long* yp = (n_long*)&yplane_[offset];
	u_short* up;
	u_short* vp;
	if (uplane_) {
		up = (u_short*)&uplane_[offset / 2];
		vp = (u_short*)&vplane_[offset / 2];
	} else {
		up = 0;
		vp = 0;
	}
	/*XXX*/
	if (bp + 128 > end) {
		/*XXX*/
		return (end);
	}
	for (int i = 8; --i >= 0; ) {
		yp[0] = *(u_int*)bp;
		bp += 4;
		yp[1] = *(u_int*)bp;
		bp += 4;

		if (up != 0) {
			*(u_int*)up = *(u_int*)bp;
			bp += 4;
			*(u_int*)vp = *(u_int*)bp;
			bp += 4;
			up += width / 2 / sizeof(*up);
			vp += width / 2 / sizeof(*vp);
		}
		yp += width / sizeof(*yp);
	}
	rvts_[RV_INDEX(x0, y0)] = now_;
	return (bp);
}

void ScrDecoder::decode(u_char* p, int len, nvhdr* ph)
{
	int width = ntohs(ph->width) & 0x7fff;
	int height = ntohs(ph->height);

	if ((unsigned)width >= 1000 || (unsigned)height >= 1000) {
		/*XXX*/
		return;
	}
	if (width != inw_ || height != inh_)
		resize(width, height);

	u_char* end = p + len;
	while (p < end && p != 0)
		p = decode_block(p, end, width);
}
