/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/decoder.C,v 1.4 1995/08/26 21:15:48 olson Exp (LBL)";

#include <osfcn.h>
#include <sys/param.h>
#include <sys/time.h>
#include "inet.h"
#include "packet.h"
#include "video.h"
#include "decoder.h"
#include "Tcl.h"
#include "renderer.h"

#include "decoder-nv.h"
#include "decoder-scr.h"
#ifdef JVIDEO
#include "decoder-jv.h"
#endif
#ifdef MME
#include "decoder-mme.h"
#endif
#include "decoder-jpeg.h"
#include "decoder-h261.h"
#include "decoder-cellb.h"

extern "C" {
#ifndef NOSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#if defined(sun) && !defined(__svr4__)
int shmget(key_t, int, int);
char *shmat(int, char*, int);
int shmdt(char*);
int shmctl(int, int, struct shmid_ds*);
#endif
#endif
}

int Decoder::use_hardware_;

Decoder::Decoder(int hdrlen)
	: PacketHandler(hdrlen), color_(1), ntoss_(0), valid_(1),
	  inw_(0), inh_(0), update_stamp_(1), analog_(0)
{
	update_ = atoi(Tcl::instance().attr("stampInterval"));
	ssched(0);
}

Decoder::~Decoder()
{
}

void Decoder::toggle_analog()
{
    analog_ = 1 - analog_;
#ifdef DEBUG
    printf("%x::toggle_analog(): analog_=%d\n", this, analog_);
#endif
}

void Decoder::info(char* wrk) const
{
	*wrk = 0;
}

void Decoder::counter_names(char* wrk) const
{
	*wrk = 0;
}

int Decoder::counter_query(const char*, u_int&) const
{
	return (-1);
}

void Decoder::setcolor(int)
{
}

void Decoder::timeout()
{
	update_stamp();
	msched(update_);
}

void Decoder::update_stamp()
{
	update_stamp_ = 1;
}

void Decoder::allocshm(dmabuf& d, int size, int flag)
{
#ifndef NOSHM
	d.shmid = shmget(IPC_PRIVATE, size, IPC_CREAT|0777);
	if (d.shmid < 0) {
		perror("shmget");
		exit(1);
	}
	d.bp = (u_char *)shmat(d.shmid, 0, flag);
	if (d.bp == (u_char*)-1) {
		perror("shmat");
		exit(1);
	}
#endif
}

/*XXX not used*/
void Decoder::freeshm(dmabuf& d)
{
#ifndef NOSHM
	if (shmdt((char*)d.bp) < 0)
		perror("shmdt");
#endif
}

/*
 * Return time of day in microseconds.
 */
double Decoder::gettimeofday()
{
	timeval tv;
	::gettimeofday(&tv, 0);
	return (1e6 * double(tv.tv_sec) + double(tv.tv_usec));
}

void Decoder::scale_rate(int level)
{
}

/*
 * The following implement a 'null' decoder for use when we don't
 * recognize an encoding but need to construct a window for a site.
 */
int Decoder::recv(rtphdr*, u_char* vh, u_char* pkt, int len, int jitter)
{	
	return (0);
}

void Decoder::attach(VideoWindow* w, int stamp)
{
}

void Decoder::detach(VideoWindow* w)
{
	/*
	 * Clear out the image so that the Video class does
	 * not try to use it before another window gets attached
	 * and replaces the image.
	 */
	w->setimage(0);
}

Decoder* Decoder::create(int fmt, const u_char* pkthdr)
{
	Decoder* p;
	/*
	 * Fire up a new source.
	 */
	switch (fmt) {

	case RTP_PT_JPEG:
		/*
		 * Call the instantiate tcl command in vic.tcl
		 * to create the window.
		 */
		/*XXX*/
		p = 0;
#ifdef notdef
#ifdef MME
		p = new MmeDecoder(pkthdr);
#endif
#endif
#ifdef JVIDEO
		if (use_hardware_ && (p == 0 || !p->valid())) {
			delete p;
			p = new JvDecoder(pkthdr);
		}
#endif
		if (p == 0 || !p->valid()) {
			delete p;
			p = new MotionJpegDecoder(pkthdr);
		}
		break;
		
	case RTP_PT_NV:
		p = new NvDecoder(pkthdr);
		break;

	case RTP_PT_SCR:
		p = new ScrDecoder(pkthdr);
		break;

	case RTP_PT_H261:
		p = new H261Decoder(pkthdr);
		break;

	case RTP_PT_CELLB:
		p = new CellbDecoder(pkthdr);
		break;

	case RTP_PT_MPEG:
		/*XXX*/
	default:
		p = 0;
	}
	if (p == 0 || !p->valid()) {
		delete p;
		p = new Decoder(0);
	}
	return (p);
}

int Decoder::colorhist(u_int*) const
{
	return (0);
}

void Decoder::redraw()
{
}

SoftwareDecoder::SoftwareDecoder(int hlen) :
	Decoder(hlen), engines_(0), now_(1), stamp_(0)
{
	memset(rvts_, 0, sizeof(rvts_));
}

SoftwareDecoder::~SoftwareDecoder()
{
	/* XXX spin through renderers and delete them */
	SoftwareRenderer* p = engines_;
	while (p != 0) {
		SoftwareRenderer* n = p->next_;
		delete p;
		p = n;
	}
}

void SoftwareDecoder::attach(VideoWindow* w, int stamp)
{
    SoftwareRenderer *p;
	for (p = engines_; p != 0; p = p->next_) {
		if (p->match(w)) {
			p->attach(w);
			break;
		}
	}
	if (p == 0) {
#ifdef __svr4__
		p = this->create(w);
#else
		p = create(w);
#endif
		p->next_ = engines_;
		engines_ = p;
	}
	if (stamp)
		stamp_ = p;
	w->damage();
	redraw(p);
}

void SoftwareDecoder::detach(VideoWindow* w)
{
	SoftwareRenderer** np = &engines_;
	SoftwareRenderer* p;
	for (;;) {
		p = *np;
		/*XXX*/
		if (p == 0)
			return;
		if (p->contains(w))
			break;
		np = &p->next_;
	}
	p->detach(w);
	/*
	 * Clear out the image so that the Video class does
	 * not try to use it before another window gets attached
	 * and replaces the image.
	 */
	w->setimage(0);
	if (p->empty()) {
		*np = p->next_;
		delete p;
	}
}

void SoftwareDecoder::redraw()
{
	for (SoftwareRenderer* p = engines_; p != 0; p = p->next_)
		redraw(p);
}

void SoftwareDecoder::setcolor(int color)
{
	if (color != color_) {
		color_ = color;
		for (SoftwareRenderer* p = engines_; p != 0; p = p->next_)
			p->color(color_);
		redraw();
	}
}

void SoftwareDecoder::colorhist_422_556(u_int* hist, const u_char* yp,
					const u_char* up, const u_char* vp,
					int width, int h) const
{
	while (--h >= 0) {
		for (int w = width; w > 0; w -= 2) {
			/* 5V:5U:6Y */
			int uv = vp[0] >> 3 << 11 | up[0] >> 3 << 6;
			++hist[uv | yp[0] >> 2];
			++hist[uv | yp[1] >> 2];
			yp += 2;
			up += 1;
			vp += 1;
		}
	}
}

void SoftwareDecoder::colorhist_411_556(u_int* hist, const u_char* yp,
					const u_char* up, const u_char* vp,
					int width, int h) const
{
	for (; h > 0; h -= 2) {
		for (int w = width; w > 0; w -= 2) {
			/* 5V:5U:6Y */
			int uv = vp[0] >> 3 << 11 | up[0] >> 3 << 6;
			++hist[uv | yp[0] >> 2];
			++hist[uv | yp[1] >> 2];
			++hist[uv | yp[width] >> 2];
			++hist[uv | yp[width + 1] >> 2];
			yp += 2;
			up += 1;
			vp += 1;
		}
		yp += width;
	}
}

/* XXX separete rendering and now update */
void SoftwareDecoder::render_frame(const u_char* yp,
				   const u_char* up,
				   const u_char* vp)
{
	u_char* ts = rvts_;
	/*XXX this can be better */
	int ymin = inh_;
	int ymax = 0;
	for (int y = 0; y < inh_; y += 8) {
		for (int x = 0; x < inw_; ) {
			if (ts[x >> 3] == 0) {
				x += 8;
				continue;
			}
			ts[x >> 3] = 0;
			int sx = x;
			do {
				x += 8;
			} while (x < inw_ && ts[x >> 3] == now_);

			if (y < ymin)
				ymin = y;
			if (y > ymax)
				ymax = y;

			for (SoftwareRenderer* p = engines_;
			     p != 0; p = p->next_) {
				if (p == stamp_)
					continue;
				p->dither(sx, y, x - sx, 8,
					  yp, up, vp);
			}
		}
		ts += RV_STRIDE;
	}
	SoftwareRenderer *p;
	for (p = engines_; p != 0; p = p->next_)
		if (p != stamp_)
			p->render(ymin, ymax + 8);
	for (p = engines_; p != 0; p = p->next_)
		if (p != stamp_)
			p->purge();

	if (update_stamp_) {
		update_stamp_ = 0;
		stamp_->dither(0, 0, inw_, inh_,
			       yp, up, vp);
		stamp_->render();
		stamp_->purge();
	}
	/*XXX*/
#ifdef notyet
	++now_;
#endif
}

PlaneDecoder::PlaneDecoder(int hlen) : SoftwareDecoder(hlen), yplane_(0)
{
}

PlaneDecoder::~PlaneDecoder()
{
	delete yplane_;
}

SoftwareRenderer* PlaneDecoder::create(VideoWindow* w)
{
	return (new SoftwareRenderer(w, inw_, inh_, color_));
}

void PlaneDecoder::redraw(SoftwareRenderer *p)
{
	p->dither(0, 0, inw_, inh_, yplane_, uplane_, vplane_);
	p->render();

}

void PlaneDecoder::resize(int width, int height)
{
	delete yplane_;

	inw_ = width;
	inh_ = height;
	int size = inw_ * inh_;
	yplane_ = new u_char[2 * size];
	uplane_ = yplane_ + size;
	vplane_ = uplane_ + size / 2;
	/* 
	 * Initialize image to gray.
	 */
	memset(yplane_, 0x80, 2 * size);

	/*
	 * resize all the renderers
	 */
	for (SoftwareRenderer* p = engines_; p != 0; p = p->next_)
		p->resize(width, height, color_);
	/*
	 * redraw all the windows
	 * check for if we have a thumbnail, to prevent
	 * this when called from the decoder constructor
	 */
	if (stamp_ != 0)
		render_frame(yplane_, uplane_, vplane_);
}

int PlaneDecoder::colorhist(u_int* histogram) const
{
	colorhist_422_556(histogram, yplane_, uplane_, vplane_, inw_, inh_);
	return (1);
}
