/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/encoder-nv.h,v 1.2 1995/08/16 00:45:36 olson Exp (LBL)
 */

#ifndef vic_encoder_nv_h
#define vic_encoder_nv_h

#include <sys/types.h>
#include <netinet/in.h>
#include "inet.h"
#include "framer.h"

class NvBaseEncoder : public YuvEncoder {
 public:
	NvBaseEncoder(Network*, int mtu, int rtp_type);
	~NvBaseEncoder();
	virtual void size(int w, int h);
 protected:
	int flush(pktbuf*, int sync);
	int endblock();

	int nw_;
	int nh_;

	u_char* pt_;
	u_char* ep_;
};

class NvEncoder : public NvBaseEncoder {
 public:
	NvEncoder(Network*, int mtu);
	virtual int command(Tcl&, int argc, char** argv);
	virtual int encode(n_long ts, const u_char* frame, u_char* crvec);
 private:
	void putblock(const u_char* yp, const u_char* up, const u_char* vp,
		      int x, int y, int loss, u_char*& runlen);
	void VidTransform_Fwd(const u_char* yp,
			      const u_char* up,
			      const u_char* vp,
			      int width, n_long* out) const;
	int loss_;
};

class ScrEncoder : public NvBaseEncoder {
 public:
	ScrEncoder(Network*, int mtu);
	virtual int encode(n_long ts, const u_char* frame, u_char* crvec);
 private:
	void putblock(const u_char* y, const u_char* u, const u_char* v,
		      int x0, int y0);
};
#endif

