/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/framer-jpeg.C,v 1.2 1995/08/16 00:45:39 olson Exp (LBL)";
#endif

#include "net.h"
#include "framer-jpeg.h"
#include <netinet/in.h>

JpegFramer::JpegFramer(Network* net, int mtu)
	: Framer(net, mtu)
{
	format_ = RTP_PT_JPEG;
}

JpegFramer::~JpegFramer()
{
}

void JpegFramer::setparams(int q, int fw, int fh, int type)
{
	q_ = q;
	nw_ = fw >> 3;
	nh_ = fh >> 3;
	type_ = type;
}

#define HLEN (sizeof(jpeghdr) + sizeof(rtphdr))

int JpegFramer::send(u_char* bp, int n, n_long ts)
{
	/* make sure the last frame is completely transmitted */
	flush();

	int tot = n;
	int off = 0;
	int payload = mtu_ - HLEN;
	while (off < n) {
		pktbuf* pb = alloch(ts);
		rtphdr* rh = (rtphdr*)pb->hdr;
		jpeghdr* h = (jpeghdr*)(rh + 1);

		/*XXX could maintain this as one 32-bit value */
		h->q = q_;
		h->width = nw_;
		h->height = nh_;
		h->type = type_;

		int cc = n - off;
		if (cc > payload)
			cc = payload;
		else
			rh->rh_flags |= htons(RTP_M);

		h->off = htonl(off);
		pb->iov[0].iov_len = HLEN;
		pb->iov[1].iov_base = (caddr_t)bp + off;
		pb->iov[1].iov_len = cc;
		/*XXX DEC C++ needs scope qualifier here */
		Framer::send(pb);
		tot += HLEN;
		off += cc;
	}
	return (tot);
}
