/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Network Research
 *	Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/framer.C,v 1.2 1995/08/16 00:45:40 olson Exp (LBL)";

#include <stdio.h>
#include <osfcn.h>
#include <errno.h>
#include <string.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <sys/file.h>
#include <sys/stat.h>
#include "framer.h"
#include "net.h"
#include "source.h"
#include "decoder.h"
#include "Tcl.h"

#if (defined(sun) && !defined(__svr4__)) || (defined(_AIX) && !defined(_AIX41))
extern "C" writev(int, iovec*, int);
#endif

Framer::pktbuf* Framer::freehdrs_;
Framer::buffer* Framer::freebufs_;
int Framer::nbufs_;
int Framer::nhdrs_;

Framer::Framer(Network* n, int m)
	: net_(n), mtu_(m),
	  format_(0), seqno_(1),
	  nf_(0), nb_(0), np_(0),
	  head_(0), tail_(0), busy_(0),
	  pkttime_(0.)
{
	memset((char*)&mh_, 0, sizeof(mh_));
	mh_.msg_iovlen = 2;
}

Framer::~Framer()
{
}

int Framer::command(Tcl& t, int argc, char** argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "nb") == 0) {
			char* p = t.buffer();
			sprintf(p, "%u", 8 * nb_);
			t.result(p);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "nf") == 0) {
			char* p = t.buffer();
			sprintf(p, "%u", nf_);
			t.result(p);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "np") == 0 ||
		    strcmp(argv[1], "ns") == 0) {
			char* p = t.buffer();
			sprintf(p, "%u", np_);
			t.result(p);
			return (TCL_OK);
		}
	}
	Tcl_AppendResult(t.interp(), "encoder: bad option", 0);
	return (TCL_ERROR);
}

void Framer::loopback(pktbuf* pb)
{
	rtphdr* rh = (rtphdr*)pb->hdr;
	int cc = pb->iov[0].iov_len + pb->iov[1].iov_len;
	/*
	 * Update statistics.
	 */
	nb_ += cc;
	++np_;
	int sync = ntohs(rh->rh_flags) & RTP_M;
	if (sync)
		++nf_;

	SourceManager& sm = SourceManager::instance();
	Source& s = *sm.localsrc();
	s.touch_data();
	s.datantp(rh->rh_ts);
	s.np(1);
	s.nb(cc);
	s.cs((u_short)ntohs(rh->rh_seqno));

	if (s.mute()) {
		if (sync)
			s.nf(1);
		return;
	}

	u_char* vh = (u_char*)(rh + 1);
	/*
	 * Handle initialization of loopback decoder
	 * and changes in the stream.
	 */
	PacketHandler* h = s.handler();
	if (h == 0)
		h = s.activate(format_, vh);
	else if (s.format() != format_)
		h = s.change_format(format_, vh);

	if (h->recv(rh, vh, (u_char*)pb->iov[1].iov_base,
		    pb->iov[1].iov_len, 0))
		s.nf(1);
}

int Framer::dumpfd_ = -1;
void Framer::dump(int fd)
{
	dumpfd_ = fd;
#define MAGIC "RTPCLIP 1.0"
	(void)write(fd, MAGIC, sizeof(MAGIC));
}

void Framer::dump(int fd, iovec* iov, int iovlen) const
{
	register int length = 0;
	for (int i = iovlen; --i >= 0; )
		length += iov[i].iov_len;

	char cliphdr[4];
	*(short*)cliphdr = htons(length);
	cliphdr[2] = 0; /* data packet (i.e., not an rtcp packet) */
	cliphdr[3] = 0; /* ? */

	(void)write(fd, cliphdr, 4);
	if (writev(fd, iov, iovlen) < 0) {
		perror("write");
		exit(1);
	}
}

void Framer::send(pktbuf* pb)
{
	if (!busy_) {
		nextpkttime_ = gettimeofday() + pkttime_;
		transmit(pb);
		/*
		 * emulate a transmit interrupt --
		 * assume we will have more to send.
		 */
		usched(pkttime_);
		busy_ = 1;
	} else {
		if (head_ != 0) {
			tail_->next = pb;
			tail_ = pb;
		} else
			tail_ = head_ = pb;
		pb->next = 0;
	}
}

void Framer::timeout()
{
	double now = gettimeofday();
	for (;;) {
		pktbuf* p = head_;
		if (p != 0) {
			head_ = p->next;
			transmit(p);
			nextpkttime_ += pkttime_;
			double delta = nextpkttime_ - now;
			/* make sure we will wait more than 10ms */
			if (delta > 1000.) {
				usched(delta);
				return;
			}
		} else {
			busy_ = 0;
			break;
		}
	}
}

void Framer::flush()
{
	if (busy_) {
		busy_ = 0;
		cancel();
	}

	pktbuf* p = head_;
	while (p != 0) {
		pktbuf* n = p->next;
		transmit(p);
		p = n;
	}
	head_ = 0;
}

void Framer::transmit(pktbuf* pb)
{
	if (dumpfd_ >= 0)
		dump(dumpfd_, pb->iov, mh_.msg_iovlen);
	mh_.msg_iov = pb->iov;
	/*
	 * Check to see if we want to smash the packet for RTPv1 compat.
	 * This is an ugly an inefficient place to do this, but 
	 * compatibility won't be needed very long.
	 */
	extern int rtpv1;
	if (rtpv1) {
		rtphdr rhs = *(rtphdr*)pb->iov[0].iov_base;
		/* RTP v1 hdr */
		struct rtp1hdr {
			u_char rh_chan;		/* version/channel */
			u_char rh_flags;	/* O/S/content */
			u_short	rh_seqno;	/* sequence number */
			n_long	rh_ts;
		}* rh1;
		pb->iov[0].iov_base += 4;
		rh1 = (rtp1hdr*)pb->iov[0].iov_base;
		pb->iov[0].iov_len -= 4;
		rh1->rh_chan = 0x60; // vers 1, chan 32
		int flags = ((u_char*)&rhs)[1];
		/*XXX lose a bit in PT -> content type */
		flags = ((flags & 0x80) >> 1) | (flags & 0x3f);
		rh1->rh_flags = flags;
		rh1->rh_seqno = rhs.rh_seqno;
		rh1->rh_ts = rhs.rh_ts;
		net_->sendmsg(mh_);
		pb->iov[0].iov_len += 4;
		pb->iov[0].iov_base -= 4;
		*(rtphdr*)pb->iov[0].iov_base = rhs;
	} else
		net_->sendmsg(mh_);

	loopback(pb);
	release(pb);
}

void Framer::release(pktbuf* pb)
{
	pb->next = freehdrs_;
	freehdrs_ = pb;
	buffer* p = pb->buf;
	if (p != 0) {
		p->next = freebufs_;
		freebufs_ = p;
	}
}

YuvEncoder::YuvEncoder(Network* net, int mtu) : Framer(net, mtu)
{
}

YuvEncoder::~YuvEncoder()
{
}
