/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber-jv.C,v 1.2 1995/08/16 00:45:42 olson Exp (LBL)";
#endif

#include <stdio.h>
#include <errno.h>
#include <osfcn.h>
#include <string.h>
#include <sys/file.h>
#include <sys/stat.h>

#include "jvs.h"

#include "inet.h"
#include "Tcl.h"
#include "framer-jpeg.h"
#include "packet.h"
#include "grabber.h"
#include "iohandler.h"
#include "crdef.h"

extern "C" {
#include <sys/ipc.h>
#include <sys/shm.h>
#ifdef mips
#include <mips/cachectl.h>
int cacheflush(char* addr, int len, int cache);
#endif
}

class JvGrabber : public Grabber, public IOHandler {
protected:
	JvGrabber();
public:
	virtual ~JvGrabber();
	virtual void start();
	virtual void stop();
	virtual int command(int argc, char** argv);
	/*
	 * This structure should be private,
	 * but DEC's C++ compiler cannot hack it.
	 */
	struct dmabuf {
		int shmid;
		u_char* bp;
	};
protected:
	virtual void SetComp() = 0;
	virtual void dispatch(int mask);
	virtual void timeout();
	virtual int consume(const dmabuf&, int length) = 0;

	void kick();		/* ask jvdriver to compress next frame */
	void sync();		/* synchronize with rpc compression reply */

	int server_;		/* fd of socket to jvdriver */
	dmabuf* allocdma(int type);
	void freedma(dmabuf*);
	dmabuf* cb0_;		/* capture buffer #1 */
	dmabuf* cb1_;		/* capture buffer #2 */
	dmabuf* cb_;		/* double-buffer capture pointer */
	int inw_;		/* width of original frame */
	int inh_;		/* height of original frame */
	int x_decimate_;	/* horizontal downsampling factor */
	int y_decimate_;	/* vertical downsampling factor */
	int grabbing_;		/* true when grab request pending */

	int j300_;
};

class JvJpegGrabber : public JvGrabber {
public:
	JvJpegGrabber(JpegFramer*);
	virtual ~JvJpegGrabber();
	virtual int command(int argc, char** argv);
protected:
	virtual int consume(const dmabuf&, int length);
	virtual void SetComp();
private:
	int q_;
	JpegFramer* framer_;
};

class JvYuvCore : public JvGrabber {
public:
	JvYuvCore(YuvEncoder*);
	virtual void SetComp();
	virtual ~JvYuvCore();
protected:
	YuvEncoder* encoder_;
	u_char* frame_;
	int framesize_;
};

class JvYuvGrabber : public JvYuvCore {
public:
	JvYuvGrabber(YuvEncoder*);
	virtual int command(int argc, char** argv);
	void suppress(const u_char* devbuf, const u_char* lumbuf);
	void saveblks(u_char* in);
protected:
	virtual void SetComp();
	virtual int consume(const dmabuf&, int length);
};

class JvCIFGrabber : public JvYuvCore {
public:
	JvCIFGrabber(YuvEncoder*);
	virtual int command(int argc, char** argv);
protected:
	virtual int consume(const dmabuf&, int length);
	void ntsc_grab(u_char* bp);
	void pal_grab(u_char* bp);
	void setsize();
	int outw_;
	int outh_;
	int ntsc_;
};

class JvDevice : public Device {
 public:
	JvDevice(const char* s);
	virtual int match(const char* type) const;
	virtual Grabber* jpeg_grabber(JpegFramer*);
	virtual Grabber* yuv_grabber(YuvEncoder*);
	virtual Grabber* cif_grabber(YuvEncoder*);
	void j300(int v);
};

static JvDevice jv_device("jvdriver");

JvDevice::JvDevice(const char* s) : Device(s)
{
	default_format_ = RTP_PT_JPEG;
	/* start out assuming j300 */
	attrs_ = DEV_ATTR_RAW|DEV_ATTR_CIF|DEV_ATTR_HW_JPEG|DEV_ATTR_SMALL|
		DEV_ATTR_LARGE;
}

int JvDevice::match(const char* type) const
{
	return (strcasecmp(type, "jvideo") == 0 ||
		strcasecmp(type, "jv") == 0 ||
		strcasecmp(type, "jvdriver") == 0);
}

void JvDevice::j300(int v)
{
	attrs_ = DEV_ATTR_HW_JPEG;
	if (v)
		attrs_ |= DEV_ATTR_RAW|DEV_ATTR_CIF|
			DEV_ATTR_SMALL|DEV_ATTR_LARGE;
}

Grabber* JvDevice::jpeg_grabber(JpegFramer* p)
{
	return (new JvJpegGrabber(p));
}

Grabber* JvDevice::yuv_grabber(YuvEncoder* p)
{
	return (new JvYuvGrabber(p));
}

Grabber* JvDevice::cif_grabber(YuvEncoder* p)
{
	return (new JvCIFGrabber(p));
}

JvGrabber::JvGrabber()
{
	server_ = JvsOpen("", JVS_SOCKET);
	if (server_ < 0) {
		status_ = -1;
		return;
	}
	grabbing_ = 0;
	running_ = 0;
	x_decimate_ = 2;
	y_decimate_ = 2;

	inw_ = inh_ = 0;

	char id[JVS_IDENTITY_STRING];
	id[0] = 0;
	JvsIdentity(server_, id);
	j300_ = strcmp(id, "J300") == 0;
	jv_device.j300(j300_);
}

JvGrabber::~JvGrabber()
{
	if (server_ >= 0) {
		if (running_)
			stop();
		freedma(cb0_);
		freedma(cb1_);
		close(server_);
	}
}

JvGrabber::dmabuf* JvGrabber::allocdma(int type)
{
	dmabuf* p = new dmabuf;
	int w = 640;
	int h = 480;
	if (JvsAllocateBuf(server_, JVS_INPUT, type, w, h, &p->shmid) < 0) {
		fprintf(stderr, "vic: JvsAllocateBuf failed\n");
		exit(1);
	}
	p->bp = (u_char *)shmat(p->shmid, 0, 0);
	if (p->bp == (u_char*)-1) {
		perror("vic: shmat");
		exit(1);
	}
	/*
	 * remove the shared memory segment now -- right after
	 * it has been allocated.  the idea is that we'd like
	 * the shared memory segments to go away if we exit
	 * abonormally, and since they will still be attached to
	 * our address space the kernel won't deallocate it
	 * until we exit.  the ultrix man page doesn't seems
	 * to indicate that this is necessarily the case, but
	 * a quick glance at the kernel code for the smctl()
	 * system call verifies it.
	 */
	(void)shmctl(p->shmid, IPC_RMID, 0);

	return (p);
}

void JvGrabber::freedma(JvGrabber::dmabuf* p)
{
	(void)JvsDeallocBuf(server_, p->shmid);
	if (shmdt((char*)p->bp) < 0) {
		perror("vic: shmdt");
		exit(1);
	}
	delete p;
}

void JvGrabber::start()
{
	if (running_)
		return;

	running_ = 1;
	link(server_, TK_READABLE);
	frameclock_ = gettimeofday();
	kick();
}

void JvGrabber::stop()
{
	if (!running_)
		return;

	running_ = 0;
	unlink();
	if (grabbing_)
		sync();
}

void JvGrabber::sync()
{
	/*XXX wait for last completion so nothing is left
	  in the socket buffer */
	int length, shmid;
	(void)JvsWaitComp(server_, &shmid, &length);
	grabbing_ = 0;
}

void JvGrabber::kick()
{
	cb_ = (cb_ == cb0_) ? cb1_ : cb0_;
	if (JvsStartComp(server_, cb_->shmid) < 0) {
		fprintf(stderr, "vic: JvsStartComp failed\n");
		return;
	}
	grabbing_ = 1;
}

void JvGrabber::timeout()
{
	/*
	 * Check state.  We might have stopped since
	 * the timeout was scheduled.
	 */
	if (!grabbing_ && running_)
		kick();
}

/*
 * Called by IOHandler when rpc reply arrives
 * on server socket.
 */
void JvGrabber::dispatch(int mask)
{
	int length;
	int shmid;
	/*
	 * Since we have input waiting on this socket, this
	 * rpc should return immediately.
	 */
	if (JvsWaitComp(server_, &shmid, &length) < 0) {
		fprintf(stderr, "vic: JvsWaitComp failed\n");
		/*
		 * (maybe) the compression timed out.
		 * The hardware can be flaky.
		 * Try again.
		 */
		kick();
		return;
	}
	grabbing_ = 0;
	dmabuf* p = cb_;
#ifdef mips
	/*XXX I think jvdriver does this for us.  But is the
	 cache physical or virtual? */
	if (cacheflush((char*)p->bp, length, DCACHE) < 0) {
		perror("cacheflush");
		exit(1);
	}
#endif
	int n = consume(*p, length);
	double t = tick(n);
	/*
	 * If we're running at the max-frame rate, fire off
	 * the next compression instead of setting a timer.
	 * It would be an optimization to kick off the next
	 * capture before calling consume() above, but on a
	 * uniprocessor we're likely to run to completion
	 * before the video server runs anyway.  So kicking
	 * after consume() is in practice just as efficient.
	 */
	if (t < 34000.)
		kick();
	else
		usched(t);
}

int JvGrabber::command(int argc, char** argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 4 && strcmp(argv[2], "decimate") == 0) {
		int d = atoi(argv[3]);
		/*
		 * XXX shouldn't happen but be paranoid because
		 * jvdriver will crash
		 */
		if (d == 0) {
			Tcl_AppendResult(tcl.interp(),
					 "%s: divide by zero", argv[0], 0);
			return (TCL_ERROR);
		}
		x_decimate_ = d;
		y_decimate_ = d;
		SetComp();
	}
	return (TCL_OK);
}

JvYuvCore::JvYuvCore(YuvEncoder* e) : encoder_(e), frame_(0)
{
	if (status_ < 0)
		return;
	if (!j300_) {
		close(server_);
		server_ = -1;
		status_ = -2;
		return;
	}
	cb0_ = allocdma(JVS_YUV);
	cb1_ = allocdma(JVS_YUV);
	cb_ = cb0_;
}

JvYuvCore::~JvYuvCore()
{
	delete frame_;
}

JvYuvGrabber::JvYuvGrabber(YuvEncoder* e) : JvYuvCore(e)
{
	if (status_ < 0)
		return;

	SetComp();
	encoder_->size(inw_, inh_);
}

/*XXX*/
#ifdef notdef
#define ABS(t) ((t) - (((t) >> 31 & (t)) << 1))
#endif
#define ABS(t) t
inline int diff(const u_char* in, const u_char* frm)
{
	int t = in[0] - frm[0];
	int s = ABS(t);
	t = in[2] - frm[1];
	s += ABS(t);
	t = in[4] - frm[2];
	s += ABS(t);
	t = in[6] - frm[3];
	s += ABS(t);
	t = in[8] - frm[4];
	s += ABS(t);
	t = in[10] - frm[5];
	s += ABS(t);
	t = in[12] - frm[6];
	s += ABS(t);
	t = in[14] - frm[7];
	s += ABS(t);

	/*XXX*/
	if (s < 0)
		s = -s;

	return (s);
}

inline void saveblk(const u_char* in,
		    u_char* yp, u_char* up, u_char* vp, int stride)
{
	int is = stride << 1;
	int cs = stride >> 1;
	for (int i = 8; --i >= 0; ) {
		/* XXX would word accesses help? */
		yp[0] = in[0];
		yp[1] = in[2];
		yp[2] = in[4];
		yp[3] = in[6];
		yp[4] = in[8];
		yp[5] = in[10];
		yp[6] = in[12];
		yp[7] = in[14];
		up[0] = in[1];
		up[1] = in[5];
		up[2] = in[9];
		up[3] = in[13];
		vp[0] = in[3];
		vp[1] = in[7];
		vp[2] = in[11];
		vp[3] = in[15];
		in += is;
		yp += stride;
		up += cs;
		vp += cs;
	}
}

void JvYuvGrabber::saveblks(u_char* in)
{
	u_char* crv = crvec_;
	u_char* lum = frame_;
	int off = framesize_;
	u_char* chm = lum + off;
	off >>= 1;
	int stride = 7 * stride_;
	for (int y = 0; y < blkh_; ++y) {
		for (int x = 0; x < blkw_; ++x) {
			int s = *crv++;
			if ((s & (CR_HQ|CR_LQ)) != 0)
				saveblk(in, lum, chm, chm + off, stride_);

			in += 16;
			lum += 8;
			chm += 4;
		}
		lum += stride;
		chm += stride >> 1;
		in += stride << 1;
	}
}

void JvYuvGrabber::suppress(const u_char* devbuf, const u_char* lumbuf)
{
	SUPPRESS_BODY(devbuf, lumbuf, 2)
}

int JvYuvGrabber::consume(const JvGrabber::dmabuf& db, int length)
{
	suppress(db.bp, frame_);
	saveblks(db.bp);
	return (encoder_->encode(ntptime(), frame_, crvec_));
}

JvCIFGrabber::JvCIFGrabber(YuvEncoder* e) : JvYuvCore(e)
{
	if (status_ < 0)
		return;

	frame_ = 0;
	SetComp();
	setsize();
}

void JvCIFGrabber::setsize()
{
	delete frame_;
	frame_ = 0;

	switch (inw_) {
	case 320:
		/* 1/4 NTSC */
		ntsc_ = 1;
		outw_ = 352;
		outh_ = 288;
		break;

	case 160:
		/* 1/8 NTSC */
		ntsc_ = 1;
		outw_ = 176;
		outh_ = 144;
		break;

	case 384:
		/* 1/4 PAL */
		ntsc_ = 0;
		outw_ = 352;
		outh_ = 288;
		break;

	case 192:
		/* 1/16 PAL */
		ntsc_ = 0;
		outw_ = 176;
		outh_ = 144;
		break;

	default:
		/* XXX this shouldn't happen */
		fprintf(stderr, "vic: jvdriver CIF grabber: bad geometry\n");
		abort();
	}
	int s = outw_ * outh_;
	frame_ = new u_char[s + (s >> 1)];
	framesize_ = s;
	crinit(outw_, outh_);
	encoder_->size(outw_, outh_);
}

int JvCIFGrabber::consume(const JvGrabber::dmabuf& db, int length)
{
	if (ntsc_)
		ntsc_grab(db.bp);
	else
		pal_grab(db.bp);

	suppress(frame_, cache_);
	saveblks(frame_);
	return (encoder_->encode(ntptime(), frame_, crvec_));
}

/*
 * Grab a Q/CIF frame from a 4:2:2 NTSC input.  We dup every 10th
 * pixel horizontally and every 4th line vertically.  We also
 * discard the chroma on every other line, since CIF wants 4:1:1.
 */
void JvCIFGrabber::ntsc_grab(u_char* p)
{
	u_char* yp = frame_;
	u_char* up = frame_ + framesize_;
	u_char* vp = up + (framesize_ >> 2);

	int stride = outw_;
	/*XXX*/
	int vdup = 5;
	for (int h = 0; h < inh_; ++h) {
		int hdup = 10/2;
		for (int w = inw_; w > 0; w -= 2) {
			yp[0] = p[0];
			yp[1] = p[2];
			yp += 2;
			if ((h & 1) == 0) {
				*up++ = p[1];
				*vp++ = p[3];
			}
			p += 4;
			if (--hdup <= 0) {
				hdup = 10/2;
				yp[0] = yp[-1];
				yp += 1;
				if ((h & 1) == 0) {
					if ((w & 2) == 0) {
						up[0] = up[-1];
						++up;
						vp[0] = vp[-1];
						++vp;
					}
				}
			}
		}
		if (--vdup <= 0) {
			vdup = 5;
			/* copy previous line */
			memcpy((char*)yp, (char*)yp - stride, stride);
			yp += stride;
			if ((h & 1) == 0) {
				int s = stride >> 1;
				memcpy((char*)up, (char*)up - s, s);
				memcpy((char*)vp, (char*)vp - s, s);
				up += s;
				vp += s;
			}
		}
		
	}
}

/*
 * Grab a Q/CIF frame from a 4:2:2 PAL input.  Just throw
 * away the last 32 (16 for QCIF) pixels of each line and
 * discard the chroma from every other line.
 */
void JvCIFGrabber::pal_grab(u_char* p)
{
	u_char* yp = frame_;
	u_char* up = frame_ + framesize_;
	u_char* vp = up + (framesize_ >> 2);

	int skip = (inw_ - outw_) << 1;
	for (int h = inh_; h > 0; h -= 2) {
		for (int w = outw_; w > 0; w -= 4) {
			yp[0] = p[0];
			up[0] = p[1];
			yp[1] = p[2];
			vp[0] = p[3];
			yp[2] = p[4];
			up[1] = p[5];
			yp[3] = p[6];
			vp[1] = p[7];
			yp += 4;
			up += 2;
			vp += 2;
			p += 8;
		}
		p += skip;
		for (w = outw_; w > 0; w -= 4) {
			yp[0] = p[0];
			yp[1] = p[2];
			yp[2] = p[4];
			yp[3] = p[6];
			yp += 4;
			p += 8;
		}
		p += skip;
	}
}

int JvCIFGrabber::command(int argc, char** argv)
{
	int s = JvGrabber::command(argc, argv);
	if (s == TCL_OK && argc == 4 && strcmp(argv[2], "decimate") == 0)
		setsize();

	return (TCL_OK);
}

int JvYuvGrabber::command(int argc, char** argv)
{
	return (JvGrabber::command(argc, argv));
}

void JvYuvCore::SetComp()
{
	int grabbing = grabbing_;
	if (grabbing)
		sync();

	/*XXX c'mon Lance*/
	JvsSetCompressReq r;
	r.requestCode = JVS_SETCOMPRESS;
	r.qfactor = 32;
	r.xdec = x_decimate_;
	r.ydec = y_decimate_;
	r.frameskip = 1;
	r.type = JVS_YUV;
	JvsSetCompressRep rep;
	if (JvsSetCompRaw(server_, &r, &rep) < 0) {
		fprintf(stderr, "vic: JvsSetCompRaw failed\n");
		exit(1);
	}
	int w = rep.width;
	int h = rep.height;
	/*XXX*/
	if (x_decimate_ == 1 && y_decimate_ == 1 && 3 * w != 4 * h) {
		r.xdec = x_decimate_ = 2;
		r.ydec = y_decimate_ = 2;
		if (JvsSetCompRaw(server_, &r, &rep) < 0) {
			fprintf(stderr, "vic: JvsSetCompRaw failed\n");
			exit(1);
		}
		w = rep.width;
		h = rep.height;
	}
	inw_ = w;
	inh_ = h;
	if (grabbing)
		/*
		 * If we interrupted an in-progress grab,
		 * kick off a new request.
		 */
		kick();
}

void JvYuvGrabber::SetComp()
{
	int w = inw_;
	int h = inh_;
	JvYuvCore::SetComp();
	if (inw_ != w || inh_ != h) {
		encoder_->size(inw_, inh_);
		delete frame_;
		int s = inw_ * inh_;
		framesize_ = s;
		frame_ = new u_char[2 * s];
		crinit(inw_, inh_, 0);
	}
}

extern int quality_to_qfactor(int);

JvJpegGrabber::JvJpegGrabber(JpegFramer* f) : framer_(f), q_(50)
{
	if (status_ < 0)
		return;
	cb0_ = allocdma(JVS_JPEG);
	cb1_ = allocdma(JVS_JPEG);
	cb_ = cb0_;

	SetComp();
}

JvJpegGrabber::~JvJpegGrabber()
{
}

int JvJpegGrabber::consume(const JvGrabber::dmabuf& p, int length)
{
	/*
	 * We should get a better timestamp, probably from jvdriver.
	 */
	return (framer_->send(p.bp, length, ntptime()));
}

void JvJpegGrabber::SetComp()
{
	int grabbing = grabbing_;
	if (grabbing)
		sync();

	/*
	 * Don't bother doing anything fancy with frame subsampling.
	 * On the j300 this doesn't work anyway, and on the jvideo
	 * we set it to one so we get every other field.
	 */
	int frameskip = 1;
	int w, h;
	if (JvsSetComp(server_, q_, x_decimate_, y_decimate_,
		       frameskip, &w, &h) < 0) {
		fprintf(stderr, "vic: JvsSetComp failed\n");
		exit(1);
	}
	inw_ = w;
	inh_ = h;
	framer_->setparams(q_, w, h, 0);

	if (grabbing)
		/*
		 * If we interrupted an in-progress grab,
		 * kick off a new request.
		 */
		kick();
}

int JvJpegGrabber::command(int argc, char** argv)
{
	if (argc == 4 && strcmp(argv[2], "q") == 0) {
		/* assume value is in range */
		q_ = atoi(argv[3]);
		SetComp();
		return (TCL_OK);
	}
	return (JvGrabber::command(argc, argv));
}

