/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber-nsunrise.C,v 1.1 1995/08/25 13:52:03 olson Exp (LBL)";
#endif

#include <stdio.h>
#include <errno.h>
#include <osfcn.h>
#include <string.h>
#include <sys/file.h>
#include <sys/stat.h>

#include "inet.h"
#include "Tcl.h"
#include "framer-jpeg.h"
#include "packet.h"
#include "grabber.h"
#include "crdef.h"

#include <UMSVideoIO.xh>
#include <UMSStrings.h>

#define UMS_Success(rc) (rc == UMSVideoIO_Success)

#define NTSC_FULL_WIDTH 640
#define NTSC_FULL_HEIGHT 480

#define N_RING_BUFFERS 4

class SunriseGrabber : public Grabber {
protected:
	SunriseGrabber();
public:
	virtual ~SunriseGrabber();
	virtual int command(int argc, char** argv);
	virtual void fps(int);
protected:
	_IDL_SEQUENCE_UMSVideoIO_RingBufferElement ring_buffer;
	void allocate_ring_buffers(long n_buffers, long buffer_size);
	int decimate_;
	long width_;
	long height_;
	string port_;
	virtual void setsize() = 0;
	virtual void set_port();
	Environment *ev;
};

class SunriseJpegGrabber : public SunriseGrabber {
public:
	SunriseJpegGrabber(JpegFramer*);
	virtual ~SunriseJpegGrabber();
	virtual int command(int argc, char** argv);
	void setsize();
protected:
	void start();
	void stop();
private:
	int q_;
	JpegFramer* framer_;
	int grab();
};

class SunriseYuvCore : public SunriseGrabber {
public:
	SunriseYuvCore(YuvEncoder*);
	virtual ~SunriseYuvCore();
protected:
	virtual void start();
	virtual void stop();
	virtual int command(int argc, char** argv);
	YuvEncoder* encoder_;
	u_char* frame_;
	int framesize_;
	_IDL_SEQUENCE_octet colormap;
};

class SunriseYuvGrabber : public SunriseYuvCore {
public:
	SunriseYuvGrabber(YuvEncoder*);
	virtual int command(int argc, char** argv);
	virtual int grab();
	void setsize();
protected:
};

class SunriseCIFGrabber : public SunriseYuvCore {
public:
	SunriseCIFGrabber(YuvEncoder*);
	virtual int command(int argc, char** argv);
	virtual int grab();
	void setsize();
protected:
};

class SunriseDevice : public Device {
public:
//    SunriseDevice(const char* s);
    SunriseDevice(const char* s, const char *which);
    virtual int match(const char* type) const;
    virtual Grabber* yuv_grabber(YuvEncoder*);
    virtual Grabber* jpeg_grabber(JpegFramer*);
    virtual Grabber* cif_grabber(YuvEncoder*);
    UMSVideoIO *video;
    SunriseDevice *next;

    void set_quality_factor(int q);
    void set_video_input_connector(string port);
    void set_capture_rate(int rate);
    void set_input_image_size(long &width, long &height);

    int open_device(const char *which);
    void close_device();

    int start();
    int stop();
    
protected:
    Environment *ev;
};

class SunriseDeviceMaker
{
 public:
    SunriseDeviceMaker();
    ~SunriseDeviceMaker();
 protected:
    SunriseDevice *devices;
    void trydev(string dev);
    Environment *ev;
};

static SunriseDevice sunrise_device("sunrise", "/dev/sr0");

#if 0
static SunriseDeviceMaker maker;

SunriseDeviceMaker::SunriseDeviceMaker()
{
    string devnames[] = { "/dev/sr0", "/dev/sr1", NULL };
    int i;
    string devname;

    ev = somGetGlobalEnvironment();
    
    devices = NULL;

    for (i = 0; devnames[i] != NULL; i++)
    {
	trydev(devnames[i]);
    }
}

void SunriseDeviceMaker::trydev(string devname)
{
    UMSVideoIO_ReturnCode rc;
    UMSVideoIO *vid;
    SunriseDevice *dev;
    char *name, *s;

    vid = new UMSVideoIO;

    printf("Trying %s\n", devname);
    rc = vid->open(ev, devname);
    if (rc == UMSVideoIO_Success)
    {
	vid->close(ev);
	delete vid;
	name = new char[100];
	if (s = rindex(devname, '/'))
	    s++;
	else
	    s = devname;
	sprintf(name, "sunrise-%d", s);
	printf("Creating %s\n", s);
	dev = new SunriseDevice(s);
	dev->next = devices;
	devices = dev;
    }
}

SunriseDeviceMaker::~SunriseDeviceMaker()
{
    SunriseDevice *dev, *ndev;

    for (dev = devices; dev != NULL; )
    {
	ndev = dev->next;
	delete dev;
	dev = ndev;
    }
}
#endif

SunriseDevice::SunriseDevice(const char* s, const char *device = "/dev/sr0")
    : Device(s)
{
    UMSVideoIO_ReturnCode rc;

    default_format_ = RTP_PT_JPEG;

    attrs_ = (DEV_ATTR_RAW |
	      DEV_ATTR_HW_JPEG |
	      DEV_ATTR_CIF |
	      DEV_ATTR_PORT |
	      DEV_ATTR_SMALL |
	      DEV_ATTR_LARGE);
    
#ifdef DEBUG
    printf("SunriseDevice::SunriseDevice name=%s\n", s);
#endif /* DEBUG */

    ev = somGetGlobalEnvironment();

    video = new UMSVideoIO;

    /*
     * For now, just open sr0.
     */
    if ((rc = sunrise_device.video->open(ev, (string) device)) != UMSVideoIO_Success)
    {
	printf("error %d opening video device %s\n", rc, device);
	return;
    }

#ifdef DEBUG
    printf("opened device %s\n", device);
#endif /* DEBUG */
}

int SunriseDevice::match(const char* type) const
{
	return (strcasecmp(type, "sunrise") == 0 ||
		strcasecmp(type, "ibm") == 0 ||
		strcasecmp(type, "aix") == 0);
}

Grabber* SunriseDevice::jpeg_grabber(JpegFramer* p)
{
	return (new SunriseJpegGrabber(p));
}

Grabber* SunriseDevice::yuv_grabber(YuvEncoder* p)
{
    return (new SunriseYuvGrabber(p));
}

Grabber* SunriseDevice::cif_grabber(YuvEncoder* p)
{
	return (new SunriseCIFGrabber(p));
}

SunriseGrabber::SunriseGrabber() : port_(INPUT_COMPOSITE)
{
    ev = somGetGlobalEnvironment();

    ring_buffer._buffer = (UMSVideoIO_RingBufferElement *) NULL;
}

SunriseGrabber::~SunriseGrabber()
{
}

void SunriseGrabber::fps(int val)
{
    UMSVideoIO_ReturnCode rc;

    SunriseSuspend suspend(sunrise_device);
    
    sunrise_device.set_capture_rate(val);

    Grabber::fps(val);
}

int SunriseGrabber::command(int argc, char** argv)
{
    int i;
    Tcl& tcl = Tcl::instance();
    
#ifdef DEBUG
    printf("SunriseGrabber::command argc=%d ", argc);
    for (i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif /* DEBUG */
    if (argc == 4)
    {
	if (strcmp(argv[2], "port") == 0) {
	    int p = atoi(argv[3]);
	    string new_port = "";

	    switch (p)
	    {
	    case 1:
	    case 2:
		new_port = INPUT_COMPOSITE;
		break;

	    case 3:
		new_port = INPUT_SVIDEO;
		break;
	    }

	    if (strcmp(port_, new_port) != 0)
	    {
		port_ = new_port;
		set_port();
	    }
	}
    }

    return (TCL_OK);
}

void SunriseGrabber::set_port()
{
    int was_running = running_;
    SunriseSuspend suspend(sunrise_device);


    sunrise_device.set_video_input_connector(port_);
}

SunriseJpegGrabber::SunriseJpegGrabber(JpegFramer* f) : framer_(f), q_(50)
{
    decimate_ = 2;
    setsize();
}

void SunriseJpegGrabber::setsize()
{
    UMSVideoIO_ReturnCode rc;
    long w, h;
    int was_running;
    SunriseSuspend suspend(sunrise_device);
    
#ifdef DEBUG
    printf("SunriseJpegGrabber::setsize()\n");
#endif /* DEBUG */

    w = 640 / decimate_;
    if (decimate_ == 4 || decimate_ == 2)
	h = 240;
    else
	h = 120;

#ifdef DEBUG
    printf("set_input_image_size %dx%d\n", w, h);
#endif /* DEBUG */
    
    sunrise_device.set_input_image_size(&w, &h);

#ifdef DEBUG
    printf("after set_input_image_size w=%d h=%d rc=%d\n", w, h, rc);
#endif /* DEBUG */

    width_ = w;
    height_ = h;
    framer_->setparams(q_, width_, height_, 0);

    rc = sunrise_device.set_capture_rate(fps_);

    allocate_ring_buffers(N_RING_BUFFERS, w * h);

    rc = sunrise_device.set_quality_factor(q);
    q_ = q;

    rc = sunrise_device.video->setup_compressed_capture_buffers(ev, &ring_buffer);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("setup_compressed_capture failed with rc=%d\n", rc);
#endif /* DEBUG */
	return;
    }
}

void SunriseJpegGrabber::start()
{
    UMSVideoIO_ReturnCode rc;

    rc = sunrise_device.video->set_compression(ev, UMSVideoIO_On);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("set_compression failed with rc=%d\n", rc);
#endif /* DEBUG */
	return;
    }

    running_ = 1;
    frameclock_ = gettimeofday();

    /*
     * This sleep is VERY IMPORTANT. It lets the capture card
     * get ahead of vic, so that frames will be available when
     * vic tries to grab them. Otherwise, the grab will block
     * and vic will be permanently behind, locking out X events.
     *
     * Another fix might be to use nonblocking frame grabs.
     */
    sleep(1);
    timeout();
}

void SunriseJpegGrabber::stop()
{
    UMSVideoIO_ReturnCode rc;

    if (!running_)
	return;

    rc = sunrise_device.video->set_compression(ev, UMSVideoIO_Off);
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("set_uncompression failed with rc=%d\n", rc);
#endif /* DEBUG */
    }

    running_ = 0;
    cancel();
}

int SunriseJpegGrabber::grab()
{
    UMSVideoIO_RingBufferElement *frame;
    static UMSVideoIO_RingBufferElement *last_frame = NULL;
    long index;
    UMSVideoIO_ReturnCode rc;
    int frc;
    unsigned char *data;

    if (getenv("VIC_DEBUG"))
	printf("jpeg grab start\n");

    framer_->flush();
    if (last_frame)
	last_frame->InUseByCaller = 0;

    rc = sunrise_device.video->get_compressed_frame(ev, &index, TRUE);
    if (!UMS_Success(rc))
    {
	printf("get_compressed_frame failed with rc=%d\n", rc);
	exit(1);
    }
    
    frame = &(ring_buffer._buffer[index]);

    data = (unsigned char *) frame->Address;

    frc = framer_->send(data, frame->SizeOfDataInBuffer, ntptime());

    last_frame = frame;
    
    if (getenv("VIC_DEBUG"))
	printf("jpeg grab end %d\n", frc);

    return frc;
}

int SunriseJpegGrabber::command(int argc, char** argv)
{
    int i;
    Tcl& tcl = Tcl::instance();
    
#ifdef DEBUG
    printf("SunriseJpegGrabber::command argc=%d ", argc);
    for (i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif /* DEBUG */

    if (argc == 4)
    {
	if (strcmp(argv[2], "q") == 0) {
	    /* assume value is in range */
	    q_ = atoi(argv[3]);
	    setsize();
	    return (TCL_OK);
	}
	else if (strcmp(argv[2], "decimate") == 0) {
	    int d = atoi(argv[3]);
	    Tcl& tcl = Tcl::instance();
	    if (d <= 0) {
		Tcl_AppendResult(tcl.interp(),
				 "%s: divide by zero", argv[0], 0);
		return (TCL_ERROR);
	    }
	    decimate_ = d;
	    setsize();
	    return(TCL_OK);
	}
    }
    return SunriseGrabber::command(argc, argv);
}

SunriseJpegGrabber::~SunriseJpegGrabber()
{
}


SunriseYuvCore::SunriseYuvCore(YuvEncoder* e) : encoder_(e), frame_(0)
{
}

int SunriseYuvCore::command(int argc, char** argv)
{
    int i;
    Tcl& tcl = Tcl::instance();
    
#ifdef DEBUG
    printf("SunriseYuvCore::command argc=%d ", argc);
    for (i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif /* DEBUG */

    if (argc == 4 && strcmp(argv[2], "decimate") == 0) {
	int d = atoi(argv[3]);
	Tcl& tcl = Tcl::instance();
	if (d <= 0) {
	    Tcl_AppendResult(tcl.interp(),
			     "%s: divide by zero", argv[0], 0);
	    return (TCL_ERROR);
	}
	decimate_ = d;
	setsize();
	return(TCL_OK);
    }
    return SunriseGrabber::command(argc, argv);
}

 SunriseYuvCore::~SunriseYuvCore()
{
}

 SunriseYuvGrabber::SunriseYuvGrabber(YuvEncoder* e) : SunriseYuvCore(e)
{
    decimate_ = 2;
    setsize();
}

void SunriseYuvGrabber::setsize()
{
    long w, h, v;
    UMSVideoIO_ReturnCode rc;
    int was_running;
    
    was_running = running_;
    if (was_running)
	stop();

    w = NTSC_FULL_WIDTH / decimate_;
    h = NTSC_FULL_HEIGHT / decimate_;

    rc = sunrise_device.video->set_output_image_format(ev, YUV422);
    
    if (!UMS_Success(rc))
    {
#ifdef DEBUG
	printf("set_output_image_format failed with rc=%d\n", rc);
#endif /* DEBUG */
	return;
    }

    long iw, ih;

    if (w > 320)
	iw = 640;
    else
	iw = 320;

    if (h > 240)
	ih = 480;
    else
	ih = 240;
    
    rc = sunrise_device.video->set_input_image_size(ev, &iw, &ih);
    if (!UMS_Success(rc) && rc != UMSVideoIO_ValueChanged)
    {
#ifdef DEBUG
	printf("set_input_image_size failed with rc=%d\n", rc);
#endif /* DEBUG */
	return;
    }

#ifdef DEBUG
    printf("after set_input iw=%d ih=%d rc=%d\n", iw, ih, rc);
#endif /* DEBUG */
    
    rc = sunrise_device.video->set_uncompressed_image_size(ev, &w, &h);
    if (!UMS_Success(rc) && rc != UMSVideoIO_ValueChanged)
    {
#ifdef DEBUG
	printf("set_uncompressed_image_size failed with rc=%d\n", rc);
#endif /* DEBUG */
	return;
    }

#ifdef DEBUG
    printf("after set_uncompressed_image_size w=%d h=%d rc=%d\n", w, h, rc);
#endif /* DEBUG */


    width_ = w;
    height_ = h;
    framesize_ = w * h;
    frame_ = new u_char[2 * framesize_];
    crinit(w, h);
    encoder_->size((int) w, (int) h);

    double capture_rate = fps_;
    rc = sunrise_device.video->set_capture_rate(ev, &capture_rate);
#ifdef DEBUG
    if (!UMS_Success(rc))
    {
	printf("set_capture_rate failed with rc=%d\n", rc);
    }
    else
	printf("set capture rate to %lf\n", capture_rate);
#endif /* DEBUG */

    allocate_ring_buffers(N_RING_BUFFERS, w * h * 2);
    
    rc = sunrise_device.video->setup_uncompressed_capture_buffers(ev, &ring_buffer);

#ifdef DEBUG
    if (!UMS_Success(rc))
	printf("setup_uncompressed_capture failed with rc=%d\n", rc);
#endif /* DEBUG */

    if (was_running)
	start();
}

int SunriseYuvGrabber::command(int argc, char** argv)
{
    return (SunriseYuvCore::command(argc, argv));
}

int SunriseYuvGrabber::grab()
{
    UMSVideoIO_RingBufferElement *frame;
    long index;
    UMSVideoIO_ReturnCode rc;
    int grc;
    unsigned char *yptr, *uptr, *vptr;
    int i;

    rc = sunrise_device.video->get_uncompressed_frame(ev, &index, TRUE);
    if (!UMS_Success(rc))
    {
	printf("get_uncompressed_frame failed with rc=%d\n", rc);
	exit(1);
    }
    
    frame = &(ring_buffer._buffer[index]);


    /*
     * Pixel format: YYYYYYYY UUUUUUUU YYYYYYYY VVVVVVVV
     * U & V are shared for two pixels.
     */

    struct pixel
    {
	unsigned char y1;
	unsigned char u;
	unsigned char y2;
	unsigned char v;
    } *p, *p_end;

    yptr = frame_;
    uptr = frame_ + framesize_;
    vptr = uptr + framesize_ / 2;

    p = (struct pixel *) frame->Address;
    p_end = (struct pixel *) (frame->Address + frame->SizeOfDataInBuffer);

    while (p < p_end)
    {
	yptr[0] = p->y1;
	yptr[1] = p->y2;
	uptr[0] = p->u;
	vptr[0] = p->v;

	yptr += 2;
	uptr++;
	vptr++;
	p++;
    }

    int overruns = frame->NumberOfOverruns;
    frame->InUseByCaller = 0;

#if 0
    if (overruns > 0)
    {
	printf("Got %d overruns, flushing\n", frame->NumberOfOverruns);
	
	rc = sunrise_device.video->flush_uncompressed_capture_buffers(ev);
	if (!UMS_Success(rc))
	{
	    printf("flush_uncompressed_capture_buffers failed with rc=%d\n", rc);
	}
    }
#endif

    suppress(frame_, cache_);
    saveblks(frame_);
    grc = encoder_->encode(ntptime(), frame_, crvec_);
    return grc;
}


void SunriseYuvCore::start()
{
    UMSVideoIO_ReturnCode rc;

    if (running_)
	return;

    rc = sunrise_device.video->set_uncompression(ev, UMSVideoIO_On);
    if (!UMS_Success(rc))
    {
	printf("set_uncompression failed with rc=%d\n", rc);
	return;
    }

    running_ = 1;
    frameclock_ = gettimeofday();
    timeout();
}

void SunriseYuvCore::stop()
{
    UMSVideoIO_ReturnCode rc;

    if (!running_)
	return;

    rc = sunrise_device.video->set_uncompression(ev, UMSVideoIO_Off);
    if (!UMS_Success(rc))
    {
	printf("set_uncompression failed with rc=%d\n", rc);
    }

    running_ = 0;
    cancel();
}

void SunriseGrabber::allocate_ring_buffers(long n_buffers, long buffer_size)
{
    unsigned char **Address;
    UMSVideoIO_RingBufferElement *rbuffer1;
    int i;
    
    Address = new unsigned char * [n_buffers];
    
    ring_buffer._length = n_buffers;
    ring_buffer._maximum  = n_buffers;

#ifdef DEBUG
    printf("Allocating ring buffers: n_buffers=%d buffer_size=%d\n",
	   n_buffers, buffer_size);
#endif /* DEBUG */

    if (ring_buffer._buffer != (struct UMSVideoIO_RingBufferElement *) NULL)
	delete ring_buffer._buffer;

    ring_buffer._buffer = new UMSVideoIO_RingBufferElement[n_buffers];

    if (ring_buffer._buffer <= 0)
    {
	fprintf(stderr, "Cannot malloc somThis->ring_buffer._buffer\n");
	exit(1);
    }

    /* Set up each ring buffer */

    rbuffer1 = ring_buffer._buffer;

    for (i = 0; i < n_buffers; i++)
    {
	Address[i] = (unsigned char *)
	    malloc(buffer_size + 4096);
	if (Address[i] <= 0)
	{
	    fprintf(stderr, "Cannot malloc ring_buffer %d\n", i);
	    free(ring_buffer._buffer);
	    exit(1);
	}
	rbuffer1->Address = ( ((long) Address[i]) + 4096
					- ( ((long) Address[i]) % 4096));
	rbuffer1->AfterHeader = rbuffer1->Address;
	rbuffer1->SizeOfBuffer = buffer_size;
	rbuffer1->SizeOfDataInBuffer = 0;
	rbuffer1->InUseByCaller	= 0;
	rbuffer1->NumberOfOverruns = 0;
	rbuffer1++;
    }
}


 SunriseCIFGrabber::SunriseCIFGrabber(YuvEncoder* e) : SunriseYuvCore(e)
{
    printf("SunriseCIFGrabber::SunriseCIFGrabber\n");

    decimate_ = 2;
    setsize();
}

void SunriseCIFGrabber::setsize()
{
    long w, h, v;
    UMSVideoIO_ReturnCode rc;
    
    printf("SunriseCIFGrabber::setsize()\n");

    w = 2 * 352 / decimate_;
    h = 2 * 288 / decimate_;

    rc = sunrise_device.video->set_output_image_format(ev, YUV422);
    
    if (!UMS_Success(rc))
    {
	printf("set_output_image_format failed with rc=%d\n", rc);
	return;
    }


    long iw, ih;

    if (w > 320)
	iw = 640;
    else
	iw = 320;

    if (h > 240)
	ih = 480;
    else
	ih = 240;
    
    rc = sunrise_device.video->set_input_image_size(ev, &iw, &ih);
    if (!UMS_Success(rc) && rc != UMSVideoIO_ValueChanged)
    {
	printf("set_input_image_size failed with rc=%d\n", rc);
	return;
    }

    printf("after set_input iw=%d ih=%d rc=%d\n", iw, ih, rc);
    
    rc = sunrise_device.video->set_uncompressed_image_size(ev, &w, &h);
    if (!UMS_Success(rc) && rc != UMSVideoIO_ValueChanged)
    {
	printf("set_uncompressed_image_size failed with rc=%d\n", rc);
	return;
    }

    printf("after set_uncompressed_image_size w=%d h=%d rc=%d\n", w, h, rc);


    width_ = w;
    height_ = h;
    framesize_ = w * h;
    frame_ = new u_char[framesize_ + framesize_ / 2];
    crinit((int) w, (int) h);
    encoder_->size((int) w, (int) h);

    double capture_rate = fps_;
    rc = sunrise_device.video->set_capture_rate(ev, &capture_rate);
    if (!UMS_Success(rc))
    {
	printf("set_capture_rate failed with rc=%d\n", rc);
    }
    else
	printf("set capture rate to %lf\n", capture_rate);

    allocate_ring_buffers(N_RING_BUFFERS, w * h * 2);
    
    rc = sunrise_device.video->setup_uncompressed_capture_buffers(ev, &ring_buffer);
    if (!UMS_Success(rc))
    {
	printf("setup_uncompressed_capture failed with rc=%d\n", rc);
	return;
    }

}

int SunriseCIFGrabber::grab()
{
    UMSVideoIO_RingBufferElement *frame;
    long index;
    UMSVideoIO_ReturnCode rc;
    int grc;
    unsigned char *yptr, *uptr, *vptr;
    int i, x, y;

    rc = sunrise_device.video->get_uncompressed_frame(ev, &index, TRUE);
    if (!UMS_Success(rc))
    {
	printf("get_uncompressed_frame failed with rc=%d\n", rc);
	exit(1);
    }
    
    frame = &(ring_buffer._buffer[index]);

    /*
     * Pixel format: YYYYYYYY UUUUUUUU YYYYYYYY VVVVVVVV
     * U & V are shared for two pixels.
     */

    struct pixel
    {
	unsigned char y1;
	unsigned char u;
	unsigned char y2;
	unsigned char v;
    } *p, *p_end;

    yptr = frame_;
    uptr = frame_ + framesize_;
    vptr = uptr + framesize_ / 4;

    p = (struct pixel *) frame->Address;
    p_end = (struct pixel *) (frame->Address + frame->SizeOfDataInBuffer);

    for (y = 0; y < height_; y++)
    {
	for (x = 0; x < width_; x += 2)
	{
	    yptr[0] = p->y1;
	    yptr[1] = p->y2;
	    yptr += 2;
	    p++;

	    if ((y & 1) == 0)
	    {
		*uptr++ = p->u;
		*vptr++ = p->v;
	    }
	}
    }

    frame->InUseByCaller = 0;

    suppress(frame_, cache_);
    saveblks(frame_);
    grc = encoder_->encode(ntptime(), frame_, crvec_);
    return grc;
}

int SunriseCIFGrabber::command(int argc, char** argv)
{
    return (SunriseYuvCore::command(argc, argv));
}

