/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * pip frame grabber for decstation
 *
 * Derived from code written by Eric Anderson, UCSD Computer Systems Lab,
 * 4/92 and hacked on by Jon Kay, Vach Kompella, Kevin Fall.
 * Ported to vic 1/94, Steven McCanne (mccanne@ee.lbl.gov) March 1993.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber-tx.C,v 1.2 1995/08/16 00:45:55 olson Exp (LBL)";
#endif

#include <osfcn.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/ipc.h>
extern "C" {
#include <sys/shm.h>
}

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "framer.h"
#include "Tcl.h"

class TxGrabber : public Grabber {
 public:
	TxGrabber(Tcl&, YuvEncoder*);
	virtual ~TxGrabber();
 protected:
	virtual const char* type() const;
	int getgrabber();
	int create_capwin(int w, int  h, char *wname, char *iname);
	XImage* create_sharedImage();
	virtual int grab();
	int width_;
	int height_;
	YuvEncoder* encoder_;
	Display* dpy_;
	Window capwin_;
	XImage* image_;
	GC gc_;
	Visual* visual_;
	XvPortID grabID_;
	XShmSegmentInfo shminfo_;
	XXX change these to pointers
	u_long y_[320 * 240 / sizeof(u_long)];
	u_long u_[320 * 240 / sizeof(u_long) / 2];
	u_long v_[320 * 240 / sizeof(u_long) / 2];
};

class TxDevice : public Device {
 public:
	virtual int match(const char* type) const;
	virtual Grabber* yuv_grabber(Tcl&, YuvEncoder*);
};

static TxDevice tx_device("TX/PIP");

int TxDevice::match(const char* type) const
{
	return (strcasecmp(type, "tx") == 0 || strcasecmp(type, "pip") == 0 ||
		strcasecmp(type, "tx/pip") == 0);
}

virtual Grabber* TxDevice::yuv_grabber(Tcl& t, YuvEncoder* p)
{
	return (new TxGrabber(t, p));
}

int TxGrabber::grab()
{
	/*
	 * The brain dead video hardware requires that the window
	 * be unobscured in order to capture the video.  So we grab
	 * the server and raise the window.  This won't work if
	 * the window isn't mapped.  Also, we block signals while
	 * the server is grabbed.  Otherwise, the process could be
	 * stopped while the display is locked out.
	 */
	int oldmask = sigsetmask(~0);
	XGrabServer(dpy_);
	XRaiseWindow(dpy_, capwin_);
	XSync(dpy_, 0);
	XvPutStill(dpy_, grabID_, capwin_, gc_,
		   0, 0, 640, 480,
		   0, 0, 320, 240);
	XSync(dpy_, False);
	XUngrabServer(dpy_);
	(void)sigsetmask(oldmask);
	XImage* p = image_;
	if (p != 0)
		XShmGetImage(dpy_, capwin_, p, 0, 0, AllPlanes);
	else
		p = XGetImage(dpy_, capwin_, 0, 0,
			      320, 240, AllPlanes, ZPixmap);

	/* XXX bold assumptions here */
	u_long* sp = (u_long*)p->data;
	/*XXX*/
	u_char* yp = (u_char*)y_;
	u_char* ep = yp + 320*240;
	u_char* up = (u_char*)u_;
	u_char* vp = (u_char*)v_;

	while (yp < ep) {
		register int y0, y1;
		register int p0 = *sp++;
		register int p1 = *sp++;

		y0 = 5 * (p0 & 0xff);
		y0 += 9 * ((p0 >> 8) & 0xff);
		y0 += 2 * ((p0 >> 16) & 0xff);
		y0 >>= 4;
		*yp++ = y0;
		
		y1 = 5 * (p1 & 0xff);
		y1 += 9 * ((p1 >> 8) & 0xff);
		y1 += 2 * ((p1 >> 16) & 0xff);
		y1 >>= 4;
		*yp++ = y1;

		y0 += y1;
		y0 >>= 1;

		p0 &=~ 0x01010100;
		p1 &=~ 0x01010100;
		p0 = (p0 + p1) >> 1;

		p1 = (126 * (((p0 >> 16) & 0xff) - y0)) / 256;
		if (p1 > 127)
			p1 = 127;
		else if (p1 < -128)
			p1 = -128;
		*up++ = p1;
			
		p1 = (160 * ((p0 & 0xff) - y0)) / 256;
		if (p1 > 127)
			p1 = 127;
		else if (p1 < -128)
			p1 = -128;
		*vp++ = p1;
	}
	if (image_ == 0)
		XDestroyImage(p);

	return (encoder_->encode((u_char*)y_, (u_char*)u_,
				 (u_char*)v_, ntptime(), background()));
}

int TxGrabber::create_capwin(int w, int  h, char *wname, char *iname)
{
	int screen = DefaultScreen(dpy_);
	Window root = DefaultRootWindow(dpy_);
	XVisualInfo vi;
	Colormap colormap;
	if (DefaultDepth(dpy_, screen) == 24) {
		visual_ = DefaultVisual(dpy_, screen);
		colormap = XDefaultColormap(dpy_, screen);
	} else if (XMatchVisualInfo(dpy_, screen, 24, TrueColor, &vi)) {
		visual_ = vi.visual;
		/*XXX*/
		colormap = XCreateColormap(dpy_, root,
					    visual_, AllocNone);
	} else
		return (-1);
	
	XSetWindowAttributes attr;
	attr.colormap = colormap;
	attr.event_mask = VisibilityChangeMask;
	attr.backing_store = Always;
	attr.border_pixel = BlackPixel(dpy_, screen);
	attr.background_pixel = BlackPixel(dpy_, screen);
	attr.override_redirect = True;

	capwin_ = XCreateWindow(dpy_, root, 0, 0, w, h,
				/*XXX*/2, vi.depth, InputOutput, visual_,
				CWColormap|CWBorderPixel|CWBackPixel|
				CWEventMask|CWBackingStore|CWOverrideRedirect,
				&attr);

	/* Initialize size hint property for window manager */
	XSizeHints size_hints;
	size_hints.flags  = PSize;
	size_hints.width  = w;
	size_hints.height = h;
	size_hints.flags |= USPosition;

	/* Set properties for window manager (always before mapping) */
	XSetStandardProperties(dpy_, capwin_, wname, iname, None, 
			       (char **) 0, 0, &size_hints);

	XMapWindow(dpy_, capwin_);

	return (0);
#ifdef notdef
	/*
	 * override window manager if it doesn't follow hints 
	 * the standard says we're not supposed to do this!
	 */
	XResizeWindow(dpy_, capwin_, w, h);
	XFlush(dpy_);
	for (;;) {
		XEvent e;
		XNextEvent(dpy_, &e);
		if (e.type == VisibilityNotify)
			break;
	}
#endif
}

XImage* TxGrabber::create_sharedImage()
{
	extern int use_shm;
	if (!use_shm)
		return (0);

	XImage* p;
	p = XShmCreateImage(dpy_, visual_, 24, ZPixmap, NULL,
			    &shminfo_, 320, 240);

	shminfo_.shmid = shmget(IPC_PRIVATE, p->bytes_per_line * p->height,
				IPC_CREAT | 0777);
	if (shminfo_.shmid == -1) {
		perror("vic: shmget");
		exit(1);
	}
	shminfo_.shmaddr = p->data = shmat(shminfo_.shmid, 0, 0);
	if ((int)p->data == -1) {
		perror("vic: shmat");
		exit(1);
	}
	shminfo_.readOnly = False;
	if (XShmAttach(dpy_, &shminfo_) == 0) {
		perror("vic: XShmAttach");
		exit(1);
	}
	return (p);
}

int TxGrabber::getgrabber()
{
	int majop, eventbase, errbase;
	if (XQueryExtension(dpy_, "XVideo", &majop,
			    &eventbase, &errbase) == False)
		return (-1);
	Window root = DefaultRootWindow(dpy_);
	u_int ngrabbers;
	XvAdaptorInfo* grabbers;
	XvQueryAdaptors(dpy_, root, &ngrabbers, &grabbers);
	if (ngrabbers == 0)
		return (-1);
	if (ngrabbers > 1)
		fprintf(stderr, "vic: warning: more than one frame grabber\n");
		
	grabID_ = grabbers->base_id;
	return (0);
}

TxGrabber::TxGrabber(Tcl& tcl, YuvEncoder* e)
	: Grabber(tcl), encoder_(e)
{
	dpy_ = Tk_Display(tcl.tkmain());
	if (getgrabber() < 0) {
		status_ = -1;
		return;
	}
	if (create_capwin(320, 240, "VIC/TX", "VIC/TX") < 0)
		return;
	image_ = create_sharedImage();
	/*XXX free*/
	gc_ = XCreateGC(dpy_, capwin_, (u_long)0, (XGCValues *)0);

	encoder_->size(320, 240);
}

TxGrabber::~TxGrabber()
{
	/*XXX*/
}

const char* TxGrabber::type() const
{
	return ("tx");
}
