/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber-vl.C,v 1.2 1995/08/16 00:45:57 olson Exp (LBL)";
#endif

#include <osfcn.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <vl/vl.h>

#include "grabber.h"
#include "crdef.h"
#include "framer.h"
#include "Tcl.h"


class VLGrabber : public Grabber {
 public:
	VLGrabber(YuvEncoder*);
	virtual ~VLGrabber();
	virtual int command(int argc, char** argv);
	virtual void fps(int);
	virtual void start();
	virtual void stop();
 protected:
	virtual void startup_vl();
	virtual void shutdown_vl();
	virtual void saveblks(u_char*);
	void suppress(const u_char* , const u_char*);
	virtual int grab();
	virtual void setsize(int xsize, int ysize);
	u_char* next_frame();

	VLServer vlServer_;
	VLPath vlPath_;
	VLBuffer transferBuf_;
	VLNode src_;
	VLNode drn_;
	int xsize_;
	int ysize_;
	int decimate_;
	int current_decimate_;
	int current_fps_;

	YuvEncoder* encoder_;

	u_char* y_;
	u_char* u_;
	u_char* v_;
};

class VLCIFGrabber : public VLGrabber {
    public:
	VLCIFGrabber(YuvEncoder*);
    protected:
	virtual int grab();
	virtual void setsize(int xsize, int ysize);
	u_char* grabline(register u_char* yp, register u_char* ye,
			 register u_char* p);
	u_char* grabline(register u_char* yp, register u_char* ye,
			 register u_char* p, register u_char* up,
			 register u_char* vp);
	int outw_;
	int outh_;
};

class vlDevice : public Device {
 public:
	vlDevice(const char* name);
	virtual int match(const char* type) const;
	virtual Grabber* yuv_grabber(YuvEncoder*);
	virtual Grabber* cif_grabber(YuvEncoder*);
};

static vlDevice VL_device("SGI-VL");

vlDevice::vlDevice(const char* name) : Device(name)
{
	attrs_ = DEV_ATTR_RAW|DEV_ATTR_CIF|DEV_ATTR_SMALL|DEV_ATTR_LARGE;
#ifdef COSMO
	attrs |= DEV_ATTR_HW_JPEG;
#endif
}

int vlDevice::match(const char* type) const
{
	return (strcasecmp(type, "vl") == 0 ||
		strcasecmp(type, "sgi-vl") == 0);
}

Grabber* vlDevice::yuv_grabber(YuvEncoder* p)
{
	return (new VLGrabber(p));
}

Grabber* vlDevice::cif_grabber(YuvEncoder* p)
{
	return (new VLCIFGrabber(p));
}


u_char* VLGrabber::next_frame()
{
	register u_char* p = 0;

	if (vlServer_ && transferBuf_) {
		/*
		 * wait at most one frame time (30ms) for a frame to show up.
		 */
		VLInfoPtr info = 0;
		for (register int i = 3; --i >= 0; ) {
			info = vlGetLatestValid(vlServer_, transferBuf_);
			if (info != 0)
				break;
			sginap(1);
		}
		if (info) {
			p = (u_char*)vlGetActiveRegion(vlServer_, transferBuf_,
						      info);
			if (p == NULL)
				vlPerror("vic: vlGetActiveRegion");
		}
	}
	return (p);
}

/*XXX*/
#ifdef notdef
#define ABS(t) ((t) - (((t) >> 31 & (t)) << 1))
#endif
#define ABS(t) t
inline int diff(const u_char* in, const u_char* frm)
{
	int t, s;
	t = in[1] - frm[0];
	s = ABS(t);
	t = in[3] - frm[1];
	s += ABS(t);
	t = in[5] - frm[2];
	s += ABS(t);
	t = in[7] - frm[3];
	s += ABS(t);
	t = in[9] - frm[4];
	s += ABS(t);
	t = in[11] - frm[5];
	s += ABS(t);
	t = in[13] - frm[6];
	s += ABS(t);
	t = in[15] - frm[7];
	s += ABS(t);

	/*XXX*/
	if (s < 0)
		s = -s;

	return (s);
}

inline void saveblk(const u_char* in,
		    u_char* yp, u_char* up, u_char* vp, int stride)
{
	int is = stride << 1;
	int cs = stride >> 1;
	register u_int mask0 = 0xff000000;
	register u_int mask1 = 0x00ff0000;

	for (int i = 8; --i >= 0; ) {
		/*
		 * SGI has the corporate attitude that anything
		 * and everything having to do with their
		 * systems should be kept a deep, dark secret.
		 * So most of this code was written by guess.
		 * The 4:2:2 data format seems to be a 32 bit int
		 * containing U Y1 V Y2 (msb to lsb).  So each
		 * iteration of this loop grabs 8 Ys & 4 U/Vs.  - vj
		 */
		register u_int d1 = *(u_int*)in;
		register u_int d2 = *(u_int*)(in + 4);
		register u_int y = (d1 << 8) & mask0;
		y |= (d1 << 16) & mask1;
		register u_int u = d1 & mask0;
		register u_int v = (d1 << 16) & mask0;

		d1 = *(u_int*)(in + 8);

		y |= (d2 & mask1) >> 8;
		y |= d2 & 0xff;
		*(u_int*)(yp + 0) = y;
		u |= (d2 >> 8) & mask1;
		v |= (d2 << 8) & mask1;

		d2 = *(u_int*)(in + 12);

		y = (d1 << 8) & mask0;
		y |= (d1 << 16) & mask1;
		u |= (d1 & mask0) >> 16;
		v |= d1 & (mask1 >> 8);

		y |= (d2 & mask1) >> 8;
		y |= d2 & 0xff;
		*(u_int*)(yp + 4) = y;
		u |= d2 >> 24;
		*(u_int*)up = u;
		v |= (d2 >> 8) & 0xff;
		*(u_int*)vp = v;

		in += is;
		yp += stride;
		up += cs;
		vp += cs;
	}
}

void VLGrabber::saveblks(u_char* in)
{
	u_char* crv = crvec_;
	u_char* yp = y_;
	u_char* up = u_;
	u_int off = v_ - up;
	int stride = 7 * stride_;
	for (int y = 0; y < blkh_; ++y) {
		for (int x = 0; x < blkw_; ++x) {
			int s = *crv++;
			if ((s & (CR_HQ|CR_LQ)) != 0)
				saveblk(in, yp, up, up + off, stride_);

			in += 16;
			yp += 8;
			up += 4;
		}
		yp += stride;
		up += stride >> 1;
		in += stride << 1;
	}
}

void VLGrabber::suppress(const u_char* devbuf, const u_char* lumbuf)
{
	SUPPRESS_BODY(devbuf, lumbuf, 2)
}

int VLGrabber::grab()
{
	u_char* p = next_frame();
	if (p == 0)
		return (0);
	suppress(p, y_);
	saveblks(p);
	return (encoder_->encode(ntptime(), y_, crvec_));
}

void VLGrabber::setsize(int xsize, int ysize)
{
	if (xsize != xsize_ || ysize != ysize_) {
		int size = xsize * ysize;
		xsize_ = xsize;
		ysize_ = ysize;
		delete y_;
		y_ = new u_char[2 * size];
		u_ = y_ + size;
		v_ = u_ + size / 2;
		encoder_->size(xsize, ysize);
		crinit(xsize_, ysize_);
	}
}

int VLGrabber::command(int argc, char** argv)
{
	if (argc == 4 && strcmp(argv[2], "decimate") == 0) {
		int d = atoi(argv[3]);
		Tcl& tcl = Tcl::instance();
		if (d <= 0) {
			Tcl_AppendResult(tcl.interp(),
					 "%s: divide by zero", argv[0], 0);
			return (TCL_ERROR);
		}
		decimate_ = d;
		startup_vl();
	}
	return (TCL_OK);
}

void VLGrabber::start()
{
	if (!running_) {
		startup_vl();
		Grabber::start();
	}
}

void VLGrabber::stop()
{
	if (running_) {
		shutdown_vl();
		Grabber::stop();
	}
}

void VLGrabber::fps(int f)
{
	if (f <= 0)
		f = 30;
	if (f > 30)
		f = 30;
	Grabber::fps(f);
	startup_vl();
}

void VLGrabber::shutdown_vl()
{
	if (vlServer_) {
		if (vlPath_) {
			if (transferBuf_) {
				vlEndTransfer(vlServer_, vlPath_);
				vlDeregisterBuffer(vlServer_, vlPath_, drn_,
						   transferBuf_);
				vlDestroyBuffer(vlServer_, transferBuf_);
				transferBuf_ = 0;
			}
			vlDestroyPath(vlServer_, vlPath_);
			vlPath_ = 0;
		}
		vlCloseVideo(vlServer_);
		vlServer_ = 0;
	}
}

void VLGrabber::startup_vl()
{
	if (vlServer_ != 0 &&
	    decimate_ == current_decimate_ && fps_ == current_fps_)
		return;

	current_decimate_ = decimate_;
	current_fps_ = fps_;
	shutdown_vl();

	VLControlValue val;
	int xsize;
	int ysize;

	vlServer_ = vlOpenVideo("");
	if (vlServer_ == 0)
		goto failed;
	src_ = vlGetNode(vlServer_, VL_SRC, VL_VIDEO, VL_ANY);
	drn_ = vlGetNode(vlServer_, VL_DRN, VL_MEM, VL_ANY);
	vlPath_ = vlCreatePath(vlServer_, VL_ANY, src_, drn_);
	if (vlPath_ < 0) {
		vlPerror("vic: create path");
		goto failed;
	}
	if (vlSetupPaths(vlServer_, (VLPathList)&vlPath_, 1,
			 VL_SHARE, VL_SHARE) < 0) {
		vlPerror("vic: set up paths");
		goto failed;
	}
	val.intVal = VL_PACKING_YVYU_422_8;
	if (vlSetControl(vlServer_, vlPath_, drn_, VL_PACKING, &val) < 0) {
		vlPerror("vic: set control packing");
		goto failed;
	}

	val.fractVal.numerator = fps_;
	val.fractVal.denominator = 1;
	if (vlSetControl(vlServer_, vlPath_, drn_, VL_RATE, &val) < 0)
		vlPerror("vic: set rate");

	val.fractVal.numerator = 1;
	val.fractVal.denominator = decimate_;
	if (vlSetControl(vlServer_, vlPath_, drn_, VL_ZOOM, &val) < 0) {
		vlPerror("vic: vl set zoom");
		goto failed;
	}

	val.intVal = VL_CAPTURE_INTERLEAVED;
	if (vlSetControl(vlServer_, vlPath_, drn_, VL_CAP_TYPE, &val) < 0) {
		vlPerror("vic: vl set captype");
		goto failed;
	}

	vlGetControl(vlServer_, vlPath_, drn_, VL_SIZE, &val);
	xsize = val.xyVal.x;
	ysize = val.xyVal.y &~ 7;

	/* Create ring buffers to hold captured data of new size. */
	transferBuf_ = vlCreateBuffer(vlServer_, vlPath_, drn_, 2);
	if (transferBuf_ == 0) {
		vlPerror("vic: create Buffer");
		goto failed;
	}
	/* Tell VL to use newly created ring buffer. */
	if (vlRegisterBuffer(vlServer_, vlPath_, drn_, transferBuf_) < 0) {
		vlPerror("vic: buffer registration");
		goto failed;
	}
	/*
	 * Try to get into frame sync for the first transfer to minimize
	 * interlace artifacts in the captured picture.
	 * it would be nice to use VL_TRANSFER_MODE_AUTOTRIGGER here so
	 * we'd stay in frame sync but it seems to do nothing but make
	 * videod hang with a galileo.
	 */
#ifdef notdef
	VLTransferDescriptor xfer;
	xfer.mode = VL_TRANSFER_MODE_CONTINUOUS;
	xfer.trigger = VLFrameVerticalRetraceMask;
	xfer.delay = 0;
	xfer.count = 1;
	if (vlBeginTransfer(vlServer_, vlPath_, 1, &xfer) < 0) {
#endif
	if (vlBeginTransfer(vlServer_, vlPath_, 0, NULL) < 0) {
		vlPerror("vic: begin transfer");
		goto failed;
	}
	vlSelectEvents(vlServer_, vlPath_, VLNoEventsMask);

	setsize(xsize, ysize);
	return;
 failed:
	shutdown_vl();
	status_ = -1;
}

VLGrabber::VLGrabber(YuvEncoder* e) : encoder_(e)
{
	xsize_ = 0;
	ysize_ = 0;
	y_ = 0;
	current_decimate_ = -1;
	current_fps_ = -1;
	transferBuf_ = 0;
	Grabber::fps(2);
	decimate_ = 2;
	vlServer_ = 0;
	transferBuf_ = 0;
	vlPath_ = 0;
	return;
}

VLGrabber::~VLGrabber()
{
	shutdown_vl();
	delete y_;
}

VLCIFGrabber::VLCIFGrabber(YuvEncoder* e) : VLGrabber(e)
{
}

void VLCIFGrabber::setsize(int xsize, int ysize)
{
	if (xsize != xsize_ || ysize != ysize_) {
		xsize_ = xsize;
		ysize_ = ysize;

		int ow, oh;
		switch (xsize) {
		case 320:
			ow = 352; oh = 288;
			break;
		case 160:
			ow = 176; oh = 144;
			break;
		default:
			fprintf(stderr, "vic: vl CIF doesn't support %dx%d\n",
				xsize, ysize);
			exit(1);
		}
		outw_ = ow; outh_ = oh;
		u_int size = ow * oh;
		delete y_;
		y_ = new u_char[2 * size];
		u_ = y_ + size;
		v_ = u_ + size / 4;
		encoder_->size(ow, oh);
		crinit(ow, oh);
	}
}

u_char* VLCIFGrabber::grabline(register u_char* yp, register u_char* ye,
			       register u_char* p)
{
	while (yp < ye) {
		register u_char y;
		yp[0] = p[1];
		yp[1] = p[3];
		yp[2] = p[5];
		yp[3] = p[7];
		yp[4] = p[9];
		yp[5] = p[11];
		yp[6] = p[13];
		yp[7] = p[15];
		yp[8] = p[17];
		y     = p[19];
		yp[9] = y;
		yp[10]= y;
		p += 20;
		yp += 11;
	}
	return (p);
}

u_char* VLCIFGrabber::grabline(register u_char* yp, register u_char* ye,
			       register u_char* p, register u_char* up,
			       register u_char* vp)
{
	while (yp < ye) {
		register u_char y, u, v;
		up[0] = p[0];
		yp[0] = p[1];
		vp[0] = p[2];
		yp[1] = p[3];

		up[1] = p[4];
		yp[2] = p[5];
		vp[1] = p[6];
		yp[3] = p[7];

		up[2] = p[8];
		yp[4] = p[9];
		vp[2] = p[10];
		yp[5] = p[11];

		up[3] = p[12];
		yp[6] = p[13];
		vp[3] = p[14];
		yp[7] = p[15];

		u = p[16];
		up[4] = u;
		yp[8] = p[17];
		v = p[18];
		vp[4] = v;
		y  = p[19];
		yp[9] = y;

		up[5] = u;
		vp[5] = v;
		yp[10] = y;

		up[6]  = p[20];
		yp[11] = p[21];
		vp[6]  = p[22];
		yp[12] = p[23];

		up[7]  = p[24];
		yp[13] = p[25];
		vp[7]  = p[26];
		yp[14] = p[27];

		up[8]  = p[28];
		yp[15] = p[29];
		vp[8]  = p[30];
		yp[16] = p[31];

		up[9]  = p[32];
		yp[17] = p[33];
		vp[9]  = p[34];
		yp[18] = p[35];

		up[10] = p[36];
		yp[19] = p[37];
		vp[10] = p[38];
		y      = p[39];
		yp[20] = y;
		yp[21] = y;

		p += 40;
		yp += 22;
		up += 11;
		vp += 11;
	}
	return (p);
}

int VLCIFGrabber::grab()
{
	/*
	 * This is similar to VLGrabber:grab() except it converts
	 * an NTSC image to CIF dimensions (by duping every 10th
	 * horiz. pixel & every 4th line) and it decimates the
	 * uv info by 2 vertically to convert 4:2:2 to 4:1:1.
	 */
	register u_char* p = next_frame();
	if (p == 0)
		return (0);

	register u_char* yp = y_;
	register u_char* up = u_;
	register u_char* vp = v_;
	register u_int stride = outw_;
	register u_int s = stride >> 1;
	for (register int v = ysize_; v > 0; v -= 10) {
		register u_char* ye = yp + stride;
		p = grabline(yp, ye, p);
		yp = ye; ye += stride;
		p = grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		p = grabline(yp, ye, p);
		yp = ye; ye += stride;
		p = grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		p = grabline(yp, ye, p);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride);
		yp = ye + stride; ye = yp + stride;

		p = grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
		p = grabline(yp, ye, p);
		yp = ye; ye += stride;

		p = grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
		p = grabline(yp, ye, p);

		yp = ye; ye += stride;
		p = grabline(yp, ye, p, up, vp);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride);
		memcpy((char*)(up + s), up, s);
		memcpy((char*)(vp + s), vp, s);
		yp = ye + stride;
		up += stride; vp += stride;
	}
	vlPutFree(vlServer_, transferBuf_);
	Grabber::suppress(y_, cache_);
	Grabber::saveblks(y_);
	return (encoder_->encode(ntptime(), y_, crvec_));
}
