/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber-vpix.C,v 1.2 1995/08/16 00:45:58 olson Exp (LBL)";

#include <osfcn.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>

extern "C" {
#include <vfc_lib.h>
}

#include "grabber.h"
#include "framer.h"
#include "Tcl.h"

/*XXX*/
#define NTSC_WIDTH 320
#define NTSC_HEIGHT 240
#define PAL_WIDTH 384
#define PAL_HEIGHT 288
#define CIF_WIDTH 352
#define CIF_HEIGHT 288

/*XXX*/
#if defined(sun) && defined(__GNUC__)
extern "C" int ioctl(int fd, int request, caddr_t arg);
#endif

class VideoPixGrabber : public Grabber {
 public:
	VideoPixGrabber(YuvEncoder*);
	virtual ~VideoPixGrabber();
	virtual void start();
 protected:
	virtual int command(int argc, char** argv);
	virtual int capture();
	virtual int grab();
	virtual void NTSCgrabSmall();
	virtual void NTSCgrabMedium();
	virtual void PALgrabSmall();
	virtual void PALgrabMedium();
	void format();
	void contrast(double b);
	void normalize();
	virtual void setsize();

	int format_;		/* video format: NTSC or PAL */
	int rformat_;		/* requested format, above + AUTO */
	YuvEncoder* encoder_;
	VfcDev* vfcdev_;
	u_char* frame_;
	int framesize_;
	int port_;		/* videopix input port */
	u_int width_;
	u_int height_;
	u_int basewidth_;
	u_int baseheight_;
	u_int decimate_;
	u_int ymin_;
	u_int ymax_;
	double contrast_;
	u_char ynorm_[256];
};

class CIFVideoPixGrabber : public VideoPixGrabber {
 public:
	CIFVideoPixGrabber(YuvEncoder*);
 protected:
	virtual void setsize();
	void grablineMedium(volatile u_int* iochan, u_int* yp, u_int* ye,
		      u_short* up, u_short* vp);
	void grablineMedium(volatile u_int* iochan, u_int* yp, u_int* ye);
	void grablineSmall(volatile u_int* iochan, u_int* yp, u_int* ye,
		      u_short* up, u_short* vp);
	void grablineSmall(volatile u_int* iochan, u_int* yp, u_int* ye);
	virtual void NTSCgrabMedium();
	virtual void PALgrabMedium();
	virtual void NTSCgrabSmall();
	virtual void PALgrabSmall();

	int outw_;
	int outh_;
};

class VideoPixDevice : public Device {
 public:
	VideoPixDevice(const char*);
	virtual int match(const char* type) const;
	virtual Grabber* yuv_grabber(YuvEncoder*);
	virtual Grabber* cif_grabber(YuvEncoder*);
};

static VideoPixDevice vpix_device("VideoPix");

VideoPixDevice::VideoPixDevice(const char* name) : Device(name)
{
#ifdef notyet
	attrs_ = DEV_ATTR_RAW|DEV_ATTR_CIF|DEV_ATTR_SMALL|DEV_ATTR_LARGE|
		DEV_ATTR_PORT|DEV_ATTR_TYPE;
#else
	attrs_ = DEV_ATTR_RAW|DEV_ATTR_CIF|DEV_ATTR_SMALL|
		 DEV_ATTR_PORT|DEV_ATTR_TYPE;
#endif
}

int VideoPixDevice::match(const char* type) const
{
	return (strcasecmp(type, "vpix") == 0 ||
		strcasecmp(type, "videopix") == 0 ||
		strcasecmp(type, "vfc") == 0);
}

Grabber* VideoPixDevice::yuv_grabber(YuvEncoder* p)
{
	return (new VideoPixGrabber(p));
}

Grabber* VideoPixDevice::cif_grabber(YuvEncoder* p)
{
	return (new CIFVideoPixGrabber(p));
}

VideoPixGrabber::VideoPixGrabber(YuvEncoder* p) : encoder_(p), frame_(0)
{
	/*
	 * This is a horrible hack to prevent the VFC library from printing
	 * error messages that we would rather not print.
	 */
	int fd = dup(2);
	close(2);
	open("/dev/null", O_RDONLY);
	vfcdev_ = vfc_open("/dev/vfc0", VFC_LOCKDEV);
	dup2(fd, 2);
	close(fd);
	if (vfcdev_ == 0) {
		status_ = -1;
		return;
	}
	port_ = 1;
	/*XXX*/
	format_ = -1;
	rformat_ = VFC_AUTO;
	decimate_ = 2;
	basewidth_ = 0;
	baseheight_ = 0;
	width_ = 0;
	height_ = 0;
	ymin_ = 16;
	ymax_ = 235;
	contrast_ = 1.0;
	for (int i = 0; i < 256; ++i)
		ynorm_[i] = i;
}

VideoPixGrabber::~VideoPixGrabber()
{
	if (vfcdev_ != 0) {
		vfc_destroy(vfcdev_);
		vfcdev_ = 0;
	}
	delete frame_;
}

void VideoPixGrabber::setsize()
{
	if (format_ < 0)
		return;
	int w = basewidth_, h = baseheight_;
	w /= decimate_; h /= decimate_;
	width_ = w; height_ = h;
	framesize_ = w * h;
	delete frame_;
	frame_ = new u_char[framesize_ * 2];
	encoder_->size(w, h);
	crinit(w, h);
}

void VideoPixGrabber::format()
{
 	int format;
	vfc_set_port(vfcdev_, port_);
	vfc_set_format(vfcdev_, rformat_, &format);
	int w, h;
	switch (format) {

	default:
 	case NO_LOCK:
		if (rformat_ == VFC_AUTO)
			fprintf(stderr,
			"vic: videopix sees no signal - using ntsc.\n");
		/* fall through */
	case NTSC_COLOR:
	case NTSC_NOCOLOR:
		format_ = VFC_NTSC;
		w = NTSC_WIDTH * 2;
		h = NTSC_HEIGHT * 2;
		break;

	case PAL_COLOR:
	case PAL_NOCOLOR:
		format_ = VFC_PAL;
		w = PAL_WIDTH * 2;
		h = PAL_HEIGHT * 2;
		break;
	}
	basewidth_ = w;
	baseheight_ = h;
	setsize();
}

void VideoPixGrabber::contrast(double c)
{
	if (! running_)
		return;

	/* map min to ccir-601 black (16) & max to ccir-601 white (235) */
	u_int min = ymin_, max = ymax_;
	double dmin = min, dmax = max;
	double dslope = 219. / (dmax - dmin) * c;
	double dy = 16.;
	for (int i = 0; i < min; ++i)
		ynorm_[i] = u_char(dy);
	for ( ; i < max; ++i) {
		ynorm_[i] = u_char(dy);
		if (dy < 235.)
			dy += dslope;
	}
	for ( ; i < 256; ++i)
		ynorm_[i] = u_char(dy);

	contrast_ = c;
}

void VideoPixGrabber::normalize()
{
	if (! running_)
		return;

	/* set the intensity map to an identity map & capture a frame. */
	for (int i = 0; i < 256; ++i)
		ynorm_[i] = u_char(i);

	capture();

	/* find the min & max y values */
	u_char* fp = frame_;
	u_int min = *fp++;
	u_int max = min;
	for (i = framesize_; --i > 0; ) {
		u_int p = *fp++;
		if (p < min)
			min = p;
		else if (p > max)
			max = p;
	}
	/*
	 * if the range of the signal is small (i.e., all black) don't
	 * do anything.  Otherwise reset the min & max based on the
	 * frame & compute a new intensity map.
	 */
printf("min = %d, max = %d\n", min, max);
	if (max - min < 64)
		return;

	ymin_ = min;
	ymax_ = max;
	contrast(contrast_);
}

void VideoPixGrabber::start()
{
	if (running_)
		return;

	format();
	Grabber::start();
}

int VideoPixGrabber::command(int argc, char** argv)
{
	if (argc == 4) {
		if (strcmp(argv[2], "decimate") == 0) {
			int dec = atoi(argv[3]);
			Tcl& tcl = Tcl::instance();
			if (dec <= 0) {
				Tcl_AppendResult(tcl.interp(),
						 "%s: divide by zero",
						 argv[0], 0);
				return (TCL_ERROR);
			}
			if (dec != decimate_) {
				decimate_ = dec;
				setsize();
			}
		} else if (strcmp(argv[2], "port") == 0) {
			int p = atoi(argv[3]);
			if (p != port_) {
				port_ = p;
				vfc_set_port(vfcdev_, p);
			}
		} else if (strcmp(argv[2], "format") == 0) {
			if (strcmp(argv[3], "auto") == 0)
				rformat_ = VFC_AUTO;
			else if (strcmp(argv[3], "pal") == 0)
				rformat_ = VFC_PAL;
			else
				rformat_ = VFC_NTSC;
			if (running_)
				format();
		} else if (strcmp(argv[2], "contrast") == 0) {
			contrast(atof(argv[3]));
		}
	} else if (argc == 3) {
		if (strcmp(argv[2], "normalize") == 0)
			normalize();
		else if (strcmp(argv[2], "format") == 0) {
			Tcl& tcl = Tcl::instance();
			switch (format_) {

			case VFC_AUTO:
				tcl.result("auto");
				break;

			case VFC_NTSC:
				tcl.result("ntsc");
				break;

			case VFC_PAL:
				tcl.result("pal");
				break;

			default:
				tcl.result("");
				break;
			}
		}
	}
	return (TCL_OK);
}

#ifdef VPIX_SUN4C
typedef unsigned long long u_ll;

#define PIX0 (d0.i.h)
#define PIX1 (d0.i.l)
#define PIX2 (d1.i.h)
#define PIX3 (d1.i.l)
#define BW_CAPTURE_DEFS \
	union { \
		u_ll l; \
		struct { \
			u_int h; \
			u_int l; \
		} i; \
	} d0, d1; \
	register u_int y; \
	register u_char* ynorm = ynorm_;
#define CAPTURE_DEFS \
	BW_CAPTURE_DEFS \
	register u_short u, v;
#define NEXT4 \
	d0.l = *(volatile u_ll*)iochan; \
	d0.i.h >>= 16; \
	d0.i.l >>= 16; \
	d1.l = *(volatile u_ll*)iochan; \
	d1.i.h >>= 16; \
	d1.i.l >>= 16;
#define PRESKIP(n) \
	for (int i = ((n) - 1) / 2; --i >= 0; ) \
		*(volatile u_ll*)iochan; \
	*iochan;
#define LINESKIP(n) \
	for (int i = (n) / 2; --i >= 0; ) \
		*(volatile u_ll*)iochan;
#else
#define PIX0 (d0)
#define PIX1 (d1)
#define PIX2 (d2)
#define PIX3 (d3)
#define BW_CAPTURE_DEFS \
	register u_int y, d0, d1, d2, d3; \
	register u_char* ynorm = ynorm_;
#define CAPTURE_DEFS \
	BW_CAPTURE_DEFS \
	register u_short u, v;
#define NEXT4 \
	d0 = *iochan >> 16; \
	d1 = *iochan >> 16; \
	d2 = *iochan >> 16; \
	d3 = *iochan >> 16;
#define PRESKIP(n) \
	for (int i = (n) - 1; --i >= 0; ) \
		*iochan;
#define LINESKIP(n) \
	for (int i = (n); --i >= 0; ) \
		*iochan;
#endif

#define YSTART(r) \
	y = ynorm[(r) >> 8];
#define YADD(r) \
	y <<= 8; \
	y |= ynorm[(r) >> 8];
#define UVHIGH \
	u  = (PIX0 & 0xc0) << 8; \
	u |= (PIX1 & 0xc0) << 6; \
	u |= (PIX2 & 0xc0) << 4; \
	v  = (PIX0 & 0x30) << 10; \
	v |= (PIX1 & 0x30) << 8; \
	v |= (PIX2 & 0x30) << 6;
#define UVLOW \
	u |= (PIX0 & 0xc0); \
	u |= (PIX1 & 0xc0) >> 2; \
	u |= (PIX2 & 0xc0) >> 4; \
	v |= (PIX0 & 0x30) << 2; \
	v |= (PIX1 & 0x30); \
	v |= (PIX2 & 0x30) >> 2; \
	u ^= 0x8080; v ^= 0x8080;

void VideoPixGrabber::NTSCgrabMedium()
{
	register volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_OSKIP_NTSC)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 1;
	u_short* vp = (u_short*)(frame_ + off);

	for (int h = NTSC_HEIGHT; --h >= 0; ) {
		LINESKIP(16)
		for (int w = NTSC_WIDTH / 20; --w >= 0; ) {
			CAPTURE_DEFS

			NEXT4 YSTART(PIX0) YADD(PIX2) UVHIGH
			NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
			yp[0] = y; up[0] = u; vp[0] = v;

			NEXT4 YSTART(PIX0) YADD(PIX2) UVHIGH
			NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
			yp[1] = y; up[1] = u; vp[1] = v;

			NEXT4 YSTART(PIX0) UVHIGH /*did 9, skip 1*/
			NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
			NEXT4 YADD(PIX0)
			yp[2] = y; up[2] = u; vp[2] = v;

			YSTART(PIX2) UVHIGH
			NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
			NEXT4 YADD(PIX0)
			yp[3] = y; up[3] = u; vp[3] = v;

			YSTART(PIX2) UVHIGH
			NEXT4 YADD(PIX0) /* did 9, skip 1 */
			NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
			yp[4] = y; up[4] = u; vp[4] = v;

			yp += 5; up += 5; vp += 5;
		}
	}
}

void VideoPixGrabber::NTSCgrabSmall()
{
	register volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_OSKIP_NTSC)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 1;
	u_short* vp = (u_short*)(frame_ + off);

	for (int h = NTSC_HEIGHT / 2; --h >= 0; ) {
		LINESKIP(720 + 16)
		for (int w = NTSC_WIDTH / 40; --w >= 0; ) {
			CAPTURE_DEFS

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0)
			yp[0] = y; up[0] = u; vp[0] = v;

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0)
			yp[1] = y; up[1] = u; vp[1] = v;

			/*after 9, skip 1*/
			NEXT4 YSTART(PIX0) NEXT4 UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0)
			yp[2] = y; up[2] = u; vp[2] = v;

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0)
			yp[3] = y; up[3] = u; vp[3] = v;

			/*after 9, skip 1*/
			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0) NEXT4
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0)
			yp[4] = y; up[4] = u; vp[4] = v;

			yp += 5; up += 5; vp += 5;
		}
	}
}

void VideoPixGrabber::PALgrabMedium()
{
	volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_ESKIP_PAL)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 1;
	u_short* vp = (u_short*)(frame_ + off);

	for (int h = PAL_HEIGHT; --h >= 0; ) {
		LINESKIP(16)
		for (int w = PAL_WIDTH / 24; --w >= 0; ) {
			CAPTURE_DEFS

			NEXT4 YSTART(PIX0) YADD(PIX2) UVHIGH
			NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
			yp[0] = y; up[0] = u; vp[0] = v;

			NEXT4 YSTART(PIX0) YADD(PIX2) UVHIGH
			NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
			yp[1] = y; up[1] = u; vp[1] = v;

			NEXT4 YSTART(PIX0) YADD(PIX2) UVHIGH
			NEXT4 YADD(PIX0) YADD(PIX1) UVLOW	/*dup 11th*/
			yp[2] = y; up[2] = u; vp[2] = v;

			YSTART(PIX2) NEXT4 YADD(PIX0) UVHIGH
			YADD(PIX2) NEXT4 YADD(PIX0) UVLOW
			yp[3] = y; up[3] = u; vp[3] = v;

			YSTART(PIX2) NEXT4 YADD(PIX0) UVHIGH
			YADD(PIX2) NEXT4 YADD(PIX0) UVLOW
			yp[4] = y; up[4] = u; vp[4] = v;

			YSTART(PIX2) NEXT4 YADD(PIX0) UVHIGH
			YADD(PIX2) YADD(PIX3) UVLOW		/*dup 11th*/
			yp[5] = y; up[5] = u; vp[5] = v;

			yp += 6; up += 6; vp += 6;
		}
	}
}

void VideoPixGrabber::PALgrabSmall()
{
	volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_ESKIP_PAL)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 1;
	u_short* vp = (u_short*)(frame_ + off);

	for (int h = PAL_HEIGHT / 2; --h >= 0; ) {
		LINESKIP(720 + 16)
		for (int w = PAL_WIDTH / 48; --w >= 0; ) {
			CAPTURE_DEFS

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0) 
			yp[0] = y; up[0] = u; vp[0] = v;

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0) 
			yp[1] = y; up[1] = u; vp[1] = v;

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW YADD(PIX2)	/*dup 11th*/
			yp[2] = y; up[2] = u; vp[2] = v;

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0)
			yp[3] = y; up[3] = u; vp[3] = v;

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW NEXT4 YADD(PIX0)
			yp[4] = y; up[4] = u; vp[4] = v;

			NEXT4 YSTART(PIX0) UVHIGH NEXT4 YADD(PIX0)
			NEXT4 YADD(PIX0) UVLOW YADD(PIX2)	/*dup 11th*/
			yp[5] = y; up[5] = u; vp[5] = v;

			yp += 6; up += 6; vp += 6;
		}
	}
}

int VideoPixGrabber::capture()
{
	int cmd = CAPTRCMD;
	if (ioctl(vfcdev_->vfc_fd, VFCSCTRL, (char*)&cmd) == -1)
		return (0);

	switch (format_ | (decimate_ << 1)) {

	case VFC_PAL | (2 << 1):
		PALgrabMedium();
		break;

	case VFC_PAL | (4 << 1):
		PALgrabSmall();
		break;

	case VFC_NTSC | (2 << 1):
		NTSCgrabMedium();
		break;

	case VFC_NTSC | (4 << 1):
		NTSCgrabSmall();
		break;

	default:
		return (0);
	}
	return (1);
}

int VideoPixGrabber::grab()
{
	if (capture() == 0)
		return (0);
	suppress(frame_, cache_);
	saveblks(frame_);
	return (encoder_->encode(ntptime(), frame_, crvec_));
}


CIFVideoPixGrabber::CIFVideoPixGrabber(YuvEncoder* p)
	: VideoPixGrabber(p)
{
}

void CIFVideoPixGrabber::setsize()
{
	if (format_ < 0)
		return;
	int w = basewidth_, h = baseheight_;
	w /= decimate_; h /= decimate_;
	width_ = w; height_ = h;
	switch (w) {
	case 320:	// 1/2 size NTSC
	case 384:	// 1/2 size PAL
		outw_ = 352;
		outh_ = 288;
		break;
	case 160:	// 1/4 size NTSC
	case 192:	// 1/4 size PAL
		outw_ = 176;
		outh_ = 144;
		break;
	default:
		fprintf(stderr,
			"vic: VPIX CIF grabber doesn't support size %dx%d\n",
			width_, height_);
		exit(1);
	}
	int s = outw_ * outh_;
	delete frame_;
	frame_ = new u_char[s + (s >> 1)];
	framesize_ = s;
	crinit(outw_, outh_);
	encoder_->size(outw_, outh_);
}

void CIFVideoPixGrabber::grablineMedium(volatile u_int* iochan,
				u_int* yp, u_int* ye, u_short* up, u_short* vp)
{
	LINESKIP(16)
	while (yp < ye) {
		CAPTURE_DEFS

		NEXT4 YSTART(PIX0) YADD(PIX2) UVHIGH
		NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
		yp[0] = y; up[0] = u; vp[0] = v;

		NEXT4 YSTART(PIX0) YADD(PIX2) UVHIGH
		NEXT4 YADD(PIX0) YADD(PIX2) UVLOW
		yp[1] = y; up[1] = u; vp[1] = v;

		yp += 2; up += 2; vp += 2;
	}
}

void CIFVideoPixGrabber::grablineMedium(volatile u_int* iochan,
				u_int* yp, u_int* ye)
{
	LINESKIP(16)
	while (yp < ye) {
		BW_CAPTURE_DEFS

		NEXT4 YSTART(PIX0) YADD(PIX2)
		NEXT4 YADD(PIX0) YADD(PIX2)
		yp[0] = y;

		NEXT4 YSTART(PIX0) YADD(PIX2)
		NEXT4 YADD(PIX0) YADD(PIX2)
		yp[1] = y;

		yp += 2;
	}
}

void CIFVideoPixGrabber::grablineSmall(volatile u_int* iochan,
				u_int* yp, u_int* ye, u_short* up, u_short* vp)
{
	LINESKIP(720 + 16)
	while (yp < ye) {
		CAPTURE_DEFS

		NEXT4 YSTART(PIX0) NEXT4 YADD(PIX0) UVHIGH
		NEXT4 YADD(PIX0) NEXT4 YADD(PIX0) UVLOW
		yp[0] = y; up[0] = u; vp[0] = v;

		NEXT4 YSTART(PIX0) NEXT4 YADD(PIX0) UVHIGH
		NEXT4 YADD(PIX0) NEXT4 YADD(PIX0) UVLOW
		yp[1] = y; up[1] = u; vp[1] = v;

		yp += 2; up += 2; vp += 2;
	}
}

void CIFVideoPixGrabber::grablineSmall(volatile u_int* iochan,
				u_int* yp, u_int* ye)
{
	LINESKIP(720 + 16)
	while (yp < ye) {
		BW_CAPTURE_DEFS

		NEXT4 YSTART(PIX0) NEXT4 YADD(PIX0)
		NEXT4 YADD(PIX0) NEXT4 YADD(PIX0)
		yp[0] = y;

		NEXT4 YSTART(PIX0) NEXT4 YADD(PIX0)
		NEXT4 YADD(PIX0) NEXT4 YADD(PIX0)
		yp[1] = y;

		yp += 2;
	}
}

void CIFVideoPixGrabber::NTSCgrabMedium()
{
	volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_OSKIP_NTSC)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 2;
	u_short* vp = (u_short*)(frame_ + off);
	register u_int stride = outw_ >> 2;
	register u_int s = stride;
	for (register int v = height_; v > 0; v -= 10) {
		register u_int* ye = yp + stride;
		grablineMedium(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineMedium(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		grablineMedium(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineMedium(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		grablineMedium(iochan, yp, ye);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride << 2);
		yp = ye + stride; ye = yp + stride;

		grablineMedium(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
		grablineMedium(iochan, yp, ye);
		yp = ye; ye += stride;

		grablineMedium(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
		grablineMedium(iochan, yp, ye);

		yp = ye; ye += stride;
		grablineMedium(iochan, yp, ye, up, vp);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride << 2);
		memcpy((char*)(up + s), up, s << 1);
		memcpy((char*)(vp + s), vp, s << 1);
		yp = ye + stride;
		up += s << 1; vp += s << 1;
	}
}

void CIFVideoPixGrabber::NTSCgrabSmall()
{
	volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_OSKIP_NTSC)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 2;
	u_short* vp = (u_short*)(frame_ + off);
	register u_int stride = outw_ >> 2;
	register u_int s = stride;
	for (register int v = height_; v > 0; v -= 10) {
		register u_int* ye = yp + stride;
		grablineSmall(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineSmall(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		grablineSmall(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineSmall(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		grablineSmall(iochan, yp, ye);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride << 2);
		yp = ye + stride; ye = yp + stride;

		grablineSmall(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
		grablineSmall(iochan, yp, ye);
		yp = ye; ye += stride;

		grablineSmall(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
		grablineSmall(iochan, yp, ye);

		yp = ye; ye += stride;
		grablineSmall(iochan, yp, ye, up, vp);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride << 2);
		memcpy((char*)(up + s), up, s << 1);
		memcpy((char*)(vp + s), vp, s << 1);
		yp = ye + stride;
		up += s << 1; vp += s << 1;
	}
}

void CIFVideoPixGrabber::PALgrabMedium()
{
	volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_ESKIP_PAL)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 2;
	u_short* vp = (u_short*)(frame_ + off);
	register u_int stride = outw_ >> 2;
	register u_int s = stride;
	for (register int v = height_; v > 0; v -= 4) {
		register u_int* ye = yp + stride;

		grablineMedium(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineMedium(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		grablineMedium(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineMedium(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
	}
}

void CIFVideoPixGrabber::PALgrabSmall()
{
	volatile u_int* iochan = (u_int*)vfcdev_->vfc_port1;
	PRESKIP(VFC_ESKIP_PAL)

	u_int* yp = (u_int*)frame_;
	int off = framesize_;
	u_short* up = (u_short*)(frame_ + off);
	off += off >> 2;
	u_short* vp = (u_short*)(frame_ + off);
	register u_int stride = outw_ >> 2;
	register u_int s = stride;
	for (register int v = height_; v > 0; v -= 4) {
		register u_int* ye = yp + stride;

		grablineSmall(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineSmall(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;

		grablineSmall(iochan, yp, ye);
		yp = ye; ye += stride;
		grablineSmall(iochan, yp, ye, up, vp);
		yp = ye; ye += stride; up += s; vp += s;
	}
}
