/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber-xil.C,v 1.2 1995/08/16 00:46:00 olson Exp (LBL)";
#endif

#include <osfcn.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include "framer.h"
#include "framer-jpeg.h"
#include "framer-h261.h"
#include "Tcl.h"

#include <tk.h>
#include <sys/param.h>
#include <xil/xil.h>
#include "grabber.h"
#include "crdef.h"

class Tcl;
class YuvEncoder;
class JpegFramer;
class H261Framer;

class XILGrabber : public Grabber, public IOHandler {
 protected:
	XILGrabber();
 public:
	virtual ~XILGrabber();
	virtual void start();
	virtual void stop();
	virtual int command(int argc, char** argv);
	virtual void fps(int);
	inline int is_pal() { return (max_fps_ == 25); }
 protected:
	void dispatch(int mask);
	int attr(const char* name, int value);
	int attr(const char* name);
	const char* cattr(const char* name);
	virtual void setsize();

	int port_;		/* XIL input port */
	XilSystemState xil_;
	XilImage source_;
	XilImage image_;
	double scale_;
	double nextframetime_;
	u_int width_;
	u_int height_;
	u_int framesize_;
	u_char* frame_;		/* our location for captured frame */
	int fd_;		/* rtvc data fd */
	u_int decimate_;
	u_int basewidth_;
	u_int baseheight_;
	u_int max_fps_;		/* 25 (PAL) or 30 (NTSC) */
};

class XILYuvGrabber : public XILGrabber {
 public:
	XILYuvGrabber(YuvEncoder*);
	virtual ~XILYuvGrabber();
 protected:
	virtual int grab();
	virtual void setsize();
	void saveblk(const u_char* in, u_char* yp, u_char* up, u_char* vp,
		     int stride, int istride);
	void saveblks(const u_char* in, u_char* lum, int istride);
	void suppress(const u_char* devbuf, const u_char* lumbuf, int istride);

	YuvEncoder* encoder_;
};


class XILCIFGrabber : public XILGrabber {
 public:
	XILCIFGrabber(YuvEncoder*);
 protected:
	virtual int grab();
	virtual void setsize();
	void saveblk(const u_char* in, u_char* yp, u_char* up, u_char* vp,
		     int stride, int istride);
	void saveblks(const u_char* in, u_char* lum, int istride);
	void suppress(const u_char* devbuf, const u_char* lumbuf, int istride);

	YuvEncoder* encoder_;
	int outw_;
	int outh_;
};

class XILCIFNGrabber : public XILCIFGrabber {
 public:
	XILCIFNGrabber(YuvEncoder* e) : XILCIFGrabber(e) { }
 protected:
	virtual int grab();
	u_char* grabline(register u_char* yp, register u_char* ye,
			 register u_char* p);
	u_char* grabline(register u_char* yp, register u_char* ye,
			 register u_char* p, register u_char* up,
			 register u_char* vp);
};


class XILCodecGrabber : public XILGrabber {
 public:
	XILCodecGrabber(const char* type);
	virtual ~XILCodecGrabber();
 protected:
	int cattr(const char* name, int value);
	u_char* capture(int& length);
	XilCis cis_;
};

class XILJpegGrabber : public XILCodecGrabber {
 public:
	XILJpegGrabber(JpegFramer*);
 protected:
	virtual int command(int argc, char** argv);
	void setq(int q);
	int grab();
	JpegFramer* framer_;
};

class XILp64Grabber : public XILCodecGrabber {
 public:
	XILp64Grabber(H261Framer*);
 protected:
	int grab();
	H261Framer* framer_;
};

class XILDevice : public Device {
 public:
	XILDevice(const char* s);
	virtual int match(const char* type) const;
	virtual Grabber* yuv_grabber(YuvEncoder*);
	virtual Grabber* cif_grabber(YuvEncoder*);
	virtual Grabber* jpeg_grabber(JpegFramer*);
#ifdef notyet
	virtual Grabber* h261_grabber(H261Framer*);
#endif
};

static XILDevice xil_device("XIL");

/*
 * XXX gcc 2.5.8 needs this useless stub.
 */
XILDevice::XILDevice(const char* s) : Device(s)
{
#ifdef notyet
	attrs_ = DEV_ATTR_RAW|DEV_ATTR_CIF|DEV_ATTR_SMALL|DEV_ATTR_LARGE|
		DEV_ATTR_HW_JPEG|DEV_ATTR_PORT|DEV_ATTR_TYPE;
#else
	attrs_ = DEV_ATTR_RAW|DEV_ATTR_CIF|DEV_ATTR_SMALL|
		 DEV_ATTR_HW_JPEG|DEV_ATTR_PORT;
#endif
}

int XILDevice::match(const char* type) const
{
	return (strcasecmp(type, "xil") == 0);
}

Grabber* XILDevice::jpeg_grabber(JpegFramer* p)
{
	return (new XILJpegGrabber(p));
}

Grabber* XILDevice::yuv_grabber(YuvEncoder* p)
{
	return (new XILYuvGrabber(p));
}

Grabber* XILDevice::cif_grabber(YuvEncoder* p)
{
	XILGrabber* g = new XILCIFNGrabber(p);
	if (g->is_pal()) {
		delete g;
		g = new XILCIFGrabber(p);
	}
	return (g);
}

#ifdef notyet
Grabber* XILDevice::h261_grabber(H261Framer* p)
{
	return (new XILp64Grabber(p));
}
#endif

#define XilDevice void*
extern "C" XilDevice xil_device_create(XilSystemState, const char*);
extern "C" xil_device_set_value(XilDevice, const char*, void*);
extern "C" xil_device_destroy(XilDevice);

XILGrabber::XILGrabber()
{
	xil_ = xil_open();
	if (xil_ == 0) {
		status_ = -1;
		return;
	}
	XilDevice device = xil_device_create(xil_, "SUNWrtvc");
	if (device == 0) {
		status_ = -1;
		return;
	}
	xil_device_set_value(device, "DEVICE_NAME", (void*)0);
	source_ = xil_create_from_device(xil_, "SUNWrtvc", (XilAttribute*)device);
	if (source_ == 0) {
		xil_close(xil_);
		status_ = -1;
		return;
	}
	xil_device_destroy(device);

	attr("PORT_V", 1);
	attr("IMAGE_SKIP", 1);
	attr("MAX_BUFFERS", 1);

	XilDataType datatype;
	u_int nbands;
	xil_get_info(source_, &basewidth_, &baseheight_, &nbands, &datatype);
	width_ = basewidth_ >> 1;
	height_ = baseheight_ >> 1;
	image_ = xil_create(xil_, width_, height_, nbands, datatype);
	decimate_ = 2;
	scale_ = .5;
	frame_ = 0;

	fd_ = attr("FILE_DESCRIPTOR");
	max_fps_ = attr("FORMAT_V") == 1? 25 : 30;
}

XILGrabber::~XILGrabber()
{
	if (status_ >= 0)
		xil_close(xil_);
}

void XILGrabber::setsize()
{
}

int XILGrabber::attr(const char* name, int value)
{
	int s = xil_set_device_attribute(source_, (char*)name, (void*)value);
	if (s != XIL_SUCCESS) {
		fprintf(stderr, "vic: can't set XIL attribute: %s\n", name);
		return (-1);
	}
	return (0);
}

int XILGrabber::attr(const char* name)
{
	int value;
	int s = xil_get_device_attribute(source_, (char*)name, (void**)&value);
	if (s != XIL_SUCCESS) {
		fprintf(stderr, "vic: can't set XIL attribute: %s\n", name);
		return (-1);
	}
	return (value);
}

const char* XILGrabber::cattr(const char* name)
{
	const char* value;
	int s = xil_get_device_attribute(source_, (char*)name, (void**)&value);
	if (s != XIL_SUCCESS) {
		fprintf(stderr, "vic: can't set XIL attribute: %s\n", name);
		return (0);
	}
	return (value);
}

int XILGrabber::command(int argc, char** argv)
{
	if (argc == 4) {
		if (strcmp(argv[2], "port") == 0) {
			int newport = 1;
			switch (atoi(argv[3])) {
			case 1: 
				newport = 1;
				break;
			case 2:
				newport = 2;
				break;
			case 3:
				newport = 0;
				break;
			default:
				Tcl& tcl = Tcl::instance();
				Tcl_AppendResult(tcl.interp(),
					 "%s: invalid input port %s", argv[0],
					 argv[3], 0);
				return (TCL_ERROR);
			}
			attr("PORT_V", newport);
			/* video format may have changed when port changed */
			XilDataType datatype;
			u_int nbands;
			xil_get_info(source_, &basewidth_, &baseheight_,
				     &nbands, &datatype);
			width_ = basewidth_ / decimate_;
			height_ = baseheight_ / decimate_;
			setsize();
			/* need to kick-start frame grabbing process */
			if (running_) {
				stop();
				start();
			}
		} else if (strcmp(argv[2], "decimate") == 0) {
			int d = atoi(argv[3]);
			if (d <= 0) {
				Tcl& tcl = Tcl::instance();
				Tcl_AppendResult(tcl.interp(),
					 "%s: divide by zero", argv[0], 0);
				return (TCL_ERROR);
			}
			scale_ = 1. / double(d);
			decimate_ = d;
			width_ = basewidth_ / d;
			height_ = baseheight_ / d;
			setsize();
		}
	}
	return (TCL_OK);
}

void XILGrabber::fps(int f)
{
	if (f <= 0)
		f = 1;
	else if (f > max_fps_)
		f = max_fps_;
	/* convert to skip count then back */
	int sc = max_fps_ / f;
	attr("IMAGE_SKIP", sc - 1);
	Grabber::fps(max_fps_ / sc);
}

void XILGrabber::start()
{
	if (! running_) {
		running_ = 1;
		link(fd_, TK_READABLE);
		double now = gettimeofday();
		frameclock_ = now;
		nextframetime_ = now + tick(grab());
	}
}

void XILGrabber::stop()
{
	if (running_) {
		running_ = 0;
		unlink();
	}
}

void XILGrabber::dispatch(int mask)
{
	double now = gettimeofday();
	if (nextframetime_ > now) {
		/*
		 * the frame is too early & we want to flush it.
		 * unfortunately, the clever fellows that wrote the
		 * sunvideo driver couldn't imagine why anyone would ever
		 * want to discard a video frame so they didn't provide
		 * any mechanism for flushing one.  So, we have to do
		 * a pointless 2MB read to get rid of the sucker.
		 */
		char* buf = new char[768 * 576 * 3];
		(void)pread(fd_, (void*)buf, sizeof(buf), (off_t)0x01000000);
		delete buf;
	} else
		nextframetime_ = tick(grab()) + now;
}

XILYuvGrabber::XILYuvGrabber(YuvEncoder* p) : encoder_(p)
{
	setsize();
}

XILYuvGrabber::~XILYuvGrabber()
{
	xil_destroy(image_);
	delete frame_;
}

void XILYuvGrabber::setsize()
{
	delete frame_;
	frame_ = 0;
	int s = width_ * height_;
	framesize_ = s;
	frame_ = new u_char[2 * s];
	crinit(width_, height_);
	encoder_->size(width_, height_);
}

/*XXX */
#ifdef notdef
#define ABS(t) ((t) - (((t) >> 31 & (t)) << 1))
#endif
#define ABS(t) t
inline int 
diff(const u_char* in, const u_char* frm)
{
	int t, s;
	t = in[0] - frm[0];
	s = ABS(t);
	t = in[3] - frm[1];
	s += ABS(t);
	t = in[6] - frm[2];
	s += ABS(t);
	t = in[9] - frm[3];
	s += ABS(t);
	t = in[12] - frm[4];
	s += ABS(t);
	t = in[15] - frm[5];
	s += ABS(t);
	t = in[18] - frm[6];
	s += ABS(t);
	t = in[21] - frm[7];
	s += ABS(t);

	/*
	 * XXX
	 */
	if (s < 0)
		s = -s;

	return (s);
}

inline void 
XILYuvGrabber::saveblk(const u_char* in,
	u_char* yp, u_char* up, u_char* vp, int stride, int istride)
{
	int cs = stride >> 1;
	register u_int mask = 0xff0000ff;
	for (int i = 8; --i >= 0;) {
		register u_int t, d0, d1, d2;

		d0 = ((u_int*)in)[0];	/* y0 u0 v0 y1 */
		d1 = ((u_int*)in)[1];	/* u1 v1 y2 u2 */
		d2 = ((u_int*)in)[2];	/* v2 y3 u3 v3 */

		t = d0 & (mask << 24);
		t |= (d0 & mask) << 16;
		t |= d1 & (mask << 8);
		t |= (d2 >> 16) & mask;
		((u_int*)yp)[0] = t;

		t = (d0 >> 16) << 8;
		t |= d1 & mask;
		((u_short*)up)[0] = t;

		t = (d0 >> 8) << 8;
		t |= d2 >> 24;
		((u_short*)vp)[0] = t;

		d0 = ((u_int*)in)[3];	/* y4 u4 v4 y5 */
		d1 = ((u_int*)in)[4];	/* u5 v5 y6 u6 */
		d2 = ((u_int*)in)[5];	/* v6 y7 u7 v7 */

		t = d0 & (mask << 24);
		t |= (d0 & mask) << 16;
		t |= d1 & (mask << 8);
		t |= (d2 >> 16) & mask;
		((u_int*)yp)[1] = t;

		t = (d0 >> 16) << 8;
		t |= d1 & mask;
		((u_short*)up)[1] = t;

		t = (d0 >> 8) << 8;
		t |= d2 >> 24;
		((u_short*)vp)[1] = t;

		in += istride;
		yp += stride;
		up += cs;
		vp += cs;
	}
}

void XILYuvGrabber::saveblks(const u_char* in, u_char* lum, int is)
{
	u_char* crv = crvec_;
	int off = framesize_;
	u_char* chm = lum + off;
	off >>= 1;
	int stride = 7 * stride_;
	int istride = is * 7;
	for (int y = 0; y < blkh_; ++y) {
		for (int x = 0; x < blkw_; ++x) {
			int s = *crv++;
			if ((s & (CR_HQ | CR_LQ)) != 0)
				saveblk(in, lum, chm, chm + off, stride_, is);

			in += 24;
			lum += 8;
			chm += 4;
		}
		lum += stride;
		chm += stride >> 1;
		in += istride;
	}
}

#define XIL_SUPPRESS_BODY(devbuf, lumbuf, is) \
	int off = scan_ * stride_; \
	const u_char* in = (devbuf) + scan_ * is; \
	const u_char* lum = (lumbuf) + off; \
 \
	int w = blkw_; \
	u_char* crv = crvec_; \
	int stride = stride_; \
	int istride = 8 * is - 3 * stride; \
	stride *= 7; \
	for (int y = 0; y < blkh_; y++) { \
		for (int x = 0; x < w; x++) { \
			int pri = diff(in, lum); \
			if (pri >= 64) { \
				crv[0] = CR_LQ; \
				if (y > 0) { \
					crv[-w] = CR_LQ; \
					if (x > 0) \
						crv[-w-1] = CR_LQ; \
					if (x < w - 1) \
						crv[-w+1] = CR_LQ; \
				} \
				if (y < blkh_ - 1) { \
					crv[w] = CR_LQ; \
					if (x > 0) \
						crv[w-1] = CR_LQ; \
					if (x < w - 1) \
						crv[w+1] = CR_LQ; \
				} \
				if (x > 0) \
					crv[-1] = CR_LQ; \
				if (x < w - 1) \
					crv[1] = CR_LQ; \
			} else { \
				/* \
				 * There was no motion.  Age this block. \
				 * Once we hit the age threshold, we \
				 * will set CR_HQ as a hint to send a \
				 * high-quality version of the block. \
				 * After this the block will stop aging, \
				 * until there is motion (it still can \
				 * be sent as background fill). \
				 */ \
				int s = *crv; \
				if ((s & (CR_HQ|CR_LQ|CR_BG)) == 0) { \
					s += 1; \
					if (s >= CR_AGETHRESH) \
						s |= CR_HQ; \
					*crv = s; \
				} \
			} \
			lum += 8; \
			in += 24; \
			++crv; \
		} \
		lum += stride; \
		in += istride; \
	} \
	/* \
	 * Run background fill algorithm. \
	 */ \
	idlefill();

void XILYuvGrabber::suppress(const u_char* devbuf, const u_char* lumbuf, int is)
{
	XIL_SUPPRESS_BODY(devbuf, lumbuf, is)
}

int XILYuvGrabber::grab()
{
	/*XXX*/
	xil_scale(source_, image_, "nearest", scale_, scale_);
	if (xil_export(image_) != XIL_SUCCESS) {
		fprintf(stderr, "vic: xil_export failed\n");
		abort();
	}
	union XilMemoryStorage layout;
	if (xil_get_memory_storage(image_, &layout) == 0) {
		fprintf(stderr, "vic: xil_get_memory_storage failed\n");
		abort();
	}
	int istride = layout.byte.scanline_stride;
	if (layout.byte.pixel_stride != 3) {
		fprintf(stderr, "vic: xil: bad pixel stride\n");
		abort();
	}
	register u_char* data = layout.byte.data;
	suppress(data, frame_, istride);
	saveblks(data, frame_, istride);
	xil_import(image_, 0);
	xil_toss(image_);
	return (encoder_->encode(ntptime(), frame_, crvec_));
}

XILCIFGrabber::XILCIFGrabber(YuvEncoder* p) : encoder_(p)
{
	setsize();
}

void XILCIFGrabber::setsize()
{
	delete frame_;
	frame_ = 0;

	switch (width_) {
	case 320:	// 1/2 size NTSC
	case 384:	// 1/2 size PAL
		outw_ = 352;
		outh_ = 288;
		break;
	case 160:	// 1/4 size NTSC
	case 192:	// 1/4 size PAL
		outw_ = 176;
		outh_ = 144;
		break;
	default:
		fprintf(stderr,
			"vic: XIL CIF grabber doesn't support size %dx%d\n",
			width_, height_);
		exit(1);
	}
	int s = outw_ * outh_;
	frame_ = new u_char[s + (s >> 1)];
	framesize_ = s;
	crinit(outw_, outh_);
	encoder_->size(outw_, outh_);
}

inline void 
XILCIFGrabber::saveblk(const u_char* in,
	u_char* yp, u_char* up, u_char* vp, int stride, int istride)
{
	int cs = stride >> 1;
	register u_int mask = 0xff0000ff;
	for (int i = 4; --i >= 0;) {
		register u_int t, d0, d1, d2;

		d0 = ((u_int*)in)[0];	/* y0 u0 v0 y1 */
		d1 = ((u_int*)in)[1];	/* u1 v1 y2 u2 */
		d2 = ((u_int*)in)[2];	/* v2 y3 u3 v3 */

		t = d0 & (mask << 24);
		t |= (d0 & mask) << 16;
		t |= d1 & (mask << 8);
		t |= (d2 >> 16) & mask;
		((u_int*)yp)[0] = t;

		t = (d0 >> 16) << 8;
		t |= d1 & mask;
		((u_short*)up)[0] = t;

		t = (d0 >> 8) << 8;
		t |= d2 >> 24;
		((u_short*)vp)[0] = t;

		d0 = ((u_int*)in)[3];	/* y4 u4 v4 y5 */
		d1 = ((u_int*)in)[4];	/* u5 v5 y6 u6 */
		d2 = ((u_int*)in)[5];	/* v6 y7 u7 v7 */

		t = d0 & (mask << 24);
		t |= (d0 & mask) << 16;
		t |= d1 & (mask << 8);
		t |= (d2 >> 16) & mask;
		((u_int*)yp)[1] = t;

		t = (d0 >> 16) << 8;
		t |= d1 & mask;
		((u_short*)up)[1] = t;

		t = (d0 >> 8) << 8;
		t |= d2 >> 24;
		((u_short*)vp)[1] = t;

		in += istride;
		yp += stride;

		/* do the 2nd (y only instead of yuv) line */

		d0 = ((u_int*)in)[0];	/* y0 u0 v0 y1 */
		d1 = ((u_int*)in)[1];	/* u1 v1 y2 u2 */
		d2 = ((u_int*)in)[2];	/* v2 y3 u3 v3 */

		t = d0 & (mask << 24);
		t |= (d0 & mask) << 16;
		t |= d1 & (mask << 8);
		t |= (d2 >> 16) & mask;
		((u_int*)yp)[0] = t;

		d0 = ((u_int*)in)[3];	/* y4 u4 v4 y5 */
		d1 = ((u_int*)in)[4];	/* u5 v5 y6 u6 */
		d2 = ((u_int*)in)[5];	/* v6 y7 u7 v7 */

		t = d0 & (mask << 24);
		t |= (d0 & mask) << 16;
		t |= d1 & (mask << 8);
		t |= (d2 >> 16) & mask;
		((u_int*)yp)[1] = t;

		in += istride;
		yp += stride;
		up += cs;
		vp += cs;
	}
}

void XILCIFGrabber::saveblks(const u_char* in, u_char* lum, int is)
{
	u_char* crv = crvec_;
	int off = framesize_;
	u_char* chm = lum + off;
	off >>= 2;
	int stride = stride_;
	int cstride = 3 * stride;
	int istride = 8 * is - cstride;
	stride *= 7;
	cstride >>= 1;
	for (int y = 0; y < blkh_; ++y) {
		for (int x = 0; x < blkw_; ++x) {
			int s = *crv++;
			if ((s & (CR_HQ | CR_LQ)) != 0)
				saveblk(in, lum, chm, chm + off, stride_, is);

			in += 24;
			lum += 8;
			chm += 4;
		}
		lum += stride;
		chm += cstride;
		in += istride;
	}
}

void XILCIFGrabber::suppress(const u_char* devbuf, const u_char* lumbuf,
			      int is)
{
	XIL_SUPPRESS_BODY(devbuf, lumbuf, is)
}

int XILCIFGrabber::grab()
{
	/*
	 * This is similar to XILGrabber:grab() except it converts
	 * an PAL image to CIF dimensions (by dropping the last
	 * 32 pixels of each line) and it decimates the
	 * uv info by 2 vertically to convert 4:2:2 to 4:1:1.
	 */
	xil_scale(source_, image_, "nearest", scale_, scale_);
	if (xil_export(image_) != XIL_SUCCESS) {
		fprintf(stderr, "vic: xil_export failed\n");
		abort();
	}
	union XilMemoryStorage layout;
	if (xil_get_memory_storage(image_, &layout) == 0) {
		fprintf(stderr, "vic: xil_get_memory_storage failed\n");
		abort();
	}
	register u_int istride = layout.byte.scanline_stride;
	register u_char* p = layout.byte.data;
	suppress(p, frame_, istride);
	saveblks(p, frame_, istride);
	return (encoder_->encode(ntptime(), frame_, crvec_));
}

u_char* XILCIFNGrabber::grabline(register u_char* yp, register u_char* ye,
				register u_char* p)
{
	while (yp < ye) {
		register u_char y;
		yp[0] = p[0];
		yp[1] = p[3];
		yp[2] = p[6];
		yp[3] = p[9];
		yp[4] = p[12];
		yp[5] = p[15];
		yp[6] = p[18];
		yp[7] = p[21];
		yp[8] = p[24];
		y     = p[27];
		yp[9] = y;
		yp[10]= y;
		p += 30;
		yp += 11;
	}
	return (p);
}

u_char* XILCIFNGrabber::grabline(register u_char* yp, register u_char* ye,
			       register u_char* p, register u_char* up,
			       register u_char* vp)
{
	while (yp < ye) {
		register u_char y;
		yp[0] = p[0];
		up[0] = p[1];
		vp[0] = p[2];
		yp[1] = p[3];

		yp[2] = p[6];
		up[1] = p[7];
		vp[1] = p[8];
		yp[3] = p[9];

		yp[4] = p[12];
		up[2] = p[13];
		vp[2] = p[14];
		yp[5] = p[15];

		yp[6] = p[18];
		up[3] = p[19];
		vp[3] = p[20];
		yp[7] = p[21];

		yp[8] = p[24];
		up[4] = p[25];
		vp[4] = p[26];
		y     = p[27];
		yp[9] = y;

		yp[10]= y;
		up[5]  = p[28];
		vp[5]  = p[29];
		yp[11] = p[30];

		yp[12] = p[33];
		up[6]  = p[34];
		vp[6]  = p[35];
		yp[13] = p[36];

		yp[14] = p[39];
		up[7]  = p[40];
		vp[7]  = p[41];
		yp[15] = p[42];

		yp[16] = p[45];
		up[8]  = p[46];
		vp[8]  = p[47];
		yp[17] = p[48];

		yp[18] = p[51];
		up[9]  = p[52];
		vp[9]  = p[53];
		yp[19] = p[54];

		y      = p[57];
		yp[20] = y;
		up[10] = p[58];
		vp[10] = p[59];
		yp[21] = y;

		p += 60;
		yp += 22;
		up += 11;
		vp += 11;
	}
	return (p);
}

int XILCIFNGrabber::grab()
{
	/*
	 * This is similar to XILGrabber:grab() except it converts
	 * an NTSC image to CIF dimensions (by duping every 10th
	 * horiz. pixel & every 4th line) and it decimates the
	 * uv info by 2 vertically to convert 4:2:2 to 4:1:1.
	 */
	xil_scale(source_, image_, "nearest", scale_, scale_);
	if (xil_export(image_) != XIL_SUCCESS) {
		fprintf(stderr, "vic: xil_export failed\n");
		abort();
	}
	union XilMemoryStorage layout;
	if (xil_get_memory_storage(image_, &layout) == 0) {
		fprintf(stderr, "vic: xil_get_memory_storage failed\n");
		abort();
	}
	register u_char* yp = frame_;
	register u_char* up = yp + framesize_;
	register u_char* vp = up + (framesize_ >> 2);
	register u_int xilstride = layout.byte.scanline_stride;
	register u_char* p = layout.byte.data;
	register u_int stride = outw_;
	register u_int s = stride >> 1;
	for (register int v = height_; v > 0; v -= 10) {
		register u_char* ye = yp + stride;
		grabline(yp, ye, p);
		yp = ye; ye += stride; p += xilstride;
		grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s; p += xilstride;

		grabline(yp, ye, p);
		yp = ye; ye += stride; p += xilstride;
		grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s; p += xilstride;

		grabline(yp, ye, p);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride);
		yp = ye + stride; ye = yp + stride; p += xilstride;

		grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s; p += xilstride;
		grabline(yp, ye, p);
		yp = ye; ye += stride; p += xilstride;

		grabline(yp, ye, p, up, vp);
		yp = ye; ye += stride; up += s; vp += s; p += xilstride;
		grabline(yp, ye, p);

		yp = ye; ye += stride; p += xilstride;
		grabline(yp, ye, p, up, vp);
		/* dup previous line */
		memcpy((char*)ye, (char*)yp, stride);
		memcpy((char*)(up + s), up, s);
		memcpy((char*)(vp + s), vp, s);
		yp = ye + stride;
		up += stride; vp += stride; p += xilstride;
	}
	xil_import(image_, 0);
	xil_toss(image_);
	Grabber::suppress(frame_, cache_);
	Grabber::saveblks(frame_);
	return (encoder_->encode(ntptime(), frame_, crvec_));
}


XILCodecGrabber::XILCodecGrabber(const char* type)
{
	cis_ = xil_cis_create(xil_, (char*)type);
	xil_cis_set_keep_frames(cis_, 1);
	xil_cis_set_max_frames(cis_, 1);
}

XILCodecGrabber::~XILCodecGrabber()
{
	xil_destroy(image_);
	xil_cis_destroy(cis_);
}

int XILCodecGrabber::cattr(const char* name, int value)
{
	(void)xil_cis_set_attribute(cis_, (char*)name, (void*)value);
	return (0);
}

u_char* XILCodecGrabber::capture(int& length)
{
	xil_scale(source_, image_, "nearest", scale_, scale_);
	xil_compress(image_, cis_);
	xil_cis_sync(cis_);
	xil_toss(image_);
	if (!xil_cis_has_frame(cis_))
		return (0);

	int cc, nf;
	u_char* p = (u_char*)xil_cis_get_bits_ptr(cis_, &cc, &nf);
	length = cc;
	return (p);
}

XILJpegGrabber::XILJpegGrabber(JpegFramer* p)
	: XILCodecGrabber("Jpeg"), framer_(p)
{
	cattr("ENCODE_411_INTERLEAVED", 1);
}

int XILJpegGrabber::command(int argc, char** argv)
{
	if (argc == 4 && strcmp(argv[2], "q") == 0) {
		int q = atoi(argv[3]);
		setq(q);
		framer_->setparams(q, width_, height_, 1);
		return (TCL_OK);
	} else if (argc == 4 && strcmp(argv[2], "decimate") == 0) {
		// silently ignore decimate command
		return (TCL_OK);
	}
	return (XILGrabber::command(argc, argv));
}

extern void jpeg_chroma_qt(int q, int* qt);
extern void jpeg_luma_qt(int q, int* qt);

void XILJpegGrabber::setq(int q)
{
	/*
	 * NB - if any entry of any quantization table is < 8, XIL
	 * will refuse to use the CL4000 for compression & instead
	 * do a raw capture followed by a software jpeg conversion.
	 * This will flush performance straight down the toilet.
	 */
	int qt[8][8];
	XilJpegQTable xq;
	jpeg_luma_qt(q, (int*)qt);
	xq.table = 0;
	xq.value = qt;
	(void)xil_cis_set_attribute(cis_, "QUANTIZATION_TABLE", (void*)&xq);
	jpeg_chroma_qt(q, (int*)qt);
	xq.table = 1;
	xq.value = qt;
	(void)xil_cis_set_attribute(cis_, "QUANTIZATION_TABLE", (void*)&xq);
}

int XILJpegGrabber::grab()
{
	/*XXX can get timestamp from xil */
	int cc;
	u_char* p = capture(cc);
	if (p == 0)
		return (0);
	/* get rid of the stupid jfif header that xil gratuitously adds */
	for (u_char* ep = p + cc; p < ep; ++p) {
		if (*p == 0xff) {
			++p;
			if  (*p == 0xda) {
				/* found start-of-scan marker */
				++p;
				if (p + 2 <= ep) {
					/* skip over SOS */
					u_int t = (p[0] << 8) | p[1];
					p += t;
					break;
				}
			}
		}
	}
	if (p >= ep)
		return (0);

	return (framer_->send(p, ep - p, ntptime()));
}

XILp64Grabber::XILp64Grabber(H261Framer* p)
	: XILCodecGrabber("H261"), framer_(p)
{
}

int XILp64Grabber::grab()
{
	/*XXX can get timestamp from xil */
	int cc;
	u_char* p = capture(cc);
	return (framer_->send(p, cc, ntptime()));
}

