/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Derived from code written by Eric Anderson, UCSD Computer Systems Lab,
 * 4/92 and hacked on by Jon Kay, Vach Kompella, Kevin Fall.
 * Ported to vic 1/94, Steven McCanne (mccanne@ee.lbl.gov) March 1993.
 */

/*
 * (c) Copyright 1994 HEWLETT-PACKARD COMPANY
 * 
 * To anyone who acknowledges that this file is provided 
 * "AS IS" without any express or implied warranty:
 * permission to use, copy, modify, and distribute this 
 * file for any purpose is hereby granted without fee, 
 * provided that the above copyright notice and this 
 * notice appears in all copies, and that the name of 
 * Hewlett-Packard Company not be used in advertising or 
 * publicity pertaining to distribution of the software 
 * without specific, written prior permission.  Hewlett-
 * Packard Company makes no representations about the 
 * suitability of this software for any purpose.
 *
 */

/*
 * Apply this patch to the extensions/xv/lib/Xv.c file:
*** Xv.c~	Thu Mar 31 11:13:47 1994
--- Xv.c	Fri May 27 11:38:33 1994
***************
*** 192,197 ****
--- 192,198 ----
  	  return(XvBadAlloc);
  	}
        (void)strncpy(name, u.string, size);
+       name[size] = '\0';
        pa->name = name;
  
        u.buffer += (size + 3) & ~3;
***************
*** 351,356 ****
--- 352,358 ----
  	  return(XvBadAlloc);
  	}
        strncpy(name, u.string, size);
+       name[size] = '\0';
        pe->name = name;
        pe++;

 * It fixes a problem with strings NOT being null terminated.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber-xv.C,v 1.2 1995/08/16 00:46:02 olson Exp (LBL)";
#endif

#include <osfcn.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/signal.h>
#include <sys/ipc.h>
extern "C" {
#include <sys/shm.h>
}

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

extern "C" {
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xvlib.h>
}

#include "grabber.h"
#include "crdef.h"
#include "framer.h"
#include "Tcl.h"

/* Sick little macro which will limit x to [0..255] with logical ops */
#define UCLIMIT(x) ((t = (x)), (t &= ~(t>>31)), (t | ~((t-256) >> 31)))
 
/* A variant of above which will limit x to [-128..127] */
#define SCLIMIT(x) (UCLIMIT((x)+128))

static u_char rgb2y[32768];
static u_char rgb2u[32768];
static u_char rgb2v[32768];
static int rgb_inited = 0;

#define NTSC_WIDTH 640
#define NTSC_HEIGHT 480
#define PAL_WIDTH 768
#define PAL_HEIGHT 576
#define CIF_WIDTH 352
#define CIF_HEIGHT 288

class XvGrabber : public Grabber {
 public:
	XvGrabber(YuvEncoder*);
	virtual ~XvGrabber();
	virtual int command(int argc, char** argv);
 protected:
	virtual const char* type() const;
	void setsize();
	void setattr(Atom attr, int value);
	void setport(char *port);
	void format(char *fmt);
	int getattr(Atom attr);
	int getgrabber();
	int create_capwin(int w, int  h, char *wname, char *iname);
	void destroy_capwin(void);
	XImage* create_sharedImage(int w, int h);
	XImage* create_image(int w, int h);
	virtual int grab();
	XvEncodingInfo *lookup_encoding_by_id(XvEncodingID id);    
	XvEncodingInfo *lookup_encoding_by_name(char *name);

	YuvEncoder* encoder_;
	Display* dpy_;
	int screen_;
	Window capwin_;
	XImage* image_;
	GC gc_;
	Visual* visual_;
	XvPortID grabID_;
	XShmSegmentInfo shminfo_;

	char grabber_name[80];
	Atom XAbrightness, XAcontrast, XAhue, XAsaturation, XAencoding;
	u_int nencodings;
	XvEncodingInfo *encoding;

	u_int basewidth_;
	u_int baseheight_;

	int width_;
	int height_;

	int decimate_;
	XvEncodingID  encodingid_;

	u_char* frame_;
	int framesize_;
};

class XvDevice : public Device {
 public:
	XvDevice(const char* s);
	virtual int match(const char* type) const;
	virtual Grabber* yuv_grabber(YuvEncoder*);
};

static XvDevice xv_device("xv");

XvDevice::XvDevice(const char* s) : Device(s)
{
	attrs_ = DEV_ATTR_RAW|DEV_ATTR_SMALL|DEV_ATTR_LARGE|DEV_ATTR_PORT|DEV_ATTR_TYPE;
}

int XvDevice::match(const char* type) const
{
	return (strcasecmp(type, "xv") == 0);
}


int XvGrabber::command(int argc, char** argv)
{
#if 0
	printf("XvDevice::command: %d: ", argc);
	for (i=0; i<argc; i++)
		printf("%s ", argv[i]);
	printf("\n");
#endif

	if (argc == 4) {
		if (strcmp(argv[2], "decimate") == 0) {
			int dec = atoi(argv[3]);
			Tcl& tcl = Tcl::instance();
			if (dec <= 0) {
				Tcl_AppendResult(tcl.interp(),
						 "%s: divide by zero",
						 argv[0], 0);
				return (TCL_ERROR);
			}
			if (dec != decimate_) {
				decimate_ = dec;
				setsize();
				destroy_capwin();
				if (create_capwin(width_, height_, grabber_name, "Xv") < 0) {
					tcl.result("Failed to create capture window for grabber");
					return (TCL_ERROR);
	}
			}
		} else if (strcmp(argv[2], "port") == 0) {
			setport(argv[3]);
		} else if (strcmp(argv[2], "type") == 0) {
			format(argv[3]);
			setsize();
			destroy_capwin();
			if (create_capwin(width_, height_, grabber_name, "Xv") < 0) {
				Tcl& tcl = Tcl::instance();
				tcl.result("Failed to create capture window for grabber");
				return (TCL_ERROR);
			}
		} else if (strcmp(argv[2], "contrast") == 0) {
			setattr(XAcontrast, atoi(argv[3]));
		} else if (strcmp(argv[2], "brightness") == 0) {
			setattr(XAbrightness, atoi(argv[3]));
		} else if (strcmp(argv[2], "hue") == 0) {
			setattr(XAhue, atoi(argv[3]));
		} else if (strcmp(argv[2], "saturation") == 0) {
			setattr(XAsaturation, atoi(argv[3]));
		}
	} else if (argc == 3) {
		if (strcmp(argv[2], "normalize") == 0)
			printf("normalize??\n");
		else if (strcmp(argv[2], "format") == 0) {
			Tcl& tcl = Tcl::instance();
			XvEncodingInfo *enc = lookup_encoding_by_id(encodingid_);
			char *name = strdup(enc->name);
			tcl.result(strtok(name, "-"));
			free(name);
		}
	}
	return (TCL_OK);
}

XvEncodingInfo *XvGrabber::lookup_encoding_by_id(XvEncodingID id)
{
	for (int i=0; i<nencodings; i++) {
		if (id == encoding[i].encoding_id)
			return &encoding[i];
	}
	return NULL;
}

XvEncodingInfo *XvGrabber::lookup_encoding_by_name(char *name)
{
	for (int i=0; i<nencodings; i++) {
		if (strncmp(name, encoding[i].name, strlen(name)) == 0)
			return &encoding[i];
	}
	return NULL;
}

void XvGrabber::format(char *fmt)
{
	XvEncodingInfo *enc = lookup_encoding_by_id(encodingid_);
	char encoding_name[256];
	char *enc_str = strdup(enc->name);
	
	char *enc_fmt = strtok(enc_str, "-");
	char *enc_port = strtok(NULL, "-");

	if (strcmp(fmt, "auto") == 0)
		fmt = "ntsc";	/* XXX */
		
	if (strcmp(fmt, "pal") == 0) {
		basewidth_ = PAL_WIDTH;
		baseheight_ = PAL_HEIGHT;
	}
	else {
		basewidth_ = NTSC_WIDTH;
		baseheight_ = NTSC_HEIGHT;
	}

	if (strcmp(fmt, enc_fmt) != 0) {
		sprintf(encoding_name, "%s-%s", fmt, enc_port);
		enc = lookup_encoding_by_name(encoding_name);
		if (enc != NULL) {
			XvSetPortAttribute(dpy_, grabID_, XAencoding, (int)enc->encoding_id);
			encodingid_ = enc->encoding_id;
		}
		else
			fprintf(stderr, "Invalid encoding \"%s\"\n", encoding_name);
	}
	free(enc_str);
}

void XvGrabber::setport(char *port)
{
	printf("setport: %s\n", port);
}

void XvGrabber::setattr(Atom attr, int value)
{
	XvSetPortAttribute(dpy_, grabID_, attr, value);
}

int XvGrabber::getattr(Atom attr)
{
	int v;
	XvGetPortAttribute(dpy_, grabID_, attr, &v);
	return(v);
}

void XvGrabber::setsize()
{
	int w = basewidth_, h = baseheight_;
	w /= decimate_; h /= decimate_;

	if (width_ != w || height_ != h) {
		width_ = w; height_ = h;

		framesize_ = w * h;
		delete frame_;
		frame_ = new u_char[framesize_ * 2];

		crinit(w, h);
		encoder_->size(w, h);
	}
}

Grabber* XvDevice::yuv_grabber(YuvEncoder* p)
{
	return (new XvGrabber(p));
}

int XvGrabber::grab()
{
	/*
	 * The brain dead video hardware requires that the window
	 * be unobscured in order to capture the video.  So we grab
	 * the server and raise the window.  This won't work if
	 * the window isn't mapped.  Also, we block signals while
	 * the server is grabbed.  Otherwise, the process could be
	 * stopped while the display is locked out.
	 */
	long oldmask = sigsetmask(~0);
	XGrabServer(dpy_);
	XSync(dpy_, 0);
	XRaiseWindow(dpy_, capwin_);
	XSync(dpy_, 0);
	XImage* p = image_;

	if (p != 0 && shminfo_.shmaddr)
		XShmGetImage(dpy_, capwin_, p, 0, 0, AllPlanes);
	else
		XGetSubImage(dpy_, capwin_, 0, 0, width_, height_, AllPlanes,
		     ZPixmap, p, 0, 0);

	XUngrabServer(dpy_);
	(void)sigsetmask(oldmask);

	int x, y;
	u_char* yp = (u_char*)frame_, *ry=rgb2y;
	int off = framesize_;
	u_char* up = (u_char*)(frame_ + off), *ru=rgb2u;
	off += off >> 1;
	u_char* vp = (u_char*)(frame_ + off), *rv=rgb2v;
	u_int p0, p1, *data=(u_int *)p->data;
 
	for (y=0; y<height_; y++) {
		for (x=0; x<width_; x+=2) {
			p0 = data[0];
			p1 = data[1];
			data += 2;
 
			yp[0] = ry[((p0<<7) & 0x7c00)+((p0>>6) & 0x3e0)+((p0>>19) & 0x1f)];
			yp[1] = ry[((p1<<7) & 0x7c00)+((p1>>6) & 0x3e0)+((p1>>19) & 0x1f)];
			yp += 2;
 
			p0 &= 0xfefeff;
			p1 &= 0xfefeff;
			p0 += p1;

			*up++ = ru[((p0<<6) & 0x7c00) +
				   ((p0>>7) & 0x3e0) +
				   ((p0>>20) & 0x1f)];
			*vp++ = rv[((p0<<6) & 0x7c00) +
				   ((p0>>7) & 0x3e0) +
				   ((p0>>20) & 0x1f)];
		}
	}

	suppress(frame_, cache_);
	saveblks(frame_);

	return (encoder_->encode(ntptime(), (u_char *)frame_, crvec_));
}

void XvGrabber::destroy_capwin(void)
{
	if (capwin_) {
		XvStopVideo(dpy_, grabID_, capwin_);
		XDestroyWindow(dpy_, capwin_);
		capwin_ = 0;
	}
	if (image_) {
		XDestroyImage(image_);
		image_ = NULL;
	}
	if (gc_) {
		XFreeGC(dpy_, gc_);
		gc_ = 0;
	}
}

int XvGrabber::create_capwin(int w, int  h, char *wname, char *iname)
{
	Window root = DefaultRootWindow(dpy_);
	XVisualInfo vi;
	Colormap colormap;

	if (DefaultDepth(dpy_, screen_) == 24) {
		visual_ = DefaultVisual(dpy_, screen_);
		colormap = XDefaultColormap(dpy_, screen_);
	} else if (XMatchVisualInfo(dpy_, screen_, 24, TrueColor, &vi)) {
		visual_ = vi.visual;
		/*XXX*/
		colormap = XCreateColormap(dpy_, root,
					    visual_, AllocNone);
#ifdef XV_PSEUDO8
	} else if (DefaultDepth(dpy_, screen_) == 8) {
		visual_ = DefaultVisual(dpy_, screen_);
		colormap = XDefaultColormap(dpy_, screen_);
		vi.depth = DefaultDepth(dpy_, screen_);
#endif
	} else
		return (-1);
	
	XSetWindowAttributes attr;
	attr.colormap = colormap;
	attr.event_mask = VisibilityChangeMask;
	attr.backing_store = Always;
	attr.border_pixel = BlackPixel(dpy_, screen_);
	attr.background_pixel = BlackPixel(dpy_, screen_);

	capwin_ = XCreateWindow(dpy_, root, 0, 0, w, h,
				/*XXX*/2, vi.depth, InputOutput, visual_,
				CWColormap|CWBorderPixel|CWBackPixel|
				CWEventMask|CWBackingStore,
				&attr);

	/* Initialize size hint property for window manager */
	XSizeHints size_hints;
	size_hints.flags  = PSize;
	size_hints.width  = w;
	size_hints.height = h;
	size_hints.flags |= USSize;

	/* Set properties for window manager (always before mapping) */
	XSetStandardProperties(dpy_, capwin_, wname, iname, None, 
			       (char **) 0, 0, &size_hints);


	/* Create Ximage for capture */
#ifdef XV_USES_XSHM
	image_ = create_sharedImage(width_, height_);
#else
	image_ = create_image(width_, height_);
#endif

	gc_ = XCreateGC(dpy_, capwin_, (u_long)0, (XGCValues *)0);
	XFlush(dpy_);
	XMapWindow(dpy_, capwin_);

	/* Start the video display */
	XvPutVideo(dpy_, grabID_, capwin_, gc_, 0, 0, basewidth_, baseheight_,
		   0, 0, width_, height_);

	return (0);
}

XImage* XvGrabber::create_sharedImage(int width, int height)
{
	extern int use_shm;
	if (!use_shm)
		return (0);

	XImage* p;
	p = XShmCreateImage(dpy_, visual_, 24, ZPixmap, NULL,
			    &shminfo_, width, height);

	shminfo_.shmid = shmget(IPC_PRIVATE, p->bytes_per_line * p->height,
				IPC_CREAT | 0777);
	if (shminfo_.shmid == -1) {
		perror("vic: shmget");
		exit(1);
	}
	shminfo_.shmaddr = (char *)shmat(shminfo_.shmid, 0, 0);
	p->data = shminfo_.shmaddr;

	if ((int)p->data == -1) {
		perror("vic: shmat");
		exit(1);
	}
	shminfo_.readOnly = False;
	if (XShmAttach(dpy_, &shminfo_) == 0) {
		perror("vic: XShmAttach");
		exit(1);
	}
	return (p);
}

XImage* XvGrabber::create_image(int width, int height)
{
	XImage* p;
	int ximage_size;

	p = XCreateImage(dpy_, visual_, 24, ZPixmap, 0, NULL,
			 width, height, 32, 0);
	ximage_size = p->bytes_per_line * p->height;
	p->data = (char *) malloc(ximage_size);

	shminfo_.shmaddr = NULL;

	return (p);
}

int XvGrabber::getgrabber()
{
	int majop, eventbase, errbase;
	if (XQueryExtension(dpy_, "XVideo", &majop,
			    &eventbase, &errbase) == False)
		return (-1);

	Window root = DefaultRootWindow(dpy_);
	u_int ngrabbers;
	XvAdaptorInfo* grabbers;
	XvQueryAdaptors(dpy_, root, &ngrabbers, &grabbers);
	if (ngrabbers == 0)
		return (-1);
	if (ngrabbers > 1)
		fprintf(stderr, "vic: warning: more than one frame grabber\n");
		
	strncpy(grabber_name, grabbers[0].name, sizeof(grabber_name));
	grabID_ = grabbers[0].base_id;

	XvFreeAdaptorInfo(grabbers);

	XvQueryEncodings(dpy_, grabID_, &nencodings, &encoding);

#if 0
	printf("Encodings(%d): ", nencodings);
	for (int i=0; i<nencodings; i++)
	  printf("%s ", encoding[i].name);
	printf("\n");
#endif

	XAbrightness = XInternAtom(dpy_, "XV_BRIGHTNESS", False);
	XAcontrast = XInternAtom(dpy_, "XV_CONTRAST", False);
	XAhue = XInternAtom(dpy_, "XV_HUE", False);
	XAsaturation = XInternAtom(dpy_, "XV_SATURATION", False);
	XAencoding = XInternAtom(dpy_, "XV_ENCODING", False);

	XvGetPortAttribute(dpy_, grabID_, XAencoding, (int *)&encodingid_);

	return (0);
}

/* Initialize the RGB555 to YUV tables */
static void InitRGBTables(void)
{
    int i, t, r, g, b, y, u, v;

    i = 0;
    for (r=0; r<256; r+=8) {
	for (g=0; g<256; g+=8) {
	    for (b=0; b<256; b+=8) {
		y = (38*r+75*g+15*b+64)/128;
		u = SCLIMIT(74*(b-y)/128);
		v = SCLIMIT(93*(r-y)/128);
		rgb2y[i] = y;
		rgb2u[i] = u;
		rgb2v[i] = v;
		i++;
	    }
	}
    }
}

XvGrabber::XvGrabber(YuvEncoder* e) : encoder_(e), frame_(0)
{
	if (!rgb_inited) {
		InitRGBTables();
		rgb_inited++;
	}

	Tcl& tcl = Tcl::instance();

	dpy_ = Tk_Display(tcl.tkmain());
	screen_ = Tk_ScreenNumber(tcl.tkmain());

	decimate_ = 0;
	basewidth_ = NTSC_WIDTH;
	baseheight_ = NTSC_HEIGHT;

	image_ = NULL;
	capwin_ = 0;
	gc_ = 0;

	if (getgrabber() < 0) {
		status_ = -1;
		return;
	}

	if (XvGrabPort(dpy_, grabID_, CurrentTime) != Success) {
		tcl.result("Unable to reserve grabber.");
		fprintf(stderr, "Unable to grab port\n");
		status_ = -1;
		return;
	}

}

XvGrabber::~XvGrabber()
{
	if (status_ >= 0) {
		XvUngrabPort(dpy_, grabID_, CurrentTime);
		destroy_capwin();
		delete frame_;
	}
}

const char* XvGrabber::type() const
{
	return ("xv");
}
