/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/grabber.C,v 1.3 1995/08/25 04:37:38 olson Exp (LBL)";
#endif

#if defined(__osf__) || defined(__ultrix__)
/*XXX they didn't get this one right */
extern "C" {
#include <sys/param.h>
#include <sys/uio.h>
}
#else
#include <sys/param.h>
#ifdef _AIX
#include <sys/types.h>
#endif
#include <sys/uio.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <osfcn.h>

#include "inet.h"
#include "grabber.h"
#include "decoder.h"
#include "net.h"
#include "Tcl.h"
#include "crdef.h"

#include "grabber-null.h"
#include "framer-jpeg.h"
#include "framer-h261.h"
#include "encoder-nv.h"
#include "encoder-cellb.h"
#include "encoder-h261.h"

#ifdef __osf__
extern "C" {
int sendmsg(int, struct msghdr*, int);
}
#endif

#ifdef __ultrix__
extern "C" int writev(int, struct iovec*, int);
#endif

static struct encodingEnt {
	char* name;
	int encoding;
} encodings[] = {
	{ "jpeg", RTP_PT_JPEG },
	{ "jpg", RTP_PT_JPEG },
	{ "mjpg", RTP_PT_JPEG },
	{ "mpg", RTP_PT_MPEG },
	{ "mpeg", RTP_PT_MPEG },
	{ "h261", RTP_PT_H261 },
	{ "h.261", RTP_PT_H261 },
	{ "nv", RTP_PT_NV },
	{ "scr", RTP_PT_SCR },
	{ "cellb", RTP_PT_CELLB },

	{ 0, -1 }
};

int
lookup_format(const char* name)
{
	encodingEnt* p = encodings;
	for (; p->name != 0; ++p)
		if (strcasecmp(p->name, name) == 0)
			break;
	return (p->encoding);
}

const char* 
lookup_format(int encoding)
{
	static char buf[32];

	encodingEnt* p = encodings;
	for (; p->name != 0; ++p)
		if (encoding == p->encoding)
			return (p->name);

	sprintf(buf, "fmt-0x%x", encoding);
	return (buf);
}

Grabber* Device::grabber_;
Framer* Device::framer_;

Grabber* Device::jpeg_grabber(JpegFramer*)
{
	return (0);
}

Grabber* Device::h261_grabber(H261Framer*)
{
	return (0);
}

Grabber* Device::yuv_grabber(YuvEncoder*)
{
	return (0);
}

Grabber* Device::cif_grabber(YuvEncoder*)
{
	return (0);
}

Device::Device(const char* name) : name_(name),
	default_format_(RTP_PT_NV), attrs_(0)
{
	next_ = alldev_;
	alldev_ = this;
}

void Device::init()
{
	Tcl& tcl = Tcl::instance();
	tcl.CreateCommand("device", Device::device_cmd);
	tcl.CreateCommand("encoder", Device::encoder_cmd);

#if defined(VIDEOPIX)
	tcl.add_default("device", "videopix");
	tcl.add_default("format", "nv");
#elif defined(XIL)
	tcl.add_default("device", "xil");
	tcl.add_default("format", "nv");
#elif defined(RTVC)
	tcl.add_default("device", "rtvc");
	tcl.add_default("format", "nv");
#elif defined(PARALLAX)
	tcl.add_default("device", "parallax");
	tcl.add_default("format", "jpeg");
#elif defined(JVIDEO)
	tcl.add_default("device", "jv");
	tcl.add_default("format", "jpeg");
#elif defined(SUNRISE)
	tcl.add_default("device", "sunrise");
	tcl.add_default("format", "nv");
#elif defined(MME)
	tcl.add_default("device", "mme");
	tcl.add_default("format", "jpeg");
#elif defined(TX)
	tcl.add_default("device", "tx");
	tcl.add_default("format", "nv");
#elif defined(VL)
	tcl.add_default("device", "vl");
	tcl.add_default("format", "nv");
#elif defined(SVIDEO)
	tcl.add_default("device", "svideo");
	tcl.add_default("format", "nv");
#elif defined(XV)
	tcl.add_default("device", "xv");
	tcl.add_default("format", "nv");
#else
	tcl.add_default("device", "none");
	tcl.add_default("format", "nv");/*XXX*/
	if (nulldev_ == 0)
		nulldev_ = new NullVideoDevice("none");
#endif
}

Device* Device::alldev_;
Device* Device::nulldev_;
Device& Device::lookup(const char* type)
{
	for (Device* p = alldev_; p != 0; p = p->next_)
		if (p->match(type))
			return (*p);
	if (nulldev_ == 0) {
		nulldev_ = new NullVideoDevice("none");
#ifdef notdef
		/*XXX hack: take null device off list */
		alldev_ = alldev_->next_;
#endif
	}
	return (*nulldev_);
}

int Device::compat(int fmt) const
{
	switch (fmt) {
	case RTP_PT_SCR:
	case RTP_PT_CELLB:
	case RTP_PT_NV:
		return (attrs_ & DEV_ATTR_RAW);

	case RTP_PT_H261:
		return (attrs_ & (DEV_ATTR_CIF|DEV_ATTR_HW_H261));

	case RTP_PT_JPEG:
		return (attrs_ & DEV_ATTR_HW_JPEG);
	}
	return (0);
}

int Device::supports(const char* attr) const
{
	if (strcasecmp(attr, "large") == 0)
		return (attrs_ & DEV_ATTR_LARGE);
	else if (strcasecmp(attr, "small") == 0)
		return (attrs_ & DEV_ATTR_SMALL);
	else if (strcasecmp(attr, "port") == 0)
		return (attrs_ & DEV_ATTR_PORT);
	else if (strcasecmp(attr, "type") == 0)
		return (attrs_ & DEV_ATTR_TYPE);
	return (0);
}

/*
 * Device commands:
 *	device open $device $format
 *	device default $device
 *	device open
 *	device close
 *	device start
 *	device close
 *	device list
 *	device fps $fps
 *	device bps $bps
 *	device fillrate $blocksperframe
 *	device format $device $format
 *	device supports $device [type|port|small|large|etc]
 *	device change $format
 */
int Device::device_cmd(ClientData cd, Tcl_Interp* t, int argc, char** argv)
{
	Tcl& tcl = Tcl::instance();
	int i;

#ifdef DEBUG
    printf("Device::device_cmd argc=%d ", argc);
    for (i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
#endif

	if (argc >= 2 && strcmp(argv[1], "ctrl") == 0) {
		if (grabber_ == 0) {
			/*XXX*/
			return (TCL_OK);
			goto no_grabber;
		}
		return (grabber_->command(argc, argv));
	}
	if (argc == 2) {
		if (strcmp(argv[1], "open") == 0) {
			if (grabber_ == 0)
				tcl.result("0");
			else
				tcl.result("1");
			return (TCL_OK);
		}
		if (strcmp(argv[1], "list") == 0) {
			/*
			 * Return a list of the names of the
			 * available capture devices.
			 */
			t->result = tcl.buffer();
			char* cp = t->result;
			for (Device* p = alldev_; p != 0; p = p->next_) {
				const char* name = p->name();
				strcpy(cp, name);
				cp += strlen(name);
				*cp++ = ' ';
			}
			*cp++ = '\0';
			return (TCL_OK);
		}
		if (grabber_ == 0)
			goto no_grabber;
		if (strcmp(argv[1], "close") == 0) {
			delete grabber_;
			grabber_ = 0;
			delete framer_;
			framer_ = 0;
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "default") == 0) {
			int fmt = Device::lookup(argv[2]).default_format();
			tcl.result(lookup_format(fmt));
			return (TCL_OK);
		}
		if (grabber_ == 0)
			goto no_grabber;
		if (strcmp(argv[1], "send") == 0) {
			if (atoi(argv[2]))
				grabber_->start();
			else
				grabber_->stop();
			return (TCL_OK);
		}
		if (strcmp(argv[1], "lock") == 0) {
			grabber_->locked(atoi(argv[2]));
			return (TCL_OK);
		}
		if (strcmp(argv[1], "fps") == 0) {
			/*XXX assume value in range */
			grabber_->fps(atoi(argv[2]));
			return (TCL_OK);
		}
		if (strcmp(argv[1], "bps") == 0) {
			/*XXX assume value in range */
			int bps = atoi(argv[2]);
			grabber_->bps(bps);
			framer_->bps(bps);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "fillrate") == 0) {
			/*XXX assume value in range */
			grabber_->fillrate(atoi(argv[2]));
			return (TCL_OK);
		}
	} else if (argc == 4) {
		if (strcmp(argv[1], "open") == 0)
			return (open(argv[2], argv[3]));
		if (strcmp(argv[1], "format") == 0) {
			int fmt = lookup_format(argv[3]);
			if (Device::lookup(argv[2]).compat(fmt))
				tcl.result("1");
			else
				tcl.result("0");
			return (TCL_OK);
		}
		if (strcmp(argv[1], "supports") == 0) {
			if (Device::lookup(argv[2]).supports(argv[3]))
				tcl.result("1");
			else
				tcl.result("0");
			return (TCL_OK);
		}
	}
	Tcl_AppendResult(t, argv[0], ": unknown option: ", argv[1], 0);
	return (TCL_ERROR);
 no_grabber:
	Tcl_AppendResult(t, argv[0], ": device not open: ", argv[1], 0);
	return (TCL_ERROR);
}

int Device::encoder_cmd(ClientData cd, Tcl_Interp* t, int argc, char** argv)
{
	if (framer_ == 0) {
		Tcl_AppendResult(t, argv[0], ": no framer", 0);
		return (TCL_ERROR);
	}
	return (framer_->command(Tcl::instance(), argc, argv));
}

/*XXX*/
extern Network* net;

int Device::open(const char* devname, const char* format)
{
	Tcl& tcl = Tcl::instance();
	int fmt = lookup_format(format);
	Device& device = Device::lookup(devname);
	int mtu = atoi(tcl.attr("mtu"));

	switch (fmt) {

	case RTP_PT_JPEG:
		{
			JpegFramer* jf = new JpegFramer(net, mtu);
			grabber_ = device.jpeg_grabber(jf);
			framer_ = jf;
			break;
		}
	case RTP_PT_SCR:
		{
			ScrEncoder* scr = new ScrEncoder(net, mtu);
			grabber_ = device.yuv_grabber(scr);
			framer_ = scr;
			break;
		}
	case RTP_PT_NV:
		{
			NvEncoder* nv = new NvEncoder(net, mtu);
			grabber_ = device.yuv_grabber(nv);
			framer_ = nv;
			break;
		}
	case RTP_PT_CELLB:
		{
			CellbEncoder* cb = new CellbEncoder(net, mtu);
			grabber_ = device.yuv_grabber(cb);
			framer_ = cb;
			break;
		}
	case RTP_PT_H261:
		{
			/* XXX should make decision baseds on attrs() */
			H261Framer* p = new H261Framer(net, mtu);
			framer_ = p;
			grabber_ = device.h261_grabber(p);
			if (grabber_ == 0) {
				delete framer_;
				H261Encoder* p;
				p = new H261Encoder(net, mtu);
				grabber_ = device.cif_grabber(p);
				framer_ = p;
			}
			break;
		}
	}
	if (grabber_ == 0) {
		tcl.result("-1");
		return (TCL_OK);
	}
	if (grabber_->status() < 0) {
		char* p = tcl.buffer();
		sprintf(p, "%d", grabber_->status());
		delete grabber_;
		grabber_ = 0;
		delete framer_;
		framer_ = 0;
		tcl.result(p);
		return (TCL_OK);
	}
	tcl.result("0");
	return (TCL_OK);
}

Grabber::Grabber()
	: locked_(0), status_(0), running_(0), delta_(0.),
	  crvec_(0), cache_(0), rover_(0)
{
	bps(128);
	fps(1);

	/*XXX*/
	idle_low_ = 2;
	idle_high_ = 2;
}

Grabber::~Grabber()
{
	delete crvec_;
	delete cache_;
}

/*
 * Middle 32 bits of 64-bit ntp timestamp, per RTP spec.
 */
n_long Grabber::ntptime()
{
	timeval t;
	::gettimeofday(&t, 0);
	return (t.tv_sec << 16 | (t.tv_usec << 10) / 15625);
}

void Grabber::fps(int v)
{
	fps_ = v;
	frametime_ = 1e6 / double(v);
}

void Grabber::bps(int kbps)
{
	bps_ = 1000 * kbps;
}

int Grabber::command(int argc, char** argv)
{
    int i;
	/*
	 * ignore ctrl methods that are ignored by device subclass
	 */
    printf("Grabber::command argc=%d ", argc);
    for (i = 0; i < argc; i++)
	printf("\"%s\" ", argv[i]);
    printf("\n");
	return (TCL_OK);
}

/*
 * Advance the frame clock and return the amount of time we
 * need to wait before sending the next frame.  We compute
 * this time according to the desired bit and frame rates,
 * favoring the more restrictive metric.  If we're more than
 * 200ms behind (e.g., the cpu is saturated or we've been
 * suspended), give up and reset the frame clock.
 */
double Grabber::tick(int n)
{
    static int n_catchups = 0;
    double frametime = 8e6 * double(n) / double(bps_);
    if (frametime < frametime_) {
	if (frametime * 2. < frametime_)
	    delta_ += (frametime - delta_) * .25;
	else
	    delta_ = frametime;
	frametime = frametime_;
    } else
	delta_ = frametime;

    frameclock_ += frametime;
    double now = gettimeofday();
    double delta = frameclock_ - now;
    if (getenv("VIC_DEBUG"))
	printf("n=%d frametime=%f frametime_=%f now=%f delta=%f\n",
	       n, frametime, frametime_, now, delta);
    if (delta < -0.2e6) {
	delta = frametime;
	frameclock_ = now;
	n_catchups = 0;
    }
    else if (delta < 0 )
    {
	if (n_catchups < 10)
	{
	    /*
	     * We're not too far behind.
	     * Try to catch up.
	     */
	    delta = 0.;
	    n_catchups++;
	}
	else
	{
	    delta = frametime;
	    frameclock_ = now;
	    n_catchups = 0;
	}
    }
    else
    {
	n_catchups = 0;
    }

    return (delta);
}

void Grabber::start()
{
	if (!running_) {
		running_ = 1;
		frameclock_ = gettimeofday();
		timeout();
	}
}

void Grabber::stop()
{
	if (running_) {
		running_ = 0;
		cancel();
	}
}

void Grabber::timeout()
{
	for (;;) {
		double delta = tick(grab());
		if (delta != 0.) {
			usched(delta);
			return;
		}
		else
		{
		    if (getenv("VIC_DEBUG"))
			printf("Looping in timeout delta=%f\n", delta);
		}
	}
}

int Grabber::grab()
{
	abort();
	return (0);
}

void Grabber::crinit(int w, int h, int needcache)
{
	stride_ = w;
	blkw_ = (w + 7) >> 3;
	blkh_ = (h + 7) >> 3;
	scan_ = 0;
	nblk_ = blkw_ * blkh_;
	delete crvec_;
	crvec_ = new u_char[nblk_];
	memset((char*)crvec_, 0, nblk_);
	delete cache_;
	if (needcache) {
		int s = w * h;
		cache_ = new u_char[s];
		memset((char*)cache_, 0, s);
	} else
		cache_ = 0;
}

/*XXX*/
#define ABS(t) ((t) - (((t) >> 31 & (t)) << 1))
inline int diff(const u_char* block1, const u_char* block2)
{
#ifdef INT_64
	INT_64 m = 0xff00ff00ff00ff;
	INT_64 d = *(u_long*)block1;
	INT_64 s = d & m;
	s += (d >> 8) & m;
	s += s >> 32;
	s += s >> 16;
	int t = (int)s & 0xffff;

	d = *(INT_64*)block2;
	s = d & m;
	s += (d >> 8) & m;
	s += s >> 32;
	s += s >> 16;
	t -= (int)s & 0xffff;
#else
	n_long m = 0xff00ff;
	n_long d = *(n_long*)block1;
	n_long s = d & m;
	s += (d >> 8) & m;
	d = *(n_long*)(block1 + 4);
	s += d & m;
	s += (d >> 8) & m;
	s += s >> 16;
	int t = (int)s & 0xffff;

	d = *(n_long*)block2;
	s = d & m;
	s += (d >> 8) & m;
	d = *(n_long*)(block2 + 4);
	s += d & m;
	s += (d >> 8) & m;
	s += s >> 16;
	t -= (int)s & 0xffff;
#endif
	return (ABS(t));
}

void Grabber::suppress(const u_char* devbuf, const u_char* lumbuf)
{
	SUPPRESS_BODY(devbuf, lumbuf, 1)
}

inline void save(const u_char* lum, u_char* cache, int stride)
{
	for (int i = 8; --i >= 0; ) {
		((u_int*)cache)[0] = ((u_int*)lum)[0];
		((u_int*)cache)[1] = ((u_int*)lum)[1];
		cache += stride;
		lum += stride;
	}
}

/*
 * Default save routine -- stuff new luma blocks into cache.
 */
void Grabber::saveblks(u_char* lum)
{
	u_char* crv = crvec_;
	u_char* cache = cache_;
	int stride = stride_;
	stride = (stride << 3) - stride;
	for (int y = 0; y < blkh_; y++) {
		for (int x = 0; x < blkw_; x++) {
			if ((*crv++ & (CR_HQ|CR_LQ)) != 0)
				save(lum, cache, stride_);
			cache += 8;
			lum += 8;
		}
		lum += stride;
		cache += stride;
	}
}

void Grabber::idlefill()
{
	int blkno = rover_;
	int n = (delta_ * 2. < frametime_)? idle_high_ : idle_low_;
	while (n > 0) {
		if (crvec_[blkno] & CR_BG) {
			crvec_[blkno] = CR_HQ;
			--n;
		}
		if (++blkno >= nblk_) {
			blkno = 0;
			/* one way to guarantee loop termination */
			break;
		}
	}
	rover_ = blkno;
}

