/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and the Network Research Group at
 *	Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/main.C,v 1.4 1995/08/26 21:15:53 olson Exp (LBL)";

#include <osfcn.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/stat.h>
#ifdef sgi
#include <getopt.h>
#endif
#include <X11/Xlib.h>
#include <X11/Xutil.h>

extern "C" {
#include <tk.h>
#ifndef NOSHM
#ifdef sgi
#define XShmAttach __XShmAttach__
#endif
#include <X11/extensions/XShm.h>
#ifdef sgi
#undef XShmAttach
int XShmAttach(Display*, XShmSegmentInfo*);
#endif
#include <sys/ipc.h>
#include <sys/shm.h>
#if defined(sun) && !defined(__svr4__)
int shmget(key_t, int, int);
int shmctl(int shmid, int cmd, struct shmid_ds *);
#endif
#endif
}

#include "inet.h"
#include "packet.h"
#include "source-vic.h"
#include "session.h"
#include "mcastchan.h"
#include "video.h"
#include "Tcl.h"
#include "grabber.h"
#include "decoder.h"
#include "switcher.h"
#include "crypt.h"
#ifdef RTIP
#include "rtipchan.h"
extern "C" {
#include "rt.h"
#include "rcap.h"
}
#endif
#include "color.h"
#include "framer.h"

#ifdef RTIP
static void
rtip_params(Tcl& t, rcapTraffic& rt, rcapRequirements& rr)
{
	rt.xmin = atoi(t.attr("rtipXmin"));
	rt.xave = atoi(t.attr("rtipXave"));
	rt.I = atoi(t.attr("rtipI"));
	rt.smax = atoi(t.attr("rtipSmax"));
	rr.D = atoi(t.attr("rtipD"));
	rr.J = atoi(t.attr("rtipJ"));
	rr.Z = atoi(t.attr("rtipZ"));
	rr.W = atoi(t.attr("rtipW"));
	rr.U = atoi(t.attr("rtipU"));
	rr.type = atoi(t.attr("rtipType"));
}
#endif

static void
usage()
{
	fprintf(stderr, "\
Usage: vic [-Hs] [-A nv|ivs] [-B maxbps] [-C conf]\n\
\t[-c best|dither|ed|gray|quantize|true|truegray] [-D device]\n\
\t[-d display] [-f cellb|h261|jpeg|mpeg|nv|scr] [-F maxfps]\n\
\t[-I channel] [-K key ] [-M colormap] [-m mtu] [-N session]\n\
\t[-o clipfile] [-r send|recv|duplex] [-t ttl]\n\
\t[-U interval] [-u script] [-X resource=value] dest/port[/fmt/ttl]\n");
	exit(1);
}

static SessionManager* manager;
Network* net;

extern "C" void adios(int);

void
adios(int s)
{
	delete manager;
	delete net;
	ColorModel::release();
	exit(0);
}

static int
cmd_adios(ClientData, Tcl_Interp*, int, char**)
{
	adios(0);
	/*NOTREACHED*/
	return (0);
}

static int
cmd_havefont(ClientData, Tcl_Interp* ti, int argc, char **argv)
{
	Tcl& t = Tcl::instance();
	if (argc != 2)
		t.result("0");
	else {
		Tk_Window tk = t.tkmain();
		Tk_Uid uid = Tk_GetUid(argv[1]);
		XFontStruct* p = Tk_GetFontStruct(ti, tk, uid);
		if (p != 0) {
/* no need, we use them anyway */
#ifdef notdef
			Tk_FreeFontStruct(p);
#endif
			t.result("1");
		} else
			t.result("0");
	}
	return (TCL_OK);
}

extern "C" char version[];

static int
cmd_version(ClientData, Tcl_Interp* tcl, int, char**)
{
	tcl->result = version;
	return (TCL_OK);
}

/*
 * The following command & the associate service procedure
 * creates a `rate variable'.  Reading a rate variable returns
 * the time rate of change of its value.  Writing a rate variable
 * updates the rate of change estimate.  A rate variable can be
 * either simple or an array element.  It must be global.  This
 * routine is called as:
 *	rate_variable varname filter_const
 * where 'varname' is the variable name (e.g., "foo" or "foo(bar)")
 * and 'filter_const' is a smoothing constant for the rate estimate.
 * It should be >0 and <1 (a value of 1 will result in no smoothing).
 */
struct rv_data {
	double rate;
	double filter;
	timeval lastupdate;
	int lastval;
	char format[16];
};

static char*
update_rate_var(ClientData clientData, Tcl_Interp* tcl, char* name1,
		char* name2, int flags)
{
	rv_data* rv = (rv_data*)clientData;
	if (rv == NULL)
		return ("no clientdata for rate var");
	if (flags & TCL_TRACE_WRITES) {
		/*
		 * the variable has been written with a new value.
		 * compute the rate of change & make that its vale
		 * for subsequent reads.
		 */
		char res[128];
		flags &= TCL_GLOBAL_ONLY;
		char* cv = Tcl_GetVar2(tcl, name1, name2, flags);
		if (cv == NULL)
			return (tcl->result);
		int curval = atoi(cv);
		double rate = 0.;
		timeval tv;
		gettimeofday(&tv, 0);
		if (rv->lastupdate.tv_sec != 0) {
			/* not first time through */
			double dt = double(tv.tv_sec - rv->lastupdate.tv_sec) +
				    double(tv.tv_usec -
					rv->lastupdate.tv_usec) * 1e-6;
			if (dt <= 0.)
				/* have to wait until next clock tick */
				return NULL;
			double dv = double(curval - rv->lastval);
			if (dv >= 0.) {
				rate = rv->rate;
				rate += (dv/dt - rate) * rv->filter;
				/* work around tcl f.p. precision bug */
				if (rate < 1e-12)
					rate = 0.;
			}
		}
		rv->rate = rate;
		rv->lastupdate = tv;
		rv->lastval = curval;
		sprintf(res, rv->format, rate);
		Tcl_SetVar2(tcl, name1, name2, res, flags);
	} else if (flags & (TCL_TRACE_DESTROYED|TCL_INTERP_DESTROYED)) {
		delete rv;
	}
	return NULL;
}

static int
cmd_rate_variable(ClientData, Tcl_Interp* tcl, int argc, char **argv)
{
	const char* fmt;
	if (argc == 4)
		fmt = argv[3];
	else if (argc == 3)
		fmt = "%g";
	else {
		tcl->result = "usage: rate_variable varname filter_const";
		return (TCL_ERROR);
	}
	double fconst = atof(argv[2]);
	if (fconst <= 0. || fconst > 1.) {
		tcl->result = "rate_variable: invalid filter constant";
		return (TCL_ERROR);
	}
	rv_data* rv = new rv_data;
	rv->filter = fconst;
	rv->rate = 0.;
	rv->lastupdate.tv_sec = 0;
	rv->lastval = 0;
	strcpy(rv->format, fmt);
	int sts = Tcl_TraceVar(tcl, argv[1],
			TCL_TRACE_WRITES|TCL_TRACE_UNSETS|TCL_GLOBAL_ONLY,
			update_rate_var, (ClientData)rv);
	if (sts != TCL_OK)
		delete rv;
	return (sts);
}

#ifdef NOSHM
int use_shm = 0;
#else
int use_shm = 1;
static int
noXShm(ClientData, XErrorEvent*)
{
	use_shm = 0;
	fprintf(stderr, "vic: warning: not using shared memory\n");
	return (0);
}

/*
 * XXX this is a hack to see if we can used shared memory.
 * if the X server says we can, and we're on the same
 * host as the X server, then we are golden.
 */
static void
checkXShm(Tk_Window tk, const char*)
{
	Display* dpy = Tk_Display(tk);
	XShmSegmentInfo si;
  	if ((si.shmid = shmget(IPC_PRIVATE, 512, IPC_CREAT|0777)) < 0) {
		if (use_shm) {
 			perror("shmget");
 			exit(1);
		} else
			/*
			 * If we got here from a bad shmget() and
			 * use_shm is 0, it is probably because
			 * we caught a SIGSYS on NetBSD.  So just
			 * jump out and let "use_shm = 0" catch
			 * everything from now on.
			 */
 			return;
	}
	si.readOnly = 1;
	XSync(dpy, 0);
	Tk_ErrorHandler handler = Tk_CreateErrorHandler(dpy, -1, -1, -1,
							noXShm, 0);
	XShmAttach(dpy, &si);
	XSync(dpy, 0);
	Tk_DeleteErrorHandler(handler);
	(void)shmctl(si.shmid, IPC_RMID, 0);
}
#endif

extern "C" char *optarg;
extern "C" int optind;
extern "C" int opterr;

const char*
disparg(int argc, const char** argv, const char* optstr)
{
	const char* display = 0;
	int op;
	while ((op = getopt(argc, (char**)argv, (char*)optstr)) != -1) {
		if (op == 'd') {
			display = optarg;
			break;
		}
		else if (op == '?')
			usage();
	}
	optind = 1;
	return (display);
}

#include "bitmap/play.xbm"
#include "bitmap/stop.xbm"
#include "bitmap/rewind.xbm"
#include "bitmap/ff.xbm"
#include "bitmap/square.xbm"
void
loadbitmaps(Tcl_Interp* tcl)
{
	Tk_DefineBitmap(tcl, Tk_GetUid("play"),
			play_bits, play_width, play_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("stop"),
			stop_bits, stop_width, stop_height);

	Tk_DefineBitmap(tcl, Tk_GetUid("rewind"),
			rewind_bits, rewind_width, rewind_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("ff"),
			ff_bits, ff_width, ff_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("square"),
			square_bits, square_width, square_height);
}

#ifdef JVIDEO
int useJvColors = 1;
#else
int useJvColors = 0;
#endif

static int hostspec(const char* dst, n_long& addr, u_short& port,
		    const char*& fmt, int& ttl)
{
	char spec[256];
	strcpy(spec, dst);

	int ret = 1;
	char* cp = strchr(spec, '/');
	if (cp)
		*cp = 0;
	if ((addr = LookupHostAddr(spec)) == 0) {
		printf("vic: invalid dest host `%s'\n", spec);
		exit(1);
	}

	if (cp && *++cp != '/') {
		/* have a port number */
		port = atoi(cp);
		if (port <= 0) {
			printf("invalid port `%s'\n", cp);
			exit(1);
		}
		ret = 0;
		cp = strchr(cp, '/');
	}

	if (cp && cp[1]) {
		if (cp[1] == '/')
			++cp;
		else if (!isdigit(cp[1])) {
			fmt = cp;
			cp = strchr(++cp, '/');
			if (cp)
				*cp = 0;
		}
	}
	if (cp && *++cp) {
		ttl = atoi(cp);
		if (ttl <= 0 || ttl > 255) {
			printf("invalid ttl `%s'\n", cp);
			exit(1);
		}
	}

	return (ret);
}

char*
parse_assignment(char* cp)
{
	cp = strchr(cp, '=');
	if (cp != 0) {
		*cp = 0;
		return (cp + 1);
	} else
		return ("true");
}

extern const char tcl_tk[];
extern const char tcl_tkerror[];
extern const char tcl_local[];
extern const char tcl_vic[];

/*XXX*/
int rtpv1;

double rtcp_inv_bw;

extern "C" int Blt_TableInit(Tcl_Interp*);
extern "C" int Tk_StripchartCmd(ClientData, Tcl_Interp*, int ac, char** av);


int
main(int argc, const char** argv)
{
	const char* script = 0;
	int channel = 0;
#ifdef __svr4__
	signal(SIGINT, (void (*)(int))adios);
	signal(SIGHUP, (void (*)(int))adios);
	signal(SIGTERM, (void (*)(int))adios);
#else
#ifdef sgi
#ifndef IRIX5
	signal(SIGINT, (int (*)(int ...))adios);
	signal(SIGHUP, (int (*)(int ...))adios);
	signal(SIGTERM, (int (*)(int ...))adios);
#else
	signal(SIGINT, (void (*)(int ...))adios);
	signal(SIGHUP, (void (*)(int ...))adios);
	signal(SIGTERM, (void (*)(int ...))adios);
#endif
#elif defined(__osf__) ||  defined(__hpux__) || defined(__NetBSD__)
	signal(SIGINT, (void (*)(int))adios);
	signal(SIGHUP, (void (*)(int))adios);
	signal(SIGTERM, (void (*)(int))adios);
#ifdef __NetBSD__
	signal(SIGSYS, (void (*)(int))noXShm);
#endif
#else
	signal(SIGHUP, adios);
	signal(SIGINT, adios);
	signal(SIGTERM, adios);
#endif
#endif
	opterr = 0;
	const char* optstr = 
		"A:B:c:C:D:d:e:f:F:HI:i:K:M:m:N:o:Pp:q:r:S:sT:t:U:u:V:X:";
	const char* display = disparg(argc, argv, optstr);
	Tcl::init("vic");
	Tcl& tcl = Tcl::instance();
	tcl.evalc(tcl_tkerror);
	Tk_Window tk = Tk_CreateMainWindow(tcl.interp(),
					   (char*)display, "vic", "Vic");
	if (tk == 0) {
		fprintf(stderr, "vic: %s\n", tcl.result());
		exit(1);
	}
	tcl.tkmain(tk);
	(void)Blt_TableInit(tcl.interp());
	tcl.CreateCommand("stripchart", Tk_StripchartCmd, (ClientData)tk);

	char* conference = 0;
	int ctrlsock = 1;
	int session_type = SESSION_VIC;

	int op;
	while ((op = getopt(argc, (char**)argv, (char*)optstr)) != -1) {
		switch (op) {

		default:
			usage();

		case 'A':
			if (strcmp(optarg, "nv") == 0) {
				tcl.add_default("defaultFormat", "nv");
				ctrlsock = 0;
				rtpv1 = 1;
				session_type = SESSION_NV;
			} else if (strcmp(optarg, "ivs") == 0) {
				tcl.add_default("defaultFormat", "h.261");
				ctrlsock = 2;
				rtpv1 = 1;
				session_type = SESSION_IVS;
			}
			else if (strcmp(optarg, "vic") == 0)
				/* default */;
			else
				usage();
			break;

		case 'B':
			tcl.add_option("maxbw", optarg);
			break;

		case 'C':
			conference = optarg;
			break;

		case 'c':
			tcl.add_option("colorModel", optarg);
			break;

		case 'D':
			tcl.add_option("device", optarg);
			break;

		case 'd':
			break;

		case 'f':
			tcl.add_option("defaultFormat", optarg);
			break;

		case 'F':
			tcl.add_option("maxfps", optarg);
			break;

		case 'H':
			Decoder::use_hardware(1);
			break;

		case 'I':
			channel = atoi(optarg);
			break;

		case 'K':
			/*XXX probably do not want this in X server*/
			tcl.add_option("sessionKey", optarg);
			break;

		case 'M':
			tcl.add_option("colorFile", optarg);
			break;

		case 'm':
			tcl.add_option("mtu", optarg);
			break;

		case 'N':
			tcl.add_option("sessionName", optarg);
			break;

		case 'o':
			tcl.add_option("outfile", optarg);
			break;

		case 'P':
			ColorModel::use_private_map(1);
			break;

		case 'q':
			tcl.add_option("jpegQfactor", optarg);
			break;

		case 'r':
			tcl.add_option("rtip", optarg);
			break;

		case 's':
			use_shm = 0;
			break;

		case 't':
			tcl.add_option("defaultTTL", optarg);
			break;

		case 'T':
			tcl.add_option("softJPEGthresh", optarg);
			break;

		case 'u':
			script = optarg;
			break;

		case 'U':
			tcl.add_option("stampInterval", optarg);
			break;

		case 'V':
			tcl.add_option("softJPEGcthresh", optarg);
			break;

		case 'X':
			{
				const char* value = parse_assignment(optarg);
				tcl.add_option(optarg, value);
			}
			break;
		}
	}
#ifdef __linux__
	optind++;
	/* This is bizarre ... */
#endif
#ifndef NOSHM
	if (use_shm)
		checkXShm(tk, display);
#endif

	loadbitmaps(tcl.interp());

	tcl.evalc(tcl_tk);
	tcl.evalc(tcl_local);
	tcl.CreateCommand("adios", cmd_adios);
	tcl.CreateCommand("version", cmd_version);
	tcl.CreateCommand("havefont", cmd_havefont);
	tcl.CreateCommand("rate_variable", cmd_rate_variable);
	tcl.CreateCommand("video", VideoWindow::create);
	Device::init();

	if (script != 0)
		tcl.EvalFile(script);
	else
		tcl.evalc(tcl_vic);

	const char* dst;
	if (optind < argc && argc > 1) {
		dst = argv[optind];
		if (argc - optind > 1) {
			fprintf(stderr,
				"vic: extra arguments (starting with `%s')\n",
				argv[optind + 1]);
			exit(1);
		}
	} else if ((dst = tcl.attr("defaultHostSpec")) == 0) {
		fprintf(stderr, "vic: destination address required\n");
		exit(1);
	}
	const char* outfile = tcl.attr("outfile");
	if (outfile != 0) {
		int fd = open(outfile, O_WRONLY|O_TRUNC|O_CREAT, 0644);
		if (fd < 0) {
			fprintf(stderr, "vic: ");
			perror(outfile);
			exit(1);
		}
		Framer::dump(fd);
	}

	n_long addr;
	int ttl = atoi(tcl.attr("defaultTTL"));
	u_short port = 0;
	const char* fmt = 0;
	hostspec(dst, addr, port, fmt, ttl);
	if (port == 0) {
		fprintf(stderr, "vic: must supply a port number\n");
		exit(1);
	}
	if ( session_type == SESSION_VIC)
		/* rtpv2 spec says data goes on even port and ctrl on odd */
		port &=~ 1;
	port = htons(port);

	if (fmt != 0)
		tcl.add_option("defaultFormat", fmt);

	const char* name = tcl.attr("sessionName");
	if (name == 0) {
		char host[MAXHOSTNAMELEN];
		gethostname(host, sizeof(host));
		const char* username = tcl.attr("userName");
		if (username == 0 || *username == 0)
			username = getenv("USER");
		if (username == 0) {
			fprintf(stderr, "\
vic: Vic.sessionName, Vic.userName and your USER environment variable are\n\
vic: all undefined.  One of these must be set to run vic.\n");
			exit(1);
		}
		char* p = new char[strlen(username) + strlen(host) + 2];
		sprintf(p, "%s@%s", username, host);
		name = p;
	}
	if (conference == 0) {
		const char* cp = LookupHostName(addr);
		conference = new char[strlen(cp) + 4 + 6 + 2];
		sprintf(conference, "%s/%d", cp, htons(port));
	}
	tcl.add_option("conferenceName", conference);

#ifdef RTIP
	const char* rtip = tcl.attr("rtip");
	int rtipdir;
	if (rtip == 0)
		;
	else if (strcmp(rtip, "send") == 0)
		rtipdir = 1;
	else if (strcmp(rtip, "recv") == 0)
		rtipdir = -1;
	else if (strcmp(rtip, "duplex") == 0)
		rtipdir = 0;
	else {
		fprintf(stderr, "vic: bogus rtip option: %s\n", rtip);
		exit(1);
	}
	if (rtip) {
		ttl = 0;
		rcapTraffic traffic;
		rcapRequirements requirements;
		rtip_params(tcl, traffic, requirements);
		net = new rtipChannel(addr, port, rtipdir,
				      (void*)&traffic, (void*)&requirements);
	} else
#endif
	{
		net = new mcastChannel(addr, port, ttl, ctrlsock);
	}

	const char* cname = tcl.attr("cname");
	if (cname == 0) {
		const char* user = tcl.attr("userName");
		if (user == 0 || *user == 0)
			user = getenv("USER");
		n_long a = net->interface();
		if (a == 0) {
			a = LookupLocalAddr();
			if (a == 0) {
				/*XXX*/
				fprintf(stderr, "vic: can't form cname\n");
				exit(1);
			}
		}
		const char* addr = InetNtoa(a);
		char* p = new char[strlen(addr) + strlen(user) + 2];
		sprintf(p, "%s@%s", user, addr);
		cname = p;
	}
	Tk_MakeWindowExist(tk);
	ColorModel::init(ColorModel::atocm(tcl.attr("colorModel")));

	tcl.evalc("build.tk");
	/*
	 * if the max bandwidth wasn't set, pick one based on the
	 * session ttl.  Then use the max bandwidth to compute the
	 * fraction allocated to rtcp (control) traffic.  We have
	 * to do this before creating the session manager since it
	 * will use the result to compute a time to the first report.
	 * bw is in kbits/sec and we want bytes/ms so we divide by 8.
	 */
	int maxbw = atoi(tcl.attr("maxbw"));
	if (maxbw <= 0) {
		char buf[32];

		if (ttl <= 16 || !IN_MULTICAST(ntohl(addr)))
			maxbw = 7072;
		else if (ttl <= 64)
			maxbw = 1024;
		else if (ttl <= 128)
			maxbw = 128;
		else if (ttl <= 192)
			maxbw = 53;
		else
			maxbw = 32;
		sprintf(buf, "%d", maxbw);
		tcl.add_option("maxbw", buf);
	}
	double bw = double(maxbw) / 8.;
	rtcp_inv_bw = 1. / (bw * RTCP_SESSION_BW_FRACTION);

	vicSourceManager::init(net->srcid(), net->interface());
	Source* local = SourceManager::instance().localsrc();
	Switcher::channel(channel);
	Switcher::local(local->addr());
	local->SetName(name);
	local->SetCName(cname);

	switch (session_type) {
	case SESSION_NV:
		manager = new NvSessionManager(net);
		break;
	case SESSION_IVS:
		manager = new IvsSessionManager(net);
		break;
	default:
		manager = new SessionManager(net);
		break;
	}

	Tk_MapWindow(tk);

	/*
	 * re-nice the vic process before we start processing video
	 * so that video processing won't screw up the local audio.
	 */
	int pri = atoi(tcl.attr("priority"));
	if (pri > 0)
		nice(pri);

	/*
	 * XXX this doesn't work yet, but
	 * Ousterhout says it will in Tk 4.0
	 * Hopefully, he'll fix the event model
	 * and we won't need it; not likely.
	 */
#ifdef notyet
	while (tk_NumMainWindows > 0) {
		/*
		 * Tk doesn't give priority to the X server so file handlers
		 * (i.e., packets from the network) can starve the window
		 * system (technically, tk lets one X event through per file
		 * event, but a file event can trigger more than one X event
		 * so the X server loses to a high-rate video source).
		 * The solution is to let the X server go first by only
		 * processing file handlers when their are no pending
		 * X events.
		 */
		if (Tk_DoOneEvent(TK_X_EVENTS|TK_TIMER_EVENTS|
				  TK_DONT_WAIT) == 0)
			Tk_DoOneEvent(0);
	}
#else
	Tk_MainLoop();
#endif
	adios(0);
	return 0;
}

