/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/net.C,v 1.4 1995/08/26 21:15:57 olson Exp (LBL)";

#include <stdlib.h>
#include <unistd.h>
#include <osfcn.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/uio.h>
#include "net.h"
#include "crypt.h"

Crypt* Network::crypt_;

Network::Network(n_long addr, u_short port, int ttl)
	: addr_(addr), port_(port), ttl_(ttl), srcid_(0),
	  local_(0)
{
}

Network::~Network()
{
}

n_long Network::alloc_srcid(n_long addr) const
{
	timeval tv;
	gettimeofday(&tv, 0);
	n_long srcid = n_long(tv.tv_sec + tv.tv_usec);
	srcid += (n_long)getuid();
	srcid += (n_long)getpid();
	srcid += addr;
	return (srcid);
}

void Network::resolve_srcid_collision()
{
	/* just allocate a new random srcid */
	srcid_ = alloc_srcid(local_);
}

void Network::nonblock(int fd)
{       
        int flags = fcntl(fd, F_GETFL, 0);
#if defined(hpux) || defined(__hpux)
        flags |= O_NONBLOCK;
#else
        flags |= O_NONBLOCK|O_NDELAY;
#endif
        if (fcntl(fd, F_SETFL, flags) == -1) {
                perror("vic: fcntl: F_SETFL");
                exit(1);
        }
}

static u_char* wrkbuf;
static int wrkbuflen;

static void expand_wrkbuf(int len)
{
	if (wrkbuflen == 0)
		wrkbuf = (u_char*)malloc(len);
	else
		wrkbuf = (u_char*)realloc((u_char*)wrkbuf, len);
	wrkbuflen = len;
}

void Network::dosend(u_char* buf, int len, int fd)
{
	int cc = ::send(fd, (char*)buf, len, 0);
	if (cc < 0) {
		switch (errno) {
		case ECONNREFUSED:
			/* no one listening at some site - ignore */
#if defined(__osf__) || defined(_AIX)
			reset();
#endif
			break;

		case ENETUNREACH:
		case EHOSTUNREACH:
			/*
			 * These "errors" are totally meaningless.
			 * There is some broken host sending
			 * icmp unreachables for multicast destinations.
			 * UDP probably aborted the send because of them --
			 * try exactly once more.  E.g., the send we
			 * just did cleared the errno for the previous
			 * icmp unreachable, so we should be able to
			 * send now.
			 */
			(void)::send(ssock_, (char*)buf, len, 0);
			break;

		default:
			perror("vic: send");
			return;
		}
	}
}

void Network::send(u_char* buf, int len)
{
	if (crypt_)
		buf = crypt_->Encrypt(buf, len);
	dosend(buf, len, ssock_);
}

void Network::sendmsg(const msghdr& mh)
{
	if (crypt_) {
	    int len, i;
		for (len = 0, i = 0; i < mh.msg_iovlen; ++i)
			len += mh.msg_iov[i].iov_len;
		expand_wrkbuf(len);
		u_char* cp = wrkbuf;
		for (i = 0; i < mh.msg_iovlen; ++i) {
			len = mh.msg_iov[i].iov_len;
			memcpy(cp, mh.msg_iov[i].iov_base, len);
			cp += len;
		}
		send(wrkbuf, cp - wrkbuf);
		return;
	}
	int cc = ::sendmsg(ssock_, (msghdr *) &mh, 0);
	if (cc < 0) {
		switch (errno) {
		case ECONNREFUSED:
			/* no one listening at some site - ignore */
#if defined(__osf__) || defined(_AIX)
			reset();
#endif
			break;

		case ENETUNREACH:
		case EHOSTUNREACH:
			/*
			 * These "errors" are totally meaningless.
			 * There is some broken host sending
			 * icmp unreachables for multicast destinations.
			 * UDP probably aborted the send because of them --
			 * try exactly once more.  E.g., the send we
			 * just did cleared the errno for the previous
			 * icmp unreachable, so we should be able to
			 * send now.
			 */
			(void)::sendmsg(ssock_, (msghdr *) &mh, 0);
			break;

		default:
			perror("vic: sendmsg");
			return;
		}
	}
}

void Network::sendctrl(u_char* buf, int len)
{
	if (crypt_)
		buf = crypt_->EncryptCtrl(buf, len);
	dosend(buf, len, cssock_);
}

int Network::dorecv(u_char* buf, int len, n_long& from, int fd)
{
	sockaddr_in sfrom;
	int fromlen = sizeof(sfrom);
	int cc = ::recvfrom(fd, (char*)buf, len, 0,
			    (sockaddr*)&sfrom, &fromlen);
	if (cc < 0) {
		if (errno != EWOULDBLOCK)
			perror("vic: recvfrom");
		return (-1);
	}
	from = sfrom.sin_addr.s_addr;
	return (cc);
}

int Network::recv(u_char* buf, int len, n_long& from)
{
	if (crypt_) {
		if (len > wrkbuflen)
			expand_wrkbuf(len);
		int cc = dorecv(wrkbuf, len, from, rsock_);
		return (crypt_->Decrypt(wrkbuf, cc, buf));
	}
	return (dorecv(buf, len, from, rsock_));
}

int Network::recvctrl(u_char* buf, int len, n_long& from)
{
	if (crypt_) {
		if (len > wrkbuflen)
			expand_wrkbuf(len);
		int cc = dorecv(wrkbuf, len, from, crsock_);
		return (crypt_->DecryptCtrl(wrkbuf, cc, buf));
	}
	return (dorecv(buf, len, from, crsock_));
}

void Network::reset()
{
}

