/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Network Research
 *	Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/p64/p64.h,v 1.2 1995/08/16 00:48:21 olson Exp (LBL)
 */

#ifndef lib_p64_h
#define lib_p64_h

#include <sys/types.h>

struct huffcode;

#define MBPERGOB 33

class P64Renderer {
 public:
	virtual void render(int x, int y, int w, int h,
			    u_char* yp, u_char* up, u_char* vp) = 0;
};

class P64Decoder {
 public:
	P64Decoder(int fmt);
	virtual ~P64Decoder();
	const u_char* yplane() const { return comp_[0].back; }
	const u_char* uplane() const { return comp_[1].back; }
	const u_char* vplane() const { return comp_[2].back; }
	inline int width() const { return (width_); }
	inline int height() const { return (height_); }
	virtual int decode(u_char* bp, int cc, int sbit, int ebit, int fwd);
	void sync();
	inline void bb(int& x, int& y, int& w, int& h) {
		x = bbx_; y = bby_; w = bbw_; h = bbh_;
	};
	inline void renderer(P64Renderer* p) { renderer_ = p; }
	inline u_int bad_psc() const { return (bad_psc_); }
	inline u_int bad_bits() const { return (bad_bits_); }
	inline u_int bad_GOBno() const { return (bad_GOBno_); }
	inline u_int bad_fmt() const { return (bad_fmt_); }
 protected:
	struct hufftab {
		int maxlen;
		const short* prefix;
	};
	struct component {
		int height;
		int width;
		int hf;
		int vf;
		u_char *front;
		u_char *back;
	};
	void init();
	void initcomp();
	void inithuff();
	void initquant();
	virtual void err(const char* msg ...) const;
	int quantize(int v, int q);

#ifdef INT_64
	int parse_block(short* blk, INT_64* mask);
#else
	int parse_block(short* blk, u_int* mask);
#endif
	void decode_block(u_int tc, u_int x, u_int y, u_int stride);

	void filter(u_char* in, u_char* out, u_int stride);
	void mvblk(u_char* in, u_char* out, u_int stride);
	void mvblka(u_char*, u_char*, u_int stride);

	void swap();

	void mbcopy(u_int mba, int n);

	int parse_picture_hdr();
	void parse_gob_hdr();
	int parse_mb_hdr(u_int& cbp);

	int decode_gob(u_int gob);
	int decode_mb();

	hufftab ht_mba_;
	hufftab ht_mvd_;
	hufftab ht_cbp_;
	hufftab ht_tcoeff_;
	hufftab ht_mtype_;

	u_char* mb_sv_;
	u_char mb_skip_[16 * 64];
	/* fast quantization */
	short quant_[32 * 256];
	short* qt_;
	/* gob/mba to coordinate mappings */
	u_short base_[16 * 64];
	u_short* coord_;

	u_int width_;		/* width of Y component */
	u_int height_;		/* height of Y component */
	
	int ncomp_;
	component comp_[3];
	component* iob_;

#define IT_QCIF	0
#define IT_CIF	1
	u_int fmt_;		/* image format (CIF/QCIF) */
	u_int ngob_;		/* number of gobs (12 for CIF, 3 for QCIF) */

	u_int mt_;		/* macroblock type (flags in p64-huff.h) */
	u_int gob_;		/* current gob index */
	int mba_;		/* macroblock address (predictor) */
	int mvdh_;		/* motion vector (predictor) */
	int mvdv_;		/* motion vector (predictor) */

	/* error counters */
	int bad_psc_;
	int bad_bits_;
	int bad_GOBno_;
	int bad_fmt_;		/* # times RTP fmt != H.261 fmt */

	u_int bb_;		/* 32-bit bit buffer */
	int nbb_;		/* number bits in bit buffer */
	const u_short* bs_;	/* input bit stream (less bits in bb_) */
	const u_short* es_;	/* pointer to end if input stream */
	const u_short* ps_;	/* packet start (for error reporting) */
	int pebit_;		/* packet end bit (for error reporting) */

	/* bounding box */
	u_int minx_;
	u_int miny_;
	u_int maxx_;
	u_int maxy_;
	u_int bbx_;
	u_int bby_;
	u_int bbw_;
	u_int bbh_;

	u_int gobfrm_[12];
	u_int frmno_;

	/*XXX might want to be a ref set by constructor*/
	P64Renderer* renderer_;
};

#endif
