/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/renderer.h,v 1.2 1995/08/16 00:46:34 olson Exp (LBL)
 */

#ifndef vic_renderer_h
#define vic_renderer_h

#include "video.h"
#include "color.h"

class Colorizer;

class Renderer {
protected:
	Renderer(VideoWindow* w, int inw, int inh);
public:
	virtual ~Renderer();
        /*
	 * This structures should be protected,
	 * but DEC's C++ compiler cannot hack it.
	 */
	int match(VideoWindow*);
	int contains(VideoWindow*);
	void attach(VideoWindow*);
	void detach(VideoWindow*);
	inline int empty() { return (windows_ == 0); }
	void reset(int inq, int inw, int inh);
	void init(VideoWindow*);
	void purge();
protected:
	void resize(int w, int h);
	VideoWindow* windows_;	/* list of output windows */

	int inw_;		/* width of original image */
	int inh_;		/* height of original */
	int outw_;		/* width of target image (maybe != width_) */
	int outh_;		/* height of target image (maybe != height_) */
	int imagesize_;		/* target image size */

	int width_;		/* width of target window */
	int height_;		/* height of target window */
	int scale_;		/* log base two of inverse of scale factor */
};

inline int Renderer::match(VideoWindow* w)
{
	return (w->width() == width_ && w->height() == height_);
}

class SoftwareRenderer : public Renderer {
 public:
	SoftwareRenderer(VideoWindow* w, int inw, int inh, int color,
			 int decimation = 422);
	~SoftwareRenderer();
	void resize(int w, int h, int c);
	void color(int c);
	void damage() const;
	void render() const;
	void render(int miny, int maxy) const;
	void dither(int x, int y, int w, int h,
		    const u_char* yp, const u_char* up, const u_char* vp);
	SoftwareRenderer* next_;
 protected:
	void init(VideoWindow* w);
	SharedVideoImage* image_;
	Colorizer* colorizer_;
	int decimation_;
};

inline void SoftwareRenderer::dither(int x, int y, int w, int h,
				     const u_char* yp,
				     const u_char* up,
				     const u_char* vp)
{
	colorizer_->colorize((u_char*)image_->ximage()->data,
			     yp, up, vp, 
			     x, y, w, h);
}

/*
 * Render an image to all windows attached to this object.
 * Doesn't necessarily push the image to the frame buffer;
 * use purge() to do that.
 */
inline void SoftwareRenderer::render() const
{
	for (VideoWindow* vw = windows_; vw != 0; vw = vw->next_)
		vw->render(image_);
}

inline void SoftwareRenderer::damage() const
{
	for (VideoWindow* vw = windows_; vw != 0; vw = vw->next_)
		vw->damage();
}

/*
 * Flush all windows attached to this renderer to the frame buffer.
 * The reason we separate render and purge is so we can queue up
 * a bunch of putimage's then do one context switch with the X server
 * to copy all the images in a single go.
 */
inline void Renderer::purge()
{
	for (VideoWindow* vw = windows_; vw != 0; vw = vw->next_)
		vw->purge();
}

#endif
