#ifdef __linux__

#include <sys/types.h>
#include <sys/uio.h>

#define MSGBUF_MAX 16384

/*
 * Message header for recvmsg and sendmsg calls.
 */
struct msghdr {
        caddr_t msg_name;               /* optional address */
        int     msg_namelen;            /* size of address */
        struct  iovec *msg_iov;         /* scatter/gather array */
        int     msg_iovlen;             /* # elements in msg_iov */
        caddr_t msg_accrights;          /* access rights sent/received */
        int     msg_accrightslen;
};


int sendmsg(int fd, const struct msghdr *msg, unsigned int flags)
{
    static char msgbuf[MSGBUF_MAX];
    static char *endp = msgbuf + MSGBUF_MAX;
    char *p;
    int i, len, ilen;
    int rc;

    if (msg->msg_iovlen == 1)
    {
	rc = sendto(fd, msg->msg_iov[0].iov_base, msg->msg_iov[0].iov_len,
		    flags, msg->msg_name, msg->msg_namelen);
	printf("single sendmsg returns %d\n", rc);
    }
    else
    {

	for (p = msgbuf, len = 0, i = 0; i < msg->msg_iovlen; i++)
	{
	    ilen = msg->msg_iov[i].iov_len;
	    if (p + ilen > endp)
	    {
		printf("sendmsg() ran out of space!\n");
		exit(1);
	    }
	    bcopy(msg->msg_iov[i].iov_base, p, msg->msg_iov[i].iov_len);
	    len += msg->msg_iov[i].iov_len;
	    p += msg->msg_iov[i].iov_len;
	}
	rc = sendto(fd, msgbuf, len, flags, msg->msg_name, msg->msg_namelen);
	printf("multiple sendmsg returns %d\n", rc);
    }
    return rc;
}

#endif
