/*-
 * Copyright (c) 1993-1994 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/session.C,v 1.2 1995/08/16 00:46:40 olson Exp (LBL)";

#include <osfcn.h>
#include <math.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#ifdef _AIX
extern "C" long random();
#endif
#include "session.h"
#include "net.h"
#include "source.h"
#include "Tcl.h"

extern double rtcp_inv_bw;

SessionNotifier::SessionNotifier(SessionManager& m)
	: manager_(m)
{
}

void SessionNotifier::dispatch(int)
{
	manager_.rtcp_recv();
}

SessionManager::SessionManager(Network* n)
	: net_(n), sn_(*this), last_np_(0), rtcp_avg_size_(128.),
	  badversion_(0), badoptions_(0), nrunt_(0)
{
	Tcl& tcl = Tcl::instance();
	tcl.CreateCommand("session", cmd_session, (ClientData)this);
	/*XXX*/
	pktbuf_ = new u_char[2 * VIC_MTU];

	if (n->rchannel() >= 0)
		link(n->rchannel(), TK_READABLE);
	if (n->rchannel() != n->rctrl() && n->rctrl() >= 0)
		sn_.link(n->rctrl(), TK_READABLE);

	/*
	 * schedule a timer for our first report using half the
	 * min rtcp interval.  This gives us some time before
	 * our first report to learn about other sources so our
	 * next report interval will account for them.  The avg
	 * rtcp size was initialized to 128 bytes which is
	 * conservative (it assumes everyone else is generating
	 * SRs instead of RRs).
	 */
	double rint = rtcp_avg_size_ * rtcp_inv_bw;
	if (rint < RTCP_MIN_RPT_TIME / 2. * 1000.)
		rint = RTCP_MIN_RPT_TIME / 2. * 1000.;
	msched(int(fmod(double(random()), rint) + rint * .5 + .5));
}

SessionManager::~SessionManager()
{
	if (pktbuf_) {
		if (net_->sctrl() >= 0)
			send_bye();
		delete pktbuf_;
	}
}

int SessionManager::cmd_session(ClientData cd, Tcl_Interp* t,
				int argc, char **argv)
{
	SessionManager* p = (SessionManager*)cd;
	Network* n = p->net_;
	Tcl& tcl = Tcl::instance();
	char* cp = tcl.buffer();
	if (argc == 2) {
		if (strcmp(argv[1], "addr") == 0)
			strcpy(cp, intoa(n->addr()));
		else if (strcmp(argv[1], "port") == 0)
			sprintf(cp, "%d", ntohs(n->port()));
		else if (strcmp(argv[1], "ttl") == 0)
			sprintf(cp, "%d", n->ttl());
		else if (strcmp(argv[1], "srcid") == 0)
			sprintf(cp, "%u", ntohl(n->srcid()));
		else if (strcmp(argv[1], "name") == 0)
			strcpy(cp,
			       SourceManager::instance().localsrc()->name());
		else if (strcmp(argv[1], "active") == 0)
			SourceManager::instance().sortactive(cp);
		else {
			Tcl_AppendResult(t, "\"", argv[0],
					 "\": bad option", 0);
			return (TCL_ERROR);
		}
		t->result = cp;
		return (TCL_OK);
	} else if (argc == 3) {
		if (strcmp(argv[1], "name") == 0) {
			SourceManager::instance().localsrc()->SetName(argv[2]);
			return (TCL_OK);
		}
	}
	Tcl_AppendResult(t, "\"", argv[0], "\": malformed invocation", 0);
	return (TCL_ERROR);
}

inline n_long SessionManager::ntptime(timeval t)
{
	return (t.tv_sec << 16 | (t.tv_usec << 10) / 15625);
}

inline n_long SessionManager::ntptime()
{
	timeval tv;
	::gettimeofday(&tv, 0);
	return (ntptime(tv));
}

/*
 * convert microseconds to fraction of second * 2^32 (i.e., the lsw of
 * a 64-bit ntp timestamp).  This routine uses the factorization
 * 2^32/10^6 = 4096 + 256 - 3650/64 which results in a conversion
 * error of roughly 10^-6.
 */
static inline u_int usec2ntp(u_int usec)
{
	u_int t = (usec * 3650) >> 6;
	return ((usec << 12) + (usec << 8) - t);
}

static inline ntp64 ntp64time()
{
	ntp64 n;
	timeval tv;
	::gettimeofday(&tv, 0);
	n.upper = tv.tv_sec;
	n.lower = usec2ntp(tv.tv_usec);
	return (n);
}

/*XXX check for buffer overflow*/
/*
 * Send an RTPv2 report packet.
 */
void SessionManager::send_report()
{
	SourceManager& sm = SourceManager::instance();
	Source& s = *sm.localsrc();
	rtcphdr* rh = (rtcphdr*)pktbuf_;
	rh->rh_ssrc = net_->srcid();
	int flags = RTP_VERSION << 14;
	u_int now = ntptime();
	int we_sent = 0;
	rtcp_rr* rr;
	if (s.np() != last_np_) {
		last_np_ = s.np();
		we_sent = 1;
		flags |= RTCP_PT_SR;
		/*
		 * We've sent data since our last sender report.
		 * Send a new report.
		 */
		rtcp_sr* sr = (rtcp_sr*)(rh + 1);
		sr->sr_ntp = ntp64time();
		//now = sr->sr_ntp.upper << 16 | sr->sr_ntp.lower >> 16;
		sr->sr_ts = htonl(now);
		sr->sr_np = htonl(s.nb());
		sr->sr_nb = htonl(s.np());
		rr = (rtcp_rr*)(sr + 1);
	} else {
		flags |= RTCP_PT_RR;
		rr = (rtcp_rr*)(rh + 1);
	}
	int nrr = 0;
	int nsrc = 0;
	for (Source* sp = sm.sources(); sp != 0; sp = sp->next_) {
		++nsrc;
		if (sp->np() == sp->snp())
			continue;
		sp->snp(sp->np());
		rr->rr_srcid = sp->srcid();
		rr->rr_nr = htonl(sp->ns());
		rr->rr_np = htonl(sp->np());
		rr->rr_dv = htonl(n_long(sp->dvar()));
		rr->rr_lsr = sp->ctrlntp();
		rr->rr_dlsr = htonl(now - sp->ts_ctrl());
		++rr;
		if (++nrr >= 31)
			break;
	}
	flags |= nrr << 8;
	rh->rh_flags = htons(flags);
	int len = (u_char*)rr - pktbuf_;
	rh->rh_len = htons((len >> 2) - 1);

	/*
	 * Now build the CNAME.
	 */
	rh = (rtcphdr*)rr;
	flags = RTP_VERSION << 14 | 1 << 8 | RTCP_PT_SDES;
	rh->rh_flags = htons(flags);
	rh->rh_ssrc = net_->srcid();
	u_char* p = (u_char*)(rh + 1);

	const char* name = s.cname();
	len = strlen(name);
	p[0] = RTCP_SDES_CNAME;
	p[1] = len;
	memcpy((char*)&p[2], name, len);
	p += len + 2;

	/* XXX do this one at a lower frequency */
	name = s.name();
	len = strlen(name);
	p[0] = RTCP_SDES_NAME;
	p[1] = len;
	/*
	 * Pre-zero the last word
	 * Need to do this before the memcpy.
	 * If we landed exactly on the boundary, then this
	 * will give us a whole null word to terminate the
	 * sdes, as is needed.
	 */
	int off = (p - pktbuf_) + len + 2;
	n_long* np = (n_long*)(pktbuf_ + (off & ~3));
	*np++ = 0;

	memcpy((char*)&p[2], name, len);

	len = (u_char*)np - (u_char*)rh;
	rh->rh_len = htons((len >> 2) - 1);

	len = (u_char*)np - pktbuf_;
	net_->sendctrl(pktbuf_, len);

	rtcp_avg_size_ += RTCP_SIZE_GAIN * (double(len + 28) - rtcp_avg_size_);

	s.touch_ctrl();

	/*
	 * compute the time to the next report.  we do this here
	 * because we need to know if there were any active sources
	 * during the last report period (nrr above) & if we were
	 * a source.  The bandwidth limit for rtcp traffic was set
	 * on startup from the session bandwidth.  It is the inverse
	 * of bandwidth (ie., ms/byte) to avoid a divide below.
	 */
	double ibw = rtcp_inv_bw;
	if (nrr) {
		/* there were active sources */
		if (we_sent) {
			ibw *= 1./RTCP_SENDER_BW_FRACTION;
			nsrc = nrr;
		} else {
			ibw *= 1./RTCP_RECEIVER_BW_FRACTION;
			nsrc -= nrr;
		}
	}
if (nsrc <= 0) abort();
	double rint = rtcp_avg_size_ * double(nsrc) * ibw;
	if (rint < RTCP_MIN_RPT_TIME * 1000.)
		rint = RTCP_MIN_RPT_TIME * 1000.;
	msched(int(fmod(double(random()), rint) + rint * .5 + .5));
}

void SessionManager::timeout()
{
	/*
	 * Send an rtcp report packet (and possibly other info).
	 */
	send_report();
}

void SessionManager::send_bye()
{
	rtcphdr* rh = (rtcphdr*)pktbuf_;
	int flags = RTP_VERSION << 14 | 1 << 8 | RTCP_PT_BYE;
	rh->rh_flags = ntohs(flags);
	rh->rh_len = htons(1);
	rh->rh_ssrc = net_->srcid();

	net_->sendctrl(pktbuf_, 8);
}

/*
 * parse rtp v1 options (actually, nv options).
 */
int SessionManager::parseopts(const u_char* bp, int cc, n_long addr) const
{
	register int t = 0;
	register u_int optflags;
	do {
		optflags = bp[0];
		register u_int optlen = bp[1] << 2;
		t += optlen;
		if (t > cc || optlen == 0)
			/* bad optlen */
			return (-1);
		switch (optflags & 0x7f) {
		case 33:
			if (optlen > 8) {
				/* extract names from nv's rtp v1 SDES opt */
				char buf[256];
				u_int len = optlen - 9;
				if (len > 255)
					len = 255;
				Source* s = SourceManager::instance().lookup(addr, addr);
				strncpy(buf, (char*)bp + 8, len);
				buf[len] = 0;
				if (s->cname() == 0)
					s->SetCName(intoa(addr));
				s->SetName(buf);
				s->touch_ctrl();
			}
			break;

		case 35:	/* bye option */
			if (optlen == 4) {
				Source* s = SourceManager::instance().consult(addr);
				if (s)
					s->touch_done();
			}
			break;
		}
		bp += optlen;
	} while ((optflags & 0x80) == 0);
	return (t);
}

struct rtp1hdr {
	u_char rh_chan;		/* version/channel */
	u_char rh_flags;	/* O/S/content */
	u_short	rh_seqno;	/* sequence number */
	n_long	rh_ts;
};
#define RTP1_OPTION_BIT 0x80

void SessionManager::dispatch(int mask)
{
	n_long src;
	int cc = net_->recv(pktbuf_, 2 * VIC_MTU, src);
	if (cc <= 0)
		return;

	rtphdr* rh = (rtphdr*)pktbuf_;
	u_char* bp;
	int version = *(u_char*)rh >> 6;
	rtphdr hdr;
	switch (version) {
		int flags;
		struct rtp1hdr* rh1;

	case 1:
		/* RTP v1 hdr */
		rh1 = (rtp1hdr*)rh;
		flags = rh1->rh_flags;
		flags = (flags & 0x40) << 1 | flags & 0x3f;
		hdr.rh_flags = htons(flags);
		hdr.rh_seqno = rh1->rh_seqno;
		hdr.rh_ts = rh1->rh_ts;
		hdr.rh_ssrc = src;
		bp = pktbuf_ + sizeof(*rh1);
		cc -= sizeof(*rh1);
		if (rh1->rh_flags & RTP1_OPTION_BIT) {
			int t = parseopts(bp, cc, src);
			if (t <= 0) {
				++badoptions_;
				return;
			}
			bp += t;
			cc -= t;
			if (cc == 0)
				return;
		}
		rh = &hdr;
		break;

	case 2:
		bp = pktbuf_ + sizeof(*rh);
		cc -= sizeof(*rh);
		break;

	default:
		++badversion_;
		return;
	}
	if (cc < 0) {
		++nrunt_;
		return;
	}
	n_long srcid = rh->rh_ssrc;
	int flags = ntohs(rh->rh_flags);
	int fmt = flags & 0x7f;

	SourceManager& sm = SourceManager::instance();
	Source& s = *sm.lookup(srcid, src);
	s.touch_data();
	/*
	 * This is a data packet.  If the source needs activation,
	 * or the packet format has changed, deal with this.
	 * Then, hand the packet off to the packet handler.
	 * XXX might want to be careful about flip-flopping
	 * here when format changes due to misordered packets
	 * (easy solution -- keep rtp seqno of last fmt change).
	 */
	PacketHandler* h = s.handler();
	if (h == 0)
		h = s.activate(fmt, (u_char*)bp);
	else if (s.format() != fmt)
		h = s.change_format(fmt, (u_char*)bp);

	/*
	 * XXX bit rate doesn't include rtpv1 options;
	 * but v1 is going away anyway.
	 */
	s.cs((u_short)ntohs(rh->rh_seqno));
	s.np(1);
	s.nb(cc + sizeof(*rh));
	if (s.mute()) {
		if (flags & RTP_M)
			s.nf(1);
		return;
	}
	/*
	 * Adjust for timestamp clock skew, then compute
	 * the time difference between the local clock and
	 * the scheduled time.  If we're too far behind,
	 * we need to do something.  That something is 
	 * determined by the higher layers.  E.g., for
	 * motion jpeg we will just drop frames; for
	 * interframe algorithms we should do something
	 * smarter (e.g., for nv-encoding try turning off
	 * dithering to catch up).
	 *
	 * Note that the jitter can result from two
	 * sources.  The network (or sender) might inject
	 * delay, or the cpu might be busy doing software
	 * decompression.  In either case, we would just
	 * discard old frames for motion jpeg (provided
	 * newer ones were present).  But for other
	 * coding algorithms, we want to process all
	 * the data.
	 *
	 * When we add packet timestamping to the kernel,
	 * we will be able to distinguish these two cases.
	 */
	n_long ts = rh->rh_ts;
	s.datantp(ts);
	NTOHL(ts);
	n_long now = ntptime();
	if (flags & RTP_M)
		s.adapt(ts, now);
	int late;
#ifdef notyet
	if (s.skew() != 0) {
		ts += s.skew();
#define FUDGE (30 * 65536 / 1000)
		if ((ts + FUDGE + int(2. * s.dvar())) < now ) {
			late = 1;
			s.late(1);
		}
	} else
#endif
		late = 0;

	int hlen = h->hdrlen();
	cc -= hlen;
	if (cc < 0) {
		s.runt(1);
		return;
	}
	if (h->recv(rh, bp, bp + hlen, cc, late))
		s.nf(1);
}

void SessionManager::parse_sr(rtcphdr* rh, int flags, u_char* ep, Source* ps,
			      n_long addr)
{
	rtcp_sr* sr = (rtcp_sr*)(rh + 1);
	/*XXX*/
	Source* s;
	if (ps->srcid() != rh->rh_ssrc)
		s = SourceManager::instance().lookup(rh->rh_ssrc, addr);
	else
		s = ps;
	s->touch_ctrl();
	s->ctrlntp(sr->sr_ts);

	/*XXX parse rr here */
}

void SessionManager::parse_rr(rtcphdr* rh, int flags, u_char* ep, Source* ps,
			      n_long addr)
{
	Source* s;
	if (ps->srcid() != rh->rh_ssrc)
		s = SourceManager::instance().lookup(rh->rh_ssrc, addr);
	else
		s = ps;
	s->touch_ctrl();

	/*XXX parse rr here */
}

int SessionManager::sdesbody(n_long* p, u_char* ep, Source* ps, n_long addr)
{
	Source* s;
	n_long id = *p;
	if (ps->srcid() != id)
		s = SourceManager::instance().lookup(id, addr);
	else
		s = ps;
	u_char* cp = (u_char*)(p + 1);

	while (cp < ep) {
		char buf[256];

		u_int type = cp[0];
		if (type == 0) {
			/* end of chunk */
			return (((cp - (u_char*)p) >> 2) + 1);
		}
		u_int len = cp[1];
		u_char* eopt = cp + len + 2;
		if (eopt > ep)
			return (0);

		switch (type) {
		case RTCP_SDES_CNAME:
			memcpy(buf, (char*)&cp[2], len);
			buf[len] = 0;
			s->SetCName(buf);
			break;

		case RTCP_SDES_NAME:
			memcpy(buf, (char*)&cp[2], len);
			buf[len] = 0;
			s->SetName(buf);
			break;
		}
		cp = eopt;
	}
	return (0);
}

void SessionManager::parse_sdes(rtcphdr* rh, int flags, u_char* ep, Source* ps,
				n_long addr)
{
	int cnt = flags >> 8 & 0x1f;
	n_long* p = (n_long*)&rh->rh_ssrc;
	while (--cnt >= 0) {
		int n = sdesbody(p, ep, ps, addr);
		if (n == 0)
			break;
		p += n;
	}
	if (cnt >= 0)
		ps->badsdes(1);
}

void SessionManager::parse_bye(rtcphdr* rh, int flags, u_char* ep, Source* ps)
{
	int cnt = flags >> 8 & 0x1f;
	n_long* p = (n_long*)&rh->rh_ssrc;

	while (--cnt >= 0) {
		if (p >= (n_long*)ep) {
			ps->badbye(1);
			return;
		}
		Source* s;
		if (ps->srcid() != rh->rh_ssrc)
			s = SourceManager::instance().consult(*p);
		else
			s = ps;
		if (s != 0)
			s->touch_done();
		++p;
	}
}

/*
 * Receive an rtcp packet (from the control port).
 */
void SessionManager::rtcp_recv()
{
	n_long src;
	int cc = net_->recvctrl(pktbuf_, 2 * VIC_MTU, src);
	if (cc <= 0)
		return;

	rtcphdr* rh = (rtcphdr*)pktbuf_;
	if (cc < sizeof(*rh)) {
		++nrunt_;
		return;
	}
	/*
	 * try to filter out junk: first thing in packet must be
	 * sr, rr or bye & version number must be correct.
	 */
	switch(ntohs(rh->rh_flags) & 0xc0ff) {
	case RTP_VERSION << 14 | RTCP_PT_SR:
	case RTP_VERSION << 14 | RTCP_PT_RR:
	case RTP_VERSION << 14 | RTCP_PT_BYE:
		break;
	default:
		++badversion_;
		return;
	}
	/*
	 * at this point we think the packet's valid.  Update our average
	 * size estimator.  Also, there's valid ssrc so charge errors to it
	 */
	rtcp_avg_size_ += RTCP_SIZE_GAIN * (double(cc + 28) - rtcp_avg_size_);
	n_long addr = src;
	Source* ps = SourceManager::instance().lookup(rh->rh_ssrc, addr);

	u_char* epack = (u_char*)rh + cc;
	while ((u_char*)rh < epack) {
		u_int len = (ntohs(rh->rh_len) << 2) + 4;
		u_char* ep = (u_char*)rh + len;
		if (ep > epack) {
			ps->badsesslen(1);
			return;
		}
		u_int flags = ntohs(rh->rh_flags);
		if (flags >> 14 != RTP_VERSION) {
			ps->badsessver(1);
			return;
		}
		switch (flags & 0xff) {

		case RTCP_PT_SR:
			parse_sr(rh, flags, ep, ps, addr);
			break;

		case RTCP_PT_RR:
			parse_rr(rh, flags, ep, ps, addr);
			break;

		case RTCP_PT_SDES:
			parse_sdes(rh, flags, ep, ps, addr);
			break;

		case RTCP_PT_BYE:
			parse_bye(rh, flags, ep, ps);
			break;

		default:
			ps->badsessopt(1);
			break;
		}
		rh = (rtcphdr*)ep;
	}
}


IvsSessionManager::IvsSessionManager(Network* n) : SessionManager(n)
{
}

IvsSessionManager::~IvsSessionManager()
{
	if (net_->sctrl() >= 0)
		send_bye();
	/* pktbuf_ is zeroed to flag our base class not to send another bye */
	delete pktbuf_;
	pktbuf_ = 0;
}

void IvsSessionManager::rtcp_recv()
{
	/*
	 * handle in incoming ivs session packet
	 */
	n_long src;
	int cc = net_->recvctrl(pktbuf_, 2 * VIC_MTU, src);
	if (cc <= 0)
		return;

	rtp1hdr* rh = (rtp1hdr*)pktbuf_;
	if (cc < sizeof(*rh)) {
		++nrunt_;
		return;
	}
	if ((rh->rh_chan & 0xc0) != (1 << 6) ||
	    (rh->rh_flags & RTP1_OPTION_BIT) == 0) {
		++badversion_;
		return;
	}
	u_int optflags;
	u_char* bp = (u_char*)(rh + 1);
	u_char* ep = bp + cc - sizeof(*rh);;
	n_long addr = (n_long)src;
	do {
		optflags = bp[0];
		u_int optlen = bp[1] << 2;
		if (bp + optlen > ep || optlen == 0) {
			/* empty packet or bad optlen */
			++badoptions_;
			return;
		}
		switch (optflags & 0x7f) {
			Source* s;
			u_int len;
			char buf[256 + 32];

		case 34:	/* ivs's rtp v1 SDES opt */
			len = bp[17] << 2;
			if (optlen < 20 || len > optlen - 16 || len < 4) {
				++badoptions_;
				return;
			}
			len -= 2;
			s = SourceManager::instance().lookup(addr, addr);
			strncpy(buf, (char*)bp + 18, len);
			buf[len] = 0;
			if (bp[16] != 16) {
				len = strlen(buf);
				buf[len] = '@';
				strcpy(buf + len + 1, intoa(addr));
				s->SetCName(buf);
			} else {
				if (s->cname() == 0)
					s->SetCName(intoa(addr));
				s->SetName(buf);
			}
			s->touch_ctrl();
			break;

		case 35:	/* bye option */
			if (optlen != 4) {
				++badoptions_;
				return;
			}
			s = SourceManager::instance().consult(addr);
			if (s)
				s->touch_done();
		}
		bp += optlen;
	} while ((optflags & 0x80) == 0);
}

/*
 * Send an IVS report packet.
 */
void IvsSessionManager::send_report()
{
	SourceManager& sm = SourceManager::instance();
	Source& s = *sm.localsrc();
	rtp1hdr* rh = (rtp1hdr*)pktbuf_;
	rh->rh_chan = 1 << 6;
	rh->rh_flags = RTP1_OPTION_BIT | RTP_PT_H261;
	rh->rh_seqno = 0;
	rh->rh_ts = 0;
	u_char* opt = (u_char*)(rh + 1);

	opt[0] = 127;	/*APP option*/
	opt[1] = 3;	/*opt len*/
	opt[2] = 0;
	opt[3] = 0;
	opt[4] = 'I';
	opt[5] = 'V';
	opt[6] = 'S';
	opt[7] = ' ';
	opt[8] = 0x33;	/*ivs version*/
	opt[9] = (s.ts_data() && int(s.ts_data() - s.ts_ctrl()) >= 0)?
			(1 << 6) : 0;
	opt[10] = 0;
	opt[11] = 0;
	opt += 12;

	opt[0] = 34 | 0x80;	/*SDES option + Final*/
	const char* name = s.name();
	int len = strlen(name);
	int optlen = (len + (19 + 3)) >> 2;
	u_char* ep = opt + (optlen << 2);
	opt[1] = optlen;
	opt[2] = 0;		/*id*/
	opt[3] = 0;
	opt[4] = 2;		/*port*/
	opt[5] = 1;		/*port len*/
	opt[6] = ntohs(net_->cport()) >> 8;
	opt[7] = ntohs(net_->cport());
	opt[8] = 1;		/*addr*/
	opt[9] = 2;		/*addr len*/
	opt[10] = 0;
	opt[11] = 1;		/*addr type = ipv4*/
	*(u_int*)&opt[12] = net_->interface();
	opt[16] = 16;		/*class user text*/
	opt[17] = optlen - 4;
	opt += 18;
	strcpy((char*)opt, name);
	opt += len;
	while (opt < ep)
		*opt++ = 0;

	len = ep - pktbuf_;
	net_->sendctrl(pktbuf_, len);

	s.touch_ctrl();

	/* schedule a timer for the next report */
	double rint = rtcp_inv_bw * 64.;
	if (rint < RTCP_MIN_RPT_TIME * 1000.)
		rint = RTCP_MIN_RPT_TIME * 1000.;
	msched(int(fmod(double(random()), rint) + rint * .5 + .5));
}

void IvsSessionManager::send_bye()
{
	rtp1hdr* rh = (rtp1hdr*)pktbuf_;
	rh->rh_chan = 1 << 6;
	rh->rh_flags = RTP1_OPTION_BIT | RTP_PT_H261;
	rh->rh_seqno = 0;
	rh->rh_ts = 0;
	u_char* opt = (u_char*)(rh + 1);
	opt[0] = 35 | 0x80;	/*BYE option + Final*/
	opt[1] = 1;
	opt[2] = 0;		/*id*/
	opt[3] = 0;
	net_->sendctrl(pktbuf_, sizeof(*rh) + 4);
}


NvSessionManager::NvSessionManager(Network* n) : SessionManager(n)
{
}

NvSessionManager::~NvSessionManager()
{
	send_bye();
	/* pktbuf_ is zeroed to flag our base class not to send another bye */
	delete pktbuf_;
	pktbuf_ = 0;
}

/*
 * Send an NV report packet.
 */
void NvSessionManager::send_report()
{
	/* only send name if we're sending video */
	Source& s = *SourceManager::instance().localsrc();
	if (s.np() != last_np_) {
		last_np_ = s.np();
		rtp1hdr* rh = (rtp1hdr*)pktbuf_;
		rh->rh_chan = 0x60; /* v1, chan 32 */
		rh->rh_flags = RTP1_OPTION_BIT | 28;
		rh->rh_seqno = s.cs();
		rh->rh_ts = htonl(ntohl(s.datantp()) - 1);
		u_char* opt = (u_char*)(rh + 1);

		opt[0] = 33 | 0x80;	/* nv SDESC option + Final*/
		const char* name = s.name();
		int len = strlen(name);
		int optlen = ((len + 8) >> 2) + 1;
		u_char* ep = opt + (optlen << 2);
		opt[1] = optlen;
		opt[2] = 0;		/*id*/
		opt[3] = 0;
		*(u_int*)&opt[4] = net_->interface();
		opt += 8;
		strcpy((char*)opt, name);
		opt += len;
		while (opt < ep)
			*opt++ = 0;

		len = ep - pktbuf_;
		net_->send(pktbuf_, len);
	}
	s.touch_ctrl();

	/* schedule a timer for the next report */
	double rint = rtcp_inv_bw * 64.;
	if (rint < RTCP_MIN_RPT_TIME * 1000.)
		rint = RTCP_MIN_RPT_TIME * 1000.;
	msched(int(fmod(double(random()), rint) + rint * .5 + .5));
}

void NvSessionManager::send_bye()
{
	/* only send bye if we've sent video */
	Source& s = *SourceManager::instance().localsrc();
	if (s.ns() != 0) {
		rtp1hdr* rh = (rtp1hdr*)pktbuf_;
		rh->rh_chan = 0x60; /* v1, chan 32 */
		rh->rh_flags = RTP1_OPTION_BIT | 28;
		rh->rh_seqno = s.cs();
		rh->rh_ts = htonl(ntohl(s.datantp()) - 1);
		u_char* opt = (u_char*)(rh + 1);
		opt[0] = 35 | 0x80;	/*BYE option + Final*/
		opt[1] = 1;
		opt[2] = 0;		/*id*/
		opt[3] = 0;
		net_->send(pktbuf_, sizeof(*rh) + 4);
	}
}
