/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/session.h,v 1.2 1995/08/16 00:46:41 olson Exp (LBL)
 */

#ifndef vic_session_h
#define vic_session_h

extern "C" {
#include <tk.h>
}
#include "inet.h"
#include "packet.h"
#include "iohandler.h"
#include "timer.h"

/* session protocol types */
#define SESSION_VIC 1
#define SESSION_IVS 2
#define SESSION_NV 3

class Network;
class Decoder;
class Source;

class SessionManager;

class SessionNotifier : public IOHandler {
 public:
	SessionNotifier(SessionManager&);
	virtual void dispatch(int mask);
 private:
	SessionManager& manager_;
};

class SessionManager : public IOHandler, public Timer {
 public:
	SessionManager(Network*);
	virtual ~SessionManager();
	static int cmd_session(ClientData, Tcl_Interp*, int argc, char **argv);
	virtual void rtcp_recv();
 protected:
	n_long ntptime(timeval t);
	n_long ntptime();
	virtual void send_report();
	virtual void send_bye();

	void parse_sr(rtcphdr* rh, int flags, u_char* ep, Source* ps, n_long addr);
	void parse_rr(rtcphdr* rh, int flags, u_char* ep, Source* ps, n_long addr);
	int sdesbody(n_long* p, u_char* ep, Source* ps, n_long addr);
	void parse_sdes(rtcphdr* rh, int flags, u_char* ep, Source* ps, n_long addr);
	void parse_bye(rtcphdr* rh, int flags, u_char* ep, Source* ps);

	int parseopts(const u_char* bp, int cc, n_long addr) const;
	virtual void timeout();
	virtual void dispatch(int mask);
	int recv(int fd, sockaddr_in& src);
	int ckid(const char*, int len);

	Network* net_;
	SessionNotifier sn_;

	/*XXX cleanup*/
	u_int badversion_;
	u_int badoptions_;
	u_int nrunt_;

	n_long last_np_;

	double rtcp_avg_size_;	/* (estimated) average size of rtcp packets */
	u_char* pktbuf_;
};

class IvsSessionManager : public SessionManager {
 public:
	IvsSessionManager(Network*);
	virtual ~IvsSessionManager();
	virtual void rtcp_recv();
 protected:
	virtual void send_report();
	virtual void send_bye();
};

class NvSessionManager : public SessionManager {
 public:
	NvSessionManager(Network*);
	virtual ~NvSessionManager();
 protected:
	virtual void send_report();
	virtual void send_bye();
};
	
#endif
