/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/source-vic.C,v 1.4 1995/08/26 21:15:59 olson Exp (LBL)";
#endif

extern "C" {
#include <tk.h>
}

#include <stdlib.h>
#include <osfcn.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>

#include "source-vic.h"
#include "net.h"
#include "video.h"
#include "decoder.h"
#include "inet.h"
#include "packet.h"
#include "Tcl.h"
#include "switcher.h"
#include "color.h"

/*XXX this goes elsewhere */
#define CTRL_IDLE 30		/* gray out src if no ctrl msgs for 30 sec. */
#define CTRL_IDLE_DUP 10	/* gray out src if more than one participant
				   on a given host, and no ctrl msgs for
				   10 sec. */

#define SHASH(a) ((int)((((a) >> 20) ^ ((a) >> 10) ^ (a)) & (SOURCE_HASH-1)))

vicSource::vicSource(n_long srcid, n_long addr)
	: Source(srcid, addr), stamp_(0), windows_(0), vnext_(0), vprev_(0)
{
	Tcl::instance().evalf("register %s", method_);
}

vicSource::~vicSource()
{
	if (handler_ != 0)
		deactivate();
	Tcl::instance().evalf("unregister %s", method_);
}

PacketHandler* vicSource::activate(int format, u_char* pkthdr)
{
	extern const char* lookup_format(int format);

	format_ = format;
	Tcl& tcl = Tcl::instance();
	tcl.evalf("activate %s %s 1", method_, lookup_format(format));
	/*
	 * The tcl activate command returned the tk window name
	 * of the thumb nail.  Tell the decoder to output to it.
	 */
	stamp_ = VideoWindow::name_to_window(tcl.result());
	if (stamp_ == 0)
		abort();
	Decoder* d = Decoder::create(format, pkthdr);
	d->attach(stamp_, 1);

	decoder_ = d;
	handler_ = d;

	vicSourceManager::instance().insert(this);

	return (d);
}

void vicSource::deactivate()
{
	vicSource* other = vnext_;
	if (other == 0)
		/* not in list? */
		abort();

	Tcl& tcl = Tcl::instance();
	/*XXX*/
	tcl.evalf("deactivate %s", method_);
	/*
	 * We are in the active video list, which might
	 * mean that a corresponding UserWindow exists
	 * somewhere.  If there is another active sender,
	 * switch to it, otherwise, delete all UserWindow's.
	 */
	if (other == this)
		other = 0;
	UserWindow* p = windows_;
	while (p != 0) {
		UserWindow* n = p->next_;
		detach(p);
		if (other && p->mode() != USERWIN_MODE_LOCKED)
			other->attach(p);
		else {
			p->window()->ka(0);
			p->destroy();
			delete p;
		}
		p = n;
	}
	windows_ = 0;
	/*
	 * Revert things to a state where the source exists, but
	 * is not actively transmitting.  We delete the decoder
	 * state and thumbnail window, and move it from the viewing
	 * window list.
	 */
	vicSourceManager::instance().remove(this);
	delete decoder_;
	delete stamp_;
	decoder_ = 0;
	stamp_ = 0;
	handler_ = 0;
}

PacketHandler* vicSource::change_format(int format, u_char* pkthdr)
{
	format_ = format;
	Tcl& tcl = Tcl::instance();
	tcl.evalf("change_format %s", method_);
	/*
	 * XXX is this necessary?  I think we only need to 
	 * delete the decoder.
	 */
	Decoder* d = decoder();
	d->detach(stamp_);
	UserWindow *uw;
	for (uw = windows_; uw != 0; uw = uw->next_)
		d->detach(uw->window());

	delete d;
	d = Decoder::create(format, pkthdr);
	d->attach(stamp_, 1);
	for (uw = windows_; uw != 0; uw = uw->next_)
		d->attach(uw->window(), 0);

	decoder_ = d;
	handler_ = d;

	return (d);
}

void vicSource::name(const char* s)
{
	Source::name(s);
	if (*s == 0)
		return;
	for (UserWindow* uw = windows_; uw != 0; uw = uw->next_)
		uw->name(s);
}

int vicSource::command(Tcl_Interp* tcl, int argc, char **argv)
{
	char* wrk = Tcl::instance().buffer();

	if (argc == 2) {
#ifdef VIC_ANALOG
		if (strcmp(argv[1], "analog") == 0) {
#ifdef DEBUG
		        printf("Analog %s!\n", tcl_method());
#endif
			decoder_->toggle_analog();
			return (TCL_OK);
		}
#endif
		if (strcmp(argv[1], "select") == 0) {
			select();
			return (0);
		}
		if (strcmp(argv[1], "width") == 0) {
			sprintf(wrk, "%d", decoder()->width());
			tcl->result = wrk;
			return (0);
		}
		if (strcmp(argv[1], "height") == 0) {
			sprintf(wrk, "%d", decoder()->height());
			tcl->result = wrk;
			return (0);
		}
		if (strcmp(argv[1], "decoder-whichstats") == 0) {
			decoder()->counter_names(wrk);
			tcl->result = wrk;
			return (0);
		}
		if (strcmp(argv[1], "decoder-info") == 0) {
			decoder()->info(wrk);
			tcl->result = wrk;
			return (0);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "counter") == 0) {
			u_int v;
			if (decoder()->counter_query(argv[2], v) == 0) {
				sprintf(wrk, "%u", v);
				tcl->result = wrk;
				return (0);
			}
			return (-1);
		}
		if (strcmp(argv[1], "color") == 0) {
			int v = atoi(argv[2]);
			decoder()->setcolor(v);
			return (0);
		}
	}
	return (-1);
}
	
/*
 * The user hit 'q' in the window; this routine gets called
 * back from the VideoWindow object.  The window has been
 * destroyed but the C++ object hasn't.  So find the corresponding
 * UserWindow and nuke it.  Here and ~Source() are the only
 * places where UserWindows are destructed.  E.g., the Source
 * object has ultimate control of UserWindow existence.
 */
void vicSource::kill(VideoWindow* w)
{
	for (UserWindow* p = windows_; p != 0; p = p->next_) {
		if (p->window() == w) {
			detach(p);
			delete p;
			return;
		}
	}
}

void vicSource::select()
{
	attach(new UserWindow(this));
}

void vicSource::attach(UserWindow* w)
{
	w->next_ = windows_;
	windows_ = w;
	w->window()->ka(this);
	decoder()->attach(w->window(), 0);
	if (name_ != 0)
		w->name(name_);
	else if (cname_ != 0)
		w->name(cname_);
	else
		w->name(InetNtoa(addr_));
	if (w->mode() == USERWIN_MODE_SWITCHED)
		Switcher::instance().attach(w);
	w->source(this);
}

void vicSource::detach(UserWindow* w)
{
	decoder()->detach(w->window());
	UserWindow** p = &windows_;
	while (*p != w)
		p = &(*p)->next_;
	*p = w->next_;
	w->window()->ka(0);
	if (w->mode() == USERWIN_MODE_SWITCHED)
		Switcher::instance().detach(w);
	w->source(0);
}

vicSourceManager::vicSourceManager() : list_(0)
{
	/*XXX*/
	Tcl::instance().CreateCommand("srctab", command, (ClientData)this);
}

void vicSourceManager::init(n_long localid, n_long localaddr)
{
	if (instance_ != 0)
		abort();
	instance_ = new vicSourceManager;
	instance_->init(localid, localaddr);
}

void vicSourceManager::insert(vicSource* p)
{
	if (list_ == 0) {
		list_ = p;
		p->vprev_ = p->vnext_ = p;
	} else {
		p->vnext_ = list_->vnext_;
		p->vprev_ = list_;
		list_->vnext_->vprev_ = p;
		list_->vnext_ = p;
	}
}

void vicSourceManager::remove(vicSource* p)
{
	for (vicSource* s = list_; ; s = s->vnext_) {
		if (s == p) {
			if (list_ == p)
				list_ = list_->vnext_;
			if (list_ == p)
				list_ = 0;
			else {
				s->vprev_->vnext_ = s->vnext_;
				s->vnext_->vprev_ = s->vprev_;
			}
			p->vnext_ = p->vprev_ = 0;
			return;
		}
	}
}

int vicSourceManager::command(ClientData cd, Tcl_Interp* t,
			      int argc, char **argv)
{
	vicSourceManager* sm = (vicSourceManager*)cd;
	if (argc == 2) {
		if (strcmp(argv[1], "mediancut") == 0) {
			ColorModel& cm = ColorModel::instance();
			if (cm.procBusy() || !sm->list_)
				return (TCL_OK);
		
			u_int histogram[65536];
			memset(histogram, 0, sizeof(histogram));
			int havehist = 0;
			vicSource* s = sm->list_;
			vicSource* ss = s;
			do {
				if (s->mute())
					continue;
				havehist |= s->decoder()->colorhist(histogram);
			} while ((s = s->vnext_) != ss);
			if (havehist)
				cm.fork(histogram);
/*XXX*/
#ifdef notyet
			tcl->result("1");
			else
				tcl->result("0");
#endif
			return (TCL_OK);
		}
		if (strcmp(argv[1], "redraw") == 0) {
			/*
			 * Redraw all windows.
			 */
			vicSource* s = sm->list_;
			vicSource* ss = s;
			do {
				s->decoder()->redraw();
			} while ((s = s->vnext_) != ss);

			return (TCL_OK);
		}
	}
	/*XXX*/
	Tcl_AppendResult(t, "\"", argv[0], "\": arg mismatch", 0);
	return (TCL_ERROR);
}

Source* vicSourceManager::create(n_long srcid, n_long addr)
{
	return (new vicSource(srcid, addr));
}

vicSource* vicSourceManager::consultXXX(n_long addr)
{
	/*
	 * There should never be all that many active sources
	 * (though there may be many non-active "sources" in
	 * the conference), so just do a linear search.
	 */
	vicSource* s = list_;
	if (s) {
		vicSource* ss = s;
		do {
			if (s->addr() == addr)
				break;
		} while ((s = s->vnext_) != ss);
	}
	return (s);
}
