/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/source-vic.h,v 1.2 1995/08/16 00:46:44 olson Exp (LBL)
 */

#ifndef vic_source_vic_h
#define vic_source_vic_h

#include "source.h"
#include "video.h"

class Decoder;
class UserWindow;

/*
 * An object for sources actively sending video.
 * i.e., this object is one-to-one with a panel
 * in the main window (i.e., thumbnail, stats, etc)
 */
class vicSource : public KillAction, public Source {
 public:
	vicSource(n_long srcid, n_long addr);
	virtual ~vicSource();

	/* import from KillAction */
	void kill(VideoWindow*);

	/* inherited from Source */
	virtual int command(Tcl_Interp*, int ac, char** av);
	virtual PacketHandler* activate(int format, u_char* pkthdr);
	virtual PacketHandler* change_format(int format, u_char* pkthdr);
	virtual void deactivate();
	virtual void name(const char*);

	void enable();
	void disable();
	void attach(UserWindow* w);
	void detach(UserWindow* w);

	inline Decoder* decoder() const { return (decoder_); }
 protected:
	void select();
	UserWindow* windows_;	/* associated top-level windows */
	VideoWindow* stamp_;	/* thumbnail */
	Decoder* decoder_;
 public:
	vicSource* vnext_;	/* next on active video list  */
	vicSource* vprev_;	/* next on active video list  */
};

class vicSourceManager : public SourceManager {
 protected:
	vicSourceManager();
 public:
	static void init(n_long localid, n_long localaddr);
	static inline vicSourceManager& instance() {
		return (*(vicSourceManager*)instance_);
	}
	virtual Source* create(n_long srcid, n_long addr);
	vicSource* consultXXX(n_long addr);
	static int command(ClientData, Tcl_Interp*, int ac, char** av);
 protected:
	vicSource* list_;
 public:
	void insert(vicSource*);
	void remove(vicSource*);
};

#endif
