/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/source.C,v 1.3 1995/08/26 21:16:00 olson Exp (LBL)";
#endif

extern "C" {
#include <tk.h>
}

#include <stdlib.h>
#include <osfcn.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#ifdef _AIX
#include <time.h>
extern "C" int strcasecmp (const char *, const char *);
#endif

#include "source.h"
#include "inet.h"
#include "Tcl.h"

/*XXX this goes elsewhere */
#define CTRL_IDLE 30		/* gray out src if no ctrl msgs for 30 sec. */
#define CTRL_IDLE_DUP 10	/* gray out src if more than one participant
				   on a given host, and no ctrl msgs for
				   10 sec. */

#define SHASH(a) ((int)((((a) >> 20) ^ ((a) >> 10) ^ (a)) & (SOURCE_HASH-1)))

Source::Source(n_long srcid, n_long addr)
	: addr_(addr), srcid_(srcid), handler_(0),
	  hlink_(0), next_(0),
	  ts_done_(0), ts_ctrl_(0), ts_data_(0), datantp_(0), ctrlntp_(0),
	  format_(-1), mute_(0), disabled_(0), dup_(0),
	  skew_(0),
	  name_(0), cname_(0),
	  fs_(0), cs_(0), np_(0), nf_(0), nb_(0), snp_(0), nlate_(0), nrunt_(0),
	  badsesslen_(0), badsessver_(0), badsessopt_(0), badsdes_(0),
	  badbye_(0), dvar_(0.), delta_(0)
{
	sprintf(method_, "s%lu", (u_long)ntohl(srcid_));
	Tcl& tcl = Tcl::instance();
	tcl.CreateCommand(method_, command, (ClientData)this, 0);
}

Source::~Source()
{
	delete name_;
	delete cname_;
}

void Source::name(const char* s)
{
	delete name_;
	name_ = new char[strlen(s) + 1];
	strcpy(name_, s);
	if (*name_ != 0)
		Tcl::instance().evalf("change_name %s", method_);
}

void Source::cname(const char* s)
{
	delete cname_;
	cname_ = new char[strlen(s) + 1];
	strcpy(cname_, s);
	if (*cname_ != 0) {
		Tcl& tcl = Tcl::instance();
		tcl.evalf("update_source_info %s", method_);
	}
}

int Source::command(ClientData cd, Tcl_Interp* tcl, int argc, char **argv)
{
	extern const char* lookup_format(int format);

	Source* s = (Source*)cd;
	char* wrk = Tcl::instance().buffer();

	/*
	 * See if derived class wants this proc first.
	 */
	if (s->command(tcl, argc, argv) == 0)
		return (TCL_OK);

	if (argc == 2) {
		if (strcmp(argv[1], "delete") == 0) {
			SourceManager::instance().remove(s);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "deactivate") == 0) {
			s->deactivate();
			return (TCL_OK);
		}
		if (strcmp(argv[1], "ns") == 0) {
			sprintf(wrk, "%lu", (u_long)s->ns());
			tcl->result = wrk;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "dvar") == 0) {
			double ms = s->dvar() * (1000. / 65536.);
			sprintf(wrk, "%g", ms);
			tcl->result = wrk;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "np") == 0) {
			sprintf(wrk, "%lu", (u_long)s->np());
			tcl->result = wrk;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "nb") == 0) {
			sprintf(wrk, "%lu", (u_long)(8 * s->nb()));
			tcl->result = wrk;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "nf") == 0) {
			sprintf(wrk, "%lu", (u_long)s->nf());
			tcl->result = wrk;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "addr") == 0) {
			strcpy(wrk, InetNtoa(s->addr_));
			tcl->result = wrk;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "srcid") == 0) {
			sprintf(wrk, "%lu", (u_long)ntohl(s->srcid_));
			tcl->result = wrk;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "name") == 0) {
			if (s->name_ == 0)
				tcl->result = "";
			else {
				strcpy(wrk, s->name_);
				tcl->result = wrk;
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "cname") == 0) {
			if (s->cname_ == 0)
				tcl->result = "";
			else {
				strcpy(wrk, s->cname_);
				tcl->result = wrk;
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "format") == 0) {
			const char* fmt = lookup_format(s->format_);
			if (fmt == 0)
				tcl->result = "";
			else {
				strcpy(wrk, fmt);
				tcl->result = wrk;
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "whichstats") == 0) {
			tcl->result =
	"Kilobits Frames Packets Missing Late Runts Bad-S-Len Bad-S-Ver Bad-S-Opt Bad-Sdes Bad-Bye";
			return (TCL_OK);
		}
		if (strcmp(argv[1], "lastdata") == 0) {
			time_t now = s->ts_data_;
			tcl->result = (now != 0) ? ctime(&now) : "";
			return (TCL_OK);
		}
		if (strcmp(argv[1], "lastctrl") == 0) {
			time_t now = s->ts_ctrl_;
			tcl->result = (now != 0) ? ctime(&now) : "";
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "disable") == 0) {
			int v = atoi(argv[2]);
			s->mute(v);
			return (TCL_OK);
		}			
		if (strcmp(argv[1], "counter") == 0) {
			if (strcasecmp(argv[2], "packets") == 0) {
				sprintf(wrk, "%lu", (u_long)s->np());
				tcl->result = wrk;
				return (TCL_OK);
			}
			/*XXX*/
			if (strcasecmp(argv[2], "kilobits") == 0) {
				sprintf(wrk, "%lu",
					(u_long)(s->nb() >> (10-3)));
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "frames") == 0) {
				sprintf(wrk, "%lu", (u_long)s->nf());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "late") == 0) {
				sprintf(wrk, "%lu", (u_long)s->late());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "runts") == 0) {
				sprintf(wrk, "%lu", (u_long)s->runt());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "bad-s-len") == 0) {
				sprintf(wrk, "%lu", (u_long)s->badsesslen());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "bad-s-ver") == 0) {
				sprintf(wrk, "%lu", (u_long)s->badsessver());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "bad-s-opt") == 0) {
				sprintf(wrk, "%lu", (u_long)s->badsessopt());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "bad-sdes") == 0) {
				sprintf(wrk, "%lu", (u_long)s->badsdes());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "bad-bye") == 0) {
				sprintf(wrk, "%lu", (u_long)s->badbye());
				tcl->result = wrk;
				return (TCL_OK);
			}
			if (strcasecmp(argv[2], "missing") == 0) {
				int missing = s->ns() - s->np();
				if (missing < 0)
					missing = 0;
				sprintf(wrk, "%lu", (u_long)missing);
				tcl->result = wrk;
				return (TCL_OK);
			}
		}
	}				
	Tcl_AppendResult(tcl, "\"", argv[0], "\": arg mismatch", 0);
	return (TCL_ERROR);
}
	
void Source::enable()
{
	if (disabled_) {
		disabled_ = 0;
		Tcl::instance().evalf("embolden %s", method_);
	}
}

void Source::disable()
{
	if (!disabled_) {
		disabled_ = 2;
		Tcl::instance().evalf("grayout %s", method_);
	}
}

void Source::adapt(n_long t, n_long a)
{
#ifdef notdef
	t += skew_;
/*XXX fixed for now */
#define NTP_SECOND 65536
#define LGAIN (1./32.)
	late_estimator_ *= 1 - LGAIN;
	if (a > t + 0.25 * NTP_SECOND)
		late_estimator_ += 1. * LGAIN;

	int level = int(late_estimator_ * 10);
	if (level != rate_level_) {
		rate_level_ = level;
		decoder_->scale_rate(level);
	}
#endif
	int delta = delta_;
	int d = t - a;
	skew_ += (d - skew_) >> 6;
	delta_ = d;
	d -= delta;
	if (d < 0)
		d = -d;
#define DGAIN (1./16.)
	dvar_ += DGAIN * (double(d) - dvar_);
}

SourceManager* SourceManager::instance_;
SourceManager::SourceManager() : sources_(0), nsources_(0), localsrc_(0)
{
	memset((char*)hashtab_, 0, sizeof(hashtab_));

	clock_ = seconds();
	ssched(3);

	instance_ = this;
}

SourceManager::~SourceManager()
{
}

inline void SourceManager::remove_from_hashtable(Source* s)
{
	/* delete the source from hash table */
	n_long srcid = s->srcid();
	int h = SHASH(srcid);
	Source** p = &hashtab_[h];
	while (*p != s)
		p = &(*p)->hlink_;
	*p = (*p)->hlink_;
}

void SourceManager::init(n_long localid, n_long localaddr)
{
	/*
	 * create the local object.  remove it from the hash
	 * table since collision resolution changes local id
	 */
	localsrc_ = create(localid, localaddr);
	enter(localsrc_);
	remove_from_hashtable(localsrc_);
}

void SourceManager::timeout()
{
	clock_ = seconds();
	CheckActiveSources();
	ssched(3);
}

Source* SourceManager::enter(Source* s)
{
	s->next_ = sources_;
	sources_ = s;

	int h = SHASH(s->srcid());
	s->hlink_ = hashtab_[h];
	hashtab_[h] = s;

	++nsources_;

	return (s);
}


Source* SourceManager::consult(n_long srcid)
{
	int h = SHASH(srcid);
	for (Source* s = hashtab_[h]; s != 0; s = s->hlink_) {
		/*XXX pulling these values into variable seems
		  to work around a DEC c++ bug */
		n_long id = s->srcid();
		if (id == srcid)
			return (s);
	}
	return (0);
}

Source* SourceManager::lookup(n_long srcid, n_long addr)
{
	Source* s = consult(srcid);
	if (s == 0) {
		s = chkdup(srcid, addr);
		if (s == 0) {
			s = create(srcid, addr);
			enter(s);
		}
	}
	return (s);
}

/*
 * XXX hack for compatibitlity with vat (i.e., voice-activated switching)
 * until vat is ported to rtpv2 and we can use CNAMES
 */
Source* SourceManager::consult_by_addr_XXX(n_long addr)
{
        for (Source* s = sources_; s != 0; s = s->next_)
		if (s->addr() == addr)
			return (s);
	return (0);
}

void SourceManager::remove(Source* s)
{
	--nsources_;

	remove_from_hashtable(s);

	/* delete the source from list */
	Source** p = &sources_;
	while (*p != s)
		p = &(*p)->next_;
	*p = (*p)->next_;

#ifdef notyet
	chkdup(srcid);
#endif
	delete s;
}

void SourceManager::remove(n_long srcid)
{
	Source* s = consult(srcid);
	if (s != 0)
		remove(s);
}

n_long SourceManager::seconds()
{
	struct timeval tv;
	::gettimeofday(&tv, 0);
	return (n_long(tv.tv_sec));
}

void SourceManager::CheckActiveSources()
{
	n_long now = clock_;

	Source* n;
	for (Source* s = sources_; s != 0; s = n) {
		n = s->next_;
		n_long t = s->ts_done();
		if (t != 0 && (now >= t + 2 || s->dup())) {
			remove(s);
			continue;
		}
		t = s->ts_ctrl();
		if (t == 0)
			/*
			 * No session packets?  Probably ivs or nv sender.
			 * Revert to the data time stamp.
			 */
			t = s->ts_data();

		int idle = s->dup() ? CTRL_IDLE_DUP : CTRL_IDLE;
		if (now >= t + idle) {
			if (s->dup())
				remove(s);
			else
				s->disable();
		} else
			s->enable();
	}
}

Source* SourceManager::chkdup(n_long srcid, n_long addr)
{
	/*XXX - should eventually be conditioned on cname not ipaddr */
	n_long now = clock_;
	register Source *s;
	for (s = sources_; s != 0; s = s->next_) {
		if (s->addr() == addr) {
			n_long t = s->ts_done();
			if (t == 0)
				t = s->ts_data();
			if (t && int(now - t) >= 2)
				break;
		}
	}
	if (s) {
		remove_from_hashtable(s);
		s->srcid(srcid);
		s->skew(0);
		int h = SHASH(srcid);
		s->hlink_ = hashtab_[h];
		hashtab_[h] = s;
	}
	return (s);
}

int SourceManager::compare(const void* v0, const void* v1)
{
	const Source* x = *(Source**)v0;
	const Source* y = *(Source**)v1;

	const char* yn = y->name();
	if (yn == 0) {
		yn = y->cname();
		if (yn == 0)
			return (-1);
	}
	const char* xn = x->name();
	if (xn == 0) {
		xn = x->cname();
		if (xn == 0)
			return (1);
	}
	return (strcmp(xn, yn));
}

/*
 * Sort the sources by name and format a corresponding list
 * of ascii srcid's in the input buffer.
 */
void SourceManager::sortactive(char* cp) const
{
	static int ntab;
	static Source** srctab;
	if (srctab == 0) {
		ntab = 2 * nsources_;
		if (ntab < 32)
			ntab = 32;
		srctab = new Source*[ntab];
	} else if (ntab < nsources_) {
		ntab = 2 * nsources_;
		delete srctab;
		srctab = new Source*[ntab];
	}
	int n = 0;
	for (Source* s = sources_; s != 0; s = s->next_)
		if (s->handler() != 0)
			/* source is active if it has a PacketHandler */
			srctab[n++] = s;

	if (n == 0) {
		*cp = 0;
		return;
	}
	qsort(srctab, n, sizeof(*srctab), compare);
	for (int i = 0; i < n; ++i) {
		sprintf(cp, "%lu ", (u_long)ntohl(srctab[i]->srcid()));
		cp += strlen(cp);
	}
	/* nuke trailing space */
	cp[-1] = 0;
}
