/*
 * Copyright (c) 1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/source.h,v 1.2 1995/08/16 00:46:47 olson Exp (LBL)
 */

#ifndef vic_source_h
#define vic_source_h

#include <string.h>
#include <sys/param.h>
extern "C" {
#include <tcl.h>
}

#include "timer.h"
#include "inet.h"

class SourceManager;
class Tcl;

#define SOURCE_HASH 1024

class PacketHandler {
 public:
	inline PacketHandler(int hdrlen) { hdrlen_ = hdrlen; }
	inline int hdrlen() const { return (hdrlen_); }
	virtual int recv(struct rtphdr*, u_char* vh, u_char* data,
			 int len, int jitter) = 0;
 protected:
	int hdrlen_;
};

class Source {
 protected:
	Source(n_long srcid, n_long addr);
 public:
	virtual ~Source();

	virtual PacketHandler* activate(int format, u_char* pkthdr) = 0;
	virtual PacketHandler* change_format(int format, u_char* pkthdr) = 0;
	virtual void deactivate() = 0;

	inline PacketHandler* handler() const { return (handler_); }

	void touch_data();
	void touch_ctrl();
	void touch_done();
	inline int ts_data() const { return (ts_data_); }
	inline int ts_ctrl() const { return (ts_ctrl_); }
	inline int ts_done() const { return (ts_done_); }

	inline int datantp() const { return (datantp_); }
	inline void datantp(n_long n) { datantp_ = n; }
	inline int ctrlntp() const { return (ctrlntp_); }
	inline void ctrlntp(n_long n) { ctrlntp_ = n; }

	inline const char* name() const { return (name_); }
	inline const char* cname() const { return (cname_); }
	inline void SetName(const char*);
	inline void SetCName(const char*);
	inline n_long addr() const { return (addr_); }
	inline void addr(n_long a) { addr_ = a; }
	inline n_long srcid() const { return (srcid_); }
	inline void srcid(n_long s) { srcid_ = s; }
	inline int dup() { return (dup_); }
	inline void dup(int v) { dup_ = v; }
	inline void format(int v) { format_ = v; }
	inline int format() { return (format_); }
	inline void mute(int v) { mute_ = v; }
	inline int mute() { return (mute_); }
	void enable();
	void disable();

	void adapt(n_long src_ts, n_long arr_ts);
	inline void skew(n_long s) { skew_ = s; }
	inline n_long skew() const { return (skew_); }

	/*XXX should start at random values*/
	inline n_long nb() const { return (nb_); }
	inline n_long nf() const { return (nf_); }
	inline n_long np() const { return (np_); }
	inline n_long ns() const { return (cs_ - fs_); }
	inline n_long cs() const { return (cs_); }
	inline n_long snp() const { return (snp_); }
	inline n_long runt() const { return (nrunt_); }
	inline n_long late() const { return (nlate_); }
	inline n_long badsesslen() const { return (badsesslen_); }
	inline n_long badsessver() const { return (badsessver_); }
	inline n_long badsessopt() const { return (badsessopt_); }
	inline n_long badsdes() const { return (badsdes_); }
	inline n_long badbye() const { return (badbye_); }
	inline void nb(int v) { nb_ += v; }
	inline void nf(int v) { nf_ += v; }
	inline void np(int v) { np_ += v; }
	inline void snp(int v) { snp_ = v; }
	inline void fs(int v) { fs_ = v; }
	inline void runt(int v) { nrunt_ += v; }
	inline void late(int v) { nlate_ += v; }
	inline void badsesslen(int v) { badsesslen_ += v; }
	inline void badsessver(int v) { badsessver_ += v; }
	inline void badsessopt(int v) { badsessopt_ += v; }
	inline void badsdes(int v) { badsdes_ += v; }
	inline void badbye(int v) { badbye_ += v; }
	void cs(u_short v);
	/*XXX*/
	inline double dvar() const { return (dvar_); }
	inline const char* tcl_method() const { return (method_); }
 protected:
	virtual void name(const char*);
	void cname(const char*);
 public:
	Source* next_;		/* link for SourceManager source list */
	Source* hlink_;		/* link for SourceManager hash table */
 protected:
	static int command(ClientData, Tcl_Interp*, int ac, char** av);
	virtual int command(Tcl_Interp*, int ac, char** av) = 0;
	char method_[16];	/* tcl command instance name */
	char* name_;		/* source name */
	char* cname_;		/* canonical rtp name */
	int cname_hash_;	/* for quick comparisons */

	PacketHandler* handler_;

	n_long srcid_;		/* rtp global src id (net order) */
	n_long addr_;		/* address of sender (net order) */

	n_long ts_data_;	/* time of last data packet */
	n_long ts_ctrl_;	/* time of last control packet */
	n_long ts_done_;	/* time of bye packet */
	n_long datantp_;	/* sender's ntp time from last data packet */
	n_long ctrlntp_;	/* sender's ntp time from last control packet */

	n_long fs_;		/* first seq. no received */
	n_long cs_;		/* current (most recent) seq. no received */
	n_long np_;		/* no. packets received  */
	n_long nf_;		/* no. video frames received */
	n_long nb_;		/* no. bytes received */
	n_long snp_;		/* last advertised no. pkts received */
	n_long nlate_;		/* no. of late packets */
	n_long nrunt_;		/* count of packets too small */
	/* following errors are from session (rtcp) processing */
	n_long badsesslen_;	/* bad header length field */
	n_long badsessver_;	/* bad header version number */
	n_long badsessopt_;	/* unrecognized option */
	n_long badsdes_;	/* sdes cnt > available data */
	n_long badbye_;		/* bye cnt > available data */

	int format_;		/* RTP packet type */
	int mute_;		/* true if source muted */
	int disabled_;		/* true when signal lost */
	int dup_;		/* true when > one source with same addr */

	n_long skew_;		/* clock skew */
	double dvar_;		/* delay variance */
	int delta_;		/* last delay time */
};

class SourceManager : public Timer {
 protected:
	SourceManager();
 public:
	static inline SourceManager& instance() { return (*instance_); }
	~SourceManager();
	void init(n_long localid, n_long localaddr);
 protected:
	virtual Source* create(n_long srcid, n_long addr) = 0;
	static SourceManager* instance_;
 public:
	Source* lookup(n_long srcid, n_long addr);
	Source* consult(n_long srcid);
	Source* consult_by_addr_XXX(n_long addr);
	inline int nsources() const { return (nsources_); }
	inline Source* sources() const { return (sources_); }

	void CheckActiveSources();
	void ListSources();

	n_long clock() const { return (clock_); }
	inline Source* localsrc() const { return (localsrc_); }
	virtual void timeout();

	void sortactive(char*) const;
	void remove(Source*);
 private:
	static int compare(const void*, const void*);
	Source* enter(Source* s);
	void remove(n_long srcid);
	void remove_from_hashtable(Source* s);

	Source* chkdup(n_long srcid, n_long addr);

	n_long seconds();

	int nsources_;
	Source* sources_;
	Source* hashtab_[SOURCE_HASH];
	n_long clock_;
 protected:
	Source* localsrc_;
};

inline void Source::SetName(const char* s)
{
	if (name_ == 0 || strcmp(name_, s) != 0)
		name(s);
}

inline void Source::SetCName(const char* s)
{
	if (cname_ == 0 || strcmp(cname_, s) != 0)
		cname(s);
}

inline void Source::touch_ctrl()
{
	ts_ctrl_ = SourceManager::instance().clock();
}

inline void Source::touch_data()
{
	ts_data_ = SourceManager::instance().clock();
}

inline void Source::touch_done()
{
	ts_done_ = SourceManager::instance().clock();
}

inline void Source::cs(u_short v)
{
	/*
	 * This routine updates a 32 bit sequence number based on
	 * the rtp packet's 16 bit seq. no.
	 */
	register int c = cs_;
	register int d = v - c;
	if (d < -1024 || d > 1024) {
		cs_ = v;
		if (v < 512 && c > 0x10000-512) {
			/*
			 * seq no. wrapped - subtract 64k from fs to
			 * account for it.
			 */
			fs_ -= 0x10000;
		} else {
			/*
			 * the seq. no made a very large jump.  assume
			 * that the other side restarted without telling
			 * us about it so just re-sync (i.e., pretend
			 * this was the first packet).
			 */
			fs_ = v - 1;
			np_ = 0;
			nf_ = 0;
			nb_ = 0;
			snp_ = 0;
		}
	} else if (d > 0) {
		/* d <= 0 means duplicate or reordered packet - ignore */
		cs_ = v;
	}
}

#endif
