/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/switcher.h,v 1.2 1995/08/16 00:46:52 olson Exp (LBL)
 */

#ifndef vic_switcher_h
#define vic_switcher_h

#include "group-ipc.h"
#include "video.h"

class vicSource;

class UserWindow : public ResizeAction, public TclAction, public Timer {
public:
	UserWindow* next_;
	UserWindow(vicSource*);
	virtual ~UserWindow();
	virtual void resize(VideoWindow* w, int width, int height);
	virtual void command(VideoWindow* w, int argc, char** argv);
	virtual void timeout();

	void switchto(vicSource* s);
	inline VideoWindow* window() { return (window_); }
	inline int topmatch(const char* s) {
		return (strcmp(s, toplev_) == 0);
	}
	void name(const char*);
	inline int mode() { return (mode_); }
#define USERWIN_MODE_LOCKED	0
#define USERWIN_MODE_SWITCHED	1
#define USERWIN_MODE_BROWSE	2
	inline void touch(int v) { touch_ = v; }
	inline int touch() { return (touch_); }
	inline vicSource* source() const { return (source_); }
	inline void source(vicSource* s) { source_ = s; }
	void destroy();
private:
	vicSource* source_;
	VideoWindow* window_;
	int mode_;
	int touch_;
	int interval_;
	char toplev_[80];
};

class Switcher : public IPCHandler {
 protected:
	Switcher();
 public:
	static Switcher& instance();
	virtual ~Switcher();
	void ipc_input(int type, int frompid, u_char* p, int len);
	static void local(n_long v) { local_ = v; }
	void detach(vicSource*);
	void detach(UserWindow*);
	void attach(UserWindow*);
	static inline void channel(int c) { channel_ = c; }
 private:
	static Switcher* instance_;
	void detach(int);
	UserWindow* lookupLRU(n_long);
	GroupIPC* ipc_;

#define MAX_SWITCHED_WINDOWS 10
	int nwin_;
	UserWindow* windows_[MAX_SWITCHED_WINDOWS];
	int clock_;
	static n_long local_;
	static int channel_;
};

#endif
