/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) /home/mmlab1/Software/CVSMaster/vic/video.C,v 1.3 1995/08/26 21:16:02 olson Exp (LBL)";
#endif

#include <osfcn.h>
#include <string.h>
#include <sys/param.h>
#include <sys/ipc.h>
#include <tk.h>

#ifndef NOSHM

extern "C" {
#include <sys/shm.h>
#if defined(sun) && !defined(__svr4__)
int shmget(key_t, int, int);
char *shmat(int, char*, int);
int shmdt(char*);
int shmctl(int, int, struct shmid_ds*);
#endif
#ifdef __osf__
int XShmGetEventBase(struct _XDisplay *);
#else
int XShmGetEventBase(struct Display *);
#endif
#ifdef sgi
#define XShmAttach __XShmAttach__
#define XShmDetach __XShmDetach__
#define XShmPutImage __XShmPutImage__
#endif
#include <X11/extensions/XShm.h>
#ifdef sgi
#undef XShmAttach
#undef XShmDetach
#undef XShmPutImage
int XShmAttach(Display*, XShmSegmentInfo*);
int XShmDetach(Display*, XShmSegmentInfo*);
int XShmPutImage(Display*, Drawable, GC, XImage*, int, int, int, int,
		 int, int, int);
#endif
}

#endif

#include <X11/Xutil.h>
#include "video.h"
#include "color.h"
#include "Tcl.h"

VideoImage::VideoImage(Tk_Window tk, int w, int h)
	: shared_(0)
{
	int depth = Tk_Depth(tk);
	dpy_ = Tk_Display(tk);
	image_ = XCreateImage(dpy_, Tk_Visual(tk),
			      /*XXX*/
			      depth, ZPixmap, 0, (char*)0,
			      w, h, depth == 24? 32 : depth, 0);
}

VideoImage::~VideoImage()
{
	/*
	 * The pixel data must be freed from aobve.
	 */
	/*XXX*/
	XSync(dpy_, 0);
	if (image_ != 0) {
		image_->data = 0;
		image_->obdata = 0;
		XDestroyImage(image_);
	}
}

#ifdef NOSHM
SharedVideoImage::SharedVideoImage(Tk_Window tk, int w, int h,
				   int size, int rmid) 
	: VideoImage(tk, w, h)
{
	int depth = Tk_Depth(tk);
	if (depth > 8) {
		if (depth == 24)
			size *= 4;
		else
			size *= depth / 8;
	}
	image_->data = (char*)malloc(size);
}
SharedVideoImage::~SharedVideoImage()
{
	free(image_->data);
}
#else
SharedVideoImage::SharedVideoImage(Tk_Window tk, int w, int h,
				   int size, int rmid) 
	: VideoImage(tk, w, h)
{
	extern int use_shm;

	int depth = Tk_Depth(tk);
	if (depth > 8) {
		if (depth == 24)
			size *= 4;
		else
			size *= depth / 8;
	}
	if (use_shm) {
		shminfo_.shmid = shmget(IPC_PRIVATE, size, IPC_CREAT|0777);
		if (shminfo_.shmid < 0) {
			perror("shmget");
			exit(1);
		}
		shminfo_.shmaddr = (char*)shmat(shminfo_.shmid, 0, 0);
		if (shminfo_.shmaddr == (char*)-1) {
			perror("shmat");
			exit(1);
		}
		shminfo_.readOnly = 1;
		XShmAttach(dpy_, &shminfo_);
		image_->obdata = (char*)&shminfo_;
		shared_ = 1;
		/*
		 * Once the X server has attached the shm segments,
		 * we rmid them so they will go away when we exit.
		 * The sync is to make the X server do the attach
		 * before we do the rmid.
		 */
		if (rmid) {
			XSync(dpy_, 0);
			(void)shmctl(shminfo_.shmid, IPC_RMID, 0);
		}
	} else {
		shminfo_.shmid = -1;
		shared_ = 0;
		shminfo_.shmaddr = (char*)malloc(size);
	}
	image_->data = shminfo_.shmaddr;
}

SharedVideoImage::SharedVideoImage(Tk_Window tk, int w, int h,
				   u_char* shmaddr, int shmid)
	: VideoImage(tk, w, h)
{
	shminfo_.shmid = shmid;
	shminfo_.shmaddr = (char*)shmaddr;
	shminfo_.readOnly = 1;
	XShmAttach(dpy_, &shminfo_);
	image_->obdata = (char*)&shminfo_;
	shared_ = 1;
	image_->data = shminfo_.shmaddr;
}

SharedVideoImage::~SharedVideoImage()
{
	if (shminfo_.shmid >= 0) {
		XShmDetach(dpy_, &shminfo_);
		if (shmdt(shminfo_.shmaddr) < 0)
			perror("vic: shmdt");
	} else
		free(shminfo_.shmaddr);
}
#endif

GC VideoWindow::gc_;

VideoWindow::VideoWindow(const char* name, int width, int height)
	: vi_(0),
	  ra_(0), ka_(0), ta_(0),
	  hoff_(0), voff_(0),
	  width_(width), height_(height), damage_(0),
	  outboard_(0)
{
	Tcl& tcl = Tcl::instance();
	tk_ = Tk_CreateWindowFromPath(tcl.interp(), tcl.tkmain(),
				      (char*)name, 0);
	if (tk_ == 0)
		abort();
	const ColorModel& cm = ColorModel::instance();
	Tk_SetWindowVisual(tk_, cm.visual(), cm.depth(), cm.colormap());
	Tk_SetClass(tk_, "Vic");
	Tk_GeometryRequest(tk_, width, height);

#ifndef NOSHM
	completion_event_ = XShmGetEventBase(Tk_Display(tk_)) + ShmCompletion;
	/*
	 * Set up a generic handler to filter out completion events.
	 * We need to do this because tk might throw away a completion
	 * event before we have a chance to do the XIfEvent, which
	 * would cause us to block.
	 */
	Tk_CreateGenericHandler(filter, (ClientData)this);
#endif
	Tk_CreateEventHandler(tk_, ExposureMask|StructureNotifyMask,
			      handle, (ClientData)this);

	if (gc_ == 0) {
#if TK_MAJOR_VERSION < 4
		XColor* c = Tk_GetColor(tcl.interp(), tk_, Tk_Colormap(tk_),
					Tk_GetUid("gray50"));
#else
		XColor* c = Tk_GetColor(tcl.interp(), tk_,
					Tk_GetUid("gray50"));
#endif
		if (c == 0)
			abort();
		XGCValues v;
		v.background = c->pixel;
		v.foreground = c->pixel;
		const u_long mask = GCForeground|GCBackground;
		gc_ = Tk_GetGC(tk_, mask, &v);
	}
	callback_pending_ = 0;
	completion_pending_ = 0;
	tcl.CreateCommand(Tk_PathName(tk_), command, (ClientData)this);

	y0_ = 0;
	y1_ = height;
}

VideoWindow::~VideoWindow()
{
	if (callback_pending_)
		Tk_CancelIdleCall(display, (ClientData)this);
	
	/*
	 * Probably don't have to do this (e.g., it gets
	 * destroyed with the rest of the window)
	 */
	Tk_DeleteEventHandler(tk_, ExposureMask|StructureNotifyMask,
			      handle, (ClientData)this);
	/*
	 * We do have to delete this one.  Tk won't because
	 * it's not associated with any window (i.e., the
	 * window destruction won't delete this).
	 */
	Tk_DeleteGenericHandler(filter, (ClientData)this);

	for (VideoWindow** p = &allwin; *p != 0; p = &(*p)->next_) {
		if (*p == this) {
			*p = next_;
			break;
		}
	}
}

VideoWindow* VideoWindow::allwin;

VideoWindow* VideoWindow::name_to_window(const char* name)
{
    VideoWindow* p;
	for (p = allwin; p != 0; p = p->next_) {
		if (strcmp(Tk_PathName(p->tk_), name) == 0)
			break;
	}
	return (p);
}

int VideoWindow::command(ClientData cd, Tcl_Interp* tcl, int argc, char** argv)
{
	VideoWindow* vw = (VideoWindow*)cd;
	if (argc == 4 && strcmp(argv[1], "resize") == 0) {
		int w = atoi(argv[2]);
		int h = atoi(argv[3]);
		Tk_GeometryRequest(vw->tk_, w, h);
		return (TCL_OK);
	}
	if (argc == 3 && strcmp(argv[1], "outboard") == 0) {
		vw->outboard_ = atoi(argv[2]);
		return (TCL_OK);
	}
	/*XXX TclAction should return result*/
	if (vw->ta_ != 0)
		vw->ta_->command(vw, argc, argv);
	return (TCL_OK);
}

/*
 * Called via tcl 'video' command.
 */
int VideoWindow::create(ClientData cd, Tcl_Interp* tcl, int argc, char** argv)
{
	if (argc != 4) {
		Tcl_AppendResult(tcl, "\"", argv[0], "\": arg mismatch", 0);
		return (TCL_ERROR);
	}
	const char* name = argv[1];
	int width = atoi(argv[2]);
	int height = atoi(argv[3]);
	VideoWindow* p = new VideoWindow(name, width, height);
	p->next_ = allwin;
	allwin = p;

	return (TCL_OK);
}

void VideoWindow::display(ClientData cd)
{
	VideoWindow* vw = (VideoWindow*)cd;
	vw->callback_pending_ = 0;
	vw->draw(0, 0, vw->height_);
}

void VideoWindow::sync()
{
	XSync(Tk_Display(tk_), 0);
}

void VideoWindow::draw(int want_completion, int y0, int y1)
{
	if (callback_pending_) {
		callback_pending_ = 0;
		Tk_CancelIdleCall(display, (ClientData)this);
	}
	if (!Tk_IsMapped(tk_))
		return;

	Display *dpy = Tk_Display(tk_);
	Window window = Tk_WindowId(tk_);

	if (vi_ == 0) {
		XFillRectangle(dpy, window, gc_, 0, 0, width_, height_);
		return;
	}
	int hoff = (width_ - vi_->width()) >> 1;
	int voff = (height_ - vi_->height()) >> 1;
	if (damage_) {
		damage_ = 0;
		if (hoff > 0) {
			XFillRectangle(dpy, window, gc_,
				       0, 0, hoff, height_ + 1);
			XFillRectangle(dpy, window, gc_,
				       width_ - hoff, 0, hoff, height_ + 1);
		}
		if (voff > 0) {
			XFillRectangle(dpy, window, gc_,
				       0, 0, width_, voff + 1);
			XFillRectangle(dpy, window, gc_,
				       0, height_ - voff, width_, voff + 1);
		}
	}
	int h = y1 - y0;
	if (h > vi_->height())
		h = vi_->height();
	else if (h < 0)
		return;

#ifndef NOSHM
	if (vi_->shared()) {
		/*XXX shouldn't happen (or rarely if it does)*/
		if (completion_pending_)
			return;
		completion_pending_ |= want_completion;
		XShmPutImage(dpy, window, gc_, vi_->ximage(),
			     0, y0, hoff, y0 + voff,
			     vi_->width(), h, want_completion);
	} else
#endif
		XPutImage(dpy, window, gc_, vi_->ximage(), 
			  0, y0, hoff, y0 + voff,
			  vi_->width(), h);
}

int VideoWindow::filter(ClientData cd, XEvent* ep)
{
	VideoWindow* w = (VideoWindow*)cd;
#ifndef NOSHM
	if (ep->type == w->completion_event_) {
		XShmCompletionEvent* e = (XShmCompletionEvent*)ep;
		if (e->drawable == Tk_WindowId(w->tk_)) {
			w->completion_pending_ = 0;
			return (1);
		}
	}
#endif
	return (0);
}

int VideoWindow::match(Display*, XEvent* ep, char* cd)
{
	return (filter((ClientData)cd, ep));
}

void VideoWindow::purge()
{
	if (completion_pending_) {
		XEvent e;
		XIfEvent(Tk_Display(tk_), &e, match, (char*)this);
	}
}

void VideoWindow::handle(ClientData cd, XEvent* ep)
{
	VideoWindow* w = (VideoWindow*)cd;
	
	switch (ep->type) {
	case Expose:
		w->damage_ = 1;
		if (ep->xexpose.count == 0)
			w->redraw();
		break;

	case DestroyNotify:
		/* This will delete the VideoWindow as well.
		   XXX can handle be called again after this
		   happens?  E.g., do we have to use defereed
		   freeing?  This will be tricky to incorporate
		   with delete. */
		if (w->ka_ != 0)
			w->ka_->kill(w);
		break;

	case MapNotify:
		/*
		 * The ShmCompletion event gets lost when 
		 * we unmap.  This can cause completion_pending_
		 * to never get set back to zero.
		 */
		w->completion_pending_ = 0;
		break;

	case ConfigureNotify:
		if (w->ra_ != 0 &&
		    (w->width_ != ep->xconfigure.width ||
		     w->height_ != ep->xconfigure.height))
			w->resize(ep->xconfigure.width, ep->xconfigure.height);
		break;
	}
}

void VideoWindow::resize(int w, int h)
{
	if (w != width_ || h != height_) {
		width_ = w;
		height_ = h;
		ra_->resize(this, width_, height_);
	}
}

/*XXX*/
void VideoWindow::redraw()
{
	if (!callback_pending_) {
		callback_pending_ = 1;
		Tk_DoWhenIdle(display, (ClientData)this);
	}
}

void VideoWindow::render(const VideoImage* v)
{
	vi_ = v;
	draw(1, 0, v->height());
}

void VideoWindow::render(const VideoImage* v, int miny, int maxy)
{
	vi_ = v;
	draw(1, miny, maxy);
}

Tk_Window VideoWindow::tkmain() const
{
	return (Tcl::instance().tkmain());
}
