/*
 * Copyright (c) 1993-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and the Network Research Group at
 *      Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) /home/mmlab1/Software/CVSMaster/vic/video.h,v 1.2 1995/08/16 00:46:57 olson Exp (LBL)
 */

#ifndef vic_video_h
#define vic_video_h

#include <X11/Xlib.h>
#ifndef NOSHM
extern "C" {
#include <X11/extensions/XShm.h>
}
#endif
extern "C" {
#include <tk.h>
}

class VideoWindow;

class ResizeAction {
 public:
	virtual void resize(VideoWindow*, int pwidth, int pheight) = 0;
};

class KillAction {
 public:
	virtual void kill(VideoWindow*) = 0;
};

class TclAction {
 public:
	virtual void command(VideoWindow*, int argc, char** argv) = 0;
};

/*
 * A wrapper for an XImage class.  The bitmap array (ximage->data)
 * doesn't actually get allocated here.
 */
class VideoImage {
 public:
	VideoImage(Tk_Window, int width, int height);
	virtual ~VideoImage();
	inline int shared() const { return shared_; }
	inline XImage* ximage() const { return (image_); }
	inline int width() const { return (image_->width); }
	inline int height() const { return (image_->height); }
 protected:
	Display* dpy_;
	XImage* image_;
	int shared_;
};

#ifdef NOSHM
class SharedVideoImage : public VideoImage {
public:
	SharedVideoImage(Tk_Window, int width, int height, int size,
			 int rmid = 1);
	virtual ~SharedVideoImage();
};
#else
/*
 * A class for ximages, which will be allocate in shared memory
 * if available.  The constructor takes the width and height
 * of the bitmap, plus a size which must be >= width*height.
 * The size might want to be bigger than width*height for decoders
 * that might walk off the end of an image because of bit errors,
 * but the distance they walk off the end is bounded.
 */
class SharedVideoImage : public VideoImage {
public:
	SharedVideoImage(Tk_Window, int width, int height, int size,
			 int rmid = 1);
	SharedVideoImage(Tk_Window, int width, int height,
			 u_char* shmaddr, int shmid);
	virtual ~SharedVideoImage();
	inline int shmid() const { return (shminfo_.shmid); }
	inline char* buffer() const { return (shminfo_.shmaddr); }
protected:
	XShmSegmentInfo	shminfo_;
};
#endif

class VideoWindow {
public:
	VideoWindow(const char* name, int width, int height);
	~VideoWindow();
	VideoWindow* next_;
	Tk_Window tkmain() const;
	inline Tk_Window tkwin() { return (tk_); }
	inline int width() { return (width_); }
	inline int height() { return (height_); }

	void setdim(int width, int height);
	void grow();
	void shrink();
	void ra(ResizeAction* a) { ra_ = a; }
	void ka(KillAction* a) { ka_ = a; }
	void ta(TclAction* a) { ta_ = a; }
	void setimage(VideoImage* v) { vi_ = v; }
	void render(const VideoImage* vi);
	void render(const VideoImage* vi, int miny, int maxy);
	void sync();
	void purge();
	void redraw();
	void destroy();
	inline void damage() { damage_ = 1; }

	/* hack to hide J-Video artifacts off window */
	void voff(int v) { voff_ = v; }
	void hoff(int v) { hoff_ = v; }

	static void display(ClientData);
	static void handle(ClientData, XEvent*);
	static int filter(ClientData, XEvent*);
	static int match(Display*, XEvent* ep, char* cd);
	static int instance(ClientData, Tcl_Interp*, int argc, char** argv);

	/*
	 * Return the VideoWindow object associated with
	 * the tk window identified by name.  Name has the
	 * usual tk form, e.g., ".top.video.win0".
	 */
	static VideoWindow* name_to_window(const char* name);

	static VideoWindow* allwin;
	static int create(ClientData, Tcl_Interp*, int argc, char **argv);
	static int command(ClientData, Tcl_Interp*, int argc, char **argv);

	/*XXX*/
	inline int outboard() const { return (outboard_); }
private:
	void draw(int want_completion, int miny, int maxy);
	void resize(int width, int height);

	Tk_Window tk_;
	static GC gc_;
	ResizeAction* ra_;
	KillAction* ka_;
	TclAction* ta_;
	const VideoImage* vi_;
	int callback_pending_;
	int completion_pending_;
	int completion_event_;
	int damage_;

	int width_;
	int height_;
	int voff_;
	int hoff_;
	int y0_;
	int y1_;

	int outboard_;
};

#endif
